/************************************************************
*
*  Copyright (c) 2002-2004 Jakub Schwarzmeier
*
************************************************************/

/*
*  Based on following papers:
*
*	J. E. Barnes, P. Hut, Nature, 1986, vol. 324, p. 446.
*	U. Becciani et al., J. Comput. Phys., 2000, vol. 163, p. 118
*/


#pragma warning(disable:4786)

#include "../misc.h"
#include "bh_distrib.h"
#include <algorithm>  //  for_each, copy, swap

using namespace JakubSchwarzmeier::BarnesHut;

extern double theta;

LocalTree::LocalTree( RList<JakubSchwarzmeier::BarnesHut::Leave> & leavesList,
		    RemoteTree & remoteTree ): leavesList(leavesList),
			remoteTree(remoteTree), g(0)
{}


void LocalTree::construct()
{
	//  setup & initialize variables
    //
	//  adaptive bisect -- find greatest value of spatial coordinate in absolut value
	//
	//  both level & bisect can be computed from the other, bisect was previously defined as bisect = DIMENSION_LENGTH

	double bisect = 0;
	for(int i = 0; i < leavesList.getCount(); i++)
	{
		for(int c = 0; c < 3; c++)
		{
			double val = fabs(leavesList.getAt(i).position[c]);
			if ( val > bisect )  bisect = val;
		}

		//  clear old costzones load info
		leavesList[i].load = 0;
	}

	//  increase by one (because of condition that compares corners of cube with values of particles)
	//  it must be farther then the furthest particle
	bisect++;

	//  setup root node
	root = new Node( *(new Vector<>) );
	root->bisect = bisect;
	Node * actual;
	for(i = 0; i < 3; i++)	root->corner[i] = -bisect;
	for(i = 0; i < leavesList.getCount(); i++) root->leaves.Add( &( leavesList[i] ) );

	//  Phase 1: whole LocalTree construction
	//

	int level = 0;
	mostLeftAncestors.Add( new RList<Node>(100, false) );
	mostLeftAncestors[level].Add( root );

	while( level < mostLeftAncestors.getCount() )
	{

		#ifdef _DEBUG
		std::cout << "level: " << level;
		#endif

		int pos = 0;
		while( pos < mostLeftAncestors[level].getCount() )
		{
			actual = &( mostLeftAncestors[level][pos] );	

			//  partition actual node & create its children
			//

			//  create list for corners, corners are freed in part of code that scans if concrete leave belongs to some sub-cube
			RList< Vector<> > initCorners(8, false);
			//  add corners and init their values in ordinary cube
			initCubeCorners( actual->corner, initCorners, bisect );

			//  create list for Hilbert's order
			RList< Vector<> > finalCorner(8, false);
			for(int i = 0; i < 8; i++)  finalCorner.Add( &(initCorners[i]) );	//  values have no meaning now, only for the right size of list

			//  order corners according to Hilbert's curve
			//

			//  make from ordinary cube Hilbert's cube
			RList< Vector<> >::iterator initIter = finalCorner.begin();
			initIter += initCorners[7];	initIter += initCorners[5];
			initIter += initCorners[1];	initIter += initCorners[3];
			initIter += initCorners[2];	initIter += initCorners[0];
			initIter += initCorners[4];	initIter = initCorners[6];
			
			//  order curve according to angles
			
			//  x rotation
			for(i = 0; i < actual->angles.x; i=i+90 )
			{
				//  make tmp copy
//				std::copy( finalCorner.begin(), finalCorner.end(), initCorners.begin() );
				for(int j = 0; j < 8; j++)  initCorners.storage[j] = finalCorner.storage[j];

				RList< Vector<> >::iterator iterator = finalCorner.begin();
				iterator += initCorners[1];	iterator += initCorners[2];
				iterator += initCorners[3];	iterator += initCorners[0];
				iterator += initCorners[7];	iterator += initCorners[4];
				iterator += initCorners[5];	iterator = initCorners[6];
			}

			//  y rotation
			for(i = 0; i < actual->angles.y; i=i+90 )
			{
				//  make tmp copy	... nejde protoze on dela hlubokou kopii (nekopiruje pointery, ale hodnoty)
				// plati i pro vyse a nize
				//  nelze totiz ani
				// for(int j = 0; j < 8; j++)  initCorners[j] = finalCorner[j];
				//  on nezkopiruje pointery, ale udela hlubokou kopii
//				std::copy( finalCorner.begin(), finalCorner.end(), initCorners.begin() );
				for(int j = 0; j < 8; j++)  initCorners.storage[j] = finalCorner.storage[j];

				RList< Vector<> >::iterator iterator = finalCorner.begin();
				iterator += initCorners[7];	iterator += initCorners[6];
				iterator += initCorners[1];	iterator += initCorners[0];
				iterator += initCorners[3];	iterator += initCorners[2];
				iterator += initCorners[5];	iterator = initCorners[4];
			}

			//  z rotation
			for(i = 0; i < actual->angles.z; i=i+90 )
			{
				//  make tmp copy
//				std::copy( finalCorner.begin(), finalCorner.end(), initCorners.begin() );
				for(int j = 0; j < 8; j++)  initCorners.storage[j] = finalCorner.storage[j];

				RList< Vector<> >::iterator iterator = finalCorner.begin();
				iterator += initCorners[1];	iterator += initCorners[6];
				iterator += initCorners[5];	iterator += initCorners[2];
				iterator += initCorners[3];	iterator += initCorners[4];
				iterator += initCorners[7];	iterator = initCorners[0];
			}

			//
			//  order corners according to Hilbert's curve

			//  space partitioning according to Hilbert's curve
			//

			for(i = 0; i < 8; i++)
			{
				//  we get rightCorner by adding bisect to every component of corner vector
				Vector<> rightCorner = finalCorner[i];
				std::for_each( rightCorner.begin(), rightCorner.end(), AddValue<> (bisect) );

				int count = 0;
				int pos = 0;	//  nested, it don't interfere with pos outside this scope

				Node * node = NULL;

				//  find out how many leaves belong to this partition, so iterate over all of them
				for(int k = 0; k < actual->leaves.getCount(); k++)
				{
					//  look at operators >= and < ! they determinate which leave belong to this partition
					if ( (actual->leaves[k].position >= finalCorner[i]) && (actual->leaves[k].position < rightCorner) )
					{  //  leave leaves[k] belongs to partion given by corner vector
						if ( count == 0 )
						{
							pos = k;
							count++;
						}
						else
						{
							if ( count == 1)		//  true only ones
							//	at least two leaves -> create new node
							{
								Node::Angles angles;
								switch (i)	//  i identifies present corner of Hilbert's cube
								{
								case 0:
									angles.x = actual->angles.getXinc90();
									angles.y = actual->angles.getYdec90();
									angles.z = actual->angles.z;
									break;
								
								case 1:
								case 2:
									angles.x = actual->angles.getXdec90();
									angles.y = actual->angles.y;
									angles.z = actual->angles.getZinc90();
									break;
								
								case 3:
								case 4:
									angles.x = actual->angles.getXinc180();
									angles.y = actual->angles.y;
									angles.z = actual->angles.z;
									break;

								case 5:
								case 6:
									angles.x = actual->angles.getXdec90();
									angles.y = actual->angles.getYdec90();
									angles.z = actual->angles.z;
									break;

								case 7:
									angles.x = actual->angles.getXinc90();
									angles.y = actual->angles.getYinc90();
									angles.z = actual->angles.z;
									break;
								}

								node = new Node( finalCorner[i], angles );

								if ( mostLeftAncestors.getCount() == level+1 )
								{
									mostLeftAncestors.Add( new RList<Node>(100, false) );
								}

								actual->childList.Add ( node );	//  and finaly add this node between actual's children
								node->parent = actual;
								node->bisect = bisect;
								node->grouped = actual->grouped;
								node->childNo = actual->childList.getCount()-1;
								
								mostLeftAncestors[level+1].Add( node );			//  add to LocalTree

								node->leaves.Add( &( actual->leaves[pos] ) );	//  add previous
							}

							node->leaves.Add( &( actual->leaves[k] ) );
							count++;	
						}
					} 
				}  //  end of cycling through collection of positions

				if (count == 1)	//  only one child node so that's leave
				{
					actual->childList.Add( &actual->leaves[pos] );
					if (!actual->grouped)  groupsList.Add( new ExtGroup( &actual->leaves[pos] ) );  //  WD99, only one leave in group
					delete &(finalCorner[i]);	//  Hilbert
				}
				else
				if ( (count > 1) && (count <= 32) && (!actual->grouped) )	//  WD99
				{
					node->grouped = true;
					groupsList.Add( new ExtGroup(node) ); 
				}
				else
				if (count == 0)					//  Hilbert's corner
				{
					delete &(finalCorner[i]);
				}
			
			}  //  end of cycling through corners
			//  
			//	space partitioning according to Hilbert's curve

			
			//  only one child -> leads to long tails, don't want this (happens when all leaves belongs to only one of eight sub-cubes)
			if ( ( actual->childList.getCount() == 1 ) && ( actual != root ) && ( actual->grouped != true) )
			{
				#ifdef _DEBUG
				std::cout << "Jedinacek" << std::endl;
				#endif		
				
				//  remove "actual" from "parents" "childList"
				actual->parent->childList.own(false);
				actual->parent->childList.removeAt(actual->childNo);
				actual->parent->childList.own(true);
				//  update position of all "children" from "parent's" childList
				for( int i = actual->childNo; i < actual->parent->childList.getCount(); i++)
				{
					if ( typeid(actual->parent->childList[i]) == typeid(Node) )
					{
						static_cast<Node&>(actual->parent->childList[i]).childNo -= 1;
					}
				}
				//  add the only actual's child into "parent's" "childList" and position in this list save in childs "member value"
				static_cast<Node&>(actual->childList[0]).childNo = actual->parent->childList.Add( &( actual->childList[0] ) );
				//  set child's new "parent"
				static_cast<Node&>(actual->childList[0]).parent = actual->parent;
				//  set actual's "childList" so, that is not deleted
				actual->childList.own(false);
				//  remove actual from list and from memory
				mostLeftAncestors[level].own(true);
				mostLeftAncestors[level].removeAt(pos);
				mostLeftAncestors[level].own(false);
				//  last node in this level, perform some speciality
				if ( mostLeftAncestors[level].getCount() == 0 )
				{
					for(int i = 0; i < mostLeftAncestors[level+1].getCount(); i++)
					{
						mostLeftAncestors[level].Add ( &( mostLeftAncestors[level+1][i] ) );
					}
					mostLeftAncestors.removeAt(level+1);
					pos = 0;
					bisect /= 2;

				}
	
			}
			else
			{
				pos++;
			}

		}	//  end while( pos < mostLeftAncestors[level].getCount() )

		#ifdef _DEBUG
		std::cout << ", pocet pozic:" << pos << std::endl;
		#endif


		bisect /= 2;
		level++;
	
	}	//  end while( level < mostLeftAncestors.getCount() )

	#ifdef _DEBUG
	std::cout << "Phase 1 END" << std::endl;
	#endif

	//
	//  end phase 1, LocalTree construction done

	//  Phase 2: computing vector "center of mass" and scalar "mass"
	//

	//  for all levels including root
	for(level = mostLeftAncestors.getCount()-1; level >= 0; level--)
	{
		
		for( int pos = 0; pos < mostLeftAncestors[level].getCount(); pos++)
		{
			actual = &( mostLeftAncestors[level][pos] );
			Vector<> numerator = 0;

			for(int i = 0; i < actual->childList.getCount(); i++)
			{
				actual->mass += actual->childList[i].mass;
				//  multiplication of vector with number
				numerator += actual->childList[i].mass * actual->childList[i].position;
			}
			//  quotient of vector by number
			actual->position = numerator / actual->mass;	//  "center of mass" (geometric quantity)
		}

	}

	#ifdef _DEBUG
	std::cout << "Phase 2 END" << std::endl;
	#endif

	//
	//  end phase 2, computing "center of mass" and "mass"
}

bool LocalTree::evaluateNext()
{

	//  Phase 3: force evaluation
	//

	//  local tree

	if ( g == groupsList.getCount() )  return false;

	int pos = 0;
	Node * cell = root;

	std::stack<Node*> openedStack;
	std::stack<int> positionStack;

	group = &( groupsList[g] );
	extern double epsilon;		
	
	bool done;

	#ifdef _DEBUG
	std::cout << "Evaluating group/particle no." << g << std::endl;
	#endif

	do
	{
		while ( pos < cell->childList.getCount() )	//  walk through all children
		{
			Vector<> r = cell->childList[pos].position - group->element->position;
			double d = r.length();

			if ( (void*)&cell->childList[pos] == (void*)group->element ) { pos++; }	//  group (leave/node) don't act on themselves (and 1/(r=0)=-1.#IND00000 )	A NEMEL BYCH KONTROLOVAT r? (nekdy se castice mohou dostat tak blizko a pak by to zkolabovalo!!!)
			else
			if ( (typeid( cell->childList[pos] ) == typeid( Leave )) || (d * theta >= dynamic_cast<Node&>(cell->childList[pos]).bisect) )
			{
				real sphere = 3 * group->getCellSize() * sqrt(3) / 2;
				if ( d > sphere )
				{
					group->acceleration += r * cell->childList[pos].mass / pow( d + epsilon, 3 );  //  far list
					cell->childList[pos].load++;	//  costzones
				}
				else
				{	//  group is a leave or node that is too close to group
					group->nearList.Add( & (cell->childList[pos]) );
				}
				
				pos++;
			}
			else
			{
				openedStack.push( cell );
				positionStack.push( pos );

				#ifdef _DEBUG
				std::cout << "Particles in stack: "<< positionStack.size()<<std::endl;
				#endif

				cell = static_cast<Node*>( &(cell->childList[pos]) );
				pos = 0;
			}
		}
		
		if ( !openedStack.empty() )
		{
			cell = openedStack.top(); openedStack.pop();
			pos = positionStack.top(); positionStack.pop();
			pos++;
			done = false;
		}
		else
		{
			done = true;
		}

	} while (!done);

	
	//  remote tree
	
	evaluateRemote(0);

	g++;
	return g < groupsList.getCount();	
}

void LocalTree::evaluateRemote(int call)
{
	//  we can:
	//  * evaluate remote tree for actual group -- call = 0 -- call from evaluateNext, call = 1 -- call from computing.cpp
	//  * evaluate remote tree for actual group and for every group in stack

	if ( ( g != groupsList.getCount() ) && (call == 0) )	group = &( groupsList[g] );		//  pokud se pocita k aktualni skupina
	else
	if (!waitingGroups.empty()) 	//  pokud jsou vsechny skupiny spocteny a neco je ve stacku
	{
		group = waitingGroups.top();
		waitingGroups.pop();
	}
	else return;		//  v tom pripade je jiz vse hotovo a balime

	int pos = 0;

	std::stack<int> positionStack;
	extern double epsilon;		

	bool done;

	if ( remoteTree.rootsArrived ) 
	{
		do	//  } while ( group != NULL );
		{
			//  distinct from localTree computation beacause if we make request we must undo accelerations done from current remote tree
			//  walk through every remote tree
			for(int j = 0; j < remoteTree.root->childList.getCount(); j++)
			{
				RemoteNode * remoteCell = dynamic_cast<RemoteNode*>(&remoteTree.root->childList[j]);
				std::stack<RemoteNode*> openedStack;
					
				pos = 0;
				//  remember current acceleration lists
				Vector<> acceleration = group->acceleration;
					
				do  //  } while (!done);
				{
					while ( pos < remoteCell->childList.getCount() )	//  walk through all children
					{
						Vector<> r = remoteCell->childList[pos].position - group->element->position;
						double d = r.length();

						if ( (typeid( remoteCell->childList[pos] ) == typeid( Leave )) || (d * theta >= dynamic_cast<RemoteNode&>(remoteCell->childList[pos]).bisect) )
						{
							real sphere = 3 * group->getCellSize() * sqrt(3) / 2;
							if ( d > sphere )
							{
								group->acceleration += r * remoteCell->childList[pos].mass / pow( d + epsilon, 3 );  //  far list
							}
							else
							{
								group->remoteNearList.Add( & (remoteCell->childList[pos]) );
							}
							
							pos++;
						}
						else
						{
							//  go deeper

							openedStack.push( remoteCell );
							positionStack.push( pos );

							remoteCell = static_cast<RemoteNode*>( &(remoteCell->childList[pos]) );
							pos = 0;
							if ( ( remoteCell->childList.getCount() == 0 ) ) 
							{
								// not enough remote nodes, make remote request
									
								//  clear stacks
								while ( !openedStack.empty() )
								{
									openedStack.pop();
									positionStack.pop();	
								}

								//  set original acceleration
								group->acceleration = acceleration;

								extern long nRequests;
								InterlockedIncrement( &nRequests );		//  for checking that all requests are satisfied before end in computing.cpp

								message Message( (5*sizeof(double))+sizeof(int) );
									
								Message.Initialize(CM_COMPUTING_REQUEST);
								Message.Add( group->element->mass );
								Message.Add( group->element->position[0] );
								Message.Add( group->element->position[1] );
								Message.Add( group->element->position[2] );
								Message.Add( group->getCellSize() );
								Message.Add( (int)group );
								Message.Finish();
								
								remoteCell->Client->WriteFile( Message.GetBuffer(), Message.GetBufferSize() );
							}

						}	//  end else
						
					}	//  end while ( pos < remoteCell->childList.getCount() )	
						
					if ( !openedStack.empty() )
					{
						remoteCell = openedStack.top(); openedStack.pop();
						pos = positionStack.top(); positionStack.pop();
						pos++;
						done = false;
					}
					else
					{
						done = true;
					}

				} while (!done);
				
			}	//  end for(int j = 0; j < remoteTree.root->childList.getCount(); j++)
				
		
			//  compute rest groups (their particles) from list with freshly arrived roots
			if ( !waitingGroups.empty() )
			{
				group = waitingGroups.top();
				waitingGroups.pop();
			}
			else
			{
				group = NULL;
			}
			
		} while ( group != NULL );

	
	}	//  end true if ( remoteTree.rootsArrived )
	else
	{
		//  waiting for remote roots, push actual group onto stack

		waitingGroups.push( group );
	}

}

void LocalTree::integrate()
{
	extern double epsilon;		

	//  finish acceleration computing

	for(int i = 0; i < groupsList.getCount(); i++)	
	{

		group = &(groupsList[i]);

		if ( typeid( *group->element ) == typeid(Leave) )	//  only one particle in cell
		{
			//  add far forces acceleration
			//
			dynamic_cast<Leave*>(group->element)->acceleration += group->acceleration;

			//  compute near forces
			//
			int j;
			
			for(j = 0; j < group->nearList.getCount(); j++)
			{
				Vector<> r = group->nearList[j].position - group->element->position;
				dynamic_cast<Leave*>(group->element)->acceleration += r * group->nearList[j].mass / pow( r.length() + epsilon, 3 );
				group->nearList[j].load++;	//  costzones
			}

			for(j = 0; j < group->remoteNearList.getCount(); j++)
			{
				Vector<> r = group->remoteNearList[j].position - group->element->position;
				dynamic_cast<Leave*>(group->element)->acceleration += r * group->remoteNearList[j].mass / pow( r.length() + epsilon, 3 );
			}

			for(j = 0; j < group->remoteremoteNearList.getCount(); j++)
			{
				Vector<> r = group->remoteremoteNearList[j].position - group->element->position;
				dynamic_cast<Leave*>(group->element)->acceleration += r * group->remoteremoteNearList[j].mass / pow( r.length() + epsilon, 3 );
			}

		}
		else	//  more particles in cell
		{
			for(int k = 0; k < dynamic_cast<Node*>(group->element)->leaves.getCount(); k++)
			{
				//  for leave "k"			
				
				//  add far forces
				dynamic_cast<Node*>(group->element)->leaves[k].acceleration += group->acceleration;

				int j;

				//  add near forces
				for(j = 0; j < group->nearList.getCount(); j++)
				{
					Vector<> r = group->nearList[j].position - dynamic_cast<Node*>(group->element)->leaves[k].position;
					dynamic_cast<Node*>(group->element)->leaves[k].acceleration += r * group->nearList[j].mass / pow( r.length() + epsilon, 3 );
					group->nearList[j].load++;	//  costzones
				}

				for(j = 0; j < group->remoteNearList.getCount(); j++)
				{
					Vector<> r = group->remoteNearList[j].position - dynamic_cast<Node*>(group->element)->leaves[k].position;
					dynamic_cast<Node*>(group->element)->leaves[k].acceleration += r * group->remoteNearList[j].mass / pow( r.length() + epsilon, 3 );
				}

				for(j = 0; j < group->remoteremoteNearList.getCount(); j++)
				{
					Vector<> r = group->remoteremoteNearList[j].position - dynamic_cast<Node*>(group->element)->leaves[k].position;
					dynamic_cast<Node*>(group->element)->leaves[k].acceleration += r * group->remoteremoteNearList[j].mass / pow( r.length() + epsilon, 3 );
				}

				//  add acceleration from leaves in grouping cell
				for(j = 0; j < dynamic_cast<Node*>(group->element)->leaves.getCount(); j++)
				{
					if (k != j)
					{
						Vector<> r = dynamic_cast<Node*>(group->element)->leaves[j].position - dynamic_cast<Node*>(group->element)->leaves[k].position;
						dynamic_cast<Node*>(group->element)->leaves[k].acceleration += r * dynamic_cast<Node*>(group->element)->leaves[j].mass / pow( r.length() + epsilon, 3 );
						dynamic_cast<Node*>(group->element)->leaves[j].load++;
					}
				}
			}
			
		}

	} //  evaluate all leaves


	//  integrate, leap-frog

	#ifdef _DEBUG
	std::cout << "Phase 3 END" << std::endl;
	#endif

	//
	//  end phase 3: force evaluation

	//  integrate, leap-frog

	extern double dt;
	for( i = 0; i < leavesList.getCount(); i++ )
	{
		leavesList[i].velocity += leavesList[i].acceleration * dt;
		leavesList[i].position += leavesList[i].velocity * dt;
	}
	
	//  prepare acceleration for next step
	for(i = 0; i < leavesList.getCount(); i++)
	{
		leavesList[i].acceleration = 0;
	}

	//  propagate costzones load info to leaves
	for(int level = 0; level < mostLeftAncestors.getCount(); level++)
	{
		
		for(int pos = 0; pos < mostLeftAncestors[level].getCount(); pos++)
		{
			for(int i = 0; i <  mostLeftAncestors[level][pos].childList.getCount(); i++)
			{
				mostLeftAncestors[level][pos].childList[i].load += mostLeftAncestors[level][pos].load; // / mostLeftAncestors[level][pos].childList.getCount();
			}
		}
	}

	delete root;
}

void LocalTree::initCubeCorners( Vector<> & superiorCorner, RList< Vector<> > & initCorners, real & bisect)
{
	static int i = -1;
	Vector<> corner = superiorCorner;	//  calls "copy" constructor & performs "deep" copy

	i++;	//  actual component
	for(int j = 0; j < 2; j++)	//  bisection; every component divided into two parts
	{
		corner[i] += j*bisect;
		if (i+1 < 3)	//  3 = DIMENSIONS
		{	//  still not at the end of recursion, go deeper :-))
			initCubeCorners(corner, initCorners, bisect);
		}
		else
		{	//  one of final solutions, final coordinates of corner
			initCorners.Add( new Vector<> );
			initCorners [ initCorners.getCount()-1 ] = corner;
		}  //  end of code for one of final solutions
	
	}  // end of bisection cycle

	i--;
}

void LocalTree::fillMessage( message & msg )
{
	//  send root + two next branches
	UINT max_level = 2;

	Node * actual	= root;
	std::stack<Node*> openedStack;

	std::stack<int> positionStack;

	int pos = 0;
	bool done;

	//  insert root node
	msg.Add( (int)0 );	//  node
	msg.Add( actual->mass );
	msg.Add( actual->position[0] );
	msg.Add( actual->position[1] );
	msg.Add( actual->position[2] );
	msg.Add( actual->bisect );

	do
	{
		while ( pos < actual->childList.getCount() )	//  walk through all children
		{
			if ( typeid(actual->childList[pos]) == typeid(Leave) )
			{
				msg.Add( (int)1 );	//  leave
				
				msg.Add( actual->childList[pos].mass );
				msg.Add( actual->childList[pos].position[0] );
				msg.Add( actual->childList[pos].position[1] );
				msg.Add( actual->childList[pos].position[2] );

				pos++;
			}
			else
			{
				if ( openedStack.size() < max_level )	//  only to defined level
				{

					openedStack.push( actual );
					positionStack.push( pos );

					msg.Add( (int)0 );	//  node
					msg.Add( actual->childList[pos].mass );
					msg.Add( actual->childList[pos].position[0] );
					msg.Add( actual->childList[pos].position[1] );
					msg.Add( actual->childList[pos].position[2] );

					msg.Add( dynamic_cast<Node*>(&actual->childList[pos])->bisect );

					actual = dynamic_cast<Node*>(&actual->childList[pos]);
					pos = 0;
				}
				else
				{
					pos++;
				}
			}
		}
		
		if ( !openedStack.empty() )
		{
			msg.Add( (int)2 );	//  up

			actual = openedStack.top(); openedStack.pop();
			pos = positionStack.top(); positionStack.pop();
			pos++;
			done = false;
		}
		else
		{
			done = true;
		}

	} while (!done);

}

RemoteTree::RemoteTree()
{
	root = new RemoteNode;
	rootsArrived = false;
	root->bisect;
}

RemoteTree::~RemoteTree()
{
	delete root;
}

void RemoteTree::getMessage( crackmessage & CrackMessage, client * n_Client )
{
	RemoteNode * actual = root;
	std::stack<RemoteNode*> stack;

	while ( CrackMessage.NextData() )
	{
		switch ( CrackMessage.GetInt() )
		{
		case 0:		//  node

			stack.push(actual);

			RemoteNode * newNode;
			newNode = new RemoteNode;
			actual->childList.Add( newNode );
			actual = newNode;
			
			actual->mass = CrackMessage.GetDouble();
			actual->position[0] = CrackMessage.GetDouble();
			actual->position[1] = CrackMessage.GetDouble();
			actual->position[2] = CrackMessage.GetDouble();
			actual->bisect = CrackMessage.GetDouble();

			actual->Client = n_Client;

			break;

		case 1:		//  leave

			JakubSchwarzmeier::BarnesHut::Leave * newLeave;
			newLeave = new JakubSchwarzmeier::BarnesHut::Leave;
			actual->childList.Add( newLeave );
			newLeave->mass = CrackMessage.GetDouble();
			newLeave->position[0] = CrackMessage.GetDouble();
			newLeave->position[1] = CrackMessage.GetDouble();
			newLeave->position[2] = CrackMessage.GetDouble();

			break;

		case 2:

			if (stack.empty())  MessageBox(TEXT("A je to tady"));

			actual = stack.top(); 
			stack.pop();

			break;

		default:

			MessageBox(TEXT("Chybna data"));
		}
	}
}

void RequestsList::Add( RequestGroup * reqNode )
{
	requestNodes.Add( reqNode );
}

void RequestsList::evaluateNext()
{
//	const double kappa=6.67259e-11;


	int pos = 0;
	Node * cell = localTree->root;

	std::stack<Node*> openedStack;
	std::stack<int> positionStack;

	RequestGroup * group = & (requestNodes.getAt( requestNodes.getCount()-1 ) );
	bool done;
	extern double epsilon;		

	do
	{
		while ( pos < cell->childList.getCount() )	//  walk through all children
		{
			Vector<> r = cell->childList[pos].position - group->position;
			double d = r.length();

			if ( (typeid( cell->childList[pos] ) == typeid( JakubSchwarzmeier::BarnesHut::Leave )) || (d * theta >= dynamic_cast<JakubSchwarzmeier::BarnesHut::Node&>(cell->childList[pos]).bisect) )
			{
				real sphere = 3 * group->cellSize * sqrt(3) / 2;
				if ( d > sphere )
				{
					group->acceleration += r * cell->childList[pos].mass / pow( d + epsilon, 3 );  //  far list
					cell->childList[pos].load++;	//  costzones
				}
				else
				{
					group->nearList.Add( & (cell->childList[pos]) );
				}
				
				pos++;
			}
			else
			{
				openedStack.push( cell );
				positionStack.push( pos );

				#ifdef _DEBUG
				std::cout << "Particles in stack: "<< positionStack.size()<<std::endl;
				#endif

				cell = static_cast<JakubSchwarzmeier::BarnesHut::Node*>( &(cell->childList[pos]) );
				pos = 0;
			}
		}
		
		if ( !openedStack.empty() )
		{
			cell = openedStack.top(); openedStack.pop();
			pos = positionStack.top(); positionStack.pop();
			pos++;
			done = false;
		}
		else
		{
			done = true;
		}

	} while (!done);


	//  send back to requestor

	MEMORYSTATUS ms;
	ms.dwLength = sizeof(ms);
	GlobalMemoryStatus(&ms);

	while(ms.dwMemoryLoad > 99)
	{
		Sleep(0);
		GlobalMemoryStatus(&ms);
	}

	message Message(3*sizeof(double)+sizeof(int)+4*sizeof(double)*group->nearList.getCount());
	Message.Initialize(CM_REQUEST_SATISFIED);

	Message.Add( (int)group->pointer );
	Message.Add( (double)group->acceleration[0] );
	Message.Add( (double)group->acceleration[1] );
	Message.Add( (double)group->acceleration[2] );

	//  bacha na preteceni
	for(int i = 0; i < group->nearList.getCount(); i++)
	{
		Message.Add( (double)group->nearList[i].mass );
		Message.Add( (double)group->nearList[i].position[0] );
		Message.Add( (double)group->nearList[i].position[1] );
		Message.Add( (double)group->nearList[i].position[2] );
	}

	Message.Finish();

	group->Client->WriteFile( Message.GetBuffer(), Message.GetBufferSize() );
	
	requestNodes.removeAt( requestNodes.getCount()-1 );

}