/************************************************************
*
*  Copyright (c) 2002-2004 Jakub Schwarzmeier
*
************************************************************/

#include "general.h"
#include "data/xml.h"

struct tagSharedEncounterParameters
{
	bool keplerian;			//  will be there keplerian collision?

	//  parameters of keplerian encounter
	double numericEccenricity;
	double pericentricSeparation;
	double timeToPericenter;

	//  parameters of direct encounter

	//  common parameters
	double velocityOfCompanion;
	double initialDistanceToCenter;
	double distanceToCenterDuringPassage;

	double inclinationOfCentral;
	double argumentOfPericenterOfCentral;

	double inclinationOfCompanion;
	double argumentOfPericenterOfCompanion;

	//  input galaxy filenames
	TCHAR centralGalaxyFileName[MAX_PATH], companionGalaxyFileName[MAX_PATH];
	//  storage related filenames
	TCHAR szFolder[MAX_PATH], szName[MAX_PATH], szDescription[2048];

} SharedEncounterParameters;

void rotate( JakubSchwarzmeier::Data::RList< JakubSchwarzmeier::BarnesHut::Leave > & list,
			 double incl, double omega )
{
	for(int i = 0; i < list.getCount(); i++)
	{
		//  positions
		double x = list[i].position[0];
		double y = list[i].position[1];
		double z = list[i].position[2];
		list[i].position[0] = z * sin_angle(incl) + cos_angle(incl)*(x*cos_angle(omega) - y*sin_angle(omega));
		list[i].position[1] = x*sin_angle(omega)+y*cos_angle(omega);
		list[i].position[2] = z*cos_angle(incl)-sin_angle(incl)*(x*cos_angle(omega)-y*sin_angle(omega));

		//  velocities
		double vx = list[i].velocity[0];
		double vy = list[i].velocity[1];
		double vz = list[i].velocity[2];
		list[i].velocity[0] = vz * sin_angle(incl) + cos_angle(incl)*
			(vx*cos_angle(omega) - vy*sin_angle(omega));
		list[i].velocity[1] = vx*sin_angle(omega)+vy*cos_angle(omega);
		list[i].velocity[2] = vz*cos_angle(incl)-sin_angle(incl)*
			(vx*cos_angle(omega)-vy*sin_angle(omega));
	}
}


void CreateEncounterSimulation( tagSharedEncounterParameters & pdata )
{
	//  load bodies from files into lists
	JakubSchwarzmeier::Data::RList< JakubSchwarzmeier::BarnesHut::Leave > centralGalaxy(10000, false);
	JakubSchwarzmeier::Data::RList< JakubSchwarzmeier::BarnesHut::Leave > companionGalaxy(10000, false);
	LoadBodiesFromFile(pdata.centralGalaxyFileName, centralGalaxy);
	LoadBodiesFromFile(pdata.companionGalaxyFileName, companionGalaxy);

	oneDiscBodyMass = companionGalaxy[0].mass;
	oneBulgeBodyMass = centralGalaxy[0].mass;

	int i;		//  for cycles

	if (pdata.keplerian)
	{
		//  central galaxy mass (galaxy as sun)
		double ms = 0;
		for(i = 0; i < centralGalaxy.getCount(); i++)
		{
			ms += centralGalaxy[i].mass;
		}

		//  companion galaxy mass (galaxy as planet)
		double mp = 0;
		for(i = 0; i < companionGalaxy.getCount(); i++)
		{
			mp += companionGalaxy[i].mass;
		}

		//  works fine to tp = 0.5 * T ;  but it doesn't matter, because back it's the same
		double epsilon = pdata.numericEccenricity;
		double rp = pdata.pericentricSeparation;
		double tp = pdata.timeToPericenter;

		//  determination of a, b and e
		double a = rp / (1-epsilon);	//  semimajor axis
		double e = a * epsilon;			//  linear eccentricity (or how you call this you english-speaking boys)
		double b = sqrt( (pow(a, 2) - pow(e, 2)) );	//  semiminor axis

		//  determination of eccentric anomaly E from Kepler's equation
			
		double T = sqrt( (a*a*a*4*M_PI*M_PI)/(ms+mp) ); //  period from Kepler's third extended law
		double M = (360/T) * (tp - T);		//  mean anomaly

		//  iteratively compute E
		//

		//  first approximation
		double E = M + (M + epsilon*sin_angle(M) - M)/(1 - epsilon*cos(M) );		//  helio
		double E0 = E + (M + epsilon*sin_angle(E) - E)/(1 - epsilon*cos_angle(E) );	//  helio

		do
		{
			if ( fabs(E - E0) > 0.0000000001 )  E0 = E;

			E = E0 + (M + epsilon * sin_angle(E0) - E0) / (1 - epsilon * cos_angle(E0));		//  method encke, sky & telescope

		} while ( fabs(E - E0) > 0.0000000001 );

		if (E < 0)  E = 360 + E;

		//
		//  iteratively compute E
		
		//  determination of fi and r
		double fi = acos_angle( (cos_angle(E)-epsilon)/(1-(epsilon*cos_angle(E)) ));	//  true anomaly
		if (fi < 0)  fi = 360 + fi;		

		double r = (pow(b, 2)/a)/(1+(epsilon*cos_angle(fi)));

		//  companion x, y positions and velocity

		double companionX = epsilon + r * cos_angle(fi);
		double companionY = r * sin_angle(fi);

		double vel = sqrt( ((epsilon+1)*ms)/r );
		
		//  convert velocity to components
		double vx = -vel * sin_angle( fi );
		double vy = vel * cos_angle( fi );

		//  central x position, velocity is conserved
		double centralX = a - rp;

		//  change orientation of central galaxy according to inclination and arg. of pericenter
		//

		//  find center of mass of central gallaxy
		JakubSchwarzmeier::Math::Vector<> numerator = 0;
		double mass = 0;
		for(i = 0; i < centralGalaxy.getCount(); i++)
		{
			mass += centralGalaxy[i].mass;
			//  multiplication of vector with number
			numerator += centralGalaxy[i].mass * centralGalaxy[i].position;
		}
		JakubSchwarzmeier::Math::Vector<> cm = numerator / mass;	//  "center of mass" (geometric quantity)

		//  position galaxy according to galaxy's center of mass
		for(i = 0; i < centralGalaxy.getCount(); i++)
		{
			centralGalaxy[i].position = centralGalaxy[i].position - cm;
		}

		//  perform rotations
		rotate(centralGalaxy, pdata.inclinationOfCentral, pdata.argumentOfPericenterOfCentral);

		//  position galaxy according keplerian orbit
		for(i = 0; i < centralGalaxy.getCount(); i++)
		{
			centralGalaxy[i].position[0] += centralX;
		}

		//  note: no changes in velocity for central galaxy (except that in rotation)

		//  change orientation of companion galaxy according to inclination and arg. of pericenter
		//

		//  find center of mass of companion gallaxy
		numerator = 0;
		mass = 0;
		for(i = 0; i < companionGalaxy.getCount(); i++)
		{
			mass += companionGalaxy[i].mass;
			//  multiplication of vector with number
			numerator += companionGalaxy[i].mass * companionGalaxy[i].position;
		}
		cm = numerator / mass;	//  "center of mass" (geometric quantity)

		//  position galaxy according to galaxy's center of mass
		for(i = 0; i < companionGalaxy.getCount(); i++)
		{
			companionGalaxy[i].position = companionGalaxy[i].position - cm;
		}

		//  perform rotations
		rotate(companionGalaxy, pdata.inclinationOfCompanion, pdata.argumentOfPericenterOfCompanion);

		//  position galaxy according keplerian orbit
		for(i = 0; i < companionGalaxy.getCount(); i++)
		{
			companionGalaxy[i].position[0] += companionX;
			companionGalaxy[i].position[1] += companionY;
			companionGalaxy[i].velocity[0] += vx;		
			companionGalaxy[i].velocity[1] += vy;
		}

	}
	else
	{
		//  find center of mass of central gallaxy
		//
		JakubSchwarzmeier::Math::Vector<> numerator = 0;
		double mass = 0;
		for(i = 0; i < centralGalaxy.getCount(); i++)
		{
			mass += centralGalaxy[i].mass;
			//  multiplication of vector with number
			numerator += centralGalaxy[i].mass * centralGalaxy[i].position;
		}
		JakubSchwarzmeier::Math::Vector<> cm = numerator / mass;	//  "center of mass" (geometric quantity)

		//  position galaxy according to galaxy's center of mass
		for(i = 0; i < centralGalaxy.getCount(); i++)
		{
			centralGalaxy[i].position = centralGalaxy[i].position - cm;
		}

		//  perform rotations
		rotate(centralGalaxy, pdata.inclinationOfCentral, pdata.argumentOfPericenterOfCentral);

		//  find center of mass of companion gallaxy
		//
		numerator = 0;
		mass = 0;
		for(i = 0; i < companionGalaxy.getCount(); i++)
		{
			mass += companionGalaxy[i].mass;
			//  multiplication of vector with number
			numerator += companionGalaxy[i].mass * companionGalaxy[i].position;
		}
		cm = numerator / mass;	//  "center of mass" (geometric quantity)

		//  position galaxy according to galaxy's center of mass
		for(i = 0; i < companionGalaxy.getCount(); i++)
		{
			companionGalaxy[i].position = companionGalaxy[i].position - cm;
		}

		//  perform rotations
		rotate(companionGalaxy, pdata.inclinationOfCompanion, pdata.argumentOfPericenterOfCompanion);

		//  move galaxy to initial distance from center and set distance to center during passage
		//  add velocity
		for(i = 0; i < companionGalaxy.getCount(); i++)
		{
			companionGalaxy[i].position[2] += pdata.distanceToCenterDuringPassage;
			companionGalaxy[i].position[0] += pdata.initialDistanceToCenter;
			companionGalaxy[i].velocity[0] -= pdata.velocityOfCompanion;
		}
		
		
	}

	unsortedBodiesList = new JakubSchwarzmeier::Data::RList< JakubSchwarzmeier::BarnesHut::Leave >;

	//  insert new bodies
	for(i = 0; i < centralGalaxy.getCount(); i++)
	{
		unsortedBodiesList->Add( &centralGalaxy[i] );
	}
	for(i = 0; i < companionGalaxy.getCount(); i++)
	{
		unsortedBodiesList->Add( &companionGalaxy[i] );
	}

	//  set initial load to one for evry body
	for(i = 0; i < unsortedBodiesList->getCount(); i++)  unsortedBodiesList->getAt(i).load = 1;

 	nTotalBodies = unsortedBodiesList->getCount();
	openGL->InitializeMasses();

	//  initiate files writing
	
	TCHAR szDataFileName[MAX_PATH], szVideoFileName[MAX_PATH];
	lstrcpy( szDataFileName, pdata.szFolder );
	lstrcat( szDataFileName, TEXT("\\") );
	lstrcat( szDataFileName, pdata.szName );
	lstrcpy( szVideoFileName, szDataFileName );

	lstrcat( szDataFileName, TEXT(".nbody") );
	lstrcat( szVideoFileName, TEXT(".avi") );

	//  create data and video simulation files
	auxInfo = new AuxInfo( szDataFileName, szVideoFileName );

	//  first writing of data and video data
	auxInfo->WriteData();


	//  write xml file
	//

	//  get logged user's name
	TCHAR szUserName[512];
	DWORD dwUserSize = sizeof(szUserName)/ sizeof(*szUserName);
	GetUserName( szUserName,  &dwUserSize );

	using namespace JakubSchwarzmeier::Data;

	XmlDocument xmlDoc;

	XmlProcessingInstruction pi;
	pi = xmlDoc.CreateProcessingInstruction(TEXT("xml"), TEXT("version='1.0'"));
	xmlDoc.AppendChild( pi );
	
	XmlElement root = xmlDoc.CreateElement(TEXT("simulation"));
	xmlDoc.AppendChild( root );
	root.SetAttribute(TEXT("type"), TEXT("rendezvous"));

	XmlElement element;
	XmlText text;

	element = xmlDoc.CreateElement(TEXT("count"));
	text = xmlDoc.CreateTextNode( nTotalBodies );
	element.AppendChild( text );
	root.AppendChild( element );
	
	element = xmlDoc.CreateElement(TEXT("format"));
	text = xmlDoc.CreateTextNode(TEXT("float"));
	element.AppendChild( text );
	root.AppendChild( element );

	element = xmlDoc.CreateElement(TEXT("description"));
	text = xmlDoc.CreateTextNode( pdata.szDescription );
	element.AppendChild( text );
	root.AppendChild( element );

	XmlElement parameters = xmlDoc.CreateElement(TEXT("parameters"));
	if (pdata.keplerian)
		parameters.SetAttribute( TEXT("class"), TEXT("keplerian") );
	else
		parameters.SetAttribute( TEXT("class"), TEXT("direct") );
	root.AppendChild( parameters );

	XmlElement components = xmlDoc.CreateElement(TEXT("components"));
	parameters.AppendChild( components );

	XmlElement central = xmlDoc.CreateElement(TEXT("central"));
	components.AppendChild( central );

	element = xmlDoc.CreateElement(TEXT("omega"));
	text = xmlDoc.CreateTextNode( pdata.argumentOfPericenterOfCentral );
	element.AppendChild( text );
	central.AppendChild( element );

	element = xmlDoc.CreateElement(TEXT("incl"));
	text = xmlDoc.CreateTextNode( pdata.inclinationOfCentral );
	element.AppendChild( text );
	central.AppendChild( element );


	XmlElement companion = xmlDoc.CreateElement(TEXT("companion"));
	components.AppendChild( companion );

	element = xmlDoc.CreateElement(TEXT("omega"));
	text = xmlDoc.CreateTextNode( pdata.argumentOfPericenterOfCompanion );
	element.AppendChild( text );
	companion.AppendChild( element );

	element = xmlDoc.CreateElement(TEXT("incl"));
	text = xmlDoc.CreateTextNode( pdata.inclinationOfCompanion );
	element.AppendChild( text );
	companion.AppendChild( element );

	if (pdata.keplerian)
	{
		element = xmlDoc.CreateElement(TEXT("numericeccentricity"));
		text = xmlDoc.CreateTextNode( pdata.numericEccenricity );
		element.AppendChild( text );
		parameters.AppendChild( element );

		element = xmlDoc.CreateElement(TEXT("timetopericenter"));
		text = xmlDoc.CreateTextNode( pdata.timeToPericenter );
		element.AppendChild( text );
		parameters.AppendChild( element );
		
		element = xmlDoc.CreateElement(TEXT("pericentricseparation"));
		text = xmlDoc.CreateTextNode( pdata.pericentricSeparation );
		element.AppendChild( text );
		parameters.AppendChild( element );
	}
	else
	{
		element = xmlDoc.CreateElement(TEXT("companionvelocity"));
		text = xmlDoc.CreateTextNode( pdata.velocityOfCompanion );
		element.AppendChild( text );
		parameters.AppendChild( element );
		
		element = xmlDoc.CreateElement(TEXT("distancetocenter"));
		text = xmlDoc.CreateTextNode( pdata.initialDistanceToCenter );
		element.AppendChild( text );
		parameters.AppendChild( element );
		
		element = xmlDoc.CreateElement(TEXT("distancetocenterduringpassage"));
		text = xmlDoc.CreateTextNode( pdata.distanceToCenterDuringPassage );
		element.AppendChild( text );
		parameters.AppendChild( element );
	}

	element = xmlDoc.CreateElement(TEXT("author"));
	text = xmlDoc.CreateTextNode( szUserName );
	element.AppendChild( text );
	root.AppendChild( element );

	element = xmlDoc.CreateElement(TEXT("theta"));
	text = xmlDoc.CreateTextNode( theta );
	element.AppendChild( text );
	root.AppendChild( element );

	element = xmlDoc.CreateElement(TEXT("epsilon"));
	text = xmlDoc.CreateTextNode( epsilon );
	element.AppendChild( text );
	root.AppendChild( element );

	element = xmlDoc.CreateElement(TEXT("dt"));
	text = xmlDoc.CreateTextNode( dt );
	element.AppendChild( text );
	root.AppendChild( element );

	PathRenameExtension( szDataFileName, TEXT(".xml") );
	xmlDoc.DoHumanMarkup( root );
	xmlDoc.Save( szDataFileName );
		
	//
	//  end write xml file

}

BOOL CALLBACK EncounterWelcomePageProc ( HWND hWndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	UNREFERENCED_PARAMETER(wParam);

    switch (uMsg)
    {
    case WM_INITDIALOG:
        { 
			tagSharedEncounterParameters * pdata;
			//  save shared data to window GWL_USERDATA field
			pdata = (tagSharedEncounterParameters*) ((LPPROPSHEETPAGE) lParam) -> lParam;
    		SetWindowLong(hWndDlg, GWL_USERDATA, (DWORD) pdata);

			return 0;
        }

    case WM_NOTIFY:
        {
        LPNMHDR lpnm = (LPNMHDR) lParam;

    
		switch (lpnm->code)
        {
			
		case PSN_SETACTIVE:
			//  sent each time a page is about to be made visible 
			//  initialize data & buttons

            //  enable the Next button
			PropSheet_SetWizButtons(GetParent(hWndDlg), PSWIZB_NEXT);
		    return 0;

		case PSN_WIZNEXT:
            //  Next button is clicked
			//  check entered data
            
		    return 0;
            
		case PSN_RESET:
			//  Cancel button is clicked


		    return 0;


        default:
        
		    return 0;
		}	//  end switch (lpnm->code)
        }
        
    default:
        
	    return 0;
    }

}

BOOL CALLBACK EncounterChooseKindOfEncounterPageProc( HWND hWndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	UNREFERENCED_PARAMETER(wParam);

    //  retrieve the shared user data from GWL_USERDATA, before WM_INITDIALOG there's nothing
	tagSharedEncounterParameters * pdata = (tagSharedEncounterParameters*) GetWindowLong(hWndDlg, GWL_USERDATA);

    switch (uMsg)
    {
    case WM_INITDIALOG:
        { 
			//  save shared data to window GWL_USERDATA field
			pdata = (tagSharedEncounterParameters*) ((LPPROPSHEETPAGE) lParam) -> lParam;
    		SetWindowLong(hWndDlg, GWL_USERDATA, (DWORD) pdata);

			//  select one of radio buttons
			//
			Button_SetCheck( GetDlgItem(hWndDlg, IDC_KEPLERIAN ), BST_CHECKED );

			return 0;
        }

    case WM_NOTIFY:
        {
        LPNMHDR lpnm = (LPNMHDR) lParam;

    
		switch (lpnm->code)
        {
			
		case PSN_SETACTIVE:
			//  sent each time a page is about to be made visible 
			//  initialize data & buttons

			//  enable Back as well as Next button
			PropSheet_SetWizButtons(GetParent(hWndDlg), PSWIZB_BACK | PSWIZB_NEXT);
            return 0;

		case PSN_WIZBACK:
            //  Next button is clicked
			//  check entered data

	        return FALSE;

		case PSN_WIZNEXT:
			{
            //  Next button is clicked
			//  check entered data

			WinString strValue(256);

			//  update pdata
			//

			//  keplerian or direct collision
			if ( IsDlgButtonChecked( hWndDlg, IDC_KEPLERIAN ) )
				pdata->keplerian = true;
			else
				pdata->keplerian = false;


			//  jump to correct property sheet page
			}
	        return FALSE;
            
		case PSN_RESET:
			//  Cancel button is clicked

            return 0;


        default:
        
		    return 0;
		}	//  end switch (lpnm->code)
        }
        
    default:
        
	    return 0;
    }
}

BOOL CALLBACK EncounterInitialFilesPageProc ( HWND hWndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
    //  retrieve the shared user data from GWL_USERDATA, before WM_INITDIALOG there's nothing
	tagSharedEncounterParameters * pdata = (tagSharedEncounterParameters*) GetWindowLong(hWndDlg, GWL_USERDATA);

    switch (uMsg)
    {
    case WM_INITDIALOG:
        { 
			//  save shared data to window GWL_USERDATA field
			pdata = (tagSharedEncounterParameters*) ((LPPROPSHEETPAGE) lParam) -> lParam;
    		SetWindowLong(hWndDlg, GWL_USERDATA, (DWORD) pdata);

			pdata->centralGalaxyFileName[0] = '\0';
			pdata->companionGalaxyFileName[0] = '\0';

			HANDLE hBitmap;
			
			//  load central galaxy bitmap and set it
			hBitmap = LoadImage(g_hInstance, TEXT("CentralGalaxyBitmap"), IMAGE_BITMAP, 0, 0, LR_LOADTRANSPARENT);
			SendMessage( 
			  GetDlgItem(hWndDlg, IDC_ENCOUNTER_BCENTRAL),  // handle to destination window 
			  BM_SETIMAGE,         // message to send
			  (WPARAM) IMAGE_BITMAP,     // image type
			  (LPARAM) hBitmap      // handle to the image (HANDLE)
			);

			//  load companion galaxy bitmap and set it
			hBitmap = LoadImage(g_hInstance, TEXT("CompanionGalaxyBitmap"), IMAGE_BITMAP, 0, 0, LR_LOADTRANSPARENT);
			SendMessage( 
			  GetDlgItem(hWndDlg, IDC_ENCOUNTER_BCOMPANION),  // handle to destination window 
			  BM_SETIMAGE,         // message to send
			  (WPARAM) IMAGE_BITMAP,     // image type
			  (LPARAM) hBitmap      // handle to the image (HANDLE)
			);

			return 0;
        }

	case WM_DESTROY:
		
		HANDLE hBitmap;

		//  free central galaxy bitmap
		hBitmap = (HANDLE) SendMessage( 
		  GetDlgItem(hWndDlg, IDC_ENCOUNTER_BCENTRAL),  
		  BM_SETIMAGE,         
		  (WPARAM) IMAGE_BITMAP,     
		  (LPARAM) 0
		);
		DeleteObject (hBitmap);

		//  free companion galaxy bitmap
		hBitmap = (HANDLE) SendMessage( 
		  GetDlgItem(hWndDlg, IDC_ENCOUNTER_BCOMPANION),  
		  BM_SETIMAGE,         
		  (WPARAM) IMAGE_BITMAP,     
		  (LPARAM) 0
		);

		DeleteObject (hBitmap);


		return 0;

    case WM_NOTIFY:
        {
        LPNMHDR lpnm = (LPNMHDR) lParam;

    
		switch (lpnm->code)
        {
			
		case PSN_SETACTIVE:
			//  sent each time a page is about to be made visible 
			//  initialize data & buttons

			//  enable Back as well as Next button
			PropSheet_SetWizButtons(GetParent(hWndDlg), PSWIZB_BACK | PSWIZB_NEXT);
            return 0;

		case PSN_WIZNEXT:
            //  Next button is clicked
			//  check entered data
            
			if ( lstrlen(pdata->centralGalaxyFileName) == 0 || lstrlen(pdata->companionGalaxyFileName) == 0 )
			{
				MessageBox( hWndDlg, Language.GetString(IDS_NEW_GALAXY_WIZARD_FILENAMESREQUIRED), NULL, MB_OK | MB_ICONINFORMATION );
				SetWindowLong(hWndDlg, DWL_MSGRESULT, -1);
				return TRUE;
			}

			//  jump to correct property sheet page
			if (pdata->keplerian)  
				PropSheet_SetCurSel(GetParent(hWndDlg), NULL, 2);
			else
				PropSheet_SetCurSel(GetParent(hWndDlg), NULL, 3);

	        return FALSE;

		case PSN_WIZBACK:
            //  Next button is clicked
			//  check entered data
            
			return FALSE;
            
		case PSN_RESET:
			//  Cancel button is clicked

            return 0;


        default:
        
		    return 0;
		}	//  end switch (lpnm->code)
        }
        

	case WM_COMMAND:

		switch ( LOWORD( wParam) )
		{
		case IDC_ENCOUNTER_BCENTRAL:


 		    switch ( HIWORD (wParam) )
			{
			case BN_CLICKED:
				{
				//  into following buffer is saved filename (without path) entered by user (unused)
				//  I am using pdata->dataFileName for full path + filename
				TCHAR szFileName [MAX_PATH] = TEXT("\0");

				OPENFILENAME ofn;
				ZeroMemory( &ofn, sizeof(ofn) );
				ofn.lStructSize		= sizeof (OPENFILENAME);
				ofn.Flags			= OFN_FILEMUSTEXIST | OFN_HIDEREADONLY;
				ofn.hwndOwner		= hWndDlg;
			    ofn.lpstrFilter     = Language.GetFileFilterString (IDS_NBODY_FILTERSTRING);
				ofn.nFilterIndex    = 1L;
				ofn.lpstrFile       = pdata->centralGalaxyFileName;
				ofn.nMaxFile        = sizeof (pdata->centralGalaxyFileName); //  fix this?
				ofn.lpstrFileTitle  = szFileName;
				ofn.nMaxFileTitle   = sizeof (szFileName);			//  fix this?

				GetOpenFileName (&ofn);

				//  show chosen name in dialog box static text field
//				SetDlgItemText(hWndDlg, IDC_NEW_GALAXY_DATA_FILENAME, pdata->dataFileName);
				}
				
				return 0;
			
			}	 //  end switch ( HIWORD (wParam) )


			return 0;
		
		case IDC_ENCOUNTER_BCOMPANION:

			switch ( HIWORD (wParam) )
			{
			case BN_CLICKED:

				TCHAR szFileName [MAX_PATH] = TEXT("\0");

				OPENFILENAME ofn;
				ZeroMemory( &ofn, sizeof(ofn) );
				ofn.lStructSize		= sizeof (OPENFILENAME);
				ofn.Flags			= OFN_FILEMUSTEXIST | OFN_HIDEREADONLY;
				ofn.hwndOwner		= hWndDlg;
			    ofn.lpstrFilter     = Language.GetString (IDS_NBODY_FILTERSTRING);
				ofn.nFilterIndex    = 1L;
				ofn.lpstrFile       = pdata->companionGalaxyFileName;
				ofn.nMaxFile        = sizeof (pdata->companionGalaxyFileName); //  fix this?
				ofn.lpstrFileTitle  = szFileName;
				ofn.nMaxFileTitle   = sizeof (szFileName);			//  fix this?

				GetOpenFileName (&ofn);

				//  show chosen name in dialog box static text field
//				SetDlgItemText(hWndDlg, IDC_NEW_GALAXY_VIDEO_FILENAME, pdata->videoFileName);
				return 0;
			}	//  end switch ( HIWORD (wParam) )

			return 0;
		
		
		}	//  end switch ( LOWORD( wParam) )

			
		return 0;	//  return WM_COMMAND

    default:
        
	    return 0;
    }
}

BOOL CALLBACK KeplerianEncounterParametersPageProc ( HWND hWndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	UNREFERENCED_PARAMETER(wParam);

    //  retrieve the shared user data from GWL_USERDATA, before WM_INITDIALOG there's nothing
	tagSharedEncounterParameters * pdata = (tagSharedEncounterParameters*) GetWindowLong(hWndDlg, GWL_USERDATA);

    switch (uMsg)
    {
    case WM_INITDIALOG:
        { 
			//  save shared data to window GWL_USERDATA field
			pdata = (tagSharedEncounterParameters*) ((LPPROPSHEETPAGE) lParam) -> lParam;
    		SetWindowLong(hWndDlg, GWL_USERDATA, (DWORD) pdata);

		    return 0;
        }

    case WM_NOTIFY:
        {
        LPNMHDR lpnm = (LPNMHDR) lParam;

    
		switch (lpnm->code)
        {
			
		case PSN_SETACTIVE:
			//  sent each time a page is about to be made visible 
			//  initialize data & buttons

			//  enable Back as well as Next button
			PropSheet_SetWizButtons(GetParent(hWndDlg), PSWIZB_BACK | PSWIZB_NEXT);
            return 0;

		case PSN_WIZBACK:
            //  Next button is clicked
			//  check entered data

	        return FALSE;

		case PSN_WIZNEXT:
			{
            //  Next button is clicked
			//  check entered data

			WinString strValue(256);

			//  update pdata
			//

			GetDlgItemText( hWndDlg, IDC_ENCOUNTER_ETHETA, strValue.GetBuffer(), strValue.GetMaximumLength() );
			pdata->numericEccenricity = _tstof(strValue);

			GetDlgItemText( hWndDlg, IDC_ENCOUNTER_ERP, strValue.GetBuffer(), strValue.GetMaximumLength() );
			pdata->pericentricSeparation = _tstof(strValue);

			GetDlgItemText( hWndDlg, IDC_ENCOUNTER_ETP, strValue.GetBuffer(), strValue.GetMaximumLength() );
			pdata->timeToPericenter = _tstof(strValue);

			//  jump to correct property sheet page
			PropSheet_SetCurSel(GetParent(hWndDlg), NULL, 4);
			}
	        return FALSE;
            
		case PSN_RESET:
			//  Cancel button is clicked

            return 0;


        default:
        
		    return 0;
		}	//  end switch (lpnm->code)
        }
        
    default:
        
	    return 0;
    }
}

BOOL CALLBACK DirectEncounterParametersPageProc ( HWND hWndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	UNREFERENCED_PARAMETER(wParam);

    //  retrieve the shared user data from GWL_USERDATA, before WM_INITDIALOG there's nothing
	tagSharedEncounterParameters * pdata = (tagSharedEncounterParameters*) GetWindowLong(hWndDlg, GWL_USERDATA);

    switch (uMsg)
    {
    case WM_INITDIALOG:
        { 
			//  save shared data to window GWL_USERDATA field
			pdata = (tagSharedEncounterParameters*) ((LPPROPSHEETPAGE) lParam) -> lParam;
    		SetWindowLong(hWndDlg, GWL_USERDATA, (DWORD) pdata);

		    return 0;
        }

    case WM_NOTIFY:
        {
        LPNMHDR lpnm = (LPNMHDR) lParam;

    
		switch (lpnm->code)
        {
			
		case PSN_SETACTIVE:
			//  sent each time a page is about to be made visible 
			//  initialize data & buttons

			//  enable Back as well as Next button
			PropSheet_SetWizButtons(GetParent(hWndDlg), PSWIZB_BACK | PSWIZB_NEXT);
            return 0;

		case PSN_WIZBACK:
            //  Next button is clicked
			//  check entered data

			//  jump to correct property sheet page
			PropSheet_SetCurSel(GetParent(hWndDlg), NULL, 3);
	        return FALSE;

		case PSN_WIZNEXT:
			{
            //  Next button is clicked
			//  check entered data

			WinString strValue(256);

			//  update pdata
			//

			GetDlgItemText( hWndDlg, IDC_VELOCITY_OF_COMPANION, strValue.GetBuffer(), strValue.GetMaximumLength() );
			pdata->velocityOfCompanion = _tstof(strValue);

			GetDlgItemText( hWndDlg, IDC_INITIAL_DISTANCE_TO_CENTER, strValue.GetBuffer(), strValue.GetMaximumLength() );
			pdata->initialDistanceToCenter = _tstof(strValue);

			GetDlgItemText( hWndDlg, IDC_DISTANCE_TO_CENTER_DURING_PASSAGE, strValue.GetBuffer(), strValue.GetMaximumLength() );
			pdata->distanceToCenterDuringPassage = _tstof(strValue);

			}
	        return FALSE;
            
		case PSN_RESET:
			//  Cancel button is clicked

            return 0;


        default:
        
		    return 0;
		}	//  end switch (lpnm->code)
        }
        
    default:
        
	    return 0;
    }
}

BOOL CALLBACK EncounterCentralGalaxyPageProc ( HWND hWndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	UNREFERENCED_PARAMETER(wParam);

    //  retrieve the shared user data from GWL_USERDATA, before WM_INITDIALOG there's nothing
	tagSharedEncounterParameters * pdata = (tagSharedEncounterParameters*) GetWindowLong(hWndDlg, GWL_USERDATA);

    switch (uMsg)
    {
    case WM_INITDIALOG:
        { 
			//  save shared data to window GWL_USERDATA field
			pdata = (tagSharedEncounterParameters*) ((LPPROPSHEETPAGE) lParam) -> lParam;
    		SetWindowLong(hWndDlg, GWL_USERDATA, (DWORD) pdata);

		    return 0;
        }

    case WM_NOTIFY:
        {
        LPNMHDR lpnm = (LPNMHDR) lParam;

    
		switch (lpnm->code)
        {
			
		case PSN_SETACTIVE:
			//  sent each time a page is about to be made visible 
			//  initialize data & buttons

			//  enable Back as well as Next button
			PropSheet_SetWizButtons(GetParent(hWndDlg), PSWIZB_BACK | PSWIZB_NEXT);
            return 0;

		case PSN_WIZBACK:
            //  Next button is clicked
			//  check entered data

	        return FALSE;

		case PSN_WIZNEXT:
			{
            //  Next button is clicked
			//  check entered data

			WinString strValue(256);

			//  update pdata
			//

			GetDlgItemText( hWndDlg, IDC_ENCOUNTER_EINCLINATION, strValue.GetBuffer(), strValue.GetMaximumLength() );
			pdata->inclinationOfCentral = _tstof(strValue);

			//  softening length epsilon
			GetDlgItemText( hWndDlg, IDC_ENCOUNTER_EARGOFPERICENTRE, strValue.GetBuffer(), strValue.GetMaximumLength() );
			pdata->argumentOfPericenterOfCentral = _tstof(strValue);

			//  jump to correct property sheet page
			}
	        return FALSE;
            
		case PSN_RESET:
			//  Cancel button is clicked

            return 0;


        default:
        
		    return 0;
		}	//  end switch (lpnm->code)
        }
        
    default:
        
	    return 0;
    }
}

BOOL CALLBACK EncounterCompanyGalaxyPageProc ( HWND hWndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	UNREFERENCED_PARAMETER(wParam);

    //  retrieve the shared user data from GWL_USERDATA, before WM_INITDIALOG there's nothing
	tagSharedEncounterParameters * pdata = (tagSharedEncounterParameters*) GetWindowLong(hWndDlg, GWL_USERDATA);

    switch (uMsg)
    {
    case WM_INITDIALOG:
        { 
			//  save shared data to window GWL_USERDATA field
			pdata = (tagSharedEncounterParameters*) ((LPPROPSHEETPAGE) lParam) -> lParam;
    		SetWindowLong(hWndDlg, GWL_USERDATA, (DWORD) pdata);

		    return 0;
        }

    case WM_NOTIFY:
        {
        LPNMHDR lpnm = (LPNMHDR) lParam;

    
		switch (lpnm->code)
        {
			
		case PSN_SETACTIVE:
			//  sent each time a page is about to be made visible 
			//  initialize data & buttons

			//  enable Back as well as Next button
			PropSheet_SetWizButtons(GetParent(hWndDlg), PSWIZB_BACK | PSWIZB_NEXT);
            return 0;

		case PSN_WIZBACK:
            //  Next button is clicked
			//  check entered data

	        return FALSE;

		case PSN_WIZNEXT:
			{
            //  Next button is clicked
			//  check entered data

			WinString strValue(256);

			//  update pdata
			//

			GetDlgItemText( hWndDlg, IDC_ENCOUNTER_EINCLINATION, strValue.GetBuffer(), strValue.GetMaximumLength() );
			pdata->inclinationOfCompanion = _tstof(strValue);

			//  softening length epsilon
			GetDlgItemText( hWndDlg, IDC_ENCOUNTER_EARGOFPERICENTRE, strValue.GetBuffer(), strValue.GetMaximumLength() );
			pdata->argumentOfPericenterOfCompanion = _tstof(strValue);

			//  jump to correct property sheet page
			}
	        return FALSE;
            
		case PSN_RESET:
			//  Cancel button is clicked

            return 0;


        default:
        
		    return 0;
		}	//  end switch (lpnm->code)
        }
        
    default:
        
	    return 0;
    }
}

BOOL CALLBACK EncounterComputationParametersProc ( HWND hWndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	UNREFERENCED_PARAMETER(wParam);

    //  retrieve the shared user data from GWL_USERDATA, before WM_INITDIALOG there's nothing
	tagSharedEncounterParameters * pdata = (tagSharedEncounterParameters*) GetWindowLong(hWndDlg, GWL_USERDATA);

    switch (uMsg)
    {
    case WM_INITDIALOG:
        { 
			//  save shared data to window GWL_USERDATA field
			pdata = (tagSharedEncounterParameters*) ((LPPROPSHEETPAGE) lParam) -> lParam;
    		SetWindowLong(hWndDlg, GWL_USERDATA, (DWORD) pdata);

		    return 0;
        }

    case WM_NOTIFY:
        {
        LPNMHDR lpnm = (LPNMHDR) lParam;

    
		switch (lpnm->code)
        {
			
		case PSN_SETACTIVE:
			//  sent each time a page is about to be made visible 
			//  initialize data & buttons

			//  enable Back as well as Next button
			PropSheet_SetWizButtons(GetParent(hWndDlg), PSWIZB_BACK | PSWIZB_NEXT);
            return 0;

		case PSN_WIZBACK:
            //  Next button is clicked
			//  check entered data

	        return FALSE;

		case PSN_WIZNEXT:
			{
            //  Next button is clicked
			//  check entered data

			WinString strValue(256);

			//  update pdata
			//

			//  opening angle theta
			GetDlgItemText( hWndDlg, IDC_NEW_GALAXY_ETHETA, strValue.GetBuffer(), strValue.GetMaximumLength() );
			theta = _tstof(strValue);

			//  softening length epsilon
			GetDlgItemText( hWndDlg, IDC_NEW_GALAXY_EEPSILON, strValue.GetBuffer(), strValue.GetMaximumLength() );
			epsilon = _tstof(strValue);

			//  time step, diferential time
			GetDlgItemText( hWndDlg, IDC_NEW_GALAXY_EDT, strValue.GetBuffer(), strValue.GetMaximumLength() );
			dt = _tstof(strValue);

			//  jump to correct property sheet page
			}
	        return FALSE;
            
		case PSN_RESET:
			//  Cancel button is clicked

            return 0;


        default:
        
		    return 0;
		}	//  end switch (lpnm->code)
        }
        
    default:
        
	    return 0;
    }
}

BOOL CALLBACK EncounterFileNamesProc ( HWND hWndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
    //  retrieve the shared user data from GWL_USERDATA, before WM_INITDIALOG there's nothing
	tagSharedEncounterParameters * pdata = (tagSharedEncounterParameters*) GetWindowLong(hWndDlg, GWL_USERDATA);

    switch (uMsg)
    {
    case WM_INITDIALOG:
        { 
			//  save shared data to window GWL_USERDATA field
			pdata = (tagSharedEncounterParameters*) ((LPPROPSHEETPAGE) lParam) -> lParam;
    		SetWindowLong(hWndDlg, GWL_USERDATA, (DWORD) pdata);
			return 0;
        }

    case WM_NOTIFY:
        {
        LPNMHDR lpnm = (LPNMHDR) lParam;

    
		switch (lpnm->code)
        {
			
		case PSN_SETACTIVE:
			//  sent each time a page is about to be made visible 
			//  initialize data & buttons

			//  enable Back as well as Next button
			PropSheet_SetWizButtons(GetParent(hWndDlg), PSWIZB_BACK | PSWIZB_NEXT);
            return 0;

		case PSN_WIZNEXT:
            //  Next button is clicked
			//  check entered data
			{
			WinString strValue(256);

            GetDlgItemText( hWndDlg, IDC_NEW_GALAXY_NAME, pdata->szName, MAX_PATH );
            GetDlgItemText( hWndDlg, IDC_NEW_GALAXY_DIRECTORY, pdata->szFolder, MAX_PATH );
            GetDlgItemText( hWndDlg, IDC_NEW_GALAXY_DESCRIPTION, pdata->szDescription, lstrlen(pdata->szDescription) );

/*
			if ( lstrlen(pdata->dataFileName) == 0 || lstrlen(pdata->videoFileName) == 0 )
			{
				MessageBox( hWndDlg, Language.GetString(IDS_NEW_GALAXY_WIZARD_FILENAMESREQUIRED), NULL, MB_OK | MB_ICONINFORMATION );
				SetWindowLong(hWndDlg, DWL_MSGRESULT, -1);
				return TRUE;
			}
*/
			}
	        return FALSE;

		case PSN_WIZBACK:
            //  Next button is clicked
			//  check entered data
            
			return FALSE;
            
		case PSN_RESET:
			//  Cancel button is clicked

            return 0;


        default:
        
		    return 0;
		}	//  end switch (lpnm->code)
        }
        

	case WM_COMMAND:

		switch ( LOWORD( wParam) )
		{
		case IDC_NEW_GALAXY_BUTTON_BROWSE_DIRECTORY:

 		    switch ( HIWORD (wParam) )
			{
			case BN_CLICKED:
				{
				BROWSEINFO bi; 
				bi.hwndOwner = hWndDlg;
				bi.pidlRoot = NULL;
				bi.pszDisplayName = pdata->szFolder;
				bi.lpszTitle = Language.GetString (IDS_NEW_GALAXY_WIZARD_CHOOSE_DIRECTORY);
				bi.ulFlags = BIF_RETURNONLYFSDIRS;
				bi.lpfn = NULL;
				bi.lParam = 0;

				LPITEMIDLIST pidlSelected;
				pidlSelected = SHBrowseForFolder(&bi);

				if (SHGetPathFromIDList(pidlSelected, pdata->szFolder))
				{
				   SetDlgItemText(hWndDlg, IDC_NEW_GALAXY_DIRECTORY, pdata->szFolder);
				}

				}
				
				return 0;
			
			}	 //  end switch ( HIWORD (wParam) )


			return 0;
		
		}	//  end switch ( LOWORD( wParam) )

			
		return 0;	//  return WM_COMMAND

    default:
        
	    return 0;
    }
}

BOOL CALLBACK EncounterFinalPageProc ( HWND hWndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	UNREFERENCED_PARAMETER(wParam);

    //  retrieve the shared user data from GWL_USERDATA, before WM_INITDIALOG there's nothing
	tagSharedEncounterParameters * pdata = (tagSharedEncounterParameters*) GetWindowLong(hWndDlg, GWL_USERDATA);

    switch (uMsg)
    {
    case WM_INITDIALOG:
        { 
			//  save shared data to window GWL_USERDATA field
			pdata = (tagSharedEncounterParameters*) ((LPPROPSHEETPAGE) lParam) -> lParam;
    		SetWindowLong(hWndDlg, GWL_USERDATA, (DWORD) pdata);

		    return 0;
        }

    case WM_NOTIFY:
        {
        LPNMHDR lpnm = (LPNMHDR) lParam;

    
		switch (lpnm->code)
        {
			
		case PSN_SETACTIVE:
			//  sent each time a page is about to be made visible 
			//  initialize data & buttons

			//  enable Back and replace Next with Finish button
			PropSheet_SetWizButtons(GetParent(hWndDlg), PSWIZB_BACK | PSWIZB_FINISH);
            return 0;

		case PSN_WIZFINISH:
            //  Finish button is clicked
			//  check entered data

			CreateEncounterSimulation( *pdata );
			
	        return FALSE;

		case PSN_WIZBACK:
            //  Next button is clicked
			//  check entered data
            
	        return FALSE;
            
		case PSN_RESET:
			//  Cancel button is clicked

            return 0;


        default:
        
		    return 0;
		}	//  end switch (lpnm->code)
        }
        
    default:
        
	    return 0;
    }
}


void ShowEncounterSimulationDlg( HWND hWnd )
{
	//  create property sheet pages
	//

	PROPSHEETPAGE psp;
    HPROPSHEETPAGE ahpsp[10];

	WinString strTitle(256);
	WinString strHeaderTitle(256);

	strTitle.SetString( Language.GetString (IDS_ENCOUNTER_WIZARD_MAIN_TITLE) );

	//  welcoming property sheet page

	psp.dwSize		= sizeof(psp);
	psp.hInstance	= Language.GetResourceInstance();
	psp.lParam		= (LPARAM) &SharedEncounterParameters;  //  the shared data structure
	psp.dwFlags		= PSP_DEFAULT | PSP_HIDEHEADER  | PSP_USETITLE;
    psp.pszTemplate	= MAKEINTRESOURCE(IDD_ENCOUNTER_PAGE_WELCOME);
	psp.pfnDlgProc	= EncounterWelcomePageProc;
	psp.pszTitle	= strTitle;

    ahpsp[0] = CreatePropertySheetPage(&psp);

    //  first interior page, choosing kind of encounter (keplerian, direct)

	strHeaderTitle.SetString( Language.GetString (IDS_ENCOUNTER_WIZARD_HEADER_TITLE1) );

    psp.dwFlags			= PSP_DEFAULT | PSP_USEHEADERTITLE | PSP_USETITLE;
    psp.pszTemplate		= MAKEINTRESOURCE(IDD_ENCOUNTER_CHOOSE_KIND);
    psp.pfnDlgProc		= EncounterChooseKindOfEncounterPageProc;
    psp.pszHeaderTitle	= strHeaderTitle;

    ahpsp[1] = CreatePropertySheetPage(&psp);

    //  second interior page, initial files

	strHeaderTitle.SetString( Language.GetString (IDS_ENCOUNTER_WIZARD_HEADER_TITLE2) );

    psp.dwFlags			= PSP_DEFAULT | PSP_USEHEADERTITLE | PSP_USETITLE;
    psp.pszTemplate		= MAKEINTRESOURCE(IDD_ENCOUNTER_PAGE_INITIAL_FILES);
    psp.pfnDlgProc		= EncounterInitialFilesPageProc;
    psp.pszHeaderTitle	= strHeaderTitle;

    ahpsp[2] = CreatePropertySheetPage(&psp);

    //  third interior page, keplerian encounter parameters

	strHeaderTitle.SetString( Language.GetString (IDS_ENCOUNTER_WIZARD_HEADER_TITLE31) );

    psp.dwFlags			= PSP_DEFAULT | PSP_USEHEADERTITLE | PSP_USETITLE;
    psp.pszTemplate		= MAKEINTRESOURCE(IDD_ENCOUNTER_PAGE_ENCOUNTER_PARAMETERS);
    psp.pfnDlgProc		= KeplerianEncounterParametersPageProc;
    psp.pszHeaderTitle	= strHeaderTitle;

    ahpsp[3] = CreatePropertySheetPage(&psp);

    //  third interior page, direct encounter parameters

	strHeaderTitle.SetString( Language.GetString (IDS_ENCOUNTER_WIZARD_HEADER_TITLE32) );

    psp.dwFlags			= PSP_DEFAULT | PSP_USEHEADERTITLE | PSP_USETITLE;
    psp.pszTemplate		= MAKEINTRESOURCE(IDD_ENCOUNTER_DIRECT_PARAMETER);
    psp.pfnDlgProc		= DirectEncounterParametersPageProc;
    psp.pszHeaderTitle	= strHeaderTitle;

    ahpsp[4] = CreatePropertySheetPage(&psp);

    //  fourth interior page, central galaxy parameters

	strHeaderTitle.SetString( Language.GetString (IDS_ENCOUNTER_WIZARD_HEADER_TITLE4) );

    psp.dwFlags			= PSP_DEFAULT | PSP_USEHEADERTITLE | PSP_USETITLE;
    psp.pszTemplate		= MAKEINTRESOURCE(IDD_ENCOUNTER_PAGE_CENTRAL_GALAXY_PARAMETERS);
    psp.pfnDlgProc		= EncounterCentralGalaxyPageProc;
    psp.pszHeaderTitle	= strHeaderTitle;

    ahpsp[5] = CreatePropertySheetPage(&psp);

    //  fifth interior page, companion galaxy parameters

	strHeaderTitle.SetString( Language.GetString (IDS_ENCOUNTER_WIZARD_HEADER_TITLE5) );

    psp.dwFlags			= PSP_DEFAULT | PSP_USEHEADERTITLE | PSP_USETITLE;
    psp.pszTemplate		= MAKEINTRESOURCE(IDD_ENCOUNTER_PAGE_COMPANION_GALAXY_PARAMETERS);
    psp.pfnDlgProc		= EncounterCompanyGalaxyPageProc;
    psp.pszHeaderTitle	= strHeaderTitle;

    ahpsp[6] = CreatePropertySheetPage(&psp);

    //  sixth interior page, set parameters of computation

	strHeaderTitle.SetString( Language.GetString (IDS_NEW_GALAXY_WIZARD_HEADER_TITLE6) );

    psp.dwFlags			= PSP_DEFAULT | PSP_USEHEADERTITLE | PSP_USETITLE;
    psp.pszTemplate		= MAKEINTRESOURCE(IDD_PAGE_COMPUTATION);
    psp.pfnDlgProc		= EncounterComputationParametersProc;
    psp.pszHeaderTitle	= strHeaderTitle;

    ahpsp[7] = CreatePropertySheetPage(&psp);

    //  seventh interior page, set file names, common page

	strHeaderTitle.SetString( Language.GetString (IDS_NEW_GALAXY_WIZARD_HEADER_TITLE7) );

    psp.dwFlags			= PSP_DEFAULT | PSP_USEHEADERTITLE | PSP_USETITLE;
    psp.pszTemplate		= MAKEINTRESOURCE(IDD_PAGE_FILENAMES);
    psp.pfnDlgProc		= EncounterFileNamesProc;
    psp.pszHeaderTitle	= strHeaderTitle;

    ahpsp[8] = CreatePropertySheetPage(&psp);

    //  final property sheet page

    psp.dwFlags			= PSP_DEFAULT | PSP_HIDEHEADER |  PSP_USETITLE;
    psp.pszTemplate		= MAKEINTRESOURCE(IDD_ENCOUNTER_PAGE_FINAL);
    psp.pfnDlgProc		= EncounterFinalPageProc;

    ahpsp[9] = CreatePropertySheetPage(&psp);

	//  create property sheet dialog
	//

    PROPSHEETHEADER psh;

	psh.dwSize			= sizeof(psh);
    psh.hInstance		= g_hInstance;
    psh.hwndParent		= hWnd;
    psh.phpage			= ahpsp;
    psh.dwFlags			= PSH_WIZARD97 | PSH_WATERMARK | PSH_HEADER;
    psh.pszbmWatermark	= TEXT("WatermarkBitmapEncounter");
    psh.pszbmHeader		= TEXT("HeaderBitmap");
    psh.nStartPage		= 0;
    psh.nPages			= 10;

    PropertySheet(&psh);
}

