/************************************************************
*
*  Copyright (c) 2002-2004 Jakub Schwarzmeier
*
************************************************************/

#include "general.h"
#include "../../distribution/distributions.h"

#include "data/xml.h"

int nTotalBodies;

double oneDiscBodyMass;
double oneBulgeBodyMass;
double oneHaloBodyMass;
double oneDarkHaloBodyMass;

struct tagSharedSimulationParameters
{
	int nTotalBodies;			//  total number of bodies 
	bool disc, bulge, halo, darkhalo;	//  included components

	//  count of bodies and total mass of components
	int discBodies;
	int bulgeBodies;
	int haloBodies;
	int darkhaloBodies;

	double discTotalMass;
	double bulgeTotalMass;
	double haloTotalMass;
	double darkhaloTotalMass;

	//  disc
	int iDiscPositionModel;
	double rd;
	double ad;
	double z0;

	int iDiscVelocityModel;
	bool bWarmDisc;
	double QToomre;
	double zDisp;

	//  bulge
	int iBulgePositionModel;
	double rb;
	double ab;

	int iBulgeVelocityModel;

	//  halo
	int iHaloPositionModel;
	double rh;
	double ah;

	int iHaloVelocityModel;

	//  dark halo
	int iDarkHaloPositionModel;
	double rdh;
	double adh;
	double gamma;

	int iDarkHaloVelocityModel;

	//  storage related
	TCHAR szFolder[MAX_PATH], szName[MAX_PATH], szDescription[2048];

} SharedSimulationParameters;


void CreateNewGalaxySimulation( tagSharedSimulationParameters & data )
{
	unsortedBodiesList = new JakubSchwarzmeier::Data::RList< JakubSchwarzmeier::BarnesHut::Leave >;

	//  starting sequence
	if ( data.disc )
	{
		DiskModel::Initialize(data.discTotalMass, data.rd, data.ad, data.z0, data.QToomre, data.zDisp, DiskModel::Model1);
		DiskModel::Create( *unsortedBodiesList, data.discBodies );
		oneDiscBodyMass = data.discTotalMass / data.discBodies;
	}

	if ( data.bulge )
	{
		BulgeModel::Initialize(data.bulgeTotalMass, data.rb, data.ab);
		BulgeModel::Create( *unsortedBodiesList, data.bulgeBodies );
		oneBulgeBodyMass = data.bulgeTotalMass / data.bulgeBodies;
	}

	if ( data.halo )
	{
		HaloModel::Initialize(data.haloTotalMass, data.rh, data.ah);
		HaloModel::Create( *unsortedBodiesList, data.haloBodies );
		oneHaloBodyMass = data.haloTotalMass / data.haloBodies;
	}
	
	if ( data.darkhalo )
	{
		DarkHaloModel::Initialize(data.darkhaloTotalMass, data.rdh, data.adh, data.gamma);
		DarkHaloModel::Create( *unsortedBodiesList, data.darkhaloBodies );
		oneDarkHaloBodyMass = data.darkhaloTotalMass / data.darkhaloBodies;
	}


	//  set initial load to one for evry body
	for(int i = 0; i < unsortedBodiesList->getCount(); i++)  unsortedBodiesList->getAt(i).load = 1;

 	nTotalBodies = unsortedBodiesList->getCount();
	openGL->InitializeMasses();

	//  initiate files writing
	
	TCHAR szDataFileName[MAX_PATH], szVideoFileName[MAX_PATH];
	lstrcpy( szDataFileName, data.szFolder );
	lstrcat( szDataFileName, TEXT("\\") );
	lstrcat( szDataFileName, data.szName );
	lstrcpy( szVideoFileName, szDataFileName );

	lstrcat( szDataFileName, TEXT(".nbody") );
	lstrcat( szVideoFileName, TEXT(".avi") );

	//  create data and video simulation files
	auxInfo = new AuxInfo( szDataFileName, szVideoFileName );
	
	//  first writing of data and video data
	auxInfo->WriteData();

	//  write xml file
	//

	//  get logged user's name
	TCHAR szUserName[512];
	DWORD dwUserSize = sizeof(szUserName)/ sizeof(*szUserName);
	GetUserName( szUserName,  &dwUserSize );


//	following comment for NT4

	using namespace JakubSchwarzmeier::Data;

	XmlDocument xmlDoc;

	XmlProcessingInstruction pi;
	pi = xmlDoc.CreateProcessingInstruction(TEXT("xml"), TEXT("version='1.0'"));
	xmlDoc.AppendChild( pi );
	
	XmlElement root = xmlDoc.CreateElement(TEXT("simulation"));
	xmlDoc.AppendChild( root );
	root.SetAttribute(TEXT("type"), TEXT("single"));

	XmlElement element;
	XmlText text;

	element = xmlDoc.CreateElement(TEXT("count"));
	text = xmlDoc.CreateTextNode( data.nTotalBodies );
	element.AppendChild( text );
	root.AppendChild( element );
	
	element = xmlDoc.CreateElement(TEXT("format"));
	text = xmlDoc.CreateTextNode(TEXT("double"));
	element.AppendChild( text );
	root.AppendChild( element );

	element = xmlDoc.CreateElement(TEXT("description"));
	text = xmlDoc.CreateTextNode( data.szDescription );
	element.AppendChild( text );
	root.AppendChild( element );

	XmlElement parameters = xmlDoc.CreateElement(TEXT("parameters"));
	root.AppendChild( parameters );

	XmlElement components = xmlDoc.CreateElement(TEXT("components"));
	parameters.AppendChild( components );

	if ( data.disc )
	{
		//  disc
		XmlElement disc = xmlDoc.CreateElement( TEXT("disc") );
		components.AppendChild( disc );

		element = xmlDoc.CreateElement( TEXT("count") );
		text = xmlDoc.CreateTextNode( data.discBodies );
		element.AppendChild( text );
		disc.AppendChild( element );
		
		element = xmlDoc.CreateElement( TEXT("totalmass") );
		text = xmlDoc.CreateTextNode( data.discTotalMass );
		element.AppendChild( text );
		disc.AppendChild( element );
		
		element = xmlDoc.CreateElement( TEXT("maximumradius") );
		text = xmlDoc.CreateTextNode( data.rd );
		element.AppendChild( text );
		disc.AppendChild( element );
		
		element = xmlDoc.CreateElement( TEXT("scalelength") );
		text = xmlDoc.CreateTextNode( data.ad );
		element.AppendChild( text );
		disc.AppendChild( element );
		
		element = xmlDoc.CreateElement( TEXT("z0") );
		text = xmlDoc.CreateTextNode( data.z0 );
		element.AppendChild( text );
		disc.AppendChild( element );
		
		XmlElement velocity = xmlDoc.CreateElement( TEXT("velocity") );
		disc.AppendChild( velocity );
		
		if (data.bWarmDisc)
		{
			velocity.SetAttribute(TEXT("warm"), TEXT("true"));
			
			element = xmlDoc.CreateElement( TEXT("QToomre") );
			text = xmlDoc.CreateTextNode( data.QToomre );
			element.AppendChild( text );
			velocity.AppendChild( element );
			
			element = xmlDoc.CreateElement( TEXT("zdispersion") );
			text = xmlDoc.CreateTextNode( data.zDisp );
			element.AppendChild( text );
			velocity.AppendChild( element );
		}
		else
		{
			velocity.SetAttribute(TEXT("warm"), TEXT("false"));
		}
		
	}

	if ( data.bulge )
	{
		//  bulge
		XmlElement bulge = xmlDoc.CreateElement( TEXT("bulge") );
		components.AppendChild( bulge );

		element = xmlDoc.CreateElement( TEXT("count") );
		text = xmlDoc.CreateTextNode( data.bulgeBodies );
		element.AppendChild( text );
		bulge.AppendChild( element );
		
		element = xmlDoc.CreateElement( TEXT("totalmass") );
		text = xmlDoc.CreateTextNode( data.bulgeTotalMass );
		element.AppendChild( text );
		bulge.AppendChild( element );
		
		element = xmlDoc.CreateElement( TEXT("maximumradius") );
		text = xmlDoc.CreateTextNode( data.rb );
		element.AppendChild( text );
		bulge.AppendChild( element );
		
		element = xmlDoc.CreateElement( TEXT("scalelength") );
		text = xmlDoc.CreateTextNode( data.ab );
		element.AppendChild( text );
		bulge.AppendChild( element );
	}

	if ( data.halo )
	{
		//  halo
		XmlElement halo = xmlDoc.CreateElement( TEXT("halo") );
		components.AppendChild( halo );

		element = xmlDoc.CreateElement( TEXT("count") );
		text = xmlDoc.CreateTextNode( data.haloBodies );
		element.AppendChild( text );
		halo.AppendChild( element );
		
		element = xmlDoc.CreateElement( TEXT("totalmass") );
		text = xmlDoc.CreateTextNode( data.haloTotalMass );
		element.AppendChild( text );
		halo.AppendChild( element );
		
		element = xmlDoc.CreateElement( TEXT("maximumradius") );
		text = xmlDoc.CreateTextNode( data.rh );
		element.AppendChild( text );
		halo.AppendChild( element );
		
		element = xmlDoc.CreateElement( TEXT("scalelength") );
		text = xmlDoc.CreateTextNode( data.ah );
		element.AppendChild( text );
		halo.AppendChild( element );
	}

	if ( data.darkhalo )
	{
		//  darkhalo
		XmlElement darkhalo = xmlDoc.CreateElement( TEXT("darkhalo") );
		components.AppendChild( darkhalo );

		element = xmlDoc.CreateElement( TEXT("count") );
		text = xmlDoc.CreateTextNode( data.darkhaloBodies );
		element.AppendChild( text );
		darkhalo.AppendChild( element );
		
		element = xmlDoc.CreateElement( TEXT("totalmass") );
		text = xmlDoc.CreateTextNode( data.darkhaloTotalMass );
		element.AppendChild( text );
		darkhalo.AppendChild( element );
		
		element = xmlDoc.CreateElement( TEXT("maximumradius") );
		text = xmlDoc.CreateTextNode( data.rdh );
		element.AppendChild( text );
		darkhalo.AppendChild( element );
		
		element = xmlDoc.CreateElement( TEXT("scalelength") );
		text = xmlDoc.CreateTextNode( data.adh );
		element.AppendChild( text );
		darkhalo.AppendChild( element );
		
		element = xmlDoc.CreateElement( TEXT("coreradius") );
		text = xmlDoc.CreateTextNode( data.gamma );
		element.AppendChild( text );
		darkhalo.AppendChild( element );
	}

	element = xmlDoc.CreateElement(TEXT("author"));
	text = xmlDoc.CreateTextNode( szUserName );
	element.AppendChild( text );
	root.AppendChild( element );

	element = xmlDoc.CreateElement(TEXT("theta"));
	text = xmlDoc.CreateTextNode( theta );
	element.AppendChild( text );
	root.AppendChild( element );

	element = xmlDoc.CreateElement(TEXT("epsilon"));
	text = xmlDoc.CreateTextNode( epsilon );
	element.AppendChild( text );
	root.AppendChild( element );

	element = xmlDoc.CreateElement(TEXT("dt"));
	text = xmlDoc.CreateTextNode( dt );
	element.AppendChild( text );
	root.AppendChild( element );

	PathRenameExtension( szDataFileName, TEXT(".xml") );
	xmlDoc.DoHumanMarkup( root );
	xmlDoc.Save( szDataFileName );
		
	//
	//  end write xml file

//  end possible commention for NT4

}

BOOL CALLBACK NewGalaxyWelcomePageProc ( HWND hWndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
    UNREFERENCED_PARAMETER(wParam);

    switch (uMsg)
    {
    case WM_INITDIALOG:
        { 
			tagSharedSimulationParameters * pdata;
			//  save shared data to window GWL_USERDATA field
			pdata = (tagSharedSimulationParameters*) ((LPPROPSHEETPAGE) lParam) -> lParam;
    		SetWindowLong(hWndDlg, GWL_USERDATA, (DWORD) pdata);

			return 0;
        }

    case WM_NOTIFY:
        {
        LPNMHDR lpnm = (LPNMHDR) lParam;

    
		switch (lpnm->code)
        {
			
		case PSN_SETACTIVE:
			//  sent each time a page is about to be made visible 
			//  initialize data & buttons

            //  enable the Next button
			PropSheet_SetWizButtons(GetParent(hWndDlg), PSWIZB_NEXT);
		    return 0;

		case PSN_WIZNEXT:
            //  Next button is clicked
			//  check entered data
            
		    return 0;
            
		case PSN_RESET:
			//  Cancel button is clicked


		    return 0;


        default:
        
		    return 0;
		}	//  end switch (lpnm->code)
        }
        
    default:
        
	    return 0;
    }

}

BOOL CALLBACK NewGalaxyComponentsProc ( HWND hWndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
    //  retrieve the shared user data from GWL_USERDATA, before WM_INITDIALOG there's nothing
	tagSharedSimulationParameters * pdata = (tagSharedSimulationParameters*) GetWindowLong(hWndDlg, GWL_USERDATA);

    switch (uMsg)
    {
    case WM_INITDIALOG:
        { 
			//  save shared data to window GWL_USERDATA field
			pdata = (tagSharedSimulationParameters*) ((LPPROPSHEETPAGE) lParam) -> lParam;
    		SetWindowLong(hWndDlg, GWL_USERDATA, (DWORD) pdata);

			//  initialy disable all edit controls
			EnableWindow( GetDlgItem(hWndDlg, IDC_NEW_GALAXY_ECDISC ), FALSE );				   
			EnableWindow( GetDlgItem(hWndDlg, IDC_NEW_GALAXY_EMDISC ), FALSE );				   
			EnableWindow( GetDlgItem(hWndDlg, IDC_NEW_GALAXY_ECBULGE ), FALSE );				   
			EnableWindow( GetDlgItem(hWndDlg, IDC_NEW_GALAXY_EMBULGE ), FALSE );				   
			EnableWindow( GetDlgItem(hWndDlg, IDC_NEW_GALAXY_ECHALO ), FALSE );				   
			EnableWindow( GetDlgItem(hWndDlg, IDC_NEW_GALAXY_EMHALO ), FALSE );				   
			EnableWindow( GetDlgItem(hWndDlg, IDC_NEW_GALAXY_ECDARKHALO ), FALSE );				   
			EnableWindow( GetDlgItem(hWndDlg, IDC_NEW_GALAXY_EMDARKHALO ), FALSE );				   

			return 0;
        }

    case WM_NOTIFY:
        {
        LPNMHDR lpnm = (LPNMHDR) lParam;

    
		switch (lpnm->code)
        {
			
		case PSN_SETACTIVE:
			//  sent each time a page is about to be made visible 
			//  initialize data & buttons

			//  enable Back as well as Next button
			PropSheet_SetWizButtons(GetParent(hWndDlg), PSWIZB_BACK | PSWIZB_NEXT);
            return 0;

		case PSN_WIZNEXT:
			{
				//  Next button is clicked
				//  check entered data
            
				HWND hWndChild;
				int iResult;

				//  set shared data according to selected galaxy components
				//  disk
				hWndChild = GetDlgItem(hWndDlg, IDC_NEW_GALAXY_CHBDISC);
				iResult = Button_GetCheck (hWndChild);
				if ( iResult == 0 )  pdata->disc = false;  else  pdata->disc = true;

				//  bulge
				hWndChild = GetDlgItem(hWndDlg, IDC_NEW_GALAXY_CHBBULGE);
				iResult = Button_GetCheck (hWndChild);
				if ( iResult == 0 )  pdata->bulge = false;  else  pdata->bulge = true;

				//  halo
				hWndChild = GetDlgItem(hWndDlg, IDC_NEW_GALAXY_CHBHALO);
				iResult = Button_GetCheck (hWndChild);
				if ( iResult == 0 )  pdata->halo = false;  else  pdata->halo = true;

				//  dark halo
				hWndChild = GetDlgItem(hWndDlg, IDC_NEW_GALAXY_CHBDARKHALO);
				iResult = Button_GetCheck (hWndChild);
				if ( iResult == 0 )  pdata->darkhalo = false;  else  pdata->darkhalo = true;

				//  jump to correct property sheet page
				if (pdata->disc)  PropSheet_SetCurSel(GetParent(hWndDlg), NULL, 1);
				else
				if (pdata->bulge)  PropSheet_SetCurSel(GetParent(hWndDlg), NULL, 2);
				else
				if (pdata->halo)  PropSheet_SetCurSel(GetParent(hWndDlg), NULL, 3);
				else
				if (pdata->darkhalo)  PropSheet_SetCurSel(GetParent(hWndDlg), NULL, 4);
				else
				{
					//  if no component is selected
					WinString strMessage(512);
					strMessage.AddToString( Language.GetString (IDS_NEW_GALAXY_WIZARD_SELECT_COMPONENT) );
					MessageBox( hWndDlg, strMessage, NULL, MB_OK | MB_ICONINFORMATION );

					SetWindowLong(hWndDlg, DWL_MSGRESULT, -1);
					return TRUE;
				}

				//  fill count and total mass fields in pdata structure
				//

				WinString strText(128);

				//  disc
				GetDlgItemText(hWndDlg, IDC_NEW_GALAXY_ECDISC, strText.GetBuffer(), strText.GetMaximumLength() );
				pdata->discBodies = StrToInt (strText);
				GetDlgItemText(hWndDlg, IDC_NEW_GALAXY_EMDISC, strText.GetBuffer(), strText.GetMaximumLength() );
				pdata->discTotalMass = _tstof ( strText );

				//  bulge
				GetDlgItemText(hWndDlg, IDC_NEW_GALAXY_ECBULGE, strText.GetBuffer(), strText.GetMaximumLength() );
				pdata->bulgeBodies = StrToInt (strText);
				GetDlgItemText(hWndDlg, IDC_NEW_GALAXY_EMBULGE, strText.GetBuffer(), strText.GetMaximumLength() );
				pdata->bulgeTotalMass = _tstof ( strText );

				//  halo
				GetDlgItemText(hWndDlg, IDC_NEW_GALAXY_ECHALO, strText.GetBuffer(), strText.GetMaximumLength() );
				pdata->haloBodies = StrToInt (strText);
				GetDlgItemText(hWndDlg, IDC_NEW_GALAXY_EMHALO, strText.GetBuffer(), strText.GetMaximumLength() );
				pdata->haloTotalMass = _tstof ( strText );

				//  dark halo
				GetDlgItemText(hWndDlg, IDC_NEW_GALAXY_ECDARKHALO, strText.GetBuffer(), strText.GetMaximumLength() );
				pdata->darkhaloBodies = StrToInt (strText);
				GetDlgItemText(hWndDlg, IDC_NEW_GALAXY_EMDARKHALO, strText.GetBuffer(), strText.GetMaximumLength() );
				pdata->darkhaloTotalMass = _tstof ( strText );
			}

	        return FALSE;
            
		case PSN_RESET:
			//  Cancel button is clicked

            return 0;

        default:
        
		    return 0;
		}	//  end switch (lpnm->code)
        }
        
	case WM_COMMAND:

		switch ( LOWORD( wParam) )
		{
	    //	checkbox handlers
	    //
		case IDC_NEW_GALAXY_CHBDISC:

 		   switch ( HIWORD (wParam) )
		   {
		   case BN_CLICKED:

			   if ( IsDlgButtonChecked( hWndDlg, IDC_NEW_GALAXY_CHBDISC ) )
			   {
				   EnableWindow( GetDlgItem(hWndDlg, IDC_NEW_GALAXY_ECDISC ), TRUE );				   
				   EnableWindow( GetDlgItem(hWndDlg, IDC_NEW_GALAXY_EMDISC ), TRUE );
			   }
			   else
			   {
				   EnableWindow( GetDlgItem(hWndDlg, IDC_NEW_GALAXY_ECDISC ), FALSE );				   
				   EnableWindow( GetDlgItem(hWndDlg, IDC_NEW_GALAXY_EMDISC ), FALSE );

				   //  clear entered values	
				   SetDlgItemText( hWndDlg, IDC_NEW_GALAXY_ECDISC, TEXT("") );
				   SetDlgItemText( hWndDlg, IDC_NEW_GALAXY_EMDISC, TEXT("") );
			   }

			   return 0;
		   }

		case IDC_NEW_GALAXY_CHBBULGE:

 		   switch ( HIWORD (wParam) )
		   {
		   case BN_CLICKED:

			   if ( IsDlgButtonChecked( hWndDlg, IDC_NEW_GALAXY_CHBBULGE ) )
			   {
				   EnableWindow( GetDlgItem(hWndDlg, IDC_NEW_GALAXY_ECBULGE ), TRUE );				   
				   EnableWindow( GetDlgItem(hWndDlg, IDC_NEW_GALAXY_EMBULGE ), TRUE );				   
			   }
			   else
			   {
				   EnableWindow( GetDlgItem(hWndDlg, IDC_NEW_GALAXY_ECBULGE ), FALSE );				   
				   EnableWindow( GetDlgItem(hWndDlg, IDC_NEW_GALAXY_EMBULGE ), FALSE );				   

				   //  clear entered values	
				   SetDlgItemText( hWndDlg, IDC_NEW_GALAXY_ECBULGE, TEXT("") );
				   SetDlgItemText( hWndDlg, IDC_NEW_GALAXY_EMBULGE, TEXT("") );
			   }

			   return 0;

		   }	//  end switch ( HIWORD (wParam) )

		case IDC_NEW_GALAXY_CHBHALO:

 		   switch ( HIWORD (wParam) )
		   {
		   case BN_CLICKED:

			   if ( IsDlgButtonChecked( hWndDlg, IDC_NEW_GALAXY_CHBHALO ) )
			   {
				   EnableWindow( GetDlgItem(hWndDlg, IDC_NEW_GALAXY_ECHALO ), TRUE );				   
				   EnableWindow( GetDlgItem(hWndDlg, IDC_NEW_GALAXY_EMHALO ), TRUE );				   
			   }
			   else
			   {
				   EnableWindow( GetDlgItem(hWndDlg, IDC_NEW_GALAXY_ECHALO ), FALSE );				   
				   EnableWindow( GetDlgItem(hWndDlg, IDC_NEW_GALAXY_EMHALO ), FALSE );				   

				   //  clear entered values	
				   SetDlgItemText( hWndDlg, IDC_NEW_GALAXY_ECHALO, TEXT("") );
				   SetDlgItemText( hWndDlg, IDC_NEW_GALAXY_EMHALO, TEXT("") );
			   }

			   return 0;

		   }	//  end switch ( HIWORD (wParam) )

		case IDC_NEW_GALAXY_CHBDARKHALO:

 		   switch ( HIWORD (wParam) )
		   {
		   case BN_CLICKED:

			   if ( IsDlgButtonChecked( hWndDlg, IDC_NEW_GALAXY_CHBDARKHALO ) )
			   {
				   EnableWindow( GetDlgItem(hWndDlg, IDC_NEW_GALAXY_ECDARKHALO ), TRUE );				   
				   EnableWindow( GetDlgItem(hWndDlg, IDC_NEW_GALAXY_EMDARKHALO ), TRUE );				   
			   }
			   else
			   {
				   EnableWindow( GetDlgItem(hWndDlg, IDC_NEW_GALAXY_ECDARKHALO ), FALSE );				   
				   EnableWindow( GetDlgItem(hWndDlg, IDC_NEW_GALAXY_EMDARKHALO ), FALSE );				   

				   //  clear entered values	
				   SetDlgItemText( hWndDlg, IDC_NEW_GALAXY_ECDARKHALO, TEXT("") );
				   SetDlgItemText( hWndDlg, IDC_NEW_GALAXY_EMDARKHALO, TEXT("") );
			   }

			   return 0;

		   }	//  end switch ( HIWORD (wParam) )

		
	   //
	   //	checkbox handlers

		
		 //  edit handlers
		 //

		  case IDC_NEW_GALAXY_ECDISC:
		  case IDC_NEW_GALAXY_ECBULGE:
		  case IDC_NEW_GALAXY_ECHALO:
		  case IDC_NEW_GALAXY_ECDARKHALO:
			  {
			  WinString strValue(128);

			  //  sum count of bodies of all select components
			  GetDlgItemText( hWndDlg, IDC_NEW_GALAXY_ECDISC, strValue.GetBuffer(), strValue.GetMaximumLength() );
			  pdata->nTotalBodies = StrToInt (strValue);	//  when user enter value > 219..... --> error !
			  GetDlgItemText( hWndDlg, IDC_NEW_GALAXY_ECBULGE, strValue.GetBuffer(), strValue.GetMaximumLength() );
			  pdata->nTotalBodies += StrToInt (strValue);
			  GetDlgItemText( hWndDlg, IDC_NEW_GALAXY_ECHALO, strValue.GetBuffer(), strValue.GetMaximumLength() );
			  pdata->nTotalBodies += StrToInt (strValue);
			  GetDlgItemText( hWndDlg, IDC_NEW_GALAXY_ECDARKHALO, strValue.GetBuffer(), strValue.GetMaximumLength() );
			  pdata->nTotalBodies += StrToInt (strValue);

			  //  show pdata->nTotalBodies
			  _itot( pdata->nTotalBodies, strValue.GetBuffer(), 10 );
			  SetDlgItemText( hWndDlg, IDC_NEW_GALAXY_TCTOTAL, strValue );
			  }
			  return 0;


		  case IDC_NEW_GALAXY_EMDISC:
		  case IDC_NEW_GALAXY_EMBULGE:
		  case IDC_NEW_GALAXY_EMHALO:
		  case IDC_NEW_GALAXY_EMDARKHALO:
			  {
			  WinString strValue(128);

			  //  sum mass of all components
			  GetDlgItemText( hWndDlg, IDC_NEW_GALAXY_EMDISC, strValue.GetBuffer(), strValue.GetMaximumLength() );
			  double mass = _tstof(strValue);	//  sometimes error, when large numbers, why??? -- correct it (sometimes :-))
			  GetDlgItemText( hWndDlg, IDC_NEW_GALAXY_EMBULGE, strValue.GetBuffer(), strValue.GetMaximumLength() );
			  mass += StrToInt (strValue);
			  GetDlgItemText( hWndDlg, IDC_NEW_GALAXY_EMHALO, strValue.GetBuffer(), strValue.GetMaximumLength() );
			  mass += StrToInt (strValue);
			  GetDlgItemText( hWndDlg, IDC_NEW_GALAXY_EMDARKHALO, strValue.GetBuffer(), strValue.GetMaximumLength() );
			  mass += StrToInt (strValue);

			  //  show mass
			  _dtot( mass, strValue.GetBuffer() );
			  SetDlgItemText( hWndDlg, IDC_NEW_GALAXY_TMTOTAL, strValue );
			  }

			  return 0;

		//
		//  edit handlers

			

		}  //  end switch ( LOWORD( wParam) )

		
		return 0;	//  end WM_COMMAND

    default:
        
	    return 0;
    }
}

BOOL CALLBACK NewGalaxyDiskParametersProc ( HWND hWndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
    //  retrieve the shared user data from GWL_USERDATA, before WM_INITDIALOG there's nothing
	tagSharedSimulationParameters * pdata = (tagSharedSimulationParameters*) GetWindowLong(hWndDlg, GWL_USERDATA);

    switch (uMsg)
    {
    case WM_INITDIALOG:
        { 
			//  save shared data to window GWL_USERDATA field
			pdata = (tagSharedSimulationParameters*) ((LPPROPSHEETPAGE) lParam) -> lParam;
    		SetWindowLong(hWndDlg, GWL_USERDATA, (DWORD) pdata);


			HWND hWndChild;

			//  insert model names into combobox
			//

			//  get model combobox
			hWndChild = GetDlgItem(hWndDlg, IDC_NEW_GALAXY_POSITION_DISTRIBUTION);

			//  load model names from resource and insert them into combo
			ComboBox_InsertString( hWndChild, 0, Language.GetString(IDS_NEW_GALAXY_WIZARD_DISK_MODEL2) );
			ComboBox_InsertString( hWndChild, 0, Language.GetString(IDS_NEW_GALAXY_WIZARD_DISK_MODEL1));

			//  selects the first item in combo as actual
			ComboBox_SetCurSel( hWndChild, 0 );
		
			
			//  select one of radio buttons
			//
			hWndChild = GetDlgItem(hWndDlg, IDC_NEW_GALAXY_RBWARM);
			Button_SetCheck( hWndChild, BST_CHECKED );

		    return 0;
        }

    case WM_NOTIFY:
        {
        LPNMHDR lpnm = (LPNMHDR) lParam;

    
		switch (lpnm->code)
        {
			
		case PSN_SETACTIVE:
			//  sent each time a page is about to be made visible 
			//  initialize data & buttons

			//  enable Back as well as Next button
			PropSheet_SetWizButtons(GetParent(hWndDlg), PSWIZB_BACK | PSWIZB_NEXT);
            return 0;


		case PSN_WIZBACK:
            //  Next button is clicked
			//  check entered data

	        return FALSE;

		case PSN_WIZNEXT:
			{
            //  Next button is clicked
			//  check entered data

			HWND hWndChild;
			WinString strValue(256);

			//  update pdata
			//

			//  position distribution model
			hWndChild = GetDlgItem(hWndDlg, IDC_NEW_GALAXY_POSITION_DISTRIBUTION);
			pdata->iDiscPositionModel = ComboBox_GetCurSel (hWndChild);

			//  maximum radius rd
			GetDlgItemText( hWndDlg, IDC_NEW_GALAXY_ERD, strValue.GetBuffer(), strValue.GetMaximumLength() );
			pdata->rd = _tstof(strValue);

			//  scale length ad
			GetDlgItemText( hWndDlg, IDC_NEW_GALAXY_EAD, strValue.GetBuffer(), strValue.GetMaximumLength() );
			pdata->ad = _tstof(strValue);

			//  z0
			GetDlgItemText( hWndDlg, IDC_NEW_GALAXY_EZ0, strValue.GetBuffer(), strValue.GetMaximumLength() );
			pdata->z0 = _tstof(strValue);

			//  velocity distribution model
			hWndChild = GetDlgItem(hWndDlg, IDC_NEW_GALAXY_VELOCITY_DISTRIBUTION);
			pdata->iDiscVelocityModel = ComboBox_GetCurSel (hWndChild);

			//  velocity dispersions? warm or cold disc
			if ( IsDlgButtonChecked( hWndDlg, IDC_NEW_GALAXY_RBWARM ) )
			{
				pdata->bWarmDisc = true;

				//  get Toomre's parameter
				GetDlgItemText( hWndDlg, IDC_NEW_GALAXY_ETOOMRESQ, strValue.GetBuffer(), strValue.GetMaximumLength() );
				pdata->QToomre = _tstof(strValue);

				//  z component dispersion
				GetDlgItemText( hWndDlg, IDC_NEW_GALAXY_EZVELDISP, strValue.GetBuffer(), strValue.GetMaximumLength() );
				pdata->zDisp = _tstof(strValue);
			}
			else
			{
				pdata->bWarmDisc = false;
			}

			//
			//  update pdata

			//  jump to correct property sheet page
			if (pdata->bulge)  PropSheet_SetCurSel(GetParent(hWndDlg), NULL, 2);
			else
			if (pdata->halo)  PropSheet_SetCurSel(GetParent(hWndDlg), NULL, 3);
			else
			if (pdata->darkhalo)  PropSheet_SetCurSel(GetParent(hWndDlg), NULL, 4);
			else
				PropSheet_SetCurSel(GetParent(hWndDlg), NULL, 5);
			}
	        return FALSE;
            
		case PSN_RESET:
			//  Cancel button is clicked

            return 0;


        default:
        
		    return 0;
		}	//  end switch (lpnm->code)
        }	//  end case WM_NOTIFY

	case WM_COMMAND:

		switch ( LOWORD( wParam) )
		{
		case IDC_NEW_GALAXY_RBWARM:
		case IDC_NEW_GALAXY_RBCOLD:

 		   switch ( HIWORD (wParam) )
		   {
		   case BN_CLICKED:

				HWND hChildWnd;

				if ( IsDlgButtonChecked( hWndDlg, IDC_NEW_GALAXY_RBWARM ) )
				{
					hChildWnd = GetDlgItem( hWndDlg, IDC_NEW_GALAXY_ETOOMRESQ );
					EnableWindow( hChildWnd, TRUE );

					hChildWnd = GetDlgItem( hWndDlg, IDC_NEW_GALAXY_EZVELDISP );
					EnableWindow( hChildWnd, TRUE );
				}
				else
				{
					hChildWnd = GetDlgItem( hWndDlg, IDC_NEW_GALAXY_ETOOMRESQ );
					EnableWindow( hChildWnd, FALSE );

					hChildWnd = GetDlgItem( hWndDlg, IDC_NEW_GALAXY_EZVELDISP );
					EnableWindow( hChildWnd, FALSE );
				}
			   
				return 0;
		   }	//  end switch ( HIWORD (wParam) )

		   return 0;	//  case IDC_NEW_GALAXY_RBWARM
		
		}	//  end switch ( LOWORD (wParam) )

		
		return 0;	//  end case WM_COMMAND
        
    default:
        
	    return 0;
    
	}
}

BOOL CALLBACK NewGalaxyBulgeParametersProc ( HWND hWndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	UNREFERENCED_PARAMETER(wParam);
    
	//  retrieve the shared user data from GWL_USERDATA, before WM_INITDIALOG there's nothing
	tagSharedSimulationParameters * pdata = (tagSharedSimulationParameters*) GetWindowLong(hWndDlg, GWL_USERDATA);

    switch (uMsg)
    {
    case WM_INITDIALOG:
        { 
			//  save shared data to window GWL_USERDATA field
			pdata = (tagSharedSimulationParameters*) ((LPPROPSHEETPAGE) lParam) -> lParam;
    		SetWindowLong(hWndDlg, GWL_USERDATA, (DWORD) pdata);

			HWND hWndChild;
			WinString szString(512);

			//  insert model names into combobox
			//

			//  get model combobox
			hWndChild = GetDlgItem(hWndDlg, IDC_NEW_GALAXY_POSITION_DISTRIBUTION);

			//  load model names from resource and insert them into combo
			LoadString( g_hInstance, szString, IDS_NEW_GALAXY_WIZARD_BULGE_MODEL2 );
			ComboBox_InsertString( hWndChild, 0, szString );

			LoadString( g_hInstance, szString, IDS_NEW_GALAXY_WIZARD_BULGE_MODEL1 );
			ComboBox_InsertString( hWndChild, 0, szString);

			//  selects the first item in combo as actual
			ComboBox_SetCurSel( hWndChild, 0 );			
			
		    return 0;
        }

    case WM_NOTIFY:
        {
        LPNMHDR lpnm = (LPNMHDR) lParam;

    
		switch (lpnm->code)
        {
			
		case PSN_SETACTIVE:
			//  sent each time a page is about to be made visible 
			//  initialize data & buttons

			//  enable Back as well as Next button
			PropSheet_SetWizButtons(GetParent(hWndDlg), PSWIZB_BACK | PSWIZB_NEXT);
            return 0;

		case PSN_WIZBACK:
            //  Next button is clicked
			//  check entered data
            
			//  jump to correct property sheet page
			if (pdata->disc)  PropSheet_SetCurSel(GetParent(hWndDlg), NULL, 3);
			else
				PropSheet_SetCurSel(GetParent(hWndDlg), NULL, 2);


	        return FALSE;

		case PSN_WIZNEXT:
			{
            //  Next button is clicked
			//  check entered data

			HWND hWndChild;
			WinString strValue(256);

			//  update pdata
			//

			//  position distribution model
			hWndChild = GetDlgItem(hWndDlg, IDC_NEW_GALAXY_POSITION_DISTRIBUTION);
			pdata->iBulgePositionModel = ComboBox_GetCurSel (hWndChild);

			//  maximum radius rb
			GetDlgItemText( hWndDlg, IDC_NEW_GALAXY_ERB, strValue.GetBuffer(), strValue.GetMaximumLength() );
			pdata->rb = _tstof(strValue);

			//  scale length ab
			GetDlgItemText( hWndDlg, IDC_NEW_GALAXY_EAB, strValue.GetBuffer(), strValue.GetMaximumLength() );
			pdata->ab = _tstof(strValue);

			//  velocity distribution model
			hWndChild = GetDlgItem(hWndDlg, IDC_NEW_GALAXY_VELOCITY_DISTRIBUTION);
			pdata->iBulgeVelocityModel = ComboBox_GetCurSel (hWndChild);

			//
			//  update pdata


			//  jump to correct property sheet page
			if (pdata->halo)  PropSheet_SetCurSel(GetParent(hWndDlg), NULL, 3);
			else
			if (pdata->darkhalo)  PropSheet_SetCurSel(GetParent(hWndDlg), NULL, 4);
			else
				PropSheet_SetCurSel(GetParent(hWndDlg), NULL, 5);
			
			}

	        return FALSE;
                
		case PSN_RESET:
			//  Cancel button is clicked

            return 0;


        default:
        
		    return 0;
		}	//  end switch (lpnm->code)
        }
        
    default:
        
	    return 0;
    }
}

BOOL CALLBACK NewGalaxyHaloParametersProc ( HWND hWndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	UNREFERENCED_PARAMETER(wParam);
    
	//  retrieve the shared user data from GWL_USERDATA, before WM_INITDIALOG there's nothing
	tagSharedSimulationParameters * pdata = (tagSharedSimulationParameters*) GetWindowLong(hWndDlg, GWL_USERDATA);

    switch (uMsg)
    {
    case WM_INITDIALOG:
        { 
			//  save shared data to window GWL_USERDATA field
			pdata = (tagSharedSimulationParameters*) ((LPPROPSHEETPAGE) lParam) -> lParam;
    		SetWindowLong(hWndDlg, GWL_USERDATA, (DWORD) pdata);

			HWND hWndChild;
			WinString szString(512);

			//  insert model names into combobox
			//

			//  get model combobox
			hWndChild = GetDlgItem(hWndDlg, IDC_NEW_GALAXY_POSITION_DISTRIBUTION);

			//  load model names from resource and insert them into combo
			LoadString( g_hInstance, szString, IDS_NEW_GALAXY_WIZARD_HALO_MODEL2 );
			ComboBox_InsertString( hWndChild, 0, szString );

			LoadString( g_hInstance, szString, IDS_NEW_GALAXY_WIZARD_HALO_MODEL1 );
			ComboBox_InsertString( hWndChild, 0, szString);

			//  selects the first item in combo as actual
			ComboBox_SetCurSel( hWndChild, 0 );			

			return 0;
        }

    case WM_NOTIFY:
        {
        LPNMHDR lpnm = (LPNMHDR) lParam;

    
		switch (lpnm->code)
        {
			
		case PSN_SETACTIVE:
			//  sent each time a page is about to be made visible 
			//  initialize data & buttons

			//  enable Back as well as Next button
			PropSheet_SetWizButtons(GetParent(hWndDlg), PSWIZB_BACK | PSWIZB_NEXT);
            return 0;

		case PSN_WIZBACK:
			{
            //  Next button is clicked
			//  check entered data

			//  jump to correct property sheet page
			if (pdata->bulge)  PropSheet_SetCurSel(GetParent(hWndDlg), NULL, 4);
			else
			if (pdata->disc)  PropSheet_SetCurSel(GetParent(hWndDlg), NULL, 3);
			else
				PropSheet_SetCurSel(GetParent(hWndDlg), NULL, 1);
			}

	        return FALSE;

		case PSN_WIZNEXT:
            //  Next button is clicked
			//  check entered data
			
			{
			HWND hWndChild;
			WinString strValue(256);

			//  update pdata
			//

			//  position distribution model
			hWndChild = GetDlgItem(hWndDlg, IDC_NEW_GALAXY_POSITION_DISTRIBUTION);
			pdata->iHaloPositionModel = ComboBox_GetCurSel (hWndChild);

			//  maximum radius rh
			GetDlgItemText( hWndDlg, IDC_NEW_GALAXY_ERH, strValue.GetBuffer(), strValue.GetMaximumLength() );
			pdata->rh = _tstof(strValue);

			//  scale length ah
			GetDlgItemText( hWndDlg, IDC_NEW_GALAXY_EAH, strValue.GetBuffer(), strValue.GetMaximumLength() );
			pdata->ah = _tstof(strValue);

			//  velocity distribution model
			hWndChild = GetDlgItem(hWndDlg, IDC_NEW_GALAXY_VELOCITY_DISTRIBUTION);
			pdata->iHaloVelocityModel = ComboBox_GetCurSel (hWndChild);

			//
			//  update pdata

			
			//  jump to correct property sheet page
			if (pdata->darkhalo)  PropSheet_SetCurSel(GetParent(hWndDlg), NULL, 4);
			else
				PropSheet_SetCurSel(GetParent(hWndDlg), NULL, 5);
			
			}

	        return FALSE;
            
		case PSN_RESET:
			//  Cancel button is clicked

            return 0;


        default:
        
		    return 0;
		}	//  end switch (lpnm->code)
        }
        
    default:
        
	    return 0;
    }
}

BOOL CALLBACK NewGalaxyDarkHaloParametersProc ( HWND hWndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	UNREFERENCED_PARAMETER(wParam);

    //  retrieve the shared user data from GWL_USERDATA, before WM_INITDIALOG there's nothing
	tagSharedSimulationParameters * pdata = (tagSharedSimulationParameters*) GetWindowLong(hWndDlg, GWL_USERDATA);

    switch (uMsg)
    {
    case WM_INITDIALOG:
        { 
			//  save shared data to window GWL_USERDATA field
			pdata = (tagSharedSimulationParameters*) ((LPPROPSHEETPAGE) lParam) -> lParam;
    		SetWindowLong(hWndDlg, GWL_USERDATA, (DWORD) pdata);
			
			HWND hWndChild;
			WinString szString(512);

			//  insert model names into combobox
			//

			//  get model combobox
			hWndChild = GetDlgItem(hWndDlg, IDC_NEW_GALAXY_POSITION_DISTRIBUTION);

			//  load model names from resource and insert them into combo
			LoadString( g_hInstance, szString, IDS_NEW_GALAXY_WIZARD_DARKHALO_MODEL2 );
			ComboBox_InsertString( hWndChild, 0, szString );

			LoadString( g_hInstance, szString, IDS_NEW_GALAXY_WIZARD_DARKHALO_MODEL1 );
			ComboBox_InsertString( hWndChild, 0, szString);

			//  selects the first item in combo as actual
			ComboBox_SetCurSel( hWndChild, 0 );			

		    return 0;
        }

    case WM_NOTIFY:
        {
        LPNMHDR lpnm = (LPNMHDR) lParam;

    
		switch (lpnm->code)
        {
			
		case PSN_SETACTIVE:
			//  sent each time a page is about to be made visible 
			//  initialize data & buttons

			//  enable Back as well as Next button
			PropSheet_SetWizButtons(GetParent(hWndDlg), PSWIZB_BACK | PSWIZB_NEXT);
            return 0;

		case PSN_WIZBACK:
			{
            //  Next button is clicked

			if (pdata->halo)  PropSheet_SetCurSel(GetParent(hWndDlg), NULL, 5);
			else
			if (pdata->bulge)  PropSheet_SetCurSel(GetParent(hWndDlg), NULL, 4);
			else
			if (pdata->disc)  PropSheet_SetCurSel(GetParent(hWndDlg), NULL, 3);
			else
				PropSheet_SetCurSel(GetParent(hWndDlg), NULL, 2);
			
			}
	        return FALSE;

		case PSN_WIZNEXT:
			{
            //  Next button is clicked
			//  check entered data

			HWND hWndChild;
			WinString strValue(256);

			//  update pdata
			//

			//  position distribution model
			hWndChild = GetDlgItem(hWndDlg, IDC_NEW_GALAXY_POSITION_DISTRIBUTION);
			pdata->iDarkHaloPositionModel = ComboBox_GetCurSel (hWndChild);

			//  maximum radius rdh
			GetDlgItemText( hWndDlg, IDC_NEW_GALAXY_ERDH, strValue.GetBuffer(), strValue.GetMaximumLength() );
			pdata->rdh = _tstof(strValue);

			//  scale length adh
			GetDlgItemText( hWndDlg, IDC_NEW_GALAXY_EADH, strValue.GetBuffer(), strValue.GetMaximumLength() );
			pdata->adh = _tstof(strValue);

			//  core radius, gamma
			GetDlgItemText( hWndDlg, IDC_NEW_GALAXY_EGAMMA, strValue.GetBuffer(), strValue.GetMaximumLength() );
			pdata->gamma = _tstof(strValue);

			//  velocity distribution model
			hWndChild = GetDlgItem(hWndDlg, IDC_NEW_GALAXY_VELOCITY_DISTRIBUTION);
			pdata->iDarkHaloVelocityModel = ComboBox_GetCurSel (hWndChild);

			//
			//  update pdata
			}

			//  jump to correct property sheet page
	        return FALSE;
            
		case PSN_RESET:
			//  Cancel button is clicked

            return 0;


        default:
        
		    return 0;
		}	//  end switch (lpnm->code)
        }
        
    default:
        
	    return 0;
    }
}


BOOL CALLBACK NewGalaxyComputationParametersProc ( HWND hWndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	UNREFERENCED_PARAMETER(wParam);

    //  retrieve the shared user data from GWL_USERDATA, before WM_INITDIALOG there's nothing
	tagSharedSimulationParameters * pdata = (tagSharedSimulationParameters*) GetWindowLong(hWndDlg, GWL_USERDATA);

    switch (uMsg)
    {
    case WM_INITDIALOG:
        { 
			//  save shared data to window GWL_USERDATA field
			pdata = (tagSharedSimulationParameters*) ((LPPROPSHEETPAGE) lParam) -> lParam;
    		SetWindowLong(hWndDlg, GWL_USERDATA, (DWORD) pdata);

		    return 0;
        }

    case WM_NOTIFY:
        {
        LPNMHDR lpnm = (LPNMHDR) lParam;

    
		switch (lpnm->code)
        {
			
		case PSN_SETACTIVE:
			//  sent each time a page is about to be made visible 
			//  initialize data & buttons

			//  enable Back as well as Next button
			PropSheet_SetWizButtons(GetParent(hWndDlg), PSWIZB_BACK | PSWIZB_NEXT);
            return 0;

		case PSN_WIZBACK:
            //  Next button is clicked
			//  check entered data

			if (pdata->darkhalo)  PropSheet_SetCurSel(GetParent(hWndDlg), NULL, 6);
			else
			if (pdata->halo)  PropSheet_SetCurSel(GetParent(hWndDlg), NULL, 5);
			else
			if (pdata->bulge)  PropSheet_SetCurSel(GetParent(hWndDlg), NULL, 4);
			else
			if (pdata->disc)  PropSheet_SetCurSel(GetParent(hWndDlg), NULL, 3);
			else
				PropSheet_SetCurSel(GetParent(hWndDlg), NULL, 2);
       
	        return FALSE;

		case PSN_WIZNEXT:
			{
            //  Next button is clicked
			//  check entered data

			WinString strValue(256);

			//  update pdata
			//

			//  opening angle theta
			GetDlgItemText( hWndDlg, IDC_NEW_GALAXY_ETHETA, strValue.GetBuffer(), strValue.GetMaximumLength() );
			theta = _tstof(strValue);

			//  softening length epsilon
			GetDlgItemText( hWndDlg, IDC_NEW_GALAXY_EEPSILON, strValue.GetBuffer(), strValue.GetMaximumLength() );
			epsilon = _tstof(strValue);

			//  time step, diferential time
			GetDlgItemText( hWndDlg, IDC_NEW_GALAXY_EDT, strValue.GetBuffer(), strValue.GetMaximumLength() );
			dt = _tstof(strValue);

			//  jump to correct property sheet page
			}
	        return FALSE;
            
		case PSN_RESET:
			//  Cancel button is clicked

            return 0;


        default:
        
		    return 0;
		}	//  end switch (lpnm->code)
        }
        
    default:
        
	    return 0;
    }
}

BOOL CALLBACK NewGalaxyFileNamesProc ( HWND hWndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
    //  retrieve the shared user data from GWL_USERDATA, before WM_INITDIALOG there's nothing
	tagSharedSimulationParameters * pdata = (tagSharedSimulationParameters*) GetWindowLong(hWndDlg, GWL_USERDATA);

    switch (uMsg)
    {
    case WM_INITDIALOG:
        { 
			//  save shared data to window GWL_USERDATA field
			pdata = (tagSharedSimulationParameters*) ((LPPROPSHEETPAGE) lParam) -> lParam;
    		SetWindowLong(hWndDlg, GWL_USERDATA, (DWORD) pdata);
			return 0;
        }

    case WM_NOTIFY:
        {
        LPNMHDR lpnm = (LPNMHDR) lParam;

    
		switch (lpnm->code)
        {
			
		case PSN_SETACTIVE:
			//  sent each time a page is about to be made visible 
			//  initialize data & buttons

			//  enable Back as well as Next button
			PropSheet_SetWizButtons(GetParent(hWndDlg), PSWIZB_BACK | PSWIZB_NEXT);
            return 0;

		case PSN_WIZNEXT:
            //  Next button is clicked
			//  check entered data
			{
			WinString strValue(256);

            GetDlgItemText( hWndDlg, IDC_NEW_GALAXY_NAME, pdata->szName, MAX_PATH );
            GetDlgItemText( hWndDlg, IDC_NEW_GALAXY_DIRECTORY, pdata->szFolder, MAX_PATH );
            GetDlgItemText( hWndDlg, IDC_NEW_GALAXY_DESCRIPTION, pdata->szDescription, lstrlen(pdata->szDescription) );

/*
			if ( lstrlen(pdata->dataFileName) == 0 || lstrlen(pdata->videoFileName) == 0 )
			{
				MessageBox( hWndDlg, Language.GetString(IDS_NEW_GALAXY_WIZARD_FILENAMESREQUIRED), NULL, MB_OK | MB_ICONINFORMATION );
				SetWindowLong(hWndDlg, DWL_MSGRESULT, -1);
				return TRUE;
			}
*/
			}
	        return FALSE;

		case PSN_WIZBACK:
            //  Next button is clicked
			//  check entered data
            
			return FALSE;
            
		case PSN_RESET:
			//  Cancel button is clicked

            return 0;


        default:
        
		    return 0;
		}	//  end switch (lpnm->code)
        }
        

	case WM_COMMAND:

		switch ( LOWORD( wParam) )
		{
		case IDC_NEW_GALAXY_BUTTON_BROWSE_DIRECTORY:

 		    switch ( HIWORD (wParam) )
			{
			case BN_CLICKED:
				{
				BROWSEINFO bi; 
				bi.hwndOwner = hWndDlg;
				bi.pidlRoot = NULL;
				bi.pszDisplayName = pdata->szFolder;
				bi.lpszTitle = Language.GetString (IDS_NEW_GALAXY_WIZARD_CHOOSE_DIRECTORY);
				bi.ulFlags = BIF_RETURNONLYFSDIRS;
				bi.lpfn = NULL;
				bi.lParam = 0;

				LPITEMIDLIST pidlSelected;
				pidlSelected = SHBrowseForFolder(&bi);

				if (SHGetPathFromIDList(pidlSelected, pdata->szFolder))
				{
				   SetDlgItemText(hWndDlg, IDC_NEW_GALAXY_DIRECTORY, pdata->szFolder);
				}

				}
				
				return 0;
			
			}	 //  end switch ( HIWORD (wParam) )


			return 0;
		
		}	//  end switch ( LOWORD( wParam) )

			
		return 0;	//  return WM_COMMAND

    default:
        
	    return 0;
    }
}

BOOL CALLBACK NewGalaxyFinalPageProc ( HWND hWndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	UNREFERENCED_PARAMETER(wParam);

    //  retrieve the shared user data from GWL_USERDATA, before WM_INITDIALOG there's nothing
	tagSharedSimulationParameters * pdata = (tagSharedSimulationParameters*) GetWindowLong(hWndDlg, GWL_USERDATA);

    switch (uMsg)
    {
    case WM_INITDIALOG:
        { 
			//  save shared data to window GWL_USERDATA field
			pdata = (tagSharedSimulationParameters*) ((LPPROPSHEETPAGE) lParam) -> lParam;
    		SetWindowLong(hWndDlg, GWL_USERDATA, (DWORD) pdata);

		    return 0;
        }

    case WM_NOTIFY:
        {
        LPNMHDR lpnm = (LPNMHDR) lParam;

    
		switch (lpnm->code)
        {
			
		case PSN_SETACTIVE:
			//  sent each time a page is about to be made visible 
			//  initialize data & buttons

			//  enable Back and replace Next with Finish button
			PropSheet_SetWizButtons(GetParent(hWndDlg), PSWIZB_BACK | PSWIZB_FINISH);
            
			return 0;

		case PSN_WIZFINISH:
            //  Finish button is clicked
			//  check entered data
			{
			HWND hProgress = GetDlgItem(hWndDlg, IDC_NEW_GALAXY_PROGRESSBAR);
			
			SendDlgItemMessage( hWndDlg, IDC_NEW_GALAXY_PROGRESSBAR, PBM_SETRANGE32, 0, 20 );
			SendDlgItemMessage( hWndDlg, IDC_NEW_GALAXY_PROGRESSBAR, PBM_SETSTEP, 2, 0 );
			SetWindowLong( hProgress, GWL_STYLE,
				GetWindowLong( hProgress, GWL_STYLE ) | 0x02 );
			SendDlgItemMessage( hWndDlg, IDC_NEW_GALAXY_PROGRESSBAR, WM_USER+10, TRUE, 100 );

			CreateNewGalaxySimulation( *pdata );
			}
            
	        return FALSE;

		case PSN_WIZBACK:
            //  Next button is clicked
			//  check entered data
            
	        return FALSE;
            
		case PSN_RESET:
			//  Cancel button is clicked

            return 0;


        default:
        
		    return 0;
		}	//  end switch (lpnm->code)
        }
        
    default:
        
	    return 0;
    }
}


void ShowGalaxySimulationDlg( HWND hWnd )
{
	//  create property sheet pages
	//

	PROPSHEETPAGE psp;
    HPROPSHEETPAGE ahpsp[9];

	WinString strTitle(256);
	WinString strHeaderTitle(256);

	strTitle.SetString( Language.GetString (IDS_NEW_GALAXY_WIZARD_MAIN_TITLE) );

	//  welcoming property sheet page

	psp.dwSize		= sizeof(psp);
	psp.hInstance	= Language.GetResourceInstance();
	psp.lParam		= (LPARAM) &SharedSimulationParameters;  //  the shared data structure
	psp.dwFlags		= PSP_DEFAULT | PSP_HIDEHEADER  | PSP_USETITLE;
	psp.pszTemplate = MAKEINTRESOURCE(IDD_NEW_GALAXY_PAGE_WELCOME);
	psp.pfnDlgProc	= NewGalaxyWelcomePageProc;
	psp.pszTitle	= strTitle;

    ahpsp[0] = CreatePropertySheetPage(&psp);

    //  first interior page, components

	strHeaderTitle.SetString( Language.GetString (IDS_NEW_GALAXY_WIZARD_HEADER_TITLE1) );

    psp.dwFlags			= PSP_DEFAULT | PSP_USEHEADERTITLE | PSP_USETITLE;
    psp.pszTemplate		= MAKEINTRESOURCE(IDD_NEW_GALAXY_PAGE_COMPONENTS);
    psp.pfnDlgProc		= NewGalaxyComponentsProc;
    psp.pszHeaderTitle	= strHeaderTitle;

    ahpsp[1] = CreatePropertySheetPage(&psp);

    //  second interior page, disk parameters

	strHeaderTitle.SetString( Language.GetString (IDS_NEW_GALAXY_WIZARD_HEADER_TITLE2) );

    psp.dwFlags			= PSP_DEFAULT | PSP_USEHEADERTITLE | PSP_USETITLE;
    psp.pszTemplate		= MAKEINTRESOURCE(IDD_NEW_GALAXY_PAGE_DISC);
    psp.pfnDlgProc		= NewGalaxyDiskParametersProc;
    psp.pszHeaderTitle	= strHeaderTitle;

    ahpsp[2] = CreatePropertySheetPage(&psp);

    //  second interior page, bulge parameters

	strHeaderTitle.SetString( Language.GetString (IDS_NEW_GALAXY_WIZARD_HEADER_TITLE3) );

    psp.dwFlags			= PSP_DEFAULT | PSP_USEHEADERTITLE | PSP_USETITLE;
    psp.pszTemplate		= MAKEINTRESOURCE(IDD_NEW_GALAXY_PAGE_BULGE);
    psp.pfnDlgProc		= NewGalaxyBulgeParametersProc;
    psp.pszHeaderTitle	= strHeaderTitle;

    ahpsp[3] = CreatePropertySheetPage(&psp);

    //  third interior page, halo parameters

	strHeaderTitle.SetString( Language.GetString (IDS_NEW_GALAXY_WIZARD_HEADER_TITLE4) );

    psp.dwFlags			= PSP_DEFAULT | PSP_USEHEADERTITLE | PSP_USETITLE;
    psp.pszTemplate		= MAKEINTRESOURCE(IDD_NEW_GALAXY_PAGE_HALO);
    psp.pfnDlgProc		= NewGalaxyHaloParametersProc;
    psp.pszHeaderTitle	= strHeaderTitle;

    ahpsp[4] = CreatePropertySheetPage(&psp);

    //  fourth interior page, dark halo parameters

	strHeaderTitle.SetString( Language.GetString (IDS_NEW_GALAXY_WIZARD_HEADER_TITLE5) );

    psp.dwFlags			= PSP_DEFAULT | PSP_USEHEADERTITLE | PSP_USETITLE;
    psp.pszTemplate		= MAKEINTRESOURCE(IDD_NEW_GALAXY_PAGE_DARKHALO);
    psp.pfnDlgProc		= NewGalaxyDarkHaloParametersProc;
    psp.pszHeaderTitle	= strHeaderTitle;

    ahpsp[5] = CreatePropertySheetPage(&psp);

    //  fifth interior page, set parameters of computation

	strHeaderTitle.SetString( Language.GetString (IDS_NEW_GALAXY_WIZARD_HEADER_TITLE6) );

    psp.dwFlags			= PSP_DEFAULT | PSP_USEHEADERTITLE | PSP_USETITLE;
    psp.pszTemplate		= MAKEINTRESOURCE(IDD_PAGE_COMPUTATION);
    psp.pfnDlgProc		= NewGalaxyComputationParametersProc;
    psp.pszHeaderTitle	= strHeaderTitle;

    ahpsp[6] = CreatePropertySheetPage(&psp);

    //  seventh interior page, set file names

	strHeaderTitle.SetString( Language.GetString (IDS_NEW_GALAXY_WIZARD_HEADER_TITLE7) );

    psp.dwFlags			= PSP_DEFAULT | PSP_USEHEADERTITLE | PSP_USETITLE;
    psp.pszTemplate		= MAKEINTRESOURCE(IDD_PAGE_FILENAMES);
    psp.pfnDlgProc		= NewGalaxyFileNamesProc;
    psp.pszHeaderTitle	= strHeaderTitle;

    ahpsp[7] = CreatePropertySheetPage(&psp);

    //  final property sheet page

    psp.dwFlags			= PSP_DEFAULT | PSP_HIDEHEADER |  PSP_USETITLE;
    psp.pszTemplate		= MAKEINTRESOURCE(IDD_NEW_GALAXY_PAGE_FINAL);
    psp.pfnDlgProc		= NewGalaxyFinalPageProc;

    ahpsp[8] = CreatePropertySheetPage(&psp);

	//  create property sheet dialog
	//

    PROPSHEETHEADER psh;

	psh.dwSize			= sizeof(psh);
    psh.hInstance		= g_hInstance;
    psh.hwndParent		= hWnd;
    psh.phpage			= ahpsp;
    psh.dwFlags			= PSH_WIZARD97 | PSH_WATERMARK | PSH_HEADER;
    psh.pszbmWatermark	= TEXT("WatermarkBitmapNew");
    psh.pszbmHeader		= TEXT("HeaderBitmap");
    psh.nStartPage		= 0;	
    psh.nPages			= 9;

    PropertySheet(&psh);

/*
	//  for NT4 only
	tagSharedSimulationParameters data;

    data.disc = true;
	data.discTotalMass = 1.3;
	data.rd = 27;
	data.ad = 3.5;
	data.z0 = 0.5;
	data.bWarmDisc = false;
	data.discBodies = 25000;
	data.QToomre = 1.2;
	data.zDisp = 0.0;		//  pouziva se vzdy!!! a nemelo by

	data.bulge = true;
	data.bulgeBodies = 2500;
	data.rb = 3.5;
	data.ab = 3.5;

	data.halo = false;

	data.darkhalo = true;
	data.darkhaloBodies = 85000;
	data.rdh = 84;
	data.adh = 3.5;
	data.gamma = 3.5;


	lstrcpy( data.szFolder, TEXT("D:") );
	lstrcpy( data.szName, TEXT("simulace") );
	
	theta = 1;
	epsilon = 0.75;
	dt = 1;

	CreateNewGalaxySimulation(data);

	*/
}