#include "camera.h"

using namespace JakubSchwarzmeier::OpenGL;

//  constructor

Camera::Camera(): Matrix<double>(4,4)
{
	LoadIdentity();
	position = 0;
	
	xaxis[0] = 1;
	xaxis[1] = 0;
	xaxis[2] = 0;

	zaxis[0] = 0;
	zaxis[1] = 0;
	zaxis[2] = 1;

//	lambda = 0;  fi = 0;  theta = 0;
}

//  rotational matrices

void Camera::ActByCurrentOnNewXRotationalMatrixWithAngle(double angle)
{
	const double xRotationOperator[16]={ 1,        0,          0,               0,
										  0,  cos_angle(angle), sin_angle(angle), 0,
										  0, -sin_angle(angle), cos_angle(angle), 0,
										  0,         0,          0,               1 };


	ActOn(xRotationOperator);
}

void Camera::ActByCurrentOnNewYRotationalMatrixWithAngle(double angle)
{
	const double yRotationOperator[16]={ cos_angle(angle), 0, sin_angle(angle),  0,
											 0         , 1,         0,          0,
										   -sin_angle(angle) , 0,   cos_angle(angle),0,
											 0         , 0,         0,          1 };

	ActOn(yRotationOperator);
}

void Camera::ActByCurrentOnNewZRotationalMatrixWithAngle(double angle)
{
	const double zRotationOperator[16]={ cos_angle(angle), -sin_angle(angle), 0, 0,
									   sin_angle(angle), cos_angle(angle), 0, 0,
										 0,                 0,          1, 0,
										 0,                 0,          0, 1 };

	ActOn(zRotationOperator);
}

void Camera::ActByCurrentOnNewTranslationMatrixWithParameters(double x, double y, double z)
{
	const double positionOperator[16]={ 1, 0, 0, 0,
										0, 1, 0, 0,
										0, 0, 1, 0,
										-x, -y, -z, 1 };

	ActOn(positionOperator);
}


//  axis orientation and position validation

void Camera::UpdateMatrix()
{
	LoadIdentity();
	
	for( int i = 0; i < 3; i++)
	{
		values[i] = xaxis[i];
		values[i+8] = zaxis[i];
		values[i+12] = position[i];
	}
	
	Normalize();
}

void Camera::FinishUpdate()
{
	const double eps = 1e-10;
	
	int i;
	for( i = 0; i < 3; i++)
	{
		if ( fabs( xaxis[i] - values[i] ) > eps )  xaxis[i] = values[i];
		if ( fabs( zaxis[i] - values[i+8] ) > eps )  zaxis[i] = values[i+8];
		if ( fabs( position[i] - values[i+12] ) > eps )  position[i] = values[i+12];
	}
}

//  normalize and compute y axis

void Camera::Normalize()
{
	//  values[2] nesmi byt 0 (deleni nulou)
	//  pridal jsem ze musi byt vetsi nez 1e-10, protoze pro velmi mala cisla se vysoce zvysuje nepresnost
	if ( fabs(values[2]) > 1e-10)  values[10] = (-values[0]*values[8]-values[1]*values[9])/values[2];
	
	values[4] = values[1]*values[10]-values[2]*values[9];
	values[5] = values[2]*values[8]-values[0]*values[10];
	values[6] = values[0]*values[9]-values[1]*values[8];
}

//  user friendly functions

void Camera::ChangeLambdaBy(double angle)
{
	UpdateMatrix();

	ActByCurrentOnNewYRotationalMatrixWithAngle (angle);

	FinishUpdate();
}

void Camera::ChangeFiBy(double angle)
{
	UpdateMatrix();

	ActByCurrentOnNewXRotationalMatrixWithAngle (angle);

	FinishUpdate();
}

void Camera::ChangeThetaBy(double angle)
{
	UpdateMatrix();

	ActByCurrentOnNewZRotationalMatrixWithAngle (angle);

	FinishUpdate();
}

void Camera::ChangePositionBy(double dx, double dy, double dz)
{
	UpdateMatrix();

	ActByCurrentOnNewTranslationMatrixWithParameters (dx, dy, dz);

	FinishUpdate();
}

//  matrices multiplication, in form of field of doubles, not overloadable

void Camera::ActBy( const double rotationMatrix[] )
{
	Matrix<real> result( rows, columns );
	result.Clear();

	for(int k=0; k<16; k++)
		for(int i=0; i<4; i++)
			result[k]+=rotationMatrix[(k/4)*4+i] * values[(k%4)+4*i];

	for(k=0; k<16; k++)  values[k] = result[k];

}

void Camera::ActOn( const double rotationMatrix[] )
{
	Matrix<real> result( rows, columns );
	result.Clear();

	for(int k=0; k<16; k++)
		for(int i=0; i<4; i++)
			result[k]+=values[(k/4)*4+i] * rotationMatrix[(k%4)+4*i];

	for(k=0; k<16; k++)  values[k] = result[k];

}

//  move angle size to limits < 0 ; 360 )

void Camera::FixAngle( double & angle )
{
	while(angle >= 360)
	{
		angle -= 360;
	}
	
	while(angle < 0)
	{
		angle += 360;
	}
}

//  get class values from matrix

double Camera::GetLambda()
{
	//  backup current values
	Matrix<double> tmpMatrix(4, 4);
	tmpMatrix = *this;

	Inverse();

	double result = atan2_angle( values[8], values[10] );

	//  restore values prior to inversion of matrix
	for( int i = 0; i < 16; i++ )  values[i] = tmpMatrix[i];

	// return (result<0) ? result+360.0 : result;
	return result - floor(result/360.0) * 360.0;
}

double Camera::GetFi()
{
	//  backup current values
	Matrix<double> tmpMatrix(4, 4);
	tmpMatrix = *this;

	Inverse();
	
	double result = atan2_angle( values[9], pow( (pow(values[8],2)+pow(values[10],2)) , 0.5) );

	//  restore values prior to inversion of matrix
	for( int i = 0; i < 16; i++ )  values[i] = tmpMatrix[i];

	return 	result;
}

double Camera::GetTheta()
{
	//  backup current values
	Matrix<double> tmpMatrix(4, 4);
	tmpMatrix = *this;

	Inverse();

	double result = atan2_angle( values[1],  pow( pow(values[0],2) + pow(values[2],2), 0.5 )  );

//  restore values prior to inversion of matrix
	for( int i = 0; i < 16; i++ )  values[i] = tmpMatrix[i];
	
	return result;
}

Vector<> Camera::GetPosition()
{
	//  backup current values
	Matrix<double> tmpMatrix(4, 4);
	tmpMatrix = *this;

	Inverse();

	Vector<> returnVector;
	returnVector[0] = values[12];
	returnVector[1] = values[13];
	returnVector[2] = values[14];

//  restore values prior to inversion of matrix
	for( int i = 0; i < 16; i++ )  values[i] = tmpMatrix[i];

	return Vector<>(returnVector);
}