#include "opengl.h"

using namespace JakubSchwarzmeier::Windows;

OpenGL::OpenGL(HWND n_hWnd)
{
	active = true;
	fullscreen = false;
	hWnd = n_hWnd;
	camera = NULL;

	hMemDC = NULL;
	hRC = NULL;
	hbmOld = NULL;
	hDibSection = NULL;
}

OpenGL::~OpenGL()
{
	if (fullscreen)
	{
		ChangeDisplaySettings(NULL, 0);
		ShowCursor(TRUE);
	}

	wglMakeCurrent(NULL,NULL);
	wglDeleteContext(hRC);

	if ( hMemDC != NULL )  DeleteDC( hMemDC );
	if ( hbmOld != NULL )
	{
		HBITMAP hbmPrev = SelectBitmap(hMemDC, hbmOld);
		DeleteObject( hbmPrev );
	}

	InvalidateRect(hWnd, NULL, TRUE);	// vyvol pekreslen pozad aplikace
}

void OpenGL::SetCamera( Camera * n_camera )
{
	camera = n_camera;
}

void OpenGL::OnSize(int & cx, int & cy)
{
	Resize ( cx, cy );
}

void OpenGL::Resize(int n_width, int n_height)
{
	if ( (n_width  <= 0) || (n_height <= 0) ) 
	{
		active = false;
		return;
	}
	
	active = true;

	width = n_width;
	height = n_height;

	//  Windows specific initialization
	//

	//  DIB section preparation
					
	//  create a BITMAPINFOHEADER structure to describe the DIB
	ZeroMemory( &BIH, sizeof ( BIH ) );

	//  fill in the header info
	BIH.biSize			= sizeof ( BIH );
	BIH.biWidth			= width;
	BIH.biHeight		= height;
	BIH.biPlanes		= 1;
	BIH.biBitCount		= 24;		//  make it better with GetDeviceCaps	
	BIH.biCompression	= BI_RGB;

	if ( hMemDC != NULL )  DeleteDC( hMemDC );
	//  create a new DC, it's memory DC, that is compatible (means device depended) with screen
	hMemDC = CreateCompatibleDC(NULL);

	if (hDibSection != NULL)   DeleteObject(hDibSection);
	//  create the DIB section
	hDibSection = CreateDIBSection( hMemDC,
							(BITMAPINFO*)&BIH,
							DIB_PAL_COLORS,
							&m_pBits,
							NULL,
							0);

	//  select the new bitmap into the buffer DC
	if ( hbmOld == NULL )
	{
		hbmOld = SelectBitmap(hMemDC, hDibSection);
	}
	else
	{
		HBITMAP hPrevDibSection = SelectBitmap(hMemDC, hDibSection);
		DeleteObject( hPrevDibSection );
	}

	// set the pixel format for the DC
	PIXELFORMATDESCRIPTOR pfd;
	ZeroMemory( &pfd, sizeof ( pfd ) );

	pfd.nSize		= sizeof ( pfd );
	pfd.nVersion	= 1;
	pfd.dwFlags		= PFD_DRAW_TO_BITMAP | PFD_SUPPORT_OPENGL | PFD_SUPPORT_GDI;
	pfd.iPixelType	= PFD_TYPE_RGBA;
	pfd.cColorBits	= 24;			//  make it better with GetDeviceCaps
	pfd.cDepthBits	= 16;			//  count of bits per pixel for depth (z) buffer
	pfd.iLayerType	= PFD_MAIN_PLANE;

	int iFormat = ChoosePixelFormat( hMemDC, &pfd );
	SetPixelFormat( hMemDC, iFormat, &pfd );

	if ( hRC != NULL )  wglDeleteContext(hRC);
	// create render context (RC) and connect it with device context (DC)
	hRC = wglCreateContext( hMemDC );
	wglMakeCurrent( hMemDC, hRC );

	//
	// Windows specific initialization


	// OpenGL specific initialization
	//

	//  global properties

	glClearColor ( 0.0, 0.0, 0.0, 0.0 );

	glEnable(GL_DEPTH_TEST);			//  glClear must then have set GL_DEPTH_BUFFER_BIT flag, otherwise makes displaying errors!!!
	glDepthFunc(GL_LEQUAL);
	glClearDepth(1.0f);

	// make the shading smooth
	glShadeModel(GL_SMOOTH);

	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

	//  projection properties

	glViewport(0, 0, width, height);		//  set transformation from device to window coordinates

	glMatrixMode(GL_PROJECTION); 
	glLoadIdentity();

/*
	if (height==0)
		gluPerspective(75, (float)width, 2.0, 1000.0);
	else
		gluPerspective (90, (float)width / (float)height, 2.0, 1000.0);
*/		
	GLfloat glAspect = width/height;

	//  point of this is that eye is every time in 0,0,0 
	glFrustum (-glAspect, glAspect, -1, 1, 1.0, 10000.0);
	const unsigned char * szError = gluErrorString( glGetError() );

	glMatrixMode   ( GL_MODELVIEW );

	//
	// OpenGL specific initialization
}


void OpenGL::Render()
{
	//  clear the color and depth buffers 
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); 

	//  modeling transformation in ModelView matrix

	//  simulate viewing transformation in ModelView matrix
	glLoadMatrixd( camera->GetMatrix() );
	
	//  objects drawing

	glBegin( GL_QUADS );
		glColor3f( 1.0f, 0.0f, 0.0f ); glVertex2f( 1, 1 );
		glColor3f( 0.0f, 1.0f, 0.0f ); glVertex2f( -1, 1 );
		glColor3f( 0.0f, 0.0f, 1.0f ); glVertex2f( -1, -1 );
		glColor3f( 0.0f, 1.0f, 0.0f ); glVertex2f( 1, -1 );
	glEnd();

	glFinish();
}

void OpenGL::OnPaint (HDC hDC)
{
	BitBlt( hDC, 0, 0, width, height, hMemDC, 0, 0, SRCCOPY );
}


HBITMAP OpenGL::GetDibSection()
{
	return hDibSection;
}

HDC OpenGL::GetDC()
{
	return hMemDC;
}
