/************************************************************
*
*  Copyright (c) 2002-2004 Jakub Schwarzmeier
*
************************************************************/

/*
*  Based on following papers:
*
*	J. E. Barnes, P. Hut, Nature, 1986, vol. 324, p. 446.
*	U. Becciani et al., J. Comput. Phys., 2000, vol. 163, p. 118
*/

#pragma warning(disable:4786)

#include "bh.h"

#include <algorithm>  //  for_each, copy, swap
#include <stack>
#include <math.h>
#include <typeinfo>

#define epsilon  1

using namespace JakubSchwarzmeier::BarnesHut;

Node::RCustomList::~RCustomList()
			{
				if (owner)
				{
					if (count != 0)
					{
						for (int i = 0; i < count; i++)
						{
							// smaze objekt, na ktery nam uzivatel dal ukazatel

							if ( (storage[i] != NULL) && (typeid(*storage[i]) != typeid(Leave)) )
							{
            					delete storage[i];
							}
						}
					}
				}

				owner = false; //  zabrani destruktoru nadrazene tridy provadet jakekoliv akce  				
				//  nyni se vola destruktor nadrazene tridy
			} // ~RCustomList()



//  default constructor: n_corner = 0 
Node::Node( Vector<> & n_corner, Angles n_angles ): leaves(200, false), corner(n_corner)
{
	angles = n_angles;
    parent = NULL;
	childNo = -1;
	grouped = false;
}


Tree::Tree( RList<Leave> & n_leavesList ): leavesList(n_leavesList)

{}

// initialization of static member(s)
// const real Tree::kappa=-6.67259e-11;
 const real Tree::kappa=1;

void Tree::step()
{
	//  setup & initialize variables
    //
	
	//  adaptive bisect -- find greatest value of spatial coordinate in absolut value
	//
	//  both level & bisect can be computed from the other, bisect was previously defined as bisect = DIMENSION_LENGTH

	double bisect = 0;
	for(int i = 0; i < leavesList.getCount(); i++)
	{
		for(int c = 0; c < 3; c++)
		{
			double val = fabs(leavesList.getAt(i).position[c]);
			if ( val > bisect )  bisect = val;
		}

		//  clear old costzones load info
		leavesList[i].load = 0;
	}

	// increase by one (because of condition that compares corners of cube with values of particles)
	//  it must be farther then the furthest particle
	bisect++;

	RList< RList<Node> > mostLeftAncestors;

	RList<Group> groupsList;		//  WD99
	
	//  setup root node
	Node * root = new Node( *(new Vector<>) );
	Node * actual;
	for(i = 0; i < 3; i++)	root->corner[i] = -bisect;
	for(i = 0; i < leavesList.getCount(); i++) root->leaves.Add( &( leavesList[i] ) );

	//  Phase 1: whole tree construction
	//

	int level = 0;
	mostLeftAncestors.Add( new RList<Node>(100, false) );
	mostLeftAncestors[level].Add( root );

	while( level < mostLeftAncestors.getCount() )
	{

		#ifdef _DEBUG
		std::cout << "level: " << level;
		#endif

		int pos = 0;
		while( pos < mostLeftAncestors[level].getCount() )
		{
			actual = &( mostLeftAncestors[level][pos] );	

			//  partition actual node & create its children
			//

			//  create list for corners
			RList< Vector<> > initCorners(8, false);
			//  add corners and init their values in ordinary cube
			initCubeCorners( actual->corner, initCorners, bisect );

			//  create list for Hilbert's order
			RList< Vector<> > finalCorner(8, false);
			for(int i = 0; i < 8; i++)  finalCorner.Add( &(initCorners[i]) );	//  values have no meaning now, only for the right size of list

			//  order corners according to Hilbert's curve
			//

			//  make from ordinary cube Hilbert's cube
			RList< Vector<> >::iterator initIter = finalCorner.begin();
			initIter += initCorners[7];	initIter += initCorners[5];
			initIter += initCorners[1];	initIter += initCorners[3];
			initIter += initCorners[2];	initIter += initCorners[0];
			initIter += initCorners[4];	initIter = initCorners[6];
			
			//  order curve according to angles
			
			//  x rotation
			for(i = 0; i < actual->angles.x; i=i+90 )
			{
				//  make tmp copy
//				std::copy( finalCorner.begin(), finalCorner.end(), initCorners.begin() );
				for(int j = 0; j < 8; j++)  initCorners.storage[j] = finalCorner.storage[j];

				RList< Vector<> >::iterator iterator = finalCorner.begin();
				iterator += initCorners[1];	iterator += initCorners[2];
				iterator += initCorners[3];	iterator += initCorners[0];
				iterator += initCorners[7];	iterator += initCorners[4];
				iterator += initCorners[5];	iterator = initCorners[6];
			}

			//  y rotation
			for(i = 0; i < actual->angles.y; i=i+90 )
			{
				//  make tmp copy	... nejde protoze on dela hlubokou kopii (nekopiruje pointery, ale hodnoty)
				// plati i pro vyse a nize
				//  nelze totiz ani
				// for(int j = 0; j < 8; j++)  initCorners[j] = finalCorner[j];
				//  on nezkopiruje pointery, ale udela hlubokou kopii
//				std::copy( finalCorner.begin(), finalCorner.end(), initCorners.begin() );
				for(int j = 0; j < 8; j++)  initCorners.storage[j] = finalCorner.storage[j];

				RList< Vector<> >::iterator iterator = finalCorner.begin();
				iterator += initCorners[7];	iterator += initCorners[6];
				iterator += initCorners[1];	iterator += initCorners[0];
				iterator += initCorners[3];	iterator += initCorners[2];
				iterator += initCorners[5];	iterator = initCorners[4];
			}

			//  z rotation
			for(i = 0; i < actual->angles.z; i=i+90 )
			{
				//  make tmp copy
//				std::copy( finalCorner.begin(), finalCorner.end(), initCorners.begin() );
				for(int j = 0; j < 8; j++)  initCorners.storage[j] = finalCorner.storage[j];

				RList< Vector<> >::iterator iterator = finalCorner.begin();
				iterator += initCorners[1];	iterator += initCorners[6];
				iterator += initCorners[5];	iterator += initCorners[2];
				iterator += initCorners[3];	iterator += initCorners[4];
				iterator += initCorners[7];	iterator = initCorners[0];
			}

			//
			//  order corners according to Hilbert's curve

			//  space partitioning according to Hilbert's curve
			//

			for(i = 0; i < 8; i++)
			{
				//  we get rightCorner by adding bisect to every component of corner vector
				Vector<> rightCorner = finalCorner[i];
				std::for_each( rightCorner.begin(), rightCorner.end(), AddValue<> (bisect) );

				int count = 0;
				int pos = 0;	//  nested, it don't interfere with pos outside this scope

				Node * node = NULL;

				//  find out how many leaves belong to this partition, so iterate over all leaves
				for(int k = 0; k < actual->leaves.getCount(); k++)
				{
					//  look at operators >= and < ! they determinate which leave belong to this partition
					if ( (actual->leaves[k].position >= finalCorner[i]) && (actual->leaves[k].position < rightCorner) )
					{  //  leave leaves[k] belongs to partion given by corner vector
						if ( count == 0 )
						{
							pos = k;
							count++;
						}
						else
						{
							if ( count == 1)		//  true only ones
							//	at least two leaves -> create new node
							{
								Node::Angles angles;
								switch (i)	//  i identifies present corner of Hilbert's cube
								{
								case 0:
									angles.x = actual->angles.getXinc90();
									angles.y = actual->angles.getYdec90();
									angles.z = actual->angles.z;
									break;
								
								case 1:
								case 2:
									angles.x = actual->angles.getXdec90();
									angles.y = actual->angles.y;
									angles.z = actual->angles.getZinc90();
									break;
								
								case 3:
								case 4:
									angles.x = actual->angles.getXinc180();
									angles.y = actual->angles.y;
									angles.z = actual->angles.z;
									break;

								case 5:
								case 6:
									angles.x = actual->angles.getXdec90();
									angles.y = actual->angles.getYdec90();
									angles.z = actual->angles.z;
									break;

								case 7:
									angles.x = actual->angles.getXinc90();
									angles.y = actual->angles.getYinc90();
									angles.z = actual->angles.z;
									break;
								}

								node = new Node( finalCorner[i], angles );

								if ( mostLeftAncestors.getCount() == level+1 )
								{
									mostLeftAncestors.Add( new RList<Node>(100, false) );
								}

								actual->childList.Add ( node );	//  and finaly add this node between actual's children
								node->parent = actual;
								node->bisect = bisect;
								node->grouped = actual->grouped;
								node->childNo = actual->childList.getCount()-1;
								
								mostLeftAncestors[level+1].Add( node );			//  add to tree

								node->leaves.Add( &( actual->leaves[pos] ) );	//  add previous
							}

							node->leaves.Add( &( actual->leaves[k] ) );
							count++;	
						}
					} 
				}  //  end of cycling through collection of positions

				if (count == 1)	//  only one child node so that's leave
				{
					actual->childList.Add( &actual->leaves[pos] );
					if (!actual->grouped)  groupsList.Add( new Group( &actual->leaves[pos] ) );  //  WD99, only one leave in group
					delete &(finalCorner[i]);	//  Hilbert
				}
				else
				if ( (count > 1) && (count <= 32) && (!actual->grouped) )	//  WD99
				{
					node->grouped = true;
					groupsList.Add( new Group(node) ); 
				}
				else
				if (count == 0)					//  Hilbert's corner
				{
					delete &(finalCorner[i]);
				}
			
			}  //  end of cycling through corners
			//  
			//	space partitioning according to Hilbert's curve

			
			//  only one child -> leads to long tails, don't want this (happens when all leaves belongs to only one of eight sub-cubes)
			if ( ( actual->childList.getCount() == 1 ) && ( actual != root ) && ( actual->grouped != true) )
			{
				#ifdef _DEBUG
				std::cout << "Jedinacek" << std::endl;
				#endif		
				
				//  remove "actual" from "parents" "childList"
				actual->parent->childList.own(false);
				actual->parent->childList.removeAt(actual->childNo);
				actual->parent->childList.own(true);
				//  update position of all "children" from "parent's" childList
				for( int i = actual->childNo; i < actual->parent->childList.getCount(); i++)
				{
					if ( typeid(actual->parent->childList[i]) == typeid(Node) )
					{
						static_cast<Node&>(actual->parent->childList[i]).childNo -= 1;
					}
				}
				//  add the only actual's child into "parent's" "childList" and position in this list save in childs "member value"
				static_cast<Node&>(actual->childList[0]).childNo = actual->parent->childList.Add( &( actual->childList[0] ) );
				//  set child's new "parent"
				static_cast<Node&>(actual->childList[0]).parent = actual->parent;
				//  set actual's "childList" so, that is not deleted
				actual->childList.own(false);
				//  remove actual from list and from memory
				mostLeftAncestors[level].own(true);
				mostLeftAncestors[level].removeAt(pos);
				mostLeftAncestors[level].own(false);
				//  last node in this level, perform some speciality
				if ( mostLeftAncestors[level].getCount() == 0 )
				{
					for(int i = 0; i < mostLeftAncestors[level+1].getCount(); i++)
					{
						mostLeftAncestors[level].Add ( &( mostLeftAncestors[level+1][i] ) );
					}
					mostLeftAncestors.removeAt(level+1);
					pos = 0;
					bisect /= 2;

				}
	
			}
			else
			{
				pos++;
			}

		}	//  end while( pos < mostLeftAncestors[level].getCount() )

		#ifdef _DEBUG
		std::cout << ", pocet pozic:" << pos << std::endl;
		#endif


		bisect /= 2;
		level++;
	
	}	//  end while( level < mostLeftAncestors.getCount() )

	#ifdef _DEBUG
	std::cout << "Phase 1 END" << std::endl;
	#endif

	//
	//  end phase 1, tree construction done

	//  Phase 2: computing vector "center of mass" and scalar "mass"
	//

	//  for all levels excluding root
	for(level = mostLeftAncestors.getCount()-1; level > 0; level--)
	{
		
		for( int pos = 0; pos < mostLeftAncestors[level].getCount(); pos++)
		{
			actual = &( mostLeftAncestors[level][pos] );
			Vector<> numerator = 0;

			for(int i = 0; i < actual->childList.getCount(); i++)
			{
				actual->mass += actual->childList[i].mass;
				//  multiplication of vector with number
				numerator += actual->childList[i].mass * actual->childList[i].position;
			}
			//  quotient of vector by number
			actual->position = numerator / actual->mass;	//  "center of mass" (geometric quantity)
		}

	}

	#ifdef _DEBUG
	std::cout << "Phase 2 END" << std::endl;
	#endif

	//
	//  end phase 2, computing "center of mass" and "mass"

	//  Phase 3: force evaluation
	//

	for(i = 0; i < groupsList.getCount(); i++)	
	{
		int pos = 0;
		Node * cell = root;

		std::stack<Node*> openedStack;
		std::stack<int> positionStack;

		Group * group = &( groupsList[i] );
		
		bool done;

		#ifdef _DEBUG
		std::cout << "Evaluating group/particle no." << i << std::endl;
		#endif
		
		do
		{
			while ( pos < cell->childList.getCount() )	//  walk through all children
			{
				Vector<> r = cell->childList[pos].position - group->element->position;
				double d = r.length();

				if ( (void*)&cell->childList[pos] == (void*)group->element ) { pos++; }	//  group (leave/node) don't act on themselves (and 1/(r=0)=-1.#IND00000 )	A NEMEL BYCH KONTROLOVAT r? (nekdy se castice mohou dostat tak blizko a pak by to zkolabovalo!!!)
				else
				if ( (typeid( cell->childList[pos] ) == typeid( Leave )) || (d * FI >= dynamic_cast<Node&>(cell->childList[pos]).bisect) )
				{
					real sphere = 3 * group->getCellSize() * sqrt(3.) / 2.;
					if ( d > sphere )
					{
						group->acceleration += kappa * r * cell->childList[pos].mass / pow( d+epsilon, 3 );  //  far list
//						group->acceleration += r * cell->childList[pos].mass / pow( d+epsilon, 3 );  //  far list
						cell->childList[pos].load++;	//  costzones
					}
					else
					{	//  group is a leave or node that is too close to group
						group->nearList.Add( & (cell->childList[pos]) );
					}
					
					pos++;
				}
				else
				{
					openedStack.push( cell );
					positionStack.push( pos );

					cell = static_cast<Node*>( &(cell->childList[pos]) );
					pos = 0;
				}
			}
			
			if ( !openedStack.empty() )
			{
				cell = openedStack.top(); openedStack.pop();
				pos = positionStack.top(); positionStack.pop();
				pos++;
				done = false;
			}
			else
			{
				done = true;
			}

		} while (!done);

		//  finish acceleration computing

		if ( typeid( *group->element ) == typeid(Leave) )	//  only one particle in cell
		{
			//  add far forces acceleration
			//
			dynamic_cast<Leave*>(group->element)->acceleration += group->acceleration;

			//  compute near forces
			//

			for(int j = 0; j < group->nearList.getCount(); j++)
			{
				Vector<> r = group->nearList[j].position - group->element->position;
				dynamic_cast<Leave*>(group->element)->acceleration += kappa * r * group->nearList[j].mass / pow( r.length() + epsilon, 3 );
//				dynamic_cast<Leave*>(group->element)->acceleration += r * group->nearList[j].mass / pow( r.length() + epsilon, 3 );
				group->nearList[j].load++;	//  costzones
			}

		}
		else	//  more particles in cell
		{
			for(int k = 0; k < dynamic_cast<Node*>(group->element)->leaves.getCount(); k++)
			{
				//  for leave "k"			
				
				//  add far forces
				dynamic_cast<Node*>(group->element)->leaves[k].acceleration += group->acceleration;

				int j;

				//  add near forces
				for(j = 0; j < group->nearList.getCount(); j++)
				{
					Vector<> r = group->nearList[j].position - dynamic_cast<Node*>(group->element)->leaves[k].position;
					dynamic_cast<Node*>(group->element)->leaves[k].acceleration += kappa * r * group->nearList[j].mass / pow( r.length() + epsilon, 3 );
//					dynamic_cast<Node*>(group->element)->leaves[k].acceleration += r * group->nearList[j].mass / pow( r.length() + epsilon, 3 );
					group->nearList[j].load++;	//  costzones
				}
				//  add acceleration from leaves in grouping cell
				for(j = 0; j < dynamic_cast<Node*>(group->element)->leaves.getCount(); j++)
				{
					if (k != j)
					{
						Vector<> r = dynamic_cast<Node*>(group->element)->leaves[j].position - dynamic_cast<Node*>(group->element)->leaves[k].position;
						dynamic_cast<Node*>(group->element)->leaves[k].acceleration += kappa * r * dynamic_cast<Node*>(group->element)->leaves[j].mass / pow( r.length() + epsilon, 3 );
//						dynamic_cast<Node*>(group->element)->leaves[k].acceleration += r * dynamic_cast<Node*>(group->element)->leaves[j].mass / pow( r.length() + epsilon, 3 );
						dynamic_cast<Node*>(group->element)->leaves[j].load++;
					}
				}
			}
			
		}
		
	} //  evaluate all leaves

	#ifdef _DEBUG
	std::cout << "Phase 3 END" << std::endl;
	#endif

	//
	//  end phase 3: force evaluation

	//  integrate, leap-frog

//	real dt = 86400/3.344e13;
	real dt = 1;
	for( i = 0; i < leavesList.getCount(); i++ )
	{
		leavesList[i].velocity += leavesList[i].acceleration * dt;
		leavesList[i].position += leavesList[i].velocity * dt;
	}
	
	//  prepare acceleration for next step
	for(i = 0; i < leavesList.getCount(); i++)
	{
		leavesList[i].acceleration = 0;
	}

	//  propagate costzones load info to leaves
	for(level = 0; level < mostLeftAncestors.getCount(); level++)
	{
		
		for(int pos = 0; pos < mostLeftAncestors[level].getCount(); pos++)
		{
			for(int i = 0; i <  mostLeftAncestors[level][pos].childList.getCount(); i++)
			{
				mostLeftAncestors[level][pos].childList[i].load += mostLeftAncestors[level][pos].load / mostLeftAncestors[level][pos].childList.getCount();
			}
		}
	}

	delete root;
}

void Tree::initCubeCorners( Vector<> & superiorCorner, RList< Vector<> > & initCorners, real & bisect)
{
	static int i = -1;
	Vector<> corner = superiorCorner;	//  calls "copy" constructor & performs "deep" copy

	i++;	//  actual component
	for(int j = 0; j < 2; j++)	//  bisection; every component divided into two parts
	{
		corner[i] += j*bisect;
		if (i+1 < 3)	
		{	//  still not at the end of recursion, go deeper :-))
			initCubeCorners(corner, initCorners, bisect);
		}
		else
		{	//  one of final solutions, final coordinates of corner
			initCorners.Add( new Vector<> );
			initCorners [ initCorners.getCount()-1 ] = corner;
		}  //  end of code for one of final solutions
	
	}  // end of bisection cycle

	i--;
}

double Group::getCellSize()
{
	if ( typeid(*element) == typeid(Node) )
	{				
		return dynamic_cast<Node*>(element)->bisect;
	}
	else
	{
		return 1E10;	//  group contain only one leave
	}
}
