#include "xml.h"

#include <comdef.h>		//  _bstr_t, _variant_t classes
#include "math/math_ext.h"

using namespace JakubSchwarzmeier::Data;

XmlAttributeCollection::XmlAttributeCollection(IXMLDOMNamedNodeMap * pAttributes, IXMLDOMDocument2 * pXMLDoc)
{
	m_pAttributes = pAttributes;
	m_pXMLDoc = pXMLDoc;
}

XmlAttributeCollection::~XmlAttributeCollection()
{
	m_pAttributes->Release();
}

XmlNode XmlAttributeCollection::GetNamedItem(TCHAR * szAttribute)
{
	IXMLDOMNode * pIXMLDOMNode;

	m_pAttributes->getNamedItem(_bstr_t(szAttribute), &pIXMLDOMNode);

	return XmlNode(pIXMLDOMNode, m_pXMLDoc);
}

XmlNode XmlAttributeCollection::GetItem(long index)
{
	IXMLDOMNode * pItem;
	m_pAttributes->get_item(index, &pItem);

	return XmlNode(pItem, m_pXMLDoc);
}

long XmlAttributeCollection::GetCount()
{
	long length;
	m_pAttributes->get_length(&length);

	return length;
}

XmlAttributeCollection & XmlAttributeCollection::operator=(XmlAttributeCollection original)
{
	original.m_pAttributes->AddRef();
	m_pAttributes = original.m_pAttributes;

	return *this;
}

XmlNode::XmlNode( IXMLDOMNode * pNode, IXMLDOMDocument2 * pXMLDoc)
{
	m_pNode = pNode;
	m_pXMLDoc = pXMLDoc;
}

XmlNode::~XmlNode()
{
	//  XmlNode is base class for other classes, so this destructor is called each time
	if (m_pNode == NULL)	return;
	if (m_pNode == m_pXMLDoc)
	{
		m_pXMLDoc->Release();
		CoUninitialize();
	}
	else
	{
		if (m_pNode!=NULL)	m_pNode->Release();
	}
}

XmlNode & XmlNode::operator=(XmlNode original)
{
	original.m_pNode->AddRef();
	m_pNode = original.m_pNode;

	return *this;
}

TCHAR * XmlNode::GetText()
{
	BSTR bstr;
//	bstr = SysAllocStringLen(NULL, 512);

	m_pNode->get_text( &bstr );
#ifndef UNICODE
	lstrcpy( m_szBuffer, _com_util::ConvertBSTRToString(bstr) );
#else
	MultiByteToWideChar(
		CP_ACP, 
		0, 
		_com_util::ConvertBSTRToString(bstr),
		-1, 
		m_szBuffer,
		sizeof(m_szBuffer) 
		);
#endif
	SysFreeString(bstr);

	return m_szBuffer;
}

TCHAR * XmlNode::GetNodeValue()
{
	_variant_t var;

	#if _MSC_VER == 1300
	m_pNode->get_nodeValue( var.GetAddress() );
	#else
	var.Clear();
	m_pNode->get_nodeValue( static_cast<VARIANT*>(&var) );
	#endif // _MSC_VER == 1300
	_bstr_t bstr( &var );
	lstrcpy(m_szBuffer, bstr);

	return m_szBuffer;
}

XmlAttributeCollection XmlNode::GetAttributes()
{
	IXMLDOMNamedNodeMap * pAttributes;

	m_pNode->get_attributes( &pAttributes );

	return XmlAttributeCollection( pAttributes, m_pXMLDoc );
}

XmlNodeList XmlNode::SelectNodes( TCHAR * szExpression )
{
	IXMLDOMNodeList * pList;

	m_pNode->selectNodes( _bstr_t(szExpression), &pList );

	return XmlNodeList( pList, m_pXMLDoc );
}

XmlNode XmlNode::SelectSingleNode( TCHAR * szExpression )
{
	IXMLDOMNode * pNode;

	m_pNode->selectSingleNode( _bstr_t(szExpression), &pNode );

	return XmlNode( pNode, m_pXMLDoc );
}


XmlNodeList::XmlNodeList(IXMLDOMNodeList * pNodes, IXMLDOMDocument2 * pXMLDoc)
{
	m_pNodes = pNodes;
	m_pXMLDoc = pXMLDoc;
}

XmlNodeList::~XmlNodeList()
{
	m_pNodes->Release();
}

int XmlNodeList::GetLength()
{
	long length;
	m_pNodes->get_length( &length );
	
	return length;
}

XmlNode XmlNodeList::GetItem(int position)
{
	long index = position;

	IXMLDOMNode * pNode;
	m_pNodes->get_item( index, &pNode );

	return XmlNode( pNode, m_pXMLDoc );
}

XmlElement::XmlElement()
{
	m_pElement = NULL;
	m_pXMLDoc = NULL;
}

XmlElement::XmlElement( IXMLDOMElement * pElement, IXMLDOMDocument2 * pXMLDoc )
{
	m_pElement = pElement;
	m_pXMLDoc = pXMLDoc;
}

XmlElement::~XmlElement()
{
	if (m_pElement != NULL)
		m_pElement->Release();
}

void XmlElement::get(IXMLDOMNode ** returnPNode)
{
	*returnPNode = m_pElement;
}

void XmlElement::SetAttribute( TCHAR * name, LPCTSTR value )
{
	IXMLDOMAttribute * pa;
	m_pXMLDoc->createAttribute( _bstr_t(name), &pa );

	//  set attribute's value
	pa->put_nodeValue( _variant_t(value) );

	//  bind attribute to node
	m_pElement->setAttributeNode( pa, NULL );

	pa->Release();
}

void XmlElement::SetAttribute( TCHAR * name, double dValue )
{
	IXMLDOMAttribute * pa;
	m_pXMLDoc->createAttribute( _bstr_t(name), &pa );

	TCHAR szDoubleValue[1024];
	//  set attribute's value
	pa->put_nodeValue( _variant_t( _dtot(dValue, szDoubleValue) ) );

	//  bind attribute to node
	m_pElement->setAttributeNode( pa, NULL );

	pa->Release();
}

void XmlElement::AppendChild( XmlNode & childElement )
{
	IXMLDOMProcessingInstruction * pChild;
	childElement.get( (IXMLDOMNode**)&pChild );

	IXMLDOMNode * pNode;
    m_pElement->appendChild( pChild, &pNode );
    pNode->Release();
}


//  used when returning value through stack
XmlElement & XmlElement::operator=(XmlElement original)
{
	original.m_pElement->AddRef();
	m_pElement = original.m_pElement;
	m_pXMLDoc = original.m_pXMLDoc;

	return *this;
}

// *************

XmlProcessingInstruction::XmlProcessingInstruction( IXMLDOMProcessingInstruction * pi )
{
	m_pi = pi;
}

XmlProcessingInstruction::~XmlProcessingInstruction()
{
    m_pi->Release();
}

XmlProcessingInstruction & XmlProcessingInstruction::operator=(XmlProcessingInstruction original)
{
	original.m_pi->AddRef();
	m_pi = original.m_pi;

	return *this;
}

void XmlProcessingInstruction::get(IXMLDOMNode ** returnPNode)
{
	*returnPNode = m_pi;
}

// *************


XmlText::XmlText()
{
	m_pText = NULL;
	m_pXMLDoc = NULL;
}

XmlText::XmlText( IXMLDOMText * pText, IXMLDOMDocument2 * pXMLDoc )
{
	m_pText = pText;
	m_pXMLDoc = pXMLDoc;
}

XmlText::~XmlText()
{
	if (m_pText != NULL)
		m_pText->Release();
}

void XmlText::get(IXMLDOMNode ** returnPNode)
{
	*returnPNode = m_pText;
}

XmlText & XmlText::operator=(XmlText original)
{
	original.m_pText->AddRef();
	m_pText = original.m_pText;
	m_pXMLDoc = original.m_pXMLDoc;

	return *this;
}

// *********

XmlDocument::XmlDocument()
{
	//  initialize COM
	CoInitializeEx( NULL, COINIT_APARTMENTTHREADED );

	//  create a DOM document instance
	CoCreateInstance( __uuidof(DOMDocument), 
		NULL, 
		CLSCTX_SERVER,
		__uuidof(IXMLDOMDocument2),
		(void**)&m_pXMLDoc
		);
	
	m_pNode = m_pXMLDoc;
}

XmlProcessingInstruction XmlDocument::CreateProcessingInstruction( TCHAR * target, TCHAR * data )
{
	IXMLDOMProcessingInstruction * pi;
	m_pXMLDoc->createProcessingInstruction( _bstr_t(target), _bstr_t(data), &pi );
	return XmlProcessingInstruction(pi);
}

void XmlDocument::AppendChild( XmlNode & childElement )
{
	IXMLDOMProcessingInstruction * pChild = NULL;
	childElement.get( (IXMLDOMNode**)&pChild );

	IXMLDOMNode * pNode;
    m_pXMLDoc->appendChild( pChild, &pNode );
    pNode->Release();
}


XmlElement XmlDocument::CreateElement( TCHAR * newElement )
{
	IXMLDOMElement * pElement;
	m_pXMLDoc->createElement(_bstr_t(newElement), &pElement);
	
	return XmlElement( pElement, m_pXMLDoc );
}

XmlText XmlDocument::CreateTextNode( TCHAR * newTextNode )
{
	IXMLDOMText * pText;
	m_pXMLDoc->createTextNode(_bstr_t(newTextNode), &pText );
	
	return XmlText( pText, m_pXMLDoc );
}

XmlText XmlDocument::CreateTextNode( int newValue )
{
	TCHAR szTextNode[128];
	_itot( newValue, szTextNode, 10 );
	
	IXMLDOMText * pText;
	m_pXMLDoc->createTextNode(_bstr_t(szTextNode), &pText );
	
	return XmlText( pText, m_pXMLDoc );
}

XmlText XmlDocument::CreateTextNode( double newValue )
{
	TCHAR szTextNode[128];
	_dtot( newValue, szTextNode );
	
	IXMLDOMText * pText;
	m_pXMLDoc->createTextNode(_bstr_t(szTextNode), &pText );
	
	return XmlText( pText, m_pXMLDoc );
}

XmlDocument::~XmlDocument()
{
}

void XmlDocument::Save(LPCTSTR filename)
{
	m_pXMLDoc->save ( _variant_t(filename) );
}

void XmlDocument::Load(LPCTSTR filename)
{
	VARIANT_BOOL vBool;
	m_pXMLDoc->load ( _variant_t(filename), &vBool );
}


void XmlDocument::DoHumanMarkup( XmlElement & pRoot )
{
	IXMLDOMNode * pNode;
	pRoot.get( &pNode );
	addHumanMarkup( pNode, 0 );
}

bool XmlDocument::addHumanMarkup(IXMLDOMNode * pElement, int depth)
{
	//
	 BSTR bstr;
	 pElement->get_nodeName(&bstr);
	
	 if ( lstrcmpW( bstr, L"#text" ) == 0 )  {
		SysFreeString(bstr);
		 return false;
	 }

	SysFreeString(bstr);

	IXMLDOMText * pText;
	IXMLDOMNode * pTmp, * pTmp2;

	//  zjisti jestli child neni #text, protoze v tom pripade misto nevkladame
	VARIANT_BOOL vBool;
	pElement->hasChildNodes( &vBool );
	if (vBool)
	{
		pElement->get_firstChild( &pTmp2 );
		 pTmp2->get_nodeName(&bstr);

		 if ( lstrcmpW( bstr, L"#text" ) != 0 )  
		 {

			_bstr_t bsWhiteSpace1("\n");
			for(int i = 0; i < depth; i++)  bsWhiteSpace1 += "\t";
			m_pXMLDoc->createTextNode( bsWhiteSpace1, &pText );
			pElement->appendChild(pText, &pTmp );

			pText->Release();
			pTmp->Release();
		 }

		SysFreeString(bstr);
		pTmp2->Release();
	}

	 IXMLDOMNodeList * pNodes;
	long length;

	pElement->get_childNodes( &pNodes );
	pNodes->get_length(&length);

	for( long l = 0; l < length-1; l++ )		//  -1 protoze je tam pridanej child na konec
	{
		IXMLDOMNode * pNode;
		pNodes->get_item(l, &pNode);
		
		if ( addHumanMarkup( pNode, depth+1 ) == true )   l++;
		pNodes->get_length(&length);	//  musim aktualizovat, protoze mohl nabyt

		pNode->Release();
			
	}

	if (depth == 0)  return false;

	//  add formating characters
	_bstr_t bsWhiteSpace("\n");
	for(int i = 0; i < depth; i++)  bsWhiteSpace += "\t";
	m_pXMLDoc->createTextNode( bsWhiteSpace, &pText );

	//  get parent of pElement
	pElement->get_parentNode( &pTmp2 );
	pTmp2->insertBefore(pText, _variant_t(pElement), &pTmp );

	pTmp2->Release();
	pText->Release();
	pTmp->Release();
	return true;
}