//  matrix.h: declaration & implementation of the Matrix and Vector classes
//

#ifndef MATRIX_H
#define MATRIX_H 

#include "math_ext.h"
#include <stdexcept>
#include <iostream>
#include <iomanip>  //  setprecision

typedef double real;

namespace JakubSchwarzmeier
{
	namespace Math
	{

		//************************************************************
		//
		//  Matrix	class
		//

		template<typename Type>
		class Matrix
		{
		protected:

			int size;	//  matrix size; can be computed at any time, but this is faster
			int rows, columns;	//  number of rows & columns in matrix
			bool classStorage;	//  is (array of values) created on heap?

			#ifdef _DEBUG
			//  for internal check purposes only
			void check_matrices_sizes_equality(const Matrix& first, const Matrix& second)
			{
				if ( first.size != second.size )
				{
					throw std::runtime_error("matrices sizes aren't equal");
				}
			}
			#endif // _DEBUG


		public:

			Type * values;	//  points to values of matrix (it may be class storage or user supplied storage)

			//  default constructor
			Matrix( int rows = 16, int columns = 16)
			{
				//  create "internal" storage for values on heap
				values = new Type [rows*columns];
				classStorage = true;

				this->rows = rows;
				this->columns = columns;
				size = rows*columns;
			}

			//  as third argument takes user-supplied array of values
			Matrix( int rows, int columns, Type values[] )
			{
				//  "external" values are used (and modified)
				this->values = values;
				classStorage = false;

				this->rows = rows;
				this->columns = columns;
				size = rows*columns;
			}

			//  "copy constructor", IS inherited by derived classes
			//  eg.
			//  Vector<> pos1;
			//  Vector<> pos2 = pos1;
			Matrix( const Matrix & original )
			{
				//  this-> points to object created on heap, but member objects are
				//  not created and initilized yet; this is a constructor!

				//  create "internal" storage for values on heap
				values = new Type [original.rows*original.columns];
				classStorage = true;

				this->rows = original.rows;
				this->columns = original.columns;
				size = rows*columns;

				operator=(original);  // calls Matrix assigment operator (it's the same code)
			}

			//  "virtual" to call appropriate destructor when using "late-binding", polymorphism
			virtual ~Matrix( )
			{
				if ( classStorage ) delete [] values;
			}

			// assignment operators are not inherited by derived classes
			Matrix& operator=( const Matrix& right_side )
			{	// deep copy
				#ifdef _DEBUG
				check_matrices_sizes_equality(*this, right_side);
				#endif

				if ( this == &right_side ) return *this;
				Assign(*this, right_side);
				return *this;
			}

			Type length() const
			{
				Type result = 0;

				for(int i = 0; i < size; i++)
                    result += pow( values[i], 2 );

				return sqrt(result);
			}

			void Assign( Matrix& left, const Matrix& right )
			{
				#ifdef _DEBUG
				check_matrices_sizes_equality(left, right);
				#endif

				for ( int i = 0; i < right.size; i++ ) left[i] = right[i];
			}
			
			void Transpose()
			{
				double (*matrix)[4] = (double(*)[4])values;

				for( int r = 0; r < rows; r++)
					for( int s = r + 1; s < columns; s++ )
					{
						int tValue = matrix[r][s];
						matrix[r][s] = matrix[s][r];
						matrix[s][r] = tValue;
					}
			}

			double* GetMatrix()
			{
				return values;
			}

			void LoadIdentity()
			{
				for( int i = 0; i < 16; i++ )  values[i] = 0;
				for( i = 0; i < 16; i+=(columns+1) )  values[i] = 1;
			}


			//  compute inversion matrix -- Gauss eliminating method (GEM)
			void Inverse()
			{
				static const int max = 4;
				static const double MACHEPS = 5.0e-13;

				typedef double pole[max][max];

				double (*dat)[4] = (double(*)[4])values;

				pole A, X;
				int i, j, k, mx;
				double data, det;

				for( i = 0; i < max; i++ )
					for( j = 0; j < max; j++ )
					{
						A[i][j] = dat[i][j];
						if (i == j)  X[i][j] = 1.0;
							else X[i][j] = 0.0;
					}
				
				//  GEM

				i = 0;
				det = 1;
				
				do
				{
					if (i+1 < max)
					{
						//  maximum element
						mx = i;
						data = fabs( A[i][i] );
						k = i + 1;
						do
						{
							if ( data < fabs( A[k][i]) )
							{
								mx = k;
								data = fabs( A[k][i] );
							}
							k = k + 1;
						} while (k+1 <= max);
						
						//  row exchange
						if (mx != i)
						{
							for( j = 0; j < max; j++ )
							{
								data = A[mx][j];
								A[mx][j] = A[i][j];
								A[i][j] = data;

								data = X[mx][j];
								X[mx][j] = X[i][j];
								X[i][j] = data;
							}
							det = -det;
						}

					}
					
					//  singularity
					if ( fabs( A[i][i] ) > MACHEPS )
					{
						det = det * A[i][i];
						data = A[i][i];
						for( j = 0; j < max; j++ )
						{
							A[i][j] = A[i][j] / data;
							X[i][j] = X[i][j] / data;
						}

						for( j = 0; j < max; j++ )
						{
							data = A[j][i];
							if (j!=i)
							{
								for( k = 0; k < max; k++ )
								{
									A[j][k] = A[j][k] - data * A[i][k];
									X[j][k] = X[j][k] - data * X[i][k];
								}
							}
						}
					}
					else
					{
						//  A is singular matrix
						det = 0;
					}
					
					i = i + 1;

				} while (i+1 <= max && det != 0);

				if (det == 0)  throw std::runtime_error("determinant = 0");

				for(i = 0; i < 4; i++)
					for(j = 0; j < 4; j++)
						values[i*4+j] = X[i][j];
			}

			void Clear()
			{
				for(int i=0; i < size; i++)  values[i] = 0;
			}

			Type& operator[]( int index ) const
			{
				#ifdef _DEBUG
				if ( index >= size )
				{
					throw std::runtime_error("index out of range in []");
				}
				#endif

				return values[index];
			}

			Matrix operator+(const Matrix & other)
			{
				#ifdef _DEBUG
				check_matrices_sizes_equality(*this, other);
				#endif

				Matrix result(rows, columns);
				for (int i=0; i < size; i++) result[i] = values[i] + other.values[i];
		  		
				return result;
			}

			Matrix operator-(Matrix & other)
			{
				#ifdef _DEBUG
				check_matrices_sizes_equality(*this, other);
				#endif

				Matrix result(rows, columns);
				for (int i=0; i < size; i++) result[i] = values[i] - other.values[i];
		  		
				return result;
			}

			// Global
			friend std::ostream& operator<<(std::ostream& os, const Matrix<Type> & mat)
			{
   				os << std::setprecision(6);
				os.setf(std::ios::scientific);
				os.setf(std::ios::left, std::ios::adjustfield);

				for (int i = 0; i < mat.rows; i++)
				{
      				for (int j = 0; j < mat.columns; j++)
					{
						os.width(13);
						os << mat.values[mat.columns * i + j];
					}
						os << std::endl;
				}

				os.unsetf(std::ios::scientific | std::ios::left | std::ios::adjustfield);
				return os;
			}

			bool operator>( const Matrix & other )
			{
				#ifdef _DEBUG
				check_matrices_sizes_equality( *this, other );
				#endif
				
				for(int i = 0; i < this->size; i++)
				{
					if ( this->values[i] > other.values[i] )
					{
						continue;
					}
					else
					{
						return false;	//  loop never gets here if condition is satisfied
					}
				}

				return true;
			}

			bool operator<( const Matrix & other ) 
			{
				#ifdef _DEBUG
				check_matrices_sizes_equality( *this, other );
				#endif

				for(int i = 0; i < this->size; i++)
				{
					if ( this->values[i] < other.values[i] )
					{
						continue;
					}
					else
					{
						return false;	//  loop never gets here if condition is satisfied
					}
				}

				return true;
			}

			bool operator>=( const Matrix & other ) 
			{
				#ifdef _DEBUG
				check_matrices_sizes_equality( *this, other );
				#endif
				
				for(int i = 0; i < this->size; i++)
				{
					if ( this->values[i] >= other.values[i] )
					{
						continue;
					}
					else
					{
						return false;	//  loop never gets here if condition is satisfied
					}
				}

				return true;
			}

			Matrix & operator+=( const Matrix& matrix )
			{
				#ifdef _DEBUG
				check_matrices_sizes_equality( *this, matrix );
				#endif

				for(int i = 0; i < size; i++)
					values[i] += matrix.values[i];

				return *this;
			}

			
			// matrix = scalar * matrix;
			const friend Matrix operator*(  const Type & scalar,  const Matrix & mat )
			{
				Matrix result( mat.rows, mat.columns );
				for(int i = 0; i < mat.size; i++)
					result.values[i] = scalar * mat.values[i];
				return result;
			}

			// matrix = matrix * scalar;
			const friend Matrix operator*(  const Matrix & mat, const Type & scalar )
			{
				Matrix result( mat.rows, mat.columns );
				for(int i = 0; i < mat.size; i++)
					result.values[i] = scalar * mat.values[i];
				return result;
			}

			// matrix = matrix / scalar;
			const friend Matrix operator/( const Matrix & mat, const Type & scalar )
			{
				Matrix result( mat.rows, mat.columns );
				for(int i = 0; i < mat.size; i++)
					result.values[i] = mat.values[i] / scalar;
				return result;
			}

			// matrix = project * acton;	zatim funguje jenom pro stejne matice 4x4
			const friend Matrix operator*( const Matrix & project,  const Matrix & acton )
			{
				Matrix result( project.rows, project.columns );
				result.Clear();

				for(int k=0; k<16; k++)
					for(int i=0; i<4; i++)
						result[k]+=project.values[(k/4)*4+i] * acton.values[(k%4)+4*i];

				return result;
			}

			// matrix = matrix * matrix;	project must be a field of Type and of size acton.size
			const friend Matrix operator*(  const Type project[],  const Matrix & acton )
			{
				Matrix result( acton.rows, acton.columns );

				for(int k=0; k<16; k++)
					for(int i=0; i<4; i++)
						result[k]+=project[(k/4)*4+i] * acton.values[(k%4)+4*i];

				return result;
			}


			// je toto vubec treba ?
			class iterator;  //  declaration required
			friend class iterator;  //  make it friend
			// je toto vubec treba ?

			class iterator  //  now define it
			{
			private:

				Matrix & matrix;
				int pos;

			public:

				iterator(Matrix& n_matrix, int n_pos): matrix(n_matrix), pos(n_pos) {}

				Type& operator++() //  Prefix
				{
					return matrix.values[++pos];	 //  increment BEFORE evaluation of expression
				}

				bool operator!=(const iterator& rv) const
				{
                    return pos != rv.pos;
				}

				Type& operator*() const { return matrix.values[pos]; }

			};  //  class iterator

			iterator begin() { return iterator(*this, 0); }
			iterator end() { return iterator(*this, this->size); }

		};  //  class Matrix

		// The function object adds an element with a valueToAdd
		template <typename Type = real>
		class AddValue
		{
		private:

			Type valueToAdd;

		public:
			
			AddValue( const Type & n_valueToAdd ): valueToAdd(n_valueToAdd) {}
			void operator() ( Type & elem ) const
			{
				elem += valueToAdd;
			}
		}; //  AddValue


		//************************************************************
		//
		//  Vector	class
		//

		template<typename Type = real, const int dimensions = 3>
		class Vector: public Matrix<Type>
		{
		public:	

			Vector(): Matrix<Type>(dimensions, 1) {}
			
			Vector(Type comp1, Type comp2, Type comp3): Matrix<Type>(3, 1)
			{
				values[0] = comp1;
				values[1] = comp2;
				values[2] = comp3;
			}

			//  Vector<> position = 0;
			//  eg. in constructor declaration where i want to assign default value
			Vector( Type num ): Matrix<Type>(dimensions, 1)
			{
				operator=(num);
			}

			//  "copy constructor", IS inherited by derived classes
			//  eg.
			//  Vector<> pos1;
			//  Vector<> pos2 = pos1;
			Vector( const Matrix<Type> & original ): Matrix<Type> (original)
			{
			}

			//  assignment operators are not inherited by derived classes
			Vector& operator=( const Matrix<Type>& right_side )
			{	// deep-copy

				#ifdef _DEBUG
				check_matrices_sizes_equality(*this, right_side);
				#endif

				if ( this == &right_side ) return *this;
				Assign(*this, right_side);
				return *this;
			}

			virtual ~Vector() {}

			//  CVector<double> vec;
			//  vec = 1;
			//
			void operator=( Type num )
			{
				for (int i=0; i < size; i++) values[i] = num;
			}

		};  //  class Vector

	}  //  namespace Math

} //  namespace JakubSchwarzmeier

#endif //  MATRIX_H