/************************************************************
*
*  Copyright (c) 2002-2004 Jakub Schwarzmeier
*
************************************************************/

#include "bulge.h"
#include "math/math_ext.h"
#include "math/matrix.h"

//  static init
double BulgeModel::totalMass = 0;
double BulgeModel::rbmax = 0;
double BulgeModel::ab = 0;

double BulgeModel::rDistribution( double r )
{
	if (r < 0.001)		return 100000;	//  goes to infinity
	return 0.1 * (ab/r) * (1/pow(r+ab,3));	
}

double BulgeModel::plummer( double r )
{	//  plummer
	return pow( ( 1 + (pow(r, 2)/pow(ab, 2)) ), (-5/2) );
}


double BulgeModel::vCurve( double r )
{
	//  y=0.03+0.044* (x-0.04) * ( ( 0.091 + (x-0.04)^2))^(-3/2)
	return 0.03+(0.044* (r-0.04) * pow ( (0.091 + pow( (r-0.04), 2) ), (-3/2) ));
}

void BulgeModel::Initialize(double n_totalMass, double n_rbmax, double n_ab)
{
	totalMass = n_totalMass;
	rbmax = n_rbmax;
	ab = n_ab;
}


bool BulgeModel::Create(JakubSchwarzmeier::Data::RList< JakubSchwarzmeier::BarnesHut::Leave > & list, const int & nBodies)
{
	int bodies = 0;
	random(-100);		//  init real random number generator (Numerical Recipes in C)

	double probMax = ::FindMaximum( rDistribution, 0, rbmax, 1e-12 );
	probMax = rDistribution(probMax);

	while ( nBodies > bodies )
	{
		double r = 0.0 + random() * (rbmax - 0.0);	//  between 0 and rbmax
		double p = 0.0 + random() * (2 - 0.0);	//  between 0 and probMax (too slow), so 2

		//  find radius from radius distribution
		if ( p <= rDistribution(r) )		//  throw out points over distribution function
		{
			bodies++;

			//  Generate positions
			//

			//  randomized rotation of a cartesian system and then put only to x-axis

			double a =  0.0 + random() * (2*M_PI - 0.0);			
			double b =  0.0 + random() * (2*M_PI - 0.0);			
			double c =  0.0 + random() * (2*M_PI - 0.0);			

			Matrix<double> transform(4, 4);
			transform.LoadIdentity();
			transform[0]  = cos(a) * cos(b) - sin(a) * sin(b) * cos(c);
			transform[1]  = cos(a) * sin(b) + sin(a) * cos(b) * cos(c);
			transform[2]  = sin(a) * sin(c);
			transform[3]  = 0;
			transform[4]  = -sin(a) * cos(b) - cos(a) * sin(b) * cos(c);
			transform[5]  = -sin(a) * sin(b) + cos(a) * cos(b) * cos(c);
			transform[6]  = cos(a) * sin(c);
			transform[7]  = 0;
			transform[8]  = sin(b) * sin(c);
			transform[9]  = -cos(b) * sin(c);
			transform[10] = cos(c);
			transform[11] = 0;

			Matrix<double> pos(4,4);
			pos.LoadIdentity();
			pos[12] = r;
			pos[13] = 0;
			pos[14] = 0;

			Matrix<double> result(4,4);
			
			result = pos * transform;

			//

			Vector<> position;
			position[0] = result[12];
			position[1] = result[13];
			position[2] = result[14];

			double lambda = atan2(position[1], position[0]);
			lambda += M_PI;

			//  Generate velocity
			//
			
			double v = vCurve(r);		//  circular (tangential) velocity

			//  convert velocity to components
			double vx = -v * sin( lambda );
			double vy = v * cos( lambda );

			Vector<> velocity;
//			velocity[0] = vx;
//			velocity[1] = vy;
			velocity[0] = 0;
			velocity[1] = 0;
			velocity[2] = 0;

			list.Add( new JakubSchwarzmeier::BarnesHut::Leave(position, velocity, totalMass/nBodies) );
		}
	}

	return true;
}