/************************************************************
*
*  Copyright (c) 2002-2004 Jakub Schwarzmeier
*
************************************************************/

#include "darkhalo.h"
#include <math.h>
#include "math/math_ext.h"
#include <stdio.h>

//  static init
double DarkHaloModel::totalMass = 0;
double DarkHaloModel::rdhmax = 0;
double DarkHaloModel::adh = 0;
double DarkHaloModel::gamma = 0;

double DarkHaloModel::rDistribution( double r )
{
	double q;
	
	if (r < 1e-10)	 q = 20;	//  small r in denominator leads to big result that exp() can't handle and throws overflow exception
	else
	q = gamma / r;

	if (q > 20)  q = 20;
	
	double denom1 = sqrt(M_PI) * q * exp( pow(q,2) );
	double denom2 = 1 - erf(q);
	
	double alpha = 1 / ( 1 - (denom1*denom2) );

	double first = alpha/rdhmax;
	double second = exp( - ( pow(r, 2)/ pow(rdhmax, 2) ) );
	double third = pow(r, 2) + pow( gamma, 2);
	return 1000*first*(second/third);
}

double DarkHaloModel::vCurve( double r )
{
//  y=1.05*x*exp(x/55)*(2.4+(x+3)^2)^(-3/3.8)
	return  1.05 * r * exp( r/55 ) * pow( 2.4 + pow( r+3, 2 ), -3/3.8 );
}


void DarkHaloModel::Initialize(double n_totalMass, double n_rdhmax, double n_adh, double n_gamma)
{
	totalMass = n_totalMass;
	rdhmax = n_rdhmax;
	adh = n_adh;
	gamma = n_gamma;
}

bool DarkHaloModel::Create(JakubSchwarzmeier::Data::RList< JakubSchwarzmeier::BarnesHut::Leave > & list, const int & nBodies)
{

	int bodies = 0;
	random(-100);		//  init real random number generator (Numerical Recipes in C)

	double probMax = FindMaximum( rDistribution, 0.0, rdhmax, 1e-12 );
	probMax = rDistribution(probMax);

	while ( nBodies > bodies )
	{
		double r = 0.0 + random() * (rdhmax - 0.0);	//  between 0 and rdh 
		double p = 0.0 + random() * (0.0001 - 0.0);	//  between 0 and probMax (too slow), so 0.0001

		//  find radius from radius distribution
		if ( p <= rDistribution(r) )		//  throw out points over distribution function
		{
			bodies++;

			//  Generate positions
			//

			double a =  0.0 + random() * (2*M_PI - 0.0);			
			double b =  0.0 + random() * (2*M_PI - 0.0);			
			double c =  0.0 + random() * (2*M_PI - 0.0);			

			Matrix<double> transform(4, 4);
			transform.LoadIdentity();
			transform[0]  = cos(a) * cos(b) - sin(a) * sin(b) * cos(c);
			transform[1]  = cos(a) * sin(b) + sin(a) * cos(b) * cos(c);
			transform[2]  = sin(a) * sin(c);
			transform[3]  = 0;
			transform[4]  = -sin(a) * cos(b) - cos(a) * sin(b) * cos(c);
			transform[5]  = -sin(a) * sin(b) + cos(a) * cos(b) * cos(c);
			transform[6]  = cos(a) * sin(c);
			transform[7]  = 0;
			transform[8]  = sin(b) * sin(c);
			transform[9]  = -cos(b) * sin(c);
			transform[10] = cos(c);
			transform[11] = 0;

			Matrix<double> pos(4,4);
			pos.LoadIdentity();
			pos[12] = r;
			pos[13] = 0;
			pos[14] = 0;

			Matrix<double> result(4,4);
			
			result = pos * transform;

			//

			Vector<> position;
			position[0] = result[12];
			position[1] = result[13];
			position[2] = result[14];

			double lambda = atan2(position[1], position[0]);
			lambda += M_PI;

			//  Generate velocity
			//
			
			double v = vCurve(r);		//  circular (tangential) velocity

			//  convert velocity to components
			double vx = -v * sin( lambda );
			double vy = v * cos( lambda );

			Vector<> velocity;
//			velocity[0] = vx;
//			velocity[1] = vy;
			velocity[0] = 0;
			velocity[1] = 0;
			velocity[2] = 0;

			list.Add( new JakubSchwarzmeier::BarnesHut::Leave(position, velocity, totalMass/nBodies) );
		}
	}

	return true;
}