/************************************************************
*
*  Copyright (c) 2002-2004 Jakub Schwarzmeier
*
************************************************************/

#include "disk.h"
#include <math.h>
#include "math/math_ext.h"

//  static init
double DiskModel::totalMass = 0;
double DiskModel::rdmax = 0;
double DiskModel::ad = 0;
double DiskModel::z0 = 0;		//  maximum distance in direction perpendicular to disk plain
double DiskModel::toomreQ = 0;
double DiskModel::zVelocityDisp = 0;
double DiskModel::zignored = 2;

DiskModel::RADIAL_DISTRIBUTION DiskModel::rDistribution = Model1;
DiskModel::Z_DISTRIBUTION DiskModel::zDistribution = MyZDistr1;
DiskModel::VELOCITY_CURVE DiskModel::vCurve = Toomre;

double DiskModel::r = 0;

//  r distributions
//

// kuzmin
double DiskModel::model1( double r )
{
	return pow( ( 1 + (pow(r, 2)/pow(ad, 2)) ), (-3/2) );
}

// exponential
double DiskModel::model2( double r )
{
	return exp(-r/rdmax);
}

//
//  r distributions

//  z distributions
//

double DiskModel::zdistr1( double z )
{
	if (r >= rdmax)	return 0;		//  cannot be only rdmax, because when it's so close, function halts
	return pow( 1/cosh( z/pow( (1- (pow(r,2)/pow(rdmax,2))), 1) ) , 600*r );		//  original without multiplication with 600
}

//
//  z distributions

//  velocity curves
//
//  returns vcircular

double DiskModel::toomre( double r )
{	//  my curve

//	return 0.292 * r * pow( ( 1 + pow(r,2) ), (-3/5.3) );	//  0.292 scales to my units
	return 0.292 * r * pow( ( 1 + pow(r,2) ), (-3/5.3) );	//  0.292 scales to my units
}

double DiskModel::flat( double r )
{	//  my curve
	return 0.249 * r * pow( ( 1 + pow(r,2) ), (-3/5.9) );	
}

//
//  velocity curves

void DiskModel::Initialize(double n_totalMass, double n_rdmax, double n_ad, double n_z0, double n_toomreQ, double n_zVelocityDisp, enum RADIAL_DISTRIBUTION n_rDistribution, enum Z_DISTRIBUTION n_zDistribution /*= MyZDistr1*/, enum VELOCITY_CURVE n_vCurve /*= Toomre*/)
{
	totalMass = n_totalMass;
	rdmax = n_rdmax;
	ad = n_ad;
	z0 = n_z0;
	toomreQ = n_toomreQ;
	zVelocityDisp = n_zVelocityDisp;

	rDistribution = n_rDistribution;
	zDistribution = n_zDistribution;
	vCurve = n_vCurve;

}

bool DiskModel::Create(JakubSchwarzmeier::Data::RList< JakubSchwarzmeier::BarnesHut::Leave > & list, const int & nBodies)
{
	double (*pfnRdistribution)(double);
	double (*pfnZdistribution)(double);
	double (*pfnVcurve)(double);

	switch (rDistribution)
	{
	case Model1:

		pfnRdistribution = model1;
		break;

	case Model2:

		pfnRdistribution = model2;		
		break;
	}

	switch (zDistribution)
	{
	case MyZDistr1:

		pfnZdistribution = zdistr1;		
		break;
	}

	switch (vCurve)
	{
	case Toomre:

		pfnVcurve = toomre;
		break;
	}

	int bodies = 0;
	random(-100);		//  init real random number generator (Numerical Recipes in C)
	srand(100);			//  init integer random number generator (C RTL)

						
/*
	double rmin = 0;
	double rmin2 = rmin * rmin;
	double rmax2 = rdmax * rdmax; 

	for(int i = 0; i < nBodies; i++)
	{
		double lambda = 0.0 + random() * (2*M_PI - 0.0);

		//  convert polar -> cartesian (x-y plane)

		double r_i = sqrt( rmin2 + i * (rmax2 - rmin2) / (nBodies - 1.0) ); 

		double x = r_i * cos( lambda );
		double y = r_i * sin( lambda );

		Vector<> position;
		position[0] = x;
		position[1] = y;
		position[2] = 0;


		double acc;
		if ( r_i == 0.0 ) 
		{
			acc = 0.0;
		} 
		else
		{
			double f = 1.0/(r_i+ad); 
			double pot = -totalMass * f;
			f = pot * f / r_i;            

			acc = r_i * f;
		}
		
		double v = 0.05*sqrt( r_i * fabs(acc) );


		double vx = -v * sin( lambda );
		double vy = v * cos( lambda );
	
		Vector<> velocity;
		velocity[0] = vx;
		velocity[1] = vy;
		velocity[2] = 0;

		list.Add( new JakubSchwarzmeier::BarnesHut::Leave(position, velocity, totalMass/nBodies) );
	}
*/

	double probMax = ::FindMaximum( pfnRdistribution, 0, rdmax, 1e-12 );
	probMax = pfnRdistribution(probMax);

	while ( nBodies > bodies )
	{
		r = 0.02 + random() * (rdmax - 0.02);	//  between 0 and rdmax
		double p = 0.0 + random() * (probMax - 0.0);	//  between 0 and distribution function maxima

		//  find radius from radius distribution
		if ( p <= pfnRdistribution(r) )		//  throw out points over distribution function
		{
			bodies++;

			//  Generate positions
			//

			//  get random number <0; 2*Pi) to get polar coordinate lambda

			double lambda = 0.0 + random() * (2*M_PI - 0.0);

			//  convert polar -> cartesian (x-y plane)

			double x = r * cos( lambda );
			double y = r * sin( lambda );


			//  find z-component position from distribution
			double z;

			if ( r > zignored )
			{
				z = 0;
			}
			else
			{
				double probzMax = ::FindMaximum( pfnZdistribution, 0, z0, 1e-12 );
				probzMax = pfnZdistribution(probzMax);

				do
				{
					double zd = 0.0 + random() * (z0 - 0.0);	//  between 0 and z0
					p = 0.0 + random() * (probzMax - 0.0);	//  between 0 and distribution function maxima

					//  find radius from radius distribution
					if ( p <= pfnZdistribution(zd) )		//  throw out points over distribution function
					{
						z = zd;
						break;
					}
				} while ( true );
			}

			//  z-component distribution returns only positive values
			//  so put it sometimes to negative

			if ( (rand() % 2) == 0 )
			{
				z = -z;
			}

			Vector<> position;
			position[0] = x;
			position[1] = y;
			position[2] = z;
			
			//  Generate velocity
			//
			
			double v = pfnVcurve(r);		//  circular (tangential) velocity

			//  add tangential velocity dispersion (fluctuation)
			//

			//  angular velocity
			double omega = v/r;

			//  pozor na deleni nulou ve jmenovateli
			//  (1/(2*M_PI*pow(ad,2))) is mi zero

			double tanDisp = ( toomreQ * 3.36 * (1/(2*M_PI*pow(ad,2))) * pfnRdistribution(r) )/ (2*omega);
			
			if ( (rand() % 2) == 0 )
			{
				v = v + tanDisp;
			}
			else
			{
				v = v - tanDisp;
			}

			//  convert velocity to components
			double vx = -v * sin( lambda );
			double vy = v * cos( lambda );

			//  add radial velocity dispersion (fluctuation)
			//
			
			//  epicyclic frequency is sqrt(2)*omega
			double radDisp = ( toomreQ * 3.36 * (1/(2*M_PI*pow(ad,2))) * pfnRdistribution(r) )/ (sqrt(2)*omega);

			double radDispX = radDisp * cos( lambda );
			double radDispY = radDisp * sin( lambda );

			radDispX = 0.0 + random() * (radDispX - 0.0);	//  between 0 and radDispX
			radDispY = 0.0 + random() * (radDispY - 0.0);	//  between 0 and radDispY


			if ( (rand() % 2) == 0 )
			{
				vx += radDispX;
				vy += radDispY;
			}
			else
			{
				vx -= radDispX;
				vy -= radDispY;
			}


			//  my z-component velocity dispersion
			//

			double vz = 0;

			if ( r <= zignored )
			{
				double zDisp = 0.0 + random() * (tanDisp*zVelocityDisp - 0.0);	//  between 0 and tanDisp*zVelocityDisp

				if ( (rand() % 2) == 0 )
				{
					vz = zDisp;
				}
				else
				{
					vz = -zDisp;
				}
			}

			Vector<> velocity;
			velocity[0] = vx;
			velocity[1] = vy;
			velocity[2] = vz;

			list.Add( new JakubSchwarzmeier::BarnesHut::Leave(position, velocity, totalMass/nBodies) );
	
		}	//  end if ()
	
	}	//  end while ()

	return true;
}

