/************************************************************
*
*  Copyright (c) 2006 Jakub Schwarzmeier
*
************************************************************/

#include "stdafx.h"
#include "valtime.h"

#include <wx/datetime.h>
#include <wx/regex.h>
#include <wx/control.h>

#include <ctype.h>
#include <locale>
#include <time.h>

using namespace JakubSchwarzmeier::Astronomy;

IMPLEMENT_DYNAMIC_CLASS(wxTimeValidator, wxValidator)

wxTimeValidator::wxTimeValidator(CDateTime * val)
{
    m_pVal = val;
}

wxTimeValidator::wxTimeValidator(const wxTimeValidator& val)
    : wxValidator()
{
    Copy(val);
}

bool wxTimeValidator::Copy(const wxTimeValidator& val)
{
    wxValidator::Copy(val);

    m_pVal = val.m_pVal;

    return true;
}


// Called when the value in the window must be validated.
// This function can pop up an error message.
bool wxTimeValidator::Validate(wxWindow *parent)
{
    wxControl *control = (wxControl *) m_validatorWindow;

    // If window is disabled, simply return
    if ( !control->IsEnabled() )
        return true;

	//  Value entered by user.
    wxString val(control->GetLabel());

    bool ok = true;

    // NB: this format string should contian exactly one '%s'
    wxString errormsg;

	//  Create regular expression for this date format
	wxString strOneOrTwoDigits( wxT("(([0-9]{1})|([0-9]{2}))") );
	wxString strRegExp;
	wxChar strSeparator = CDateTime::GetTimeSeparator();

	strRegExp.Append( strOneOrTwoDigits );
	strRegExp.Append( strSeparator );
	strRegExp.Append( wxT("*") );
	strRegExp.Append( strOneOrTwoDigits );
	strRegExp.Append( wxT("*") );
	strRegExp.Append( strSeparator );
	strRegExp.Append( wxT("*") );
	strRegExp.Append( strOneOrTwoDigits );
	strRegExp.Append( wxT("*") );

	//  Perform regular expression test
	//
	wxRegEx exp( strRegExp );
	bool v = exp.IsValid();
	if ( exp.Matches( val ) )
	{
		if ( exp.GetMatch(val).Length() != val.Length() )
		{
			ok = false;
		}
	}
	else
	{
		ok = false;
	}

	if (ok)
	{
		int hour, minute, second;
		bool r = ScanfHourMinuteSecond(val, &hour, &minute, &second);
		if ( ((hour < 0) || (hour > 23)) || ((minute < 0) || (minute > 59)) ||
			((second < 0) || (second > 59)) )
		{
			ok = false;
		}
	}
		
    if ( !ok )
    {
		errormsg = _("'%s' should be in 24-hours format with locale separator and sequence hh, mm, ss.");

        wxASSERT_MSG( !errormsg.empty(), _T("you forgot to set errormsg") );

        m_validatorWindow->SetFocus();

        wxString buf;
        buf.Printf(errormsg, val.c_str());

        wxMessageBox(buf, _("Validation conflict"),
                     wxOK | wxICON_EXCLAMATION, parent);
    }

    return ok;
}

// Called to transfer data to the window
bool wxTimeValidator::TransferToWindow(void)
{
    if ( m_pVal )
    {
        wxControl * control = (wxControl *) m_validatorWindow;
		control->SetLabel( m_pVal->FormatTimeLong() );
    }

    return true;
}

// Called to transfer data to the window
bool wxTimeValidator::TransferFromWindow(void)
{
    if ( m_pVal )
    {
        wxControl *control = (wxControl *) m_validatorWindow;
        wxString val( control->GetLabel() );
		
		int hour, minute, second;
		bool r = ScanfHourMinuteSecond(val, &hour, &minute, &second);
		m_pVal->Set( m_pVal->GetDay(), m_pVal->GetMonth(), m_pVal->GetYear(), hour, minute, second );
    }

    return true;
}

bool wxTimeValidator::ScanfHourMinuteSecond(wxString val, int * hour, int * minute, int * second)
{
	wxString strD( wxT("%d") );
	wxString dateSeparator( wxT(":") );
	wxString strFormat( strD );
	strFormat.Append( dateSeparator );
	strFormat.Append( strD );
	strFormat.Append( dateSeparator );
	strFormat.Append( strD );

	*hour = 0;	*minute = 0;	*second = 0;
	int result = sscanf( val.c_str(), strFormat.c_str(), hour, minute, second );

	if (result < 1)
		return false;
	else
		return true;
}
