/************************************************************
*
*  Copyright (c) 2002-2004 Jakub Schwarzmeier
*
************************************************************/

#include "stdafx.h"

#include "spice.h"

#include "spicelib/spicelib.h"

using namespace JakubSchwarzmeier::Std::Math;
using namespace JakubSchwarzmeier::Astronomy;

CSpice::CSpice()
{
}

CSpice::~CSpice()
{
	UnloadSpkFiles();
}


void CSpice::LoadSpkFile( std::string strFileName )
{
	SpiceInt handle;
	MAP::const_iterator it = m_spkFiles.find( strFileName );
	if ( it == m_spkFiles.end() )
	{
		FILE * file = fopen( strFileName.c_str(), "r" );
		if ( file != NULL )
		{
			fclose( file );
			spklef_c( strFileName.c_str(), &handle );		//  load SPK file
			m_spkFiles.insert( MAP::value_type(strFileName, handle) );
		}
	}
}

void CSpice::UnloadSpkFile( std::string strFileName )
{
	MAP::iterator it = m_spkFiles.find( strFileName );
	if ( it != m_spkFiles.end() )
	{
		spkuef_c( it->second );		//  unload SPK file
		m_spkFiles.erase( it );
	}
}

void CSpice::UnloadSpkFiles()
{
	//  unload SPK m_spkFiles
	MAP::const_iterator it;
	for ( it = m_spkFiles.begin( ) ; it != m_spkFiles.end( ) ; it++ )
	{
		spkuef_c( it->second );		//  unload SPK file
	}
	m_spkFiles.clear();
}

void CSpice::LoadKernelFile( std::string strFileName )
{
//  load PCK (planetary constant kernel) file, text PCK cannot be unloaded
//  load LSK (leap second kernel) file
	ldpool_c ( strFileName.c_str() );
}

void CSpice::GetEphemerisEclipticRectangular( int observerId, int targetId, double utcJD, 
				JakubSchwarzmeier::Std::Math::Vector<> & position, 
				JakubSchwarzmeier::Std::Math::Vector<> & velocity )
{
	SpiceDouble state[6]; 	//  out, position & velocity
	SpiceDouble lt;		//  out, one way light time
	SpiceDouble utcJ2000 = ( utcJD - 2451545.0 ) * 86400.0;		//  ephemeris time
	SpiceDouble delta;
	SpiceDouble etJ2000;
	try
	{
		deltet_ ( &utcJ2000, "UTC", &delta, 3L );
		etJ2000 = delta + utcJ2000;
	}
	catch (...)
	{
		etJ2000 = utcJ2000;
	}

//	spkezr_c ( body, etJ2000, "J2000", "LT+S", "Sun", state, &lt );	
//	try
//	{
/*
	Planet system barycenters (i.e. 1 through 9) and the sun (10) orbit 
	the solar system barycenter (0). x99 are body centers.
	*/
		spkez_c ( targetId, etJ2000, "ECLIPJ2000", "LT+S", observerId, state, &lt );		
//	}
/*
	catch (CSpiceException * exception)
	{
		CDataInaccessible * newException = new CDataInaccessible( TRUE, CString(exception->GetErrorMessage()) );
		delete exception;
		throw newException;
	}
*/
	/*
	//  rotate frame
	//
	SpiceDouble rotate[3][3];
	SpiceDouble scpos[6];

	pxform_c ( "J2000", "ECLIPJ2000",  0, rotate );		//  from J2000 to ECLIPJ2000
	mxv_c    ( rotate,  state, scpos );    
	mxv_c    ( rotate,  &state[3], &scpos[3] ); 
	//
	//  end
*/
	//  SPICE gives position in km and velocity in km/sec --> convert them to meters and meters/sec
	
	position[0] = state[0];
	position[1] = state[1];
	position[2] = state[2];
	position = 1000 * position;

	velocity[0] = state[3];
	velocity[1] = state[4];
	velocity[2] = state[5];
	velocity = 1000 * velocity;
}

/*  Longitude of the Sun; LSUN (also known as "L-sub-s") is the planetocentric
    longitude of the sun, as seen from a specified body; LSUN is often used as 
	a measure of the season on a body: for the northern hemisphere LSUN changes 
	from 0 (spring) ---- +90 (summer) ---- +/-180 (autumn) ---- -90 (winter) ---- 
	0 (spring).  */
double CSpice::GetLSun( int observerId, double utcJD )
{
	SpiceDouble utcJ2000 = ( utcJD - 2451545.0 ) * 86400.0;		//  ephemeris time
	SpiceDouble delta;
	SpiceDouble etJ2000;
	try
	{
		deltet_ ( &utcJ2000, "UTC", &delta, 3L );
		etJ2000 = delta + utcJ2000;
	}
	catch(...)
	{
		etJ2000 = utcJ2000;
	}

	char str[1024];
	SpiceBoolean found;
	bodc2n_c ( observerId, 1024, str, &found );
    return dpr_c() * lspcn_c ( str, etJ2000, "LT+S" );
}


//  Get ephemeris in planetocentric coordinates
//  i.e. Body equator and prime meridian coordinates.
//  i.e. for the Earth, these are geocentric coordinates.
//  x -- lies in plane of equator and plane of prime meridian
//  z -- parallel with rotational axis, toward north of the invariable plane of
//       solar system (without connection to spin of body)
//  y -- completes right-handed system
// 
//  Note that planetocentric
//  coordinates differ from planetographic and cartographic coordinates in
//  that they are always right-handed, regardless of the rotation of the
//  body.
//
//  Before calling this function, you must load LSK (Leap Second Kernel) file,
//  which usually have "tls" extension. It contains the dates
//  of leap seconds and values for constants needed to
//  compute the difference between UTC and ET at any
//  epoch.
//  Also you need file (PCK) that contains IAU values needed to compute transformations 
//  from inertial (J2000) coordinates to body-fixed (pole and prime meridian).
//
void CSpice::GetEphemerisPlanetocentricLatLon( int observerId, int targetId, double utcJD, 
	double & latitude, 
	double & longitude )
{
	SpiceDouble utcJ2000 = ( utcJD - 2451545.0 ) * 86400.0;		//  ephemeris time
	SpiceDouble lt;		//  out, one way light time
	SpiceDouble state[6]; 	//  out, position & velocity

	//  Convert from UTC into ET
	//  
	//  First get DeltaET  =  ET - UTC, the increment to be applied to UTC to give ET.  
	//  This needs LSK file to be loaded.
	SpiceDouble delta;
	deltet_ ( &utcJ2000, "UTC", &delta, 3L );
	SpiceDouble etJ2000 = delta + utcJ2000;

	//  
	/*
	Compute the true state (corrected for light-time and stellar aberation)
	of the target as seen from the observer at the
	specified epoch in the target fixed reference frame.
	*/
	spkez_c ( targetId, etJ2000, "J2000", "LT+S", observerId, state, &lt );
	for(int i = 0; i < 6; i++)
		state[i] *= 1000.0 * state[i];

	/*   Return a 3x3 transformation matrix that transforms from inertial 
   coordinates to positions in body-equator-and-prime-meridian 
   coordinates at the appropriate epoch (where 5 is the length of the string "J2000" ).
   This will read constants from the planetary constants kernel (PCK) file.
   */
    SpiceDouble              tipm[3][3];	//  out, transformation matrix
	tipbod_ ( "J2000",  reinterpret_cast<SpiceInt*>( &targetId ), reinterpret_cast<SpiceDouble*>( &etJ2000 ), reinterpret_cast<SpiceDouble*>( tipm ), 5L);
	
    /*
    Put both vectors in body-fixed coordinates.
	Given a position P in the inertial reference frame, the corresponding bodyfixed 
    position is given by the matrix vector product
	tipm * P
	*/
    mxv_c ( tipm, /*IN*/state, /*OUT*/state );
    mxv_c ( tipm, &state[3], &state[3] );	//  is this transf. also valid for speed?

	//  We now have rectangular planetocentric coordinates.
	//  For planetocentric latitude and longitude we need to read radius from PCK file.
	SpiceInt dim = 3;
	SpiceDouble values[3];

	bodvar_c(targetId, "RADII", &dim, values);		//  read data from PCK
	SpiceDouble radius = values[0] * 1000.0;
	SpiceDouble lon, lat;

	/*
    Convert rectangular coordinates to planetocentric 
    latitude and longitude. 
    */
    reclat_c ( state, &radius, &lon, &lat );
    reclat_c ( &state[3], &radius, &lon, &lat );	//  velocity

	// Convert radians to degrees.
    longitude = lon * dpr_c ();
    latitude = lat * dpr_c ();

	// For values and bounds look into documentation of reclat_c SPICE function.
}

void CSpice::GetEphemerisHorizon( int observerId, int targetId, double utcJD, 
double targetLat, double targetLon, double & elevation, double & azimuth )
{
	//  Geodetic co-latitude (COLAT, the complement of latitude: 
	//  latitude + co-latitude = 90 degrees) and longitude
	//  (LONG) measured in degrees for this point. 
   /*
Prograde (or direct) rotation of a body is rotation eastward, or counter- clockwise, 
as seen from the north pole. For such bodies, east longitude is measured negatively 
to the east (0 to -360 degrees) from the prime meridian. Retrograde rotation is rotation
clockwise (westward) as seen from the north pole. East longitude is measured positively 
to the east (0 to 360 degrees) from the prime meridian. 

Exceptions are the Earth, Moon and Sun where longitude has historically been measured 
both east and west of the prime meridian 0 to 180 degrees. Though these bodies are direct
rotators, longitude is nonetheless measured positively to the east on this system, 
0 to 360 degrees, due to historical precedence. If the positive west longitude of a site
on these 3 bodies is given, it should be input here as positive east longitude, which 
would be (360 - West Longitude). If the negative east longitude is given instead, for
these exceptions only, one can input the negative east longitude. It will be converted 
to a positive east longitude on output, however. 

The following major bodies are either retrograde or exceptions and require site input
with positive east longitude: 

       Retrograde (+ east longitude):
       ------------------------------
          Venus (299), Arial (701), Umbriel (702), Titania (703),
          Oberon (704), Miranda (705), Cordelia (706), Ophelia (707),
          Bianca (708), Cressida (709), Desdemona (710), Juliet (711),
          Portia (712), Rosalind (713), Belinda (714), Puck (715),
          Uranus (799), Pluto (999), Charon (901)

       Also + east longitude (prograde exceptions): 
       --------------------------------------------
          Sun (10), Earth (399), Moon (301)

All others are prograde and must be input with negative longitude east of the adopted 
prime meridian. Since such sites are usually expressed in terms of positive west 
longitude on maps, negative east longitude would be ... 
                           ( West longitude - 360 )
*/
/*
GEODETIC, PLANETODETIC (generally this means map coordinates)
                  E-long - Geodetic east longitude (DEGREES)
                  lat    - Geodetic latitude  (DEGREES)
                  h      - Altitude above reference ellipsoid (km)

*/	
	/*

	/*
	The first angle is the negative of the longitude of the
	surface point; the second angle is the negative of the 
	point's colatitude.
	*/
	SpiceDouble colat = 90.0 - targetLat;
	SpiceDouble             angles [3]      =  { -targetLon /*-243.1945102442646*/,
												-colat /*54.7000629043147*/,
												180.0              };
	SpiceInt                axes   [3]      =  { 3, 2, 3 };
	SpiceInt                center          =    300000;
	SpiceInt                frclass         =    4;
	SpiceInt                frclsid         =    1500000;
	SpiceInt                frcode          =    1500000;

	/*
	Define the MYTOPO reference frame. 
	Note that the third argument in the pcpool_c calls is
	the length of the final string argument, including the
	terminating null character.
	*/
	pipool_c ( "FRAME_MYTOPO",            1,     &frcode   );
	pcpool_c ( "FRAME_1500000_NAME",      1, 7,  "MYTOPO"  );
	pipool_c ( "FRAME_1500000_CLASS",     1,     &frclass  );
	pipool_c ( "FRAME_1500000_CLASS_ID",  1,     &frclsid  );
	pipool_c ( "FRAME_1500000_CENTER",    1,     reinterpret_cast<const ConstSpiceInt *>(&observerId)   );

	pcpool_c ( "OBJECT_300000_FRAME",     1, 7,  "MYTOPO"  );

	pcpool_c ( "TKFRAME_MYTOPO_RELATIVE", 1, 7,  "IAU_EARTH" /* "ITRF93" */  );
	pcpool_c ( "TKFRAME_MYTOPO_SPEC",     1, 7,  "ANGLES"  );
	pcpool_c ( "TKFRAME_MYTOPO_UNITS",    1, 8,  "DEGREES" );
	pipool_c ( "TKFRAME_MYTOPO_AXES",     3,     axes      );
	pdpool_c ( "TKFRAME_MYTOPO_ANGLES",   3,     angles    );

	//  We must have loaded PCK for observer
	SpiceDouble state[6]; 	//  out, position & velocity
	SpiceDouble lt;		//  out, one way light time
	SpiceDouble utcJ2000 = ( utcJD - 2451545.0 ) * 86400.0;		//  ephemeris time
	SpiceDouble delta;
	deltet_ ( &utcJ2000, "UTC", &delta, 3L );
	SpiceDouble etJ2000 = delta + utcJ2000;

	spkez_c ( targetId, etJ2000, "MYTOPO", "LT+S", observerId, state, &lt );		

	//  convert into horizon coords, rectangular coordinates to RA and DEC
/*
   ra         is the right ascension of `rectan'.  This is the angular
              distance measured toward the east from the prime meridian
              to the meridian containing the input point. The direction
              of increasing right ascension is from the +X axis towards
              the +Y axis.

              `ra' is output in radians.  The range of `ra' is [0, 2*pi]. 


   dec        is the declination of `rectan'.  This is the angle from
              the XY plane of the ray from the origin through the
              point.

              `dec' is output in radians.  The range of `dec' is
              [-pi/2, pi/2].
*/
	SpiceDouble R;
	recrad_c(/*IN*/state, /*OUT*/&R, /*OUT*/&azimuth, /*OUT*/&elevation);
	
	// Convert radians to degrees.
    azimuth *= dpr_c ();
    elevation *= dpr_c ();
}

std::set<int> CSpice::GetSpkObjectsIDs( std::set<std::string> & files )
{
	std::set<int> bodies;

	std::set<std::string>::const_iterator it;
	for ( it = files.begin( ) ; it != files.end( ) ; it++ )
	{
		#define  MAXOBJ 1000

		//  cell is SPICE's 'safe' C-language array
		SPICEINT_CELL ( ids, MAXOBJ );	//  macros to declare ids 'SpiceCell' and a data array

		//  Find the set of objects in the SPK file. 
		spkobj_c ( it->c_str(), &ids );

		for ( int i = 0;  i < card_c( &ids );  i++  )
		{
			#define  MAXIV          1000
			#define  WINSIZ         ( 2 * MAXIV )
			SPICEDOUBLE_CELL        ( cover, WINSIZ );

			//  Find the coverage window for the current object.
			SpiceInt obj  =  SPICE_CELL_ELEM_I( &ids, i );
			spkcov_c ( it->c_str(), obj, &cover );
			bodies.insert( obj );

            //  Get the number of intervals in the coverage window.
            SpiceInt niv = card_c ( &cover ) / 2;

			//  Coverage for object 'obj'
			//
			//  Convert the coverage interval start and stop times to TDB calendar strings.
			for ( SpiceInt j = 0;  j < niv;  j++  )
			{
				SpiceDouble b;
				SpiceDouble e;
                
				//  Get the endpoints of the jth interval.
				wnfetd_c ( &cover, j, &b, &e );
				
				//  this cycle is unused now
				//  it can return valid time date range for each object for given file
			}	

		}
		
		
	}

	return bodies;
}

std::map<std::string, int> CSpice::GetSpkObjects( std::set<std::string> & files )
{
	std::map<std::string, int> res;

	std::set<int> ids = GetSpkObjectsIDs( files );

	std::set<int>::const_iterator it;
	for ( it = ids.begin( ) ; it != ids.end( ) ; it++ )
	{
		char str[1024];
		SpiceBoolean found;
		bodc2n_c( (*it), 1024, str, &found );
		if (!found)  
			sprintf(str, "%d", (*it));

		res.insert( std::map<std::string, int>::value_type( str, (*it) ) );
	}

	return res;
}

void CSpice::GetPoleEcliptic( int bodyID, double utcJD, double & longitude, double & latitude )
{
	//  nutation precession angles
	//
	bool precession = false;
	SpiceInt anglesDim = 10;
	SpiceDouble angles[10];
	SpiceInt anglesRdim;
	/*
	try
	{
		//  read data from PCK
		bodvcd_c(bodyID, "NUT_PREC_ANGLES", anglesDim, &anglesRdim, angles);
		precession = true;
	}
	catch (...)
	{
		precession = false;
	}
*/
	//  variables for RA and DEC
	SpiceInt dim = 3;
	SpiceDouble values[3];
	SpiceInt rdim;
	double et = ( utcJD - 2451545.0 );

	//  rightAscention
	bodvcd_c(bodyID, "POLE_RA", dim, &rdim, values);		//  read data from PCK
	double alpha = values[0] + (values[1]/36525.0) * et;	

	if (precession)
	{
		try
		{
			SpiceInt precDim = 5;
			SpiceDouble prec[5];
			SpiceInt rpecrdim;
			
			//  read data from PCK
			bodvcd_c(bodyID, "NUT_PREC_RA", precDim, &rpecrdim, prec);
			for(int i = 0; i < anglesRdim; i++)
				alpha += prec[i] * Functions::sin_angle( angles[i*2] + (angles[i*2 + 1]/36525.0)*et );
		}
		catch (...) {}
	}

	//  declination
	bodvcd_c(bodyID, "POLE_DEC", dim, &rdim, values);		//  read data from PCK
	double delta = values[0] + (values[1]/36525.0) * et;
	if (precession)
	{
		try
		{
			SpiceInt precDim = 5;
			SpiceDouble prec[5];
			SpiceInt rpecrdim;
			
			//  read data from PCK
			bodvcd_c(bodyID, "NUT_PREC_DEC", precDim, &rpecrdim, prec);
			for(int i = 0; i < anglesRdim; i++)
				delta += prec[i] * Functions::cos_angle( angles[i*2] + (angles[i*2 + 1]/36525.0)*et );
		}
		catch (...) {}
	}

	//  convert equatorial --> ecliptic
	double eps = 23.5;
	latitude = Functions::asin_angle( -Functions::cos_angle(delta)*Functions::sin_angle(alpha)*Functions::sin_angle(eps)+Functions::sin_angle(delta)*Functions::cos_angle(eps) );
	longitude = Functions::cos_angle(delta)*Functions::cos_angle(alpha)/Functions::cos_angle(latitude);
}

			
double CSpice::GetPM( int bodyID, double utcJD )
{
	SpiceInt dim = 3;
	SpiceDouble values[3];
	SpiceInt rdim;

	bodvcd_c(bodyID, "PM", dim, &rdim, values);		//  read data from PCK
	double initial = values[0];
	double period = values[1];
	
	double ph = initial + period * ( utcJD - 2451545.0 );

	while (ph > 360.0)
		ph -= 360;

	while (ph < 0.0)
		ph += 360;

	return ph;
}

void CSpice::GetRadii( int bodyID, double & radius )
{
	SpiceInt dim = 3;
	SpiceDouble values[3];

	bodvar_c(bodyID, "RADII", &dim, values);		//  read data from PCK
	radius = values[0]*1000;
}

void CSpice::GetMass( int bodyID, double & mass )
{
	SpiceInt dim = 1;
	SpiceDouble values;

	bodvar_c(bodyID, "GM", &dim, &values);		//  read data from PCK
	double G = 6.67259e-20;			//   (kg-1 km3 sec-2)
	mass = values/G;
}

void CSpice::GetName( int bodyID, std::string & name )
{
	char str[1024];
	SpiceBoolean found;
	bodc2n_c( bodyID, 1024, str, &found );
	if (!found)
		sprintf(str, "%d", bodyID);

	name = str;
}

std::string CSpice::GetFileType(std::string file)
{
	char arch[4];
	char type[5];
	getfat_c(file.c_str(), 4, 5, arch, type );

	return std::string( type );
}
