#include "stdafx.h"

#include "MFC.Net.h"

using namespace JakubSchwarzmeier::MFC::Net::Sockets;

IMPLEMENT_DYNAMIC( CSocket, IIOCPStream )

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

CSocket::CSocket(SOCKET n_socket /*=0*/, 
				 CSocketPool * n_pSocketPool /*=NULL*/,
				 size_t n_readBufferSize /*=256*/)
{
	m_hSocket = n_socket;
	m_pSocketPool = n_pSocketPool;

	m_readBufferSize = n_readBufferSize;
	m_readBuffer = new char[m_readBufferSize];

	m_directRouteProtocol = NULL;
}

CSocket::~CSocket()
{
	//  free reading buffers
	delete [] m_readBuffer;

	//  free all remaing writing buffers
	POSITION pos = m_writeBuffersMap.GetHeadPosition();
	while (pos != NULL) 
	{
		CRBMap<OVERLAPPEDPLUS*, char*>::CPair * cp = m_writeBuffersMap.GetNext(pos);
		delete cp->m_key;
		delete [] cp->m_value;
	}

	//  free protocols
	pos = m_protocolList.GetHeadPosition();
	while (pos != NULL) 
	{
		delete m_protocolList.GetNext(pos);
	}

	//  close socket
	ASSERT(m_hSocket != 0);
	closesocket(m_hSocket);

	m_pSocketPool->RemoveSocket( this );
}

void CSocket::ReadFile()
{
	DWORD dwRecv;
	DWORD dwFlags = 0;

	WSABUF wsabuf;
	wsabuf.len = static_cast<u_long>(m_readBufferSize);
	wsabuf.buf = static_cast<char*>(m_readBuffer);

	ZeroMemory( &m_readingOverlapped, sizeof(m_readingOverlapped) );
	m_readingOverlapped.op = Read;

	int ret = ::WSARecv(
		m_hSocket,					//  socket
		&wsabuf,					//  array of WSABUF structures
		1,							//  no. of WSABUF structures in array
		&dwRecv,					//  no. of immediatly recieved bytes
		&dwFlags,					//  flags
		&m_readingOverlapped.ol,	//  OVERLAPPED structure
		NULL						//  overlapped completion routine
		);
	
	if ( (ret != 0) && ( WSAGetLastError() != WSA_IO_PENDING ) )
	{
		delete this;
		THROW_WINDOWS_EXCEPTION()
	}

	return;

}

void CSocket::WriteFile(char * lpBuffer, DWORD dwNumberOfBytesToWrite)
{
	DWORD dwSend;
	DWORD dwFlags = 0;

	WSABUF wsabuf;
	wsabuf.len = static_cast<u_long>(dwNumberOfBytesToWrite);
	wsabuf.buf = lpBuffer;

	OVERLAPPEDPLUS * olp = new OVERLAPPEDPLUS;
	ZeroMemory( olp, sizeof(*olp) );
	olp->op = Write;

	m_criticalSection.Lock();
	m_writeBuffersMap.SetAt( olp, lpBuffer );
	m_criticalSection.Unlock();

	int ret = ::WSASend(
		m_hSocket,				//  socket
		&wsabuf,				//  array of WSABUF structures
		1,						//  no. of WSABUF structures in array
		&dwSend,				//  no. of immediatly send bytes
		dwFlags,				//  flags
		&olp->ol,				//  OVERLAPPED structure
		NULL					//  overlapped completion routine
		);

	if ( (ret != 0) && ( GetLastError() != WSA_IO_PENDING ) )
	{
		delete this;
		THROW_WINDOWS_EXCEPTION()
	}

	return;
}

CSocket::operator SOCKET() const
{ 
	return m_hSocket; 
}

CSocket::operator HANDLE() const
{
	return reinterpret_cast<HANDLE>(m_hSocket);
}


void CSocket::SetReadBufferSize(size_t n_newSize)
{
	delete [] m_readBuffer;
	m_readBufferSize = n_newSize;
	m_readBuffer = new char[m_readBufferSize];
}

void CSocket::AddProtocol( IProtocol * protocol )
{
	m_protocolList.AddTail( protocol );
}


void CSocket::OnRead(OVERLAPPEDPLUS * olp, DWORD dwSize)
{
	//  najde buffer prislusejici olp, zatim je to pevnej m_readBuffer

	DWORD dwPosition = 0;
	//  If true than a protocol ate whole buffer and waits for another buffer 
	//  that will be delivered directly to it.
	bool bContinue = false;
	POSITION pos = m_protocolList.GetHeadPosition();
	//  If false then none of the protocols used buffer. Prevent an infinite loop.
	bool bBufferUsed = false;
	DWORD dwPreviousPosition = 0;

	char * buffer = m_readBuffer;

	//  cycle through protocol list still around
	while ( (!bContinue) && (dwPosition < dwSize) && (pos != NULL) )
	{
		IProtocol * protocol;
		if (m_directRouteProtocol != NULL)
			protocol = m_directRouteProtocol;
		else
            protocol = m_protocolList.GetNext( pos );

		char * writeData = NULL;
		DWORD dwWriteDataSize = 0;
		dwPreviousPosition = dwPosition;
		bContinue = protocol->HandleRequest( buffer, dwSize, dwPosition, 
 			&writeData, dwWriteDataSize );

		if (writeData != NULL)  WriteFile( writeData, dwWriteDataSize );
		
		if (bContinue)
		{
			m_directRouteProtocol = protocol;
			break;
		}
		else
			m_directRouteProtocol = NULL;
		
		if (dwPreviousPosition != dwPosition)	bBufferUsed = true;
		if (pos == 0)
		{
			//  cycle through protocols one more time
			pos = m_protocolList.GetHeadPosition();
			if (!bBufferUsed)
			{
				//  Advance to next byte in buffer.
				buffer += 1;
				dwSize -= 1;
				dwPosition = 0;
				ATLTRACE2(traceAppMsg, 0, "CSocket: Any of registered protocols was unable to process data.\r\n");
			}
			bBufferUsed = false;
		}

		//  update variables in response to HandleRequest call
		buffer += dwPosition;
		dwSize -= dwPosition;
		dwPosition = 0;
	}

	//  zrus buffer a prislusejici olp, zatim je to pevnej m_readBuffer

	ReadFile();			//  and post read once again
}

void CSocket::OnWrite(OVERLAPPEDPLUS * olp, DWORD dwSize)
{
	m_criticalSection.Lock();

	//  find buffer according to overlapped structure pointer
	CRBMap<OVERLAPPEDPLUS*, char*>::CPair * cp = m_writeBuffersMap.Lookup(olp);

	//  remove entry from map
	m_writeBuffersMap.RemoveKey(olp);

	m_criticalSection.Unlock();
	
	//  free overlapped structure & key
	delete cp->m_key;
	delete [] cp->m_value;
}

void CSocket::OnErrorRead(OVERLAPPEDPLUS * olp, DWORD dwSize)
{
	delete this;
}

void CSocket::OnErrorWrite(OVERLAPPEDPLUS * olp, DWORD dwSize)
{
	delete this;
}

CString CSocket::GetLocalAddress()
{
	ASSERT( m_hSocket != NULL );
	
	sockaddr_in sadr;
	int iSize = sizeof(sadr);
	int i = getsockname( m_hSocket, reinterpret_cast<sockaddr *>(&sadr), &iSize );
	if ( i == SOCKET_ERROR )
	{
		THROW_WINDOWS_EXCEPTION()
	}

	CString strRet;
	strRet.Format( _T("%d.%d.%d.%d"), 
		sadr.sin_addr.S_un.S_un_b.s_b1, 
		sadr.sin_addr.S_un.S_un_b.s_b2,
		sadr.sin_addr.S_un.S_un_b.s_b3, 
		sadr.sin_addr.S_un.S_un_b.s_b4 );

	return strRet;
}

unsigned int CSocket::GetLocalPort()
{
	ASSERT( m_hSocket != NULL );

	sockaddr_in sadr;
	int iSize = sizeof(sadr);
	int i = getsockname( m_hSocket, reinterpret_cast<sockaddr *>(&sadr), &iSize );
	if ( i == SOCKET_ERROR )
	{
		THROW_WINDOWS_EXCEPTION()
	}

	return htons( sadr.sin_port );
}

CString CSocket::GetRemoteAddress()
{
	ASSERT( m_hSocket != NULL );
	
	sockaddr_in sadr;
	int iSize = sizeof(sadr);
	int i = getpeername( m_hSocket, reinterpret_cast<sockaddr *>(&sadr), &iSize );
	if ( i == SOCKET_ERROR )
	{
		THROW_WINDOWS_EXCEPTION()
	}

	CString strRet;
	strRet.Format( _T("%d.%d.%d.%d"), 
		sadr.sin_addr.S_un.S_un_b.s_b1, 
		sadr.sin_addr.S_un.S_un_b.s_b2,
		sadr.sin_addr.S_un.S_un_b.s_b3, 
		sadr.sin_addr.S_un.S_un_b.s_b4 );

	return strRet;
}

unsigned int CSocket::GetRemotePort()
{
	ASSERT( m_hSocket != NULL );

	sockaddr_in sadr;
	int iSize = sizeof(sadr);
	int i = getpeername( m_hSocket, reinterpret_cast<sockaddr *>(&sadr), &iSize );
	if ( i == SOCKET_ERROR )
	{
		THROW_WINDOWS_EXCEPTION()
	}

	return htons( sadr.sin_port );
}
