#include "stdafx.h"

#include <process.h>	//  C-runtime's _beginthreadex

#include <SFC.System.h>

using namespace JakubSchwarzmeier::SFC::System::IO;

CCompletionPort::CCompletionPort(int n_priority /*=THREAD_PRIORITY_NORMAL*/)
{
	//  create new completion port
	m_hCompletionPort = CreateIoCompletionPort( INVALID_HANDLE_VALUE, NULL, 0, 0 );
	if (m_hCompletionPort == NULL)
	{
		THROW_WINDOWS_EXCEPTION()
	}

	m_bServiceTerminating = false;

	//  create worker thread
	m_workerThread = AfxBeginThread(
		WorkerThread,		//  address of function of new thread
		this,				//  argument list
		n_priority,			//  priority
		0,					//  default stack size
		CREATE_SUSPENDED	//  thread is stopped
		);

	m_workerThread->m_bAutoDelete = false;
	m_workerThread->ResumeThread();
}

CCompletionPort::~CCompletionPort()
{
	//  terminate thread
	m_bServiceTerminating = true;
	if (m_hCompletionPort != NULL)  PostQueuedCompletionStatus( m_hCompletionPort, 0, 0, NULL );
	DWORD ret = WaitForSingleObject(m_workerThread->m_hThread, INFINITE);

	if (ret == WAIT_FAILED)
	{
		THROW_WINDOWS_EXCEPTION()
	}

	delete m_workerThread;
	CloseHandle( m_hCompletionPort );
}

CCompletionPort::operator HANDLE() const
{
	return m_hCompletionPort;
}

BOOL CCompletionPort::Associate(IIOCPStream * n_stream)
{
	//  associate socket with completion port
	HANDLE compPort = CreateIoCompletionPort( 
		*n_stream,
		m_hCompletionPort, 
		reinterpret_cast<unsigned long>(n_stream),
		0
		);

	if (compPort == NULL)
		return FALSE;
	else
		return TRUE;
}

UINT CCompletionPort::WorkerThread( LPVOID lpParameter )
{
	CCompletionPort * iocp = static_cast<CCompletionPort*>(lpParameter);
	DWORD dwBytesTransffered;

	IIOCPStream		* iocpStream;		//  points "out"
	OVERLAPPEDPLUS	* lpOverlapPlus;	//  points "out"
	OVERLAPPED		* lpOverlapped;		//  points "out", hidden in lpOverlapPlus
	while (true)
	{
		try
		{
			BOOL success = GetQueuedCompletionStatus( *iocp, &dwBytesTransffered,
				reinterpret_cast<unsigned long *>(&iocpStream), &lpOverlapped, INFINITE );

			//  terminate thread
			if ( iocp->m_bServiceTerminating )
			{
				break;		
			}

			if (success == 0)	
			{
				if (lpOverlapped == NULL)
					throw new CWindowsException(TRUE);		//  error on completion port
				//  else error on handle (lpOverlapped != NULL); handled lower
				if  (dwBytesTransffered == 0)  continue;	//  stream's handle closed
			}


			// Get the base address of the OVERLAPPEDPLUS structure 
			// containing the OVERLAPPED structure received.
			lpOverlapPlus = CONTAINING_RECORD(lpOverlapped, OVERLAPPEDPLUS, ol);
			
			//  When dwBytesTransffered = 0 and success == 1 then socket was closed
			//  on remote end.
			switch( lpOverlapPlus->op )
			{
			case Read:

				if ((success == 0) || (dwBytesTransffered == 0))
				{
					iocpStream->OnErrorRead( lpOverlapPlus, dwBytesTransffered );
				}
				else
					iocpStream->OnRead( lpOverlapPlus, dwBytesTransffered );

				break;

			case Write:

				if ((success == 0) || (dwBytesTransffered == 0))
					iocpStream->OnErrorWrite( lpOverlapPlus, dwBytesTransffered );
				else
					iocpStream->OnWrite( lpOverlapPlus, dwBytesTransffered );

				break;

			}	//  end switch 

		}	//  end try

		catch(JakubSchwarzmeier::SFC::System::CBaseException * excp)
		{
			REPORT_CATCH_EXCEPTION()
			excp->Delete();
		}

	}	//  end while(true)

	//  thread must terminate itself by calling return or _enthreadex
	return 0;		//  terminates thread, closes its handle
}
