#ifndef OCTTREE_H
#define OCTTREE_H

#include <schwarzmeier/std.h>
#include <vector>
#include <float.h>

namespace JakubSchwarzmeier
{
	namespace Std
	{
		namespace Data
		{
			using namespace JakubSchwarzmeier::Std::Math;
			using namespace JakubSchwarzmeier::Std::Graphics;


			struct ICommonNode
			{
				virtual ~ICommonNode() {}

			};	//  end struct ICommonNode

			template <typename _TCommon = ICommonNode>
			class ILeaf: public _TCommon
			{
			public:

				Vector<> m_position;
				virtual ~ILeaf() {};

			};	//  end class ILeaf

			template <typename _TCommon = ICommonNode, typename _TLeaf = ILeaf<> >
			struct INode: public _TCommon
			{
			public:

				INode() {}
				INode(INode * n_ptrParent, Vector<> * n_ptrCorner, double n_bisect, Quaternion<> * n_ptrHilbertOrientation)
				{
					m_ptrParent = n_ptrParent;
					m_ptrCorner = n_ptrCorner;
					m_bisect = n_bisect;
					m_ptrHilbertOrientation = n_ptrHilbertOrientation;
				}

				virtual ~INode()
				{
					//  delete corner & hilbert cube orientation
					delete m_ptrCorner;
					delete m_ptrHilbertOrientation;
				}

				INode *						m_ptrParent;
				Vector<> *					m_ptrCorner;	//  left corner of orthogonal area for this node
				Quaternion<> *				m_ptrHilbertOrientation;
				std::vector<_TCommon*>		m_childList;	
				std::vector<_TLeaf*>		m_leavesList;
				double						m_bisect;
                
			};	//  end class INode

			template <typename _TCommon = ICommonNode,
				typename _TLeaf = ILeaf<>, 
				typename _TNode = INode<> >
			class OctTree;

			template <typename _TCommon/*= ICommonNode*/,
				typename _TLeaf /*= ILeaf*/, 
				typename _TNode /*= INode*/>
			class OctTree
			{
			public:

				OctTree() {}

				void Construct( const std::vector<_TLeaf*> * n_ptrLeavesList )
				{
					const std::vector<_TLeaf*> & n_leavesList = *n_ptrLeavesList;

					//  Adaptive bisect -- find greatest value of spatial coordinate in absolute value
					//

					//  both level & bisect can be computed from each other
					double bisect = 0;
					for(size_t i = 0; i < n_leavesList.size(); i++)
					{
						Vector<> & position = n_leavesList[i]->m_position;
						for(int c = 0; c < 3; c++)
						{
							double val = fabs( position[c] );
							if ( val > bisect )  bisect = val;
						}
					}

					//  Increase because of condition that compares corners of cube with values of particles.
					//  It must be farther then the furthest particle.
					bisect += FLT_EPSILON;

					//
					//  end Adaptive bisect

					//  Tree construction
					//

					//  setup root node
					//
					Quaternion<> * qRoot = new Quaternion<>;
					qRoot->LoadIdentity();
					_TNode * root = new _TNode( NULL, new Vector<>(-bisect, -bisect, -bisect), 2*bisect, qRoot );
					root->m_leavesList.assign( n_leavesList.begin( ), n_leavesList.end( ) );

					size_t level = 0;
					m_octTree.push_back( new std::vector<_TNode*> );
					m_octTree[level]->push_back( root );

					//  main tree build loop
					_TNode * actualNode;
					while( level < m_octTree.size() )
					{
						size_t pos = 0;
						while( pos < m_octTree[level]->size() )
						{
							//  partition node & create its children
							//
							actualNode = m_octTree[level]->at(pos);

							//  Generate left corners. Same as generating Gray's code.
							//  Corners are then get by multiplying the vector by bisect.
							//  x y z
							// 
							//  0 0 0 leftCorners[0]
							//  0 0 1 leftCorners[1]
							//  0 1 0 leftCorners[2]
							//  0 1 1 leftCorners[3]
							//  1 0 0 leftCorners[4]
							//  1 0 1 leftCorners[5]
							//  1 1 0 leftCorners[6]
							//  1 1 1 leftCorners[7]
							//

							Vector<> leftCorners[8];

							int l = 0;
							for(int i = -1; i < 2; i+=2)
								for(int j = -1; j < 2; j+=2)
									for(int k = -1; k < 2; k+=2)
									{
/*										leftCorners[l][0] = i * (actualNode->m_bisect / 2) + actualNode->m_ptrCorner->operator[]( 0 );
                                        leftCorners[l][1] = j * (actualNode->m_bisect / 2) + actualNode->m_ptrCorner->operator[]( 1 );
                                        leftCorners[l][2] = k * (actualNode->m_bisect / 2) + actualNode->m_ptrCorner->operator[]( 2 );
*/
										leftCorners[l][0] = i;
                                        leftCorners[l][1] = j;
                                        leftCorners[l][2] = k;
										l++;
									}

							//  Rearrange corners according to basic (unrotated) Hilbert cube.
							//

							swap(leftCorners[3], leftCorners[0]);
							swap(leftCorners[3], leftCorners[2]);
							swap(leftCorners[6], leftCorners[4]);
							swap(leftCorners[6], leftCorners[5]);

                            //  Rotate Hilbert cube. This is a must.
							//

							Vector<> * finalLeftCorners[8];
							Quaternion<> * hilbertOrientations[8];

							for(int i = 0; i < 8; i++)
							{
								finalLeftCorners[i] = new Vector<>;
								hilbertOrientations[i] = new Quaternion<>;
								Quaternion<> qTemp;
								Matrix<> m(4, 4);
								Vector<> vec;

								switch (i)
								{
								case 0:
									{
									vec.Create(0, 1, 0);
									hilbertOrientations[i]->CreateFromEulerian( vec, -90.0 );
									hilbertOrientations[i]->CreateMatrix( m );
									actualNode->m_ptrHilbertOrientation->CreateMatrix( m );;
									*hilbertOrientations[i] = *hilbertOrientations[i] * *actualNode->m_ptrHilbertOrientation;
									hilbertOrientations[i]->CreateMatrix( m );
									vec.Create(1, 0, 0);
									qTemp.CreateFromEulerian( vec, +90 );
									qTemp.CreateMatrix( m );
									*hilbertOrientations[i] = qTemp * *hilbertOrientations[i];
									hilbertOrientations[i]->CreateMatrix( m );
									}
									break;

								case 1:
								case 2:
									{
									vec.Create(0, 1, 0);
									hilbertOrientations[i]->CreateFromEulerian( vec, +90 );
									hilbertOrientations[i]->CreateMatrix( m );
									actualNode->m_ptrHilbertOrientation->CreateMatrix( m );;
									*hilbertOrientations[i] = *hilbertOrientations[i] * *actualNode->m_ptrHilbertOrientation;
									hilbertOrientations[i]->CreateMatrix( m );
									vec.Create(0, 0, 1);
									qTemp.CreateFromEulerian( vec, +90 );
									qTemp.CreateMatrix( m );
									*hilbertOrientations[i] = qTemp * *hilbertOrientations[i];
									hilbertOrientations[i]->CreateMatrix( m );
									}
									break;

								case 3:
								case 4:
									{
									vec.Create(1, 0, 0);
									hilbertOrientations[i]->CreateFromEulerian( vec, -180 );
									hilbertOrientations[i]->CreateMatrix( m );
									actualNode->m_ptrHilbertOrientation->CreateMatrix( m );;
									*hilbertOrientations[i] = *hilbertOrientations[i] * *actualNode->m_ptrHilbertOrientation;
									hilbertOrientations[i]->CreateMatrix( m );
									}
									break;

								case 5:
								case 6:
									{
									vec.Create(0, 1, 0);
									hilbertOrientations[i]->CreateFromEulerian( vec, -90 );
									hilbertOrientations[i]->CreateMatrix( m );
									actualNode->m_ptrHilbertOrientation->CreateMatrix( m );;
									*hilbertOrientations[i] = *hilbertOrientations[i] * *actualNode->m_ptrHilbertOrientation;
									hilbertOrientations[i]->CreateMatrix( m );
									vec.Create(0, 0, 1);
									qTemp.CreateFromEulerian( vec, -90 );
									qTemp.CreateMatrix( m );
									*hilbertOrientations[i] = qTemp * *hilbertOrientations[i];
									hilbertOrientations[i]->CreateMatrix( m );
									}
									break;

								case 7:

									{
									vec.Create(0, 1, 0);
									hilbertOrientations[i]->CreateFromEulerian( vec, +90 );
									hilbertOrientations[i]->CreateMatrix( m );
									actualNode->m_ptrHilbertOrientation->CreateMatrix( m );;
									*hilbertOrientations[i] = *hilbertOrientations[i] * *actualNode->m_ptrHilbertOrientation;
									hilbertOrientations[i]->CreateMatrix( m );
									vec.Create(1, 0, 0);
									qTemp.CreateFromEulerian( vec, +90 );
									qTemp.CreateMatrix( m );
									*hilbertOrientations[i] = qTemp * *hilbertOrientations[i];
									hilbertOrientations[i]->CreateMatrix( m );
									}
									break;
								}
/*
								Vector<> shift;
								for(int j = 0; j < 3; j++)
								{
									shift[j] = leftCorners[i][j] + (actualNode->m_bisect / 2);
									leftCorners[i][j] -= shift[j];
								}
*/
								//  now rotate corner
								actualNode->m_ptrHilbertOrientation->CreateMatrix( m );
								ModelMatrix rm( *actualNode->m_ptrHilbertOrientation );
								rm.ChangePosition( leftCorners[i] );
								(*finalLeftCorners[i])[0] = rm[12];
								(*finalLeftCorners[i])[1] = rm[13];
								(*finalLeftCorners[i])[2] = rm[14];
								
								for(int j = 0; j < 3; j++)
								{
									if ( (*finalLeftCorners[i])[j] > 0.1 )	//  == 1
											(*finalLeftCorners[i])[j] = actualNode->m_ptrCorner->operator[]( j ) + (actualNode->m_bisect / 2.);
										else
											(*finalLeftCorners[i])[j] = actualNode->m_ptrCorner->operator[]( j );
								}
/*
								//  finalLeftCorner still contain only 1s and 0s.
								//  Add and multiply them to actual node's corner by bisect.
								*finalLeftCorners[i] *= bisect;
								*finalLeftCorners[i] += *actualNode->m_ptrCorner;
								*/
							}

							//  Space partitioning according to Hilbert's curve
							//

							for(int i = 0; i < 8; i++)
							{
								//  we get rightCorner by adding bisect to every component of corner vector
								Vector<> rightCorner = *finalLeftCorners[i];
								for(int m = 0; m < 3; m++)  rightCorner[m] += bisect;

								int count = 0;
								size_t pos = 0;	//  nested, it don't interfere with pos outside this scope

								_TNode * node = NULL;

								//  find out how many leaves belong to this partition, so iterate over all leaves
								for(size_t k = 0; k < actualNode->m_leavesList.size(); k++)
								{
									//  Look at ">=" and "<" operators. They determinate which leave belong to this partition.
									if ( (actualNode->m_leavesList[k]->m_position >= *finalLeftCorners[i]) && (actualNode->m_leavesList[k]->m_position < rightCorner) )
									{  //  leave m_leavesList[k] belongs to partion given by corner vector
										if ( count == 0 )
										{
											pos = k;
											count++;
										}
										else
										{
											if ( count == 1)		//  true only ones
											//	at least two leaves -> create new node
											{
    											node = new _TNode( actualNode, finalLeftCorners[i], bisect, hilbertOrientations[i] );

												if ( m_octTree.size() == level+1 )
												{
													m_octTree.push_back( new std::vector<_TNode*> );
												}

												actualNode->m_childList.push_back( node );	//  and finaly add this node between actualNode's children
												m_octTree[level+1]->push_back( node );		//  add to tree

												node->m_leavesList.push_back( actualNode->m_leavesList[pos] );	//  add previous
											}

											node->m_leavesList.push_back( actualNode->m_leavesList[k] );
											count++;	
										}
									} //  end if

								}  //  end for of cycling through collection of positions

								if (count == 1)	//  only one child in this node so it's leave
								{
									actualNode->m_childList.push_back( actualNode->m_leavesList[pos] );
								}
								if ( (count == 0) || (count == 1) )		//  Hilbert's corner
								{
									delete finalLeftCorners[i];
									delete hilbertOrientations[i];
								}
							
							}  //  end of cycling through corners
							//  
							//	space partitioning according to Hilbert's curve

							pos++;

						}	//  end while( pos < mostLeftAncestors[level].getCount() )

						bisect /= 2;
						level++;
					
					}	//  end while( level < mostLeftAncestors.getCount() )

					//
					//  end Tree construction					
				}

				void swap(Vector<> & s1, Vector<> & s2)
				{
					Vector<> tmp = s1;
					s1 = s2;
					s2 = tmp;
				}

				virtual ~OctTree()
				{
                    FreeTree();
				}

			protected:
				
				std::vector< std::vector<_TNode*>* > m_octTree;

				void FreeTree()
				{
					//  free memory allocated for tree rows
					for(size_t i = 0; i < m_octTree.size(); i++)
					{
						//  delete node on row
						for(size_t j = 0; j < m_octTree[i]->size(); j++)
							delete m_octTree[i]->at(j);
						//  delete a row
						delete m_octTree[i];
					}
					m_octTree.clear();
				}

			}; //  end class OctTree


		}	//  end namespace Data

	}	//  end namespace Std

}	//  end namespace JakubSchwarzmeier

#endif	//  OCTTREE_H
