/************************************************************
*
*  Copyright (c) 2002-2004 Jakub Schwarzmeier
*
************************************************************/

#include <schwarzmeier/std.h>

using namespace JakubSchwarzmeier::Std::Graphics;
using namespace JakubSchwarzmeier::Std::Math;

//  constructor

RotationMatrix::RotationMatrix(): Matrix<double>(4,4)
{
	LoadIdentity();
	
	m_q.LoadIdentity();
}

RotationMatrix::RotationMatrix( const Quaternion<> & q ): Matrix<double>(4,4)
{
	LoadIdentity();

	m_q = q;

	//  Transform rotation
	m_q.Normalize();
	m_q.CreateMatrix( m_values );

	for(int i = 0; i < 3; i++)	m_values[3 + 4 * i] = 0;
	for(int i = 0; i < 3; i++)	m_values[12 + i] = 0;
	m_values[15] = 1;
}

RotationMatrix::RotationMatrix( const Vector<> & v, const Quaternion<> & q ): Matrix<double>(4,4)
{
	LoadIdentity();

	m_q = q;

	//  Transform rotation
	m_q.Normalize();
	m_q.CreateMatrix( m_values );

	for(int i = 0; i < 3; i++)	m_values[12 + i] = v[i];
	for(int i = 0; i < 3; i++)	m_values[3 + 4 * i] = 0;
	m_values[15] = 1;
}

//  "copy constructor" is inherited by derived classes
RotationMatrix::RotationMatrix( const RotationMatrix & original ): Matrix<double>( original )
{
	// calls Matrix assigment operator (it's the same code)
	operator=(original);
}

RotationMatrix::RotationMatrix( const Matrix<> & original ): Matrix<double>( original )
{
	operator=(original);
}

RotationMatrix & RotationMatrix::operator=( const RotationMatrix & right_side )
{	
	if ( this == &right_side ) return *this;
	Matrix<double>::operator=( right_side );
	m_q = right_side.m_q;
	return *this;
}

void RotationMatrix::Transform( Quaternion<> & rot )
{
	double originalPos[4];
	for(int i = 0; i < 4; i++)	originalPos[i] = m_values[12 + i];
    Matrix<> originalPosition(1, 4, originalPos);

	//  Transform rotation
	m_q.Normalize();
	m_q.CreateMatrix( m_values );

	//  Transform position
	Matrix<> resultPosition(1, 4);
	Matrix<> rotationalMatrix(4, 4);
	
	rot.CreateMatrix( rotationalMatrix );
	for(int i = 0; i < 3; i++)	rotationalMatrix[3 + 4 * i] = 0;
	for(int i = 0; i < 3; i++)	rotationalMatrix[12 + i] = 0;
	rotationalMatrix[15] = 1;

	resultPosition = originalPosition * rotationalMatrix;

	for(int i = 0; i < 4; i++)	m_values[12 + i] = resultPosition[i];

}

void RotationMatrix::ChangeHeading( double angle )	//  y
{
	//  Create transformation matrix
	Vector<> vec(0.0, 1.0, 0.0);
	Quaternion<> rot( vec, -angle );
	m_q = rot * m_q;
	Transform(rot);
}

void RotationMatrix::ChangePitch( double angle )	//  x
{
	Vector<> vec(1, 0, 0);
	Quaternion<> rot( vec, -angle );
	m_q = rot * m_q;
	Transform(rot);
}

void RotationMatrix::ChangeRoll( double angle )		//  z
{
	Vector<> vec(0, 0, 1);
	Quaternion<> rot( vec, -angle );
	m_q = rot * m_q;
	Transform(rot);
}

void RotationMatrix::ChangeRotation( Matrix<> & rotation )
{
	*this = *this * rotation;
}

void RotationMatrix::ChangePosition( Vector<> & position )
{
	const double positionOperator[16]={ 1,	0,	0,	0,
										0,	1,	0,	0,
										0,	0,	1,	0,
										-position[0], -position[1], -position[2], 1 };
	
	*this = *this * positionOperator;
}

Vector<> RotationMatrix::GetZAxis()		//  3rd row
{
	Vector<> vec = m_q.GetZAxis();
    
	return vec;
}

Vector<> RotationMatrix::GetYAxis()		//  2nd row
{
	Vector<> vec = m_q.GetYAxis();
    
	return vec;
}

Vector<> RotationMatrix::GetXAxis()		//  1st row
{
	Vector<> vec = m_q.GetXAxis();
    
	return vec;
}

Quaternion<> RotationMatrix::GetQuaternion()
{
	return m_q;
}



void RotationMatrix::ChangeLambda( double angle )
{
	Vector<> vec(0, 1, 0);
	Quaternion<> rot( vec, -angle );
	m_q = m_q * rot;
	Transform(rot);
}

void RotationMatrix::ChangePhi( double angle )
{
	Vector<> vec(1, 0, 0);
	Quaternion<> rot( vec, -angle );
	m_q = m_q * rot;
	Transform(rot);
}

void RotationMatrix::ChangeTheta( double angle )
{
	Vector<> vec(0, 0, 1);
	Quaternion<> rot( vec, -angle );
	m_q = m_q * rot;
	Transform(rot);
}


void RotationMatrix::ResetAngles(double lambda, double phi, double theta)
{
	Vector<> pos = GetPosition();

	LoadIdentity();
	m_q.LoadIdentity();
	SetPosition( pos );

	ChangeHeading( lambda );
	ChangePitch( phi );
	ChangeRoll( theta );
}
/*
double RotationMatrix::GetLambda()
{
	Vector<> vec1( m_values[0], 0.0, m_values[2] );
	Vector<> vec2( 1.0, 0.0, 0.0 );
	double angle1 = Functions::acos_angle( scalar_product( vec1, vec2 ) / ( vec1.GetLength() * vec2.GetLength() ) );

	if (m_values[2] < 0)
	{
		angle1 = 360-angle1;
	}

	return angle1;
}

double RotationMatrix::GetPhi()
{
	Vector<> vec1( 0.0, m_values[9], m_values[10] );
	Vector<> vec2( 0.0, 0.0, 1.0 );
	double angle1 = Functions::acos_angle( scalar_product( vec1, vec2 ) / ( vec1.GetLength() * vec2.GetLength() ) );

	if (m_values[9] < 0)
	{
		angle1 = 360-angle1;
	}

	return angle1;
}

double RotationMatrix::GetTheta()
{
	Vector<> vec1( m_values[4], m_values[5], 0.0 );
	Vector<> vec2( 0.0, 1.0, 0.0 );
	double angle1 = Functions::acos_angle( scalar_product( vec1, vec2 ) / ( vec1.GetLength() * vec2.GetLength() ) );

	if (m_values[4] < 0)
	{
		angle1 = 360-angle1;
	}

	return angle1;
}
*/

Vector<> RotationMatrix::GetPosition() const
{
	Matrix<> retMatrix(4,4);
	InverseInto( retMatrix );

	return Vector<>(retMatrix[12], retMatrix[13], retMatrix[14]);
//	return Vector<>( m_values[12], m_values[13], m_values[14]);
}



void RotationMatrix::SetPosition( Vector<> & position)
{
	m_values[12] = -position[0];
	m_values[13] = -position[1];
	m_values[14] = -position[2];
}
/*
void RotationMatrix::SetLambda( double angle )
{
	RotationMatrix newCam;
	newCam.ChangeHeading( angle );
	newCam.ChangePitch( GetPhi() );
	newCam.ChangeRoll( GetTheta() );
	for ( int i = 0; i < 12; i++ ) (*this)[i] = newCam[i];
}

void RotationMatrix::SetPhi( double angle )
{
	RotationMatrix newCam;
	newCam.ChangeHeading( GetLambda() );
	newCam.ChangePitch( angle );
	newCam.ChangeRoll( GetTheta() );
	for ( int i = 0; i < 12; i++ ) (*this)[i] = newCam[i];
}

void RotationMatrix::SetTheta( double angle )
{
	RotationMatrix newCam;
	newCam.ChangeHeading( GetLambda() );
	newCam.ChangePitch( GetPhi() );
	newCam.ChangeRoll( angle );
	for ( int i = 0; i < 12; i++ ) (*this)[i] = newCam[i];
}
*/

RotationMatrix & RotationMatrix::operator=( const Matrix<> & right_side )
{	
	if ( this == &right_side ) return *this;
	Assign(*this, right_side);
	return *this;
}

