/************************************************************
*
*  Copyright (c) 2002-2004 Jakub Schwarzmeier
*
************************************************************/

#include <schwarzmeier/std.h>

using namespace JakubSchwarzmeier::Std::Graphics;
using namespace JakubSchwarzmeier::Std::Math;

//  constructor

ModelMatrix::ModelMatrix(): Matrix<double>(4,4)
{
	LoadIdentity();
	
	m_q.LoadIdentity();
}

ModelMatrix::ModelMatrix( const Quaternion<> & q ): Matrix<double>(4,4)
{
	LoadIdentity();

	m_q = q;

	//  Transform rotation
	m_q.Normalize();
	m_q.CreateMatrix( m_values );

	for(int i = 0; i < 3; i++)	m_values[3 + 4 * i] = 0;
	for(int i = 0; i < 3; i++)	m_values[12 + i] = 0;
	m_values[15] = 1;
}

//  "copy constructor" is inherited by derived classes
//  eg.
//  Vector<> pos1;
//  Vector<> pos2 = pos1;
ModelMatrix::ModelMatrix( const ModelMatrix & original ): Matrix<double>( original )
{
	// calls Matrix assigment operator (it's the same code)
	operator=(original);
}

ModelMatrix & ModelMatrix::operator=( const ModelMatrix & right_side )
{	
	if ( this == &right_side ) return *this;
	Matrix<double>::operator=( right_side );
	m_q = right_side.m_q;
	return *this;
}

void ModelMatrix::Transform( Quaternion<> & rot )
{
	//  Transform rotation
	m_q.Normalize();
	m_q.CreateMatrix( m_values );

	//  Transform position
	Matrix<> resultPosition(1, 4);
	Matrix<> rotationalMatrix(4, 4);
    Matrix<> originalPosition(1, 4, &m_values[12]);
	
	rot.CreateMatrix( rotationalMatrix );
	for(int i = 0; i < 3; i++)	rotationalMatrix[3 + 4 * i] = 0;
	for(int i = 0; i < 3; i++)	rotationalMatrix[12 + i] = 0;
	rotationalMatrix[15] = 1;

	resultPosition = originalPosition * rotationalMatrix;

	for(int i = 0; i < 4; i++)	m_values[12 + i] = resultPosition[i];
}

void ModelMatrix::ChangeHeading( double angle )	//  y
{
	//  Create transformation matrix
	Vector<> vec(0, 1, 0);
	Quaternion<> rot( vec, angle );
	m_q = m_q * rot;
	Transform(rot);
}

void ModelMatrix::ChangePitch( double angle )	//  x
{
	Vector<> vec(1, 0, 0);
	Quaternion<> rot( vec, angle );
	m_q = m_q * rot;
	Transform(rot);
}

void ModelMatrix::ChangeRoll( double angle )		//  z
{
	Vector<> vec(0, 0, 1);
	Quaternion<> rot( vec, angle );
	m_q = m_q * rot;
	Transform(rot);
}

void ModelMatrix::ChangeRotation( Matrix<> & rotation )
{
	*this = rotation * *this;
}

void ModelMatrix::ChangePosition( Vector<> & position )
{
	const double positionOperator[16]={ 1,	0,	0,	0,
										0,	1,	0,	0,
										0,	0,	1,	0,
										position[0], position[1], position[2], 1 };
	
	*this = positionOperator * *this;
}

Vector<> ModelMatrix::GetZAxis()		//  3rd row
{
	Vector<> vec = m_q.GetZAxis();
    
	return vec;
}

Vector<> ModelMatrix::GetYAxis()		//  2nd row
{
	Vector<> vec = m_q.GetYAxis();
    
	return vec;
}

Vector<> ModelMatrix::GetXAxis()		//  1st row
{
	Vector<> vec = m_q.GetXAxis();
    
	return vec;
}

Quaternion<> ModelMatrix::GetQuaternion()
{
	return m_q;
}

void ModelMatrix::ChangeLambda( double angle )
{
	// nezkouseno
	Vector<> vec(0, 1, 0);
	Quaternion<> rot( vec, -angle );
	m_q = m_q * rot;
	Transform(rot);
}

double ModelMatrix::GetLambda()
{
	Matrix<> retMatrix(4,4);
	InverseInto( retMatrix );

	double result = Math::Functions::atan2_angle( retMatrix[8], retMatrix[10] );
	while (result < 0)
		result += 360;

	return result;
}

double ModelMatrix::GetPhi()
{
	Matrix<> retMatrix(4,4);
	InverseInto( retMatrix );

	double result = - Math::Functions::atan2_angle( retMatrix[9], pow( (pow(retMatrix[8],2)+pow(retMatrix[10],2)) , 0.5) );

	return result;
}

double ModelMatrix::GetTheta()
{
	ModelMatrix calInit;
	calInit.ChangeLambda( GetLambda() );
	calInit.ChangePitch( GetPhi() );
	calInit.Inverse();
	
	Matrix<> retMatrix(4,4);
	InverseInto( retMatrix );

	Vector<> u(calInit[0], calInit[1], calInit[2]);
	Vector<> v(retMatrix[0], retMatrix[1], retMatrix[2]);
	double result = Math::Functions::acos_angle( scalar_product( u, v )/( u.GetLength()*v.GetLength() ) );

	return result;
}

Vector<> ModelMatrix::GetPosition()
{
	Matrix<> retMatrix(4,4);
	InverseInto( retMatrix );

	return Vector<>(retMatrix[12], retMatrix[13], retMatrix[14]);
}



void ModelMatrix::SetPosition( Vector<> & position)
{
	m_values[12] = -position[0];
	m_values[13] = -position[1];
	m_values[14] = -position[2];
}

void ModelMatrix::SetLambda( double angle )
{
	ModelMatrix newCam;
	newCam.ChangeHeading( angle );
	newCam.ChangePitch( GetPhi() );
	newCam.ChangeRoll( GetTheta() );
	for ( int i = 0; i < 12; i++ ) (*this)[i] = newCam[i];
}

void ModelMatrix::SetPhi( double angle )
{
	ModelMatrix newCam;
	newCam.ChangeHeading( GetLambda() );
	newCam.ChangePitch( angle );
	newCam.ChangeRoll( GetTheta() );
	for ( int i = 0; i < 12; i++ ) (*this)[i] = newCam[i];
}

void ModelMatrix::SetTheta( double angle )
{
	ModelMatrix newCam;
	newCam.ChangeHeading( GetLambda() );
	newCam.ChangePitch( GetPhi() );
	newCam.ChangeRoll( angle );
	for ( int i = 0; i < 12; i++ ) (*this)[i] = newCam[i];
}


ModelMatrix & ModelMatrix::operator=( const Matrix<> & right_side )
{	
	if ( this == &right_side ) return *this;
	Assign(*this, right_side);
	return *this;
}
