#include <schwarzmeier/std.h>

#include <exception>

using namespace JakubSchwarzmeier::Std::Math;

size_t ValuePair::GetCount()
{
	return m_x.size();
}

void ValuePair::Add( double x, double y )
{
	m_x.push_back( x );
	m_y.push_back( y );
}

std::vector<double> & ValuePair::GetX()
{
	return m_x;
}

std::vector<double> & ValuePair::GetY()
{
	return m_y;
}

// the Points property, Get method
ValuePair InterpolationApproximation::GetPoints()
{
	return m_values;
}

// the Points property, Set method
void InterpolationApproximation::SetPoints( ValuePair & values )
{
    m_values = values;
}
				
Spline::Spline()
{
		m_initialized = false;
}

Spline::~Spline()
{
	if (m_initialized)	{ delete [] m_y2; };
}

double Spline::GetValue( double point )
{
	std::vector<double> & x = m_values.GetX();
	std::vector<double> & y = m_values.GetY();
	size_t n = m_values.GetCount();

	if ( ( point < x [ 0 ] ) || ( point > (x [ n - 1 ]) ) ) throw new std::exception; // InterpolationApproximation point out of dataset.

	if ( !m_initialized )
	{
		//  this is the first run, initialize first
		m_y2 = new double [n];

		//  inputs
		double yp1, ypn;
		yp1 = ypn = 1.99e30;

		double p, qn, sig, un;
		double * u = new double [n];

		if ( yp1 > 0.99e30 ) // The lower boundary condition is set either to be "natural"
			m_y2 [0] = u [0] = 0.0;
		else               //  or else to have a specified first derivative.
		{
			m_y2 [0] = -0.5;
			u [0] = ( 3.0 / ( x [1] - x [0] ) ) * ( ( y [1] - y [0] ) / ( x [1] - x [0] ) - yp1 );
		}

		for ( int i = 1; i < n - 1; i++ )
		{   //  This is the decomposition loop of the tridiagonal algorithm.
			//  m_y2 and u are used for temporary storage of the decomposed factors. 
			sig = ( x [i] - x [i - 1] ) / ( x [i + 1] - x [i - 1] );
			p = sig * m_y2 [i - 1] + 2.0;
			m_y2 [i] = ( sig - 1.0 ) / p;
			u [i] = ( y [i + 1] - y [i] ) / ( x [i + 1] - x [i] ) - ( y [i] - y [i - 1] ) / ( x [i] - x [i - 1] );
			u [i] = ( 6.0 * u [i] / ( x [i + 1] - x [i - 1] ) - sig * u [i - 1] ) / p;
		}

		if ( ypn > 0.99e30 )  //  The upper boundary condition is set either to be "natural" 
			qn = un = 0.0;
		else                //  or else to have a specified first derivative.
		{
			qn = 0.5;
			un = ( 3.0 / ( x [n - 1] - x [n - 2] ) ) * ( ypn - ( y [n - 1] - y [n - 2] ) / ( x [n - 1] - x [n - 2] ) );
		}

		m_y2 [n - 1] = ( un - qn * u [n - 2] ) / ( qn * m_y2 [n - 2] + 1.0 );

		//  This is the backsubstitution loop of the tridiagonal algorithm. 
		for ( int i = n - 2; i >= 0; i-- )
			m_y2 [i] = m_y2 [i] * m_y2 [i + 1] + u [i];

		delete [] u;

		m_initialized = true;
	}

	//  return function value in requested point

	int klo, khi, k;
	double h, b, a;

	/*  We will find the right place in the table by means of bisection. This is optimal if sequential 
	calls to this routine are at random values of x. If sequential calls to store 
	previous values of klo and khi and test if they remain appropriate on the next call.  
*/
	klo = 0;
	khi = n - 1;
	while ( khi - klo > 1 )
	{
		k = ( khi + klo ) >> 1;
		if ( x [k] > point )
			khi = k;
		else
			klo = k;

	} //  klo and khi now bracket the input value of x.

	h = x [khi] - x [klo];
	if ( h == 0.0 ) throw new std::exception; //  "Bad xa input to routine splint" //  The xa's must be distinct.
	a = ( x [khi] - point ) / h;
	b = ( point - x [klo] ) / h; //  Cubic spline polynomial is now evaluated.

	return a * y [klo] + b * y [khi] + ( ( a * a * a - a ) * m_y2 [klo] + ( b * b * b - b ) * m_y2 [khi] ) * ( h * h ) / 6.0;
}

Spline2::Spline2()
{
	m_initialized = false;
	m_lasti = 0;
}

Spline2::~Spline2()
{
	if (m_initialized)	{ delete [] m_coef; }
}

double Spline2::GetValue( double point )
{
	if ( ( point < m_values.GetX() [ 0 ] ) || ( point > m_values.GetX() [ m_values.GetCount() - 1 ] ) ) throw new std::exception; // "InterpolationApproximation point out of dataset." 

	if ( !m_initialized )
	{
		m_coef = new double [3 * m_values.GetCount() ];
		spline( m_coef, m_values.GetX(), m_values.GetY(), m_values.GetCount() );

		m_initialized = true;
	}

	return seval( point, m_values.GetX(), m_values.GetY(), m_coef, m_values.GetCount() );
}				

/*
* SPLDIF: evaluate derivative of cubic spline.
*	double x0;              point at which to compute function
*	double x[], y[];        data points to interpolate between
*	double coef[];          storage for computed coefficients
*	int n;                number of data points
*/
double Spline2::SplDif( double point )
{
	if ( ( point < m_values.GetX() [ 0 ] ) || ( point > m_values.GetX() [ m_values.GetCount() - 1 ] ) ) throw new std::exception; // "InterpolationApproximation point out of dataset." 

	if ( !m_initialized )
	{
		m_coef = new double [3 * m_values.GetCount()];
		spline( m_coef, m_values.GetX(), m_values.GetY(), m_values.GetCount() );

		m_initialized = true;
	}

	return spldif( point, m_values.GetX(), m_values.GetY(), m_coef, m_values.GetCount() );
}

void Spline2::spline( double coef[], std::vector<double> & x, std::vector<double> & y, int n )
{
	splsub( coef, x, y, n );
//            splsub(coef[0], coef[n], coef[2 * n], x, y, n);
	m_lasti = -1;         /* at least reset the last index */
}

/*
* SPLSUB: local worker routine that sets up the spline coeffs
*/
void Spline2::splsub( double b[],
	// double [] b, double []b, double []b,   /* 1st to 3rd order polynomial coefs */
	std::vector<double> & x, std::vector<double> & y,
	int n )
{
	int i;
	double t1, tn, t;

	if ( n < 3 )
	{
//        error("spline: n=%b lt 3",n);
		return;
	}
	b [2 * n + 0] = x [1] - x [0];
	b [n + 1] = ( y [1] - y [0] ) / b [2 * n + 0];
	for ( i = 1; i <= n - 2; i++ )
	{
		b [2 * n + i] = x [i + 1] - x [i];
		b [i] = 2.0 * ( b [2 * n + i - 1] + b [2 * n + i] );
		b [n + i + 1] = ( y [i + 1] - y [i] ) / b [2 * n + i];
		b [n + i] = b [n + i + 1] - b [n + i];
	}
	b [0] = -b [2 * n + 0];
	b [n - 1] = -b [2 * n + n - 2];
	if ( n == 3 )
	{
		b [n + 0] = 0.0; b [n + n - 1] = 0.0;     /* 3b1 compiler chocked if one stmt. */
	}
	else
	{
		t1 = b [n + 2] / ( x [3] - x [1] ) - b [n + 1] / ( x [2] - x [0] );
		tn = b [n + n - 2] / ( x [n - 1] - x [n - 3] ) - b [n + n - 3] / ( x [n - 2] - x [n - 4] );
		b [n + 0] = t1 * b [2 * n + 0] * b [2 * n + 0] / ( x [3] - x [0] );
		b [n + n - 1] = -tn * b [2 * n + n - 2] * b [2 * n + n - 2] / ( x [n - 1] - x [n - 4] );
	}
	for ( i = 1; i <= n - 1; i++ )
	{
		t = b [2 * n + i - 1] / b [i - 1];
		b [i] = b [i] - t * b [2 * n + i - 1];
		b [n + i] = b [n + i] - t * b [n + i - 1];
	}
	b [n + n - 1] = b [n + n - 1] / b [n - 1];
	for ( i = n - 2; i >= 0; i-- )
		b [n + i] = ( b [n + i] - b [2 * n + i] * b [n + i + 1] ) / b [i];
	b [n - 1] = ( y [n - 1] - y [n - 2] ) / b [2 * n + n - 2] + b [2 * n + n - 2] * ( b [n + n - 2] + 2 * b [n + n - 1] );
	for ( i = 0; i <= n - 2; i++ )
	{
		b [i] = ( y [i + 1] - y [i] ) / b [2 * n + i] - b [2 * n + i] * ( b [n + i + 1] + 2 * b [n + i] );
		b [2 * n + i] = ( b [n + i + 1] - b [n + i] ) / b [2 * n + i];
		b [n + i] = 3 * b [n + i];
	}
	b [n + n - 1] = 3 * b [n + n - 1];
	b [2 * n + n - 1] = b [2 * n + n - 2];
}

/*
* SEVAL: evaluate cubic spline interpolation, using Horner's rule
*	double x0;              point at which to compute function 
*	double x[], y[];        data points to interpolate between 
*	double coef[];          storage for computed coefficients
*	int n;                number of data points 
*/

double Spline2::seval( double x0, std::vector<double> & x, std::vector<double> & y, double coef[], int n )
{
	int i;
	double u;

	i = interval( x0, x, n );
	u = x0 - x [i];
	return y [i] + u * ( coef [i] + u * ( coef [n + i] + u * coef [2 * n + i] ) );
}



double Spline2::spldif( double x0, std::vector<double> & x, std::vector<double> & y, double coef[], int n )
{

//            return (Interpolate(x0 + u) - Interpolate(x0)) / u;

	int i;
	double u;

	i = interval( x0, x, n );
	u = x0 - x [i];
	return coef [i] + u * ( 2.0 * coef [i + n] + u * 3.0 * coef [i + 2 * n] );
}

/*
* SPLDIF2: evaluate second derivative of cubic spline
*	double x0;              point at which to compute function
*	double x[], y[];        data points to interpolate between
*	double coef[];          storage for computed coefficients
*	int n;                number of data points
*/

double Spline2::spldif2( double x0, std::vector<double> & x, std::vector<double> & y, double coef[], int n )
{
	int i;
	double u;

	i = interval( x0, x, n );
	u = x0 - x [i];
	return 2.0 * coef [i + n] + u * 6.0 * coef [i + 2 * n];
}

/*
* INTERVAL: find the correct interval, optimized to first try
*           the last interval tried. ValuePair beyond the two edges
*           use the edge data
*/
int Spline2::interval( double u, std::vector<double> & x, int n )
{
	int i, j, k;

	if ( m_lasti >= 0 && m_lasti < n - 1 && x [m_lasti] <= u && u < x [m_lasti + 1] ) /* check last */
		return m_lasti;
	else if ( u < x [0] )             /* check extrapolation left */
		m_lasti = 0;
	else if ( u >= x [n - 1] )          /* and right */
		m_lasti = n - 1;
	else
	{                          /* hence u wasn't in last interval */
		i = 0;                      /* and a binary search is performed */
		k = n;
		while ( i + 1 < k )
		{
			j = ( i + k ) / 2;
			if ( x [j] <= u )
				i = j;
			else
				k = j;
		}
		m_lasti = i;                  /* and remember this interval       */
	}
	return m_lasti;
}

