/************************************************************
*
*  Copyright (c) 2002-2004 Jakub Schwarzmeier
*
************************************************************/

#include <schwarzmeier/std.h>

#include <exception>
#include <stdlib.h>	//  srand
#include <stdio.h>

using namespace JakubSchwarzmeier::Std::Math;

AdditiveIntegral::AdditiveIntegral()
{
	m_integral = 0;
	m_xi = 0;
}

double AdditiveIntegral::Integrate(Functions::Integral func, double upper, int steps)
{
	double h = (upper - m_xi) / ( double ) steps;

	for ( int k = 0; k < steps; k++ )
	{
		m_integral += h * ( func( m_xi ) + func( m_xi + h ) ) / 2;
		if ( ( m_integral == m_integral) == false )	//  m_integral is NaN
		{
			throw new std::exception;
		}

		m_xi = m_xi + h;
	}

	m_xi = m_xi - h;

	return m_integral;
}


long double Functions::sin_angle(const long double angle)
{
	long double res;
	if ((angle == 360) || (angle == 180))
		res = 0;
	else
		res = sin(angle * M_PI / 180.0);
	return res;
}

long double Functions::cos_angle(const long double angle)
{
	long double res;
	if ((angle == 90) || (angle == 270) || (angle == -90) || (angle == -270) )
		res = 0;
	else
		res = cos(angle * M_PI / 180.0);
	return res;
}


long double Functions::tan_angle(const long double angle)
{
	return tan(angle * M_PI / 180.0);
}

long double Functions::asin_angle(const double x)
{
  long double res;
  if (ceil(x) == -1.0) res = -90.0;
  else
  if (floor(x) == 1.0) res = 90;
  else
  res = asin(x) * 180.0 / M_PI;
  return res;
 
}

long double Functions::acos_angle(const double x)
{
  long double res;
  if (floor(x) == 1.0) res = 0;
  else
  if (ceil(x) == -1.0) res = 180.0;
  else
  res = acos(x) * 180.0 / M_PI;
  return res;
}

long double Functions::atan_angle(const double x)
{
  return atan(x) * 180.0 / M_PI;
}

long double Functions::atan2_angle(const double y, const double x)
{
  return atan2(y, x) * 180.0 / M_PI;
}



//  taken from Numerical recipes in C
//

//  Returns a uniform random deviate between 0.0 and 1.0.
// idum < 1 means initialization or reinitialization
double Functions::Random(long idum)	
{
	static long MBIG = 1000000000;
	static long MSEED = 161803398;
	static long MZ = 0;
	double FAC = 1.0/MBIG;

	static int inext,inextp;
	static long ma[56];
	static int iff=0;
	long mj,mk;
	int i,ii,k;
	if (idum < 0 || iff == 0)
	{ 
		iff=1;
		mj=labs(MSEED-labs(idum)); 
		mj %= MBIG;
		ma[55]=mj;
		mk=1;
		
		for (i=1;i<=54;i++)
		{
			ii=(21*i) % 55;
			ma[ii]=mk;
			mk=mj-mk;
			if (mk < MZ) mk += MBIG;
			mj=ma[ii];
		}

		for (k=1;k<=4;k++)
			for (i=1;i<=55;i++)
			{
				ma[i] -= ma[1+(i+30) % 55];
				if (ma[i] < MZ) ma[i] += MBIG;
			}
			inext=0; 
			inextp=31; 
	}

	if (++inext == 56) inext=1;
	if (++inextp == 56) inextp=1;
	mj=ma[inext]-ma[inextp]; 
	if (mj < MZ) mj += MBIG;
	ma[inext]=mj; 
	return mj*FAC; 
}
//
//	taken from from Numerical recipes in C

double Functions::TrapezoidalIntegration(Integral func, double lower, double upper, int steps )
{
	double  integral = 0;
	double h = (upper-lower) / steps;
	double xi = lower;

	for( int k = 0; k < steps; k++)
	{
		integral = integral + h * (func(xi) + func(xi+h)) / 2;
		if ( (integral == integral) == false )		//  integral is NaN
		{
			throw new std::exception;
		}

		xi = xi + h;
	}

	return integral;
}


/* Returns the integral of the function func from a to b. The parameters EPS can be set to the
desired fractional accuracy and JMAX so that 2 to the power JMAX-1 is the maximum allowed
number of steps. Integration is performed by Simpson's rule.
*/
double Functions::SimpsonIntegration( Integral f, double lower,
                    double upper, int steps)
{
	static double EPS = 1.0e-6;
	static double JMAX = 20;

	int j;
	double s, st, ost, os;
	ost = os = -1.0e30;
	for ( j = 1; j <= JMAX; j++ )
	{
		st = TrapezoidalIntegration( f, lower, upper, j );
		s = ( 4.0 * st - ost ) / 3.0; //  Compare equation (4.2.4), above.
		if ( j > 5 ) //  Avoid spurious early convergence.
			if ( fabs( s - os ) < EPS * fabs( os ) || ( s == 0.0 && os == 0.0 ) ) return s;
		os = s;
		ost = st;
	}

	throw new std::exception;
	return 0.0; ///  Never get here.
}

/**
 *
 * Purpose:      Maximize a "well-behaved" function
 *
 * Arguments:    f - Pointer to the function
 *               L - The lower bound of the search space
 *               U - The upper bound of the search space
 *               tol - The tolerance level
 *
 * Side Effects: None
 *
 * Returns:      The value of x that maximizes f(x)
 */

//  typedef double (*__func)(double);	//  takto by se slo vyhnout slozitemu zapisu a namisto
										//  prvniho parametru by pak bylo pouze __func f


double Functions::FindMaximum( double (*f)(double), double L, double U, double tol)
{

	double t_L, t_R;

  t_L = L + 0.382*(U - L);
  t_R = L + 0.618*(U - L);

  while ((U-L) > tol) {

	  if ( f(t_L) > f(t_R) ) {

      U = t_R;
      t_R = t_L;
      t_L = L + 0.382*(U - L);

    } else {

      L = t_L;
      t_L = t_R;
      t_R = L + 0.618*(U - L);
    }
  }

  return (U+L)/2.0;
}
double Functions::FindMinimum( double (*f)(double), double L, double U, double tol)
{

	double t_L, t_R;

  t_L = L + 0.382*(U - L);
  t_R = L + 0.618*(U - L);

  while ((U-L) > tol) {

	  if ( f(t_L) < f(t_R) ) {

      U = t_R;
      t_R = t_L;
      t_L = L + 0.382*(U - L);

    } else {

      L = t_L;
      t_L = t_R;
      t_R = L + 0.618*(U - L);
    }
  }

  return (U+L)/2.0;
}
/*
void SHFT2(double & a, double & b, double c)
{
 a=b;
 b=c;
}

void SHFT3(double & a, double & b, double & c, double d)
{
 a=b;
 b=c;
 c=d;
}

#define SWAP(a,b) { double itemp=(a);(a)=(b);(b)=itemp;}
#define SIGN(a,b) ((b) >= 0.0 ? fabs(a) : -fabs(a))
#define MAXINT (ULONG_MAX >> 1)
static double dmaxarg1,dmaxarg2;

template<class T>
inline const T MAX(const T &a, const T &b)
        {return b > a ? (b) : (a);}

void mnbrak(double &ax, double &bx, double &cx, double &fa, double &fb, double &fc,
	double (*func)(double))
{
	const double GOLD=1.618034,GLIMIT=100.0,TINY=1.0e-20;
	double ulim,u,r,q,fu;

	fa=func(ax);
	fb=func(bx);
	if (fb > fa) {
		SWAP(ax,bx);
		SWAP(fb,fa);
	}
	cx=bx+GOLD*(bx-ax);
	fc=func(cx);
	while (fb > fc) {
		r=(bx-ax)*(fb-fc);
		q=(bx-cx)*(fb-fa);
		u=bx-((bx-cx)*q-(bx-ax)*r)/
			(2.0*SIGN(MAX(fabs(q-r),TINY),q-r));
		ulim=bx+GLIMIT*(cx-bx);
		if ((bx-u)*(u-cx) > 0.0) {
			fu=func(u);
			if (fu < fc) {
				ax=bx;
				bx=u;
				fa=fb;
				fb=fu;
				return;
			} else if (fu > fb) {
				cx=u;
				fc=fu;
				return;
			}
			u=cx+GOLD*(cx-bx);
			fu=func(u);
		} else if ((cx-u)*(u-ulim) > 0.0) {
			fu=func(u);
			if (fu < fc) {
				SHFT3(bx,cx,u,cx+GOLD*(cx-bx));
				SHFT3(fb,fc,fu,func(u));
			}
		} else if ((u-ulim)*(ulim-cx) >= 0.0) {
			u=ulim;
			fu=func(u);
		} else {
			u=cx+GOLD*(cx-bx);
			fu=func(u);
		}
		SHFT3(ax,bx,cx,u);
		SHFT3(fa,fb,fc,fu);
	}
}

double Functions::FindMinimum( double (*f)(double), double L, double U, double tol)
{
	double ax=L;
	double bx=U;
	double cx=0;

	double dummy;
	mnbrak(ax, bx, cx, dummy, dummy, dummy, f);

	double x1, x2, f1, f2;
	double r = 0.61803399;
	double c = 1.0 - r;

	double x0 =ax;
	double x3 =cx;
	if (fabs(cx-bx) > fabs(bx-ax)) //{ Make x0 to x1 the smaller segment }
	{
		x1=bx;
		x2=bx+c*(cx-bx);
	}
	else
	{
		x2=bx;
		x1=bx-c*(bx-ax);
	}

	f1=f(x1);
	f2=f(x2);

	while (fabs(x3-x0) > tol*(fabs(x1)+fabs(x2)))
	{
		if (f2 < f1) // then { One possible outcome }
		{
			SHFT3(x0, x1, x2, r*x1+c*x3 );
			SHFT2(f1,f2,f(x2));
		}
		else //{ The other outcome }
		{
			SHFT3(x3,x2,x1,r*x2+c*x0);
			SHFT2(f2,f1,f(x1));
		}

		//  {  Back to see if we are done. }
		if (f1 < f2) //then { We are done. Output the best of the two current values. }
		{
			return x1;
		}
		else
		{
			return x2;
		}
	}
}
*/

//  hidden function for erf
double semigausdistr( double x )
{
	// y = exp(- x^2)
	return exp( - pow(x, 2) );
}

double Functions::erf( double q )
{
	return ( (2/sqrt(M_PI)) * TrapezoidalIntegration( semigausdistr, 0, q, 5000 ) );
}

//  Lagrange interpolation
//  Appeared in astronomical computing, Sky & Telescope, April, 1984
double Functions::LagrangeInterpolation(int n, double x[], double f[], double val)
{
	double * L = new double[n];
	for(int i = 0; i < n; i++)
	{
		L[i] = 1;

		for(int j = 0; j < n; j++)
		{
			if ( j==i )  continue;
			L[i] = L[i] * ( x[i] - x[j] );

		}
		
		L[i] = f[i]/L[i];
	}

	double F1 = 0;
	double F;
	for(int i = 0; i < n; i++)
	{
		if (val != x[i])  continue;
		F = f[i];
		F1 = 1;
	}

	if (F1 != 1)
	{
		double T = 1;
		F = 0;

		for(int i = 0; i < n; i++)
		{
			T = T*( val - x[i] );
		}

		for(int i = 0; i < n; i++)
		{
			F = F + L[i]*T/(val - x[i]);
		}
	}
	
	delete [] L;

	return F;
}

//  Interpolate a set of N points by fitting a polynomial of degree N-1
//  Adapted from algorithm in Numerical Recipes, Press et al. (1992), 
//  Section 3.1. 

void Functions::PolynomialInterpolation(double xa[], double ya[], int n, double x, double & y/*out*/, double & dy/*out*/)
{
	int i, m, ns = 1;
	double den, dift, ho, hp, w;

	double * c = new double[n+1];	//  bacause C recipes arrays are from 1 not 0
	double * d = new double[n+1];
	
	double dif = fabs(x-xa[1]);

	for (i = 1; i <= n; i++)
	{ 
		//  find the index of xa which is closest to x
		if ( (dift=fabs(x-xa[i])) < dif)
		{
			ns = i;
			dif = dift;
		}

		//  and initialize the tableau of c's and d's
		c[i]=ya[i];
		d[i]=ya[i];
	}	

	y = ya[ns--]; //  this is the initial approximation to y

	for (m = 1; m < n; m++)
	{ 
		//  for each column of the tableau
		for (i=1; i <= n-m; i++)
		{ 
			//  we loop over the current c's and d's and update them
			ho=xa[i]-x;
			hp=xa[i+m]-x;
			w=c[i+1]-d[i];
			//  This error can occur only if two input xa's are (to within roundof) identical
			if ( (den=ho-hp) == 0.0) throw; //std::runtime_error("Error in routine polint");
			den=w/den;
			//  Here the c's and d's are updated.
			d[i]=hp*den; 
			c[i]=ho*den;
		}

		y += (dy=(2*ns < (n-m) ? c[ns+1] : d[ns--]));
	}	

	delete [] c;
	delete [] d;
}

#ifndef TINY
	#define TINY 1.0e-25
#endif
#define FREERETURN {delete [] c;delete [] d;return;}

void Functions::RationalInterpolation(double xa[], double ya[], int n, double x, double & y, double & dy)
{
	int m,i,ns=1;
	double w,t,hh,h,dd;
	
	double * c = new double[n+1];
	double * d = new double[n+1];

	hh=fabs( xa[1]);

	for (i=1;i<=n;i++)
	{
		h=fabs(x-xa[i]);
		if (h == 0.0)
		{
			y=ya[i];
			dy=0.0;
			FREERETURN
		}
		else if (h < hh)
		{
			ns=i;
			hh=h;
		}
	
		c[i]=ya[i];
		d[i]=ya[i]+TINY; //  The TINY part is needed to prevent a rare zero-over-zero condition. 
	
	}

	y=ya[ns--];
	for (m=1;m<n;m++)
	{
		for (i=1;i<=n-m;i++)
		{
			w=c[i+1]-d[i];
			h=xa[i+m]-x; //  h will never be zero, since this was tested in the initializing loop. 
			t=(xa[i]-x)*d[i]/h;
			dd=t-c[i+1];
			//  This error condition indicates that the interpolating function has a pole at the
			//  requested value of x.
			if (dd == 0.0) throw;  //  nrerror("Error in routine ratint");
			dd=w/dd;
			d[i]=c[i+1]*dd;
			c[i]=t*dd;
		}

		y += (dy=(2*ns < (n-m) ? c[ns+1] : d[ns--]));
	}

	FREERETURN
}
