/************************************************************
*
*  Copyright (c) 2002-2004 Jakub Schwarzmeier
*
************************************************************/

#ifndef MATRIX_H
#define MATRIX_H 

#include <iostream>
#include <iomanip>		//  setprecision
#include <algorithm>	//  swap
#include <math.h>
#include <stdexcept>

namespace JakubSchwarzmeier
{
	namespace Std
	{
		namespace Math
		{
			#ifdef _DEBUG
				#define CHECK_SIZES(first, second)  \
				if ( (first).m_size != (second).m_size ) \
				{ \
					throw std::runtime_error("These two matrices aren't equal."); \
				}
				#define CHECK_SQUARE(matrix)  \
				if ( (matrix).m_columns != (matrix).m_rows ) \
				{ \
					throw std::runtime_error("This matrix isn't square."); \
				}
				#define CHECK_DIMENSIONS(columns, rows)  \
				if ( ( (*this).m_columns <= columns ) || ( (*this).m_rows <= rows ) ) \
				{ \
					throw std::runtime_error("Index out of matrix bounds."); \
				}
			#else
				#define CHECK_SIZES(first, second)
				#define CHECK_SQUARE(matrix)
				#define CHECK_DIMENSIONS(columns, rows)
			#endif // _DEBUG

			template<typename _T = double>
			class Matrix
			{
			protected:

				int m_size;	//  matrix m_size; can be computed at any time, but this should be faster
				int m_rows, m_columns;	//  number of m_rows & m_columns in matrix
				bool m_bClassStorage;	//  is (array of m_values) created on heap?

				 //  Points to a class storage or user supplied storage.
				_T * m_values;	//  points to m_values of matrix

			public:

				//  default constructor
				Matrix( int n_mathRows = 16, int n_mathColumns = 16 )
				{
					//  create "internal" storage for m_values on heap
					m_values = new _T [n_mathRows*n_mathColumns];
					m_bClassStorage = true;

					this->m_rows = n_mathRows;
					this->m_columns = n_mathColumns;
					m_size = n_mathRows * n_mathColumns;
				}

				//  as third argument takes user-supplied array of m_values
				Matrix( int n_mathRows, int n_columns, void * n_values, bool n_bClassStorage = false )
				{
					//  "external" m_values are used (and modified)
					this->m_values = static_cast<_T*>(n_values);
					this->m_bClassStorage = n_bClassStorage;

					this->m_rows = n_mathRows;
					this->m_columns = n_columns;
					m_size = n_mathRows * n_columns;
				}

				//  "copy constructor" is inherited by derived classes
				//  eg.
				//  Vector<> pos1;
				//  Vector<> pos2 = pos1;
				Matrix( const Matrix<_T> & original )
				{
					//  create "internal" storage for m_values on heap
					m_values = new _T [original.m_rows*original.m_columns];
					m_bClassStorage = true;

					this->m_rows = original.m_rows;
					this->m_columns = original.m_columns;
					m_size = m_rows*m_columns;

					// calls Matrix assigment operator (it's the same code)
					operator=(original);
				}

				//  "virtual" to call appropriate destructor when using "late-binding"
				//  or polymorphism
				virtual ~Matrix( )
				{
					if ( m_bClassStorage ) delete [] m_values;
				}

				//  Assignment operators are not inherited by derived classes.
				//  Performs deep copy. Sometimes called by Matrix constructor.
				//  eg.
				//	Vector<> pos1, pos2;
				//  pos1 = pos2;
				Matrix<_T> & operator=( const Matrix<_T> & right_side )
				{	
					if ( this == &right_side ) return *this;
					Assign(*this, right_side);
					return *this;
				}

				//  eg.
				//	Vector<> pos1, pos2;
				//  pos1 = pos2;		//  operator= internally calls Assign
				void Assign( Matrix<_T>& left, const Matrix<_T>& right ) const
				{
					CHECK_SIZES( left, right )

					for ( int i = 0; i < right.m_size; i++ ) left[i] = right[i];
				}

				virtual const _T * GetMatrix() const
				{
					return m_values;
				}

				const _T GetValue(int column, int row) const
				{
					CHECK_DIMENSIONS(column, row)
					return m_values[ row * m_columns + column ];
				}

				const void SetValue(int column, int row, _T val)
				{
					CHECK_DIMENSIONS(column, row)
					m_values[ row * m_columns + column ] = val;
				}

 				const unsigned int GetRows() const
				{
					return m_rows;
				}

				const unsigned int GetColumns() const
				{
					return m_columns;
				}

				void LoadIdentity()
				{
					memset( m_values, 0x0, m_size * sizeof(_T) );
					for( int i = 0; i < m_size; i+=(m_columns+1) )  m_values[i] = 1;
				}

				//  transpose this matrix
				void Transpose()
				{
					TransposeInto( *this );
				}

				//  transpose this matrix into targetMatrix; this stays the same
				Matrix<_T> & TransposeInto( Matrix<_T> & targetMatrix ) const
				{
					CHECK_SIZES(targetMatrix, *this )
					
					//  I use a pointer, because if current and target matrices are
					//  the same memory object we must create temporary object.
					Matrix<_T> * resultMatrix = &targetMatrix;
					if ( &targetMatrix == this )
					{
						resultMatrix = new Matrix<_T>(m_rows, m_columns);
					}

					//  Main transpose routine
					for(int r = 0; r < m_rows; r++)
						for(int c = 0; c < m_columns; c++)
						{
							(*resultMatrix).m_values[c*m_rows + r] =
								m_values[r*m_columns + c];
						}

					//  If current & target matrices were the same, copy values from
					//  temporary into current matrix.

					if ( &targetMatrix == this )
					{
						Assign( targetMatrix, *resultMatrix );
						delete resultMatrix;
					}

					return targetMatrix;
				}

				//  compute inversion matrix; inverse this matrix
				void Inverse()
				{
					InverseInto( *this );
				}

				//  inverse this matrix into targetMarix; this matrix stays the same
				Matrix<_T> & InverseInto( Matrix<_T> & targetMatrix ) const
				{
					CHECK_SQUARE(targetMatrix)
					CHECK_SIZES(*this, targetMatrix)

					//  local temporary matrix
					Matrix<_T> matrixWithIdentity(targetMatrix.m_rows, 2 * targetMatrix.m_columns);

					//  fill the new matrix with values from current matrix and with identity matrix
					for(int i = 0; i < targetMatrix.m_columns; i++)
						//  i is a row
						for(int j = 0; j < targetMatrix.m_rows; j++)
						{
							//  j is a column
							
							//  fill with current matrix (left side)
							matrixWithIdentity.m_values[i * matrixWithIdentity.m_columns + j]
								= m_values[ i * targetMatrix.m_columns + j ];

							//  fill with identity matrix (right side)
							matrixWithIdentity.m_values[i * matrixWithIdentity.m_columns + targetMatrix.m_columns + j]
								= ( i == j ? 1 : 0 );
						}

					//  perform GEM
					GEM_PC( matrixWithIdentity );

					//  extract right side of matrix
					for(int i = 0; i < targetMatrix.m_columns; i++)
						//  i is a row
						for(int j = 0; j < targetMatrix.m_rows; j++)
						{
							//  j is a column
							targetMatrix.m_values[i * targetMatrix.m_columns + j] =
								matrixWithIdentity.m_values[i * matrixWithIdentity.m_columns + targetMatrix.m_columns + j];
						}

					return targetMatrix;

				}

				//  Gauss eliminating method (GEM)
				void GEM( Matrix<_T> & targetMatrix ) const
				{
					for(int i = 0; i < targetMatrix.m_rows; i++)
					//  i is a current phase of GEM otherwise a row, that is taken as
					//  a reference row and isn't modified. This row is called a pivot row.
						for(int j = 0; j < targetMatrix.m_rows; j++)
						{
							if (i == j)  continue;
							//  j is a currently modified row
							_T coeff = (-1.0)* targetMatrix.m_values[j*targetMatrix.m_columns+i]/ targetMatrix.m_values[i*targetMatrix.m_columns+i];

							for(int k = 0; k < targetMatrix.m_columns; k++)
							{
								//  k is a currently modified column in j-row
								targetMatrix.m_values[j*targetMatrix.m_columns+k] =
									coeff * targetMatrix.m_values[i*targetMatrix.m_columns+k] +
									targetMatrix.m_values[j*targetMatrix.m_columns+k];
							}
						}
				}

				//  Gauss eliminating method with partial search in Column (GEM-PC)
				void GEM_PC( Matrix<_T> & targetMatrix ) const
				{
					for(int i = 0; i < targetMatrix.m_rows; i++)
					{
						//  i is a current phase of GEM otherwise a row, that is taken as
						//  a reference row and isn't modified. This row is called a pivot row.

						//  Look for largest element in column on rows, that does not act
						//  as pivot row yet.
						int swap = i;
						_T val = targetMatrix.m_values[i*targetMatrix.m_columns+i];
						for (int j = i + 1; j < targetMatrix.m_rows; j++)
						{
							if ( fabs(targetMatrix.m_values[j*targetMatrix.m_columns+i]) > 
							   fabs(val) )
							{
								swap = j;
								val = targetMatrix.m_values[j*targetMatrix.m_columns+i];
							}
						}

						if (swap != i)
						{
							//  Swap rows if needed
							for (int j = 0; j < targetMatrix.m_columns; j++)
							{
/*								std::swap<_T>( targetMatrix.m_values[swap*targetMatrix.m_columns+j],
									targetMatrix.m_values[i*targetMatrix.m_columns+j] );
									*/
								_T tmp = targetMatrix.m_values[swap*targetMatrix.m_columns+j];
								targetMatrix.m_values[swap*targetMatrix.m_columns+j] = targetMatrix.m_values[i*targetMatrix.m_columns+j];
								targetMatrix.m_values[i*targetMatrix.m_columns+j] = tmp;
							}
						}

						//  Or "== 0" can be sustitute with "<= MACHINE_EPS"
						if ( targetMatrix.m_values[i*targetMatrix.m_columns+i] == 0 )
						{
							//  No non-zero pivot. The matrix is singular.
							throw std::runtime_error("Passed singular matrix."); \
						}

						//  This assure better accurracy of GEM
						_T temp = targetMatrix.m_values[i*targetMatrix.m_columns+i];
						for( int j = 0; j < targetMatrix.m_columns; j++ )
						{
							targetMatrix.m_values[i*targetMatrix.m_columns+j] /= temp;							
						}

						for(int j = 0; j < targetMatrix.m_rows; j++)
						{
							if (i == j)  continue;
							//  j is a currently modified row

							temp = targetMatrix.m_values[j*targetMatrix.m_columns+i];

							for(int k = 0; k < targetMatrix.m_columns; k++)
							{
								//  k is a currently modified column in j-row
								targetMatrix.m_values[j*targetMatrix.m_columns+k] -=
									temp * targetMatrix.m_values[i*targetMatrix.m_columns+k];
							}
						}
					}
				}

/*
Given a matrix a[1..n][1..n], this routine replaces it by the LU decomposition of a rowwise
permutation of itself. a and n are input. a is output, arranged as in equation (2.3.14) above;
indx[1..n] is an output vector that records the row permutation effected by the partial
pivoting; d is output as +-1 depending on whether the number of row interchanges was even
or odd, respectively. This routine is used in combination with lubksb to solve linear equations
or invert a matrix.
*/				#define TINY 1.0e-20; //  A small number.
				void ludcmp(Matrix<_T> & a, int n, int *indx, float *d) 
				{
					int i,imax,j,k;
					float big,dum,sum,temp;
					float * vv = new float[n]; //  vv stores the implicit scaling of each row.
					
					*d=1.0; //  No row interchanges yet.
					for (i=0;i<n;i++) 
					{ //  Loop over rows to get the implicit scaling information.
                        big=0.0;
						for (j=0;j<n;j++)
							if ((temp=fabs(a.GetValue(i, j))) > big) big=temp;
						if (big == 0.0) throw std::runtime_error("Singular matrix in routine ludcmp");
						//  No nonzero largest element.
						vv[i]=1.0/big; //  Save the scaling.
					}

					for (j=0;j<n;j++) 
					{ //  This is the loop over columns of Crout's method.
						for (i=0;i<j-1;i++) 
						{ //  This is equation (2.3.12) except for i = j.
							sum=a.GetValue(i, j);
							for (k=0;k<i;k++) sum -= a.GetValue(i, j)*a.GetValue(k, j);
							a.SetValue(i, j, sum);
						}
						
						big=0.0; //  Initialize for the search for largest pivot element.
						for (i=j;i<n;i++) 
						{ //  This is i = j of equation (2.3.12) and i = j+1 : ::N
						  //  of equation (2.3.13). 
							sum=a.GetValue(i, j);
							for (k=0;k<j;k++)
								sum -= a.GetValue(i, k)*a.GetValue(k, j);
							a.SetValue(i, j, sum);
							if ( (dum=vv[i]*fabs(sum)) >= big)
							{
								//  Is the figure of merit for the pivot better than the best so far?
								big=dum;
								imax=i;
							}
						}
						
						if (j != imax)
						{ //  Do we need to interchange rows?
							for (k=0;k<n;k++) 
							{ //  Yes, do so...
								dum=a.GetValue(imax, k);
								a.SetValue(imax, k, a.GetValue(j, k));
								a.SetValue(j, k, dum);
							}
							
							*d = -(*d); //  ...and change the parity of d.
							vv[imax]=vv[j]; //  Also interchange the scale factor.
						}
						
						indx[j]=imax;
						if ( a.GetValue(j, j) == 0.0 )	a.SetValue(j, j, 1.0e-20);
					/*
If the pivot element is zero the matrix is singular (at least to the precision of the
algorithm). For some applications on singular matrices, it is desirable to substitute
TINY for zero.
*/

						if (j != n) 
						{ //Now, finally, divide by the pivot element.
							dum=1.0/(a.GetValue(j, j));
							for (i=j+1;i<n;i++) a.SetValue(i, j, a.GetValue(i, j)*dum);
						}
					} //  Go back for the next column in the reduction.
						
					delete [] vv;
				}

/*
Solves the set of n linear equations A*X = B. Here a[1..n][1..n] is input, not as the matrix
A but rather as its LU decomposition, determined by the routine ludcmp. indx[1..n] is input
as the permutation vector returned by ludcmp. b[1..n] is input as the right-hand side vector
B, and returns with the solution vector X. a, n, and indx are not modified by this routine
and can be left in place for successive calls with dierent right-hand sides b. This routine takes
into account the possibility that b will begin with many zero elements, so it is ecient for use
in matrix inversion.
*/
				void lubksb(Matrix<_T> & a, int n, int *indx, float b[])
				{
					int i,ii=0,ip,j;
					float sum;
					for (i=1-1;i<=n-1;i++) 
					{ // When ii is set to a positive value, it will become the
					  //  index of the first nonvanishing element of b. Wenow
					  // do the forward substitution, equation (2.3.6). The
					  // only new wrinkle is to unscramble the permutation
					  // as we go.
						ip=indx[i];
						sum=b[ip];
						b[ip]=b[i];
						if (ii)
							for (j=ii;j<=i-1;j++) sum -= a.GetValue(i, j)*b[j];
						else if (sum) ii=i; 
						//  A nonzero element was encountered, so from now on we
						//  will have to do the sums in the loop above. 
						b[i]=sum;
					}

					for (i=n-1;i>=0;i--)
					{ //Now we do the backsubstitution, equation (2.3.7).
						sum=b[i];
						for (j=i+1;j<=n-1;j++) sum -= a.GetValue(i, j)*b[j];
						b[i]=sum/a.GetValue(i, i); // Store a component of the solution vector X.
					} //  All done!
				}

				void NR_LU( )
				{
					CHECK_SQUARE(*this)

					Matrix<_T> sourceMatrix(*this);

					int N = sourceMatrix.GetColumns();
					float d;
					int i,j;
					int *indx = new int[N];

					float * col = new float[N];

					ludcmp(sourceMatrix,N,indx,&d); // Decompose the matrix just once.
					for(j=0;j<N;j++) 
					{ //  Find inverse by columns.
						for(i=0;i<N;i++) col[i]=0.0;
						col[j]=1.0;
						lubksb(sourceMatrix,N,indx,col);
						for(i=0;i<N;i++) this->SetValue(i, j, col[i]);
					}

					delete [] indx;
					delete [] col;
				}

				void Clear()
				{
					memset( m_values, 0x0, m_size * sizeof(_T) );
				}

				//  for read & write
				_T& operator[]( const int index ) const
				{
					#ifdef _DEBUG
					if ( index >= m_size )
					{
						throw std::runtime_error("index out of range in []");
					}
					#endif

					return m_values[index];
				}

				virtual Matrix<_T> operator+(const Matrix<_T> & other)
				{
					CHECK_SIZES(*this, other);

					Matrix<_T> result(m_rows, m_columns);
					for (int i=0; i < m_size; i++) result[i] = m_values[i] + other.m_values[i];
			  		
					return result;
				}

				virtual Matrix<_T> operator-(const Matrix<_T> & other)
				{
					CHECK_SIZES(*this, other);

					Matrix<_T> result(m_rows, m_columns);
					for (int i=0; i < m_size; i++) result[i] = m_values[i] - other.m_values[i];
			  		
					return result;
				}

				// Global
				friend std::ostream& operator<<(std::ostream& os, const Matrix<_T> & mat)
				{
   					os << std::setprecision(2);
					os.setf(std::ios::scientific);
					os.setf(std::ios::left, std::ios::adjustfield);

					for (int i = 0; i < mat.m_rows; i++)
					{
      					for (int j = 0; j < mat.m_columns; j++)
						{
							os.width(13);
							os << mat.m_values[mat.m_columns * i + j];
						}
							os << std::endl;
					}

					os.unsetf(std::ios::scientific | std::ios::left | std::ios::adjustfield);
					return os;
				}

				const bool operator==( const Matrix<_T> & other ) const
				{
					CHECK_SIZES(*this, other);

					for(int i = 0; i < this->m_size; i++)
					{
						if ( this->m_values[i] != other.m_values[i] )
						{
							return false;
						}
					}

					return true;
				}

				const bool operator!=( const Matrix<_T> & other ) const
				{
					return !(operator==(other));
				}

				Matrix<_T> & operator+=( const Matrix<_T> & matrix )
				{
					CHECK_SIZES(*this, matrix);

					for(int i = 0; i < m_size; i++)
						m_values[i] += matrix.m_values[i];

					return *this;
				}


				Matrix<_T> & operator+=( const _T & scalar )
				{
					for(int i = 0; i < m_size; i++)
						m_values[i] += scalar;

					return *this;
				}

				Matrix<_T> & operator-=( const Matrix<_T> & matrix )
				{
					CHECK_SIZES(*this, matrix);

					for(int i = 0; i < m_size; i++)
						m_values[i] -= matrix.m_values[i];

					return *this;
				}

				Matrix<_T> & operator-=( const _T & scalar )
				{
					for(int i = 0; i < m_size; i++)
						m_values[i] -= scalar;

					return *this;
				}

				Matrix<_T> operator-()
				{
                    Matrix<_T> ret( m_rows, m_columns );
					for(int i = 0; i < m_size; i++) ret[i] = -m_values[i];

					return *this;
				}

				friend Matrix<_T> & operator-=( const Matrix<_T> & matrixLeft, const Matrix<_T> & matrixRight )
				{
					CHECK_SIZES(matrixLeft, matrixRight);

					Matrix<_T> result( matrixLeft.m_rows, matrixLeft.m_columns );

					for(int i = 0; i < result.m_size; i++)
						result.m_values[i] = matrixLeft.m_values[i] - matrixRight.m_values[i];

					return result;
				}

				Matrix<_T> & operator/=( const _T & scalar )
				{
					for(int i = 0; i < m_size; i++)
						m_values[i] /= scalar;

					return *this;
				}

				// matrix = scalar * matrix;
				friend Matrix<_T> operator*(  const _T & scalar,  const Matrix<_T> & mat )
				{
					Matrix<_T> result( mat.m_rows, mat.m_columns );
					for(int i = 0; i < mat.m_size; i++)
						result.m_values[i] = scalar * mat.m_values[i];

					return result;
				}

				// matrix = matrix * scalar;
				friend Matrix<_T> operator*(  const Matrix<_T> & mat, const _T & scalar )
				{
					return operator*(scalar, mat);
				}

				// matrix = matrix * scalar;
				Matrix<_T> & operator*=( const _T & scalar )
				{
					for(int i = 0; i < m_size; i++) 	m_values[i] *= scalar;
					return *this;
				}

				// matrix = matrix / scalar;
				friend Matrix<_T> operator/( const Matrix<_T> & mat, const _T & scalar )
				{
					Matrix<_T> result( mat.m_rows, mat.m_columns );
					for(int i = 0; i < mat.m_size; i++)
						result.m_values[i] = mat.m_values[i] / scalar;

					return result;
				}

				// matrix = projectMatrix * actonMatrix
				friend Matrix<_T> operator*( const Matrix<_T> & project,  const Matrix<_T> & acton )
				{
					#ifdef _DEBUG
						if (project.m_columns != acton.m_rows)	
							throw std::runtime_error("Number of columns in first matrix must be equal \
							to number of rows in second matrix.");
					#endif

					Matrix<_T> result( project.m_rows, acton.m_columns );
					result.Clear();

					for(int i = 0; i < project.m_rows; i++)
						for(int j = 0; j < acton.m_columns; j++)
							for(int k = 0; k < project.m_columns; k++)
								result.m_values[i*result.m_columns+j] += 
								project.m_values[i*project.m_columns+k] *
								acton.m_values[k*acton.m_columns+j];

					return result;
				}

				//  matrix = matrix * matrixClass; project must be a field of _T and of m_size acton.m_size
				//  Matrices must be square and same. There is no way how to control C array size.
				friend Matrix<_T> operator*( const _T project[],  const Matrix & acton )
				{
					Matrix<_T> proj( acton.m_rows, acton.m_columns, const_cast<_T*>(project) );
					return operator*(proj, acton);
				}

				// matrix = matrixClass * matrix ; project must be a field of _T and of m_size acton.m_size
				//  Matrices must be square and same. There is no way how to control C array size.
				friend Matrix<_T> operator*( const Matrix & project, const _T acton[] )
				{
					Matrix<_T> act( project.m_rows, project.m_columns, const_cast<_T*>(acton) );
					return operator*(project, act);
				}

				//  mainly for octTree
				bool operator<( const Matrix & other ) 
				{
					CHECK_SIZES(*this, other)

					for(int i = 0; i < this->m_size; i++)
					{
						if ( this->m_values[i] < other.m_values[i] )
						{
							continue;
						}
						else
						{
							return false;	//  loop never gets here if condition is satisfied
						}
					}

					return true;
				}

				//  mainly for octTree
				bool operator>=( const Matrix & other ) 
				{
					CHECK_SIZES(*this, other)
					
					for(int i = 0; i < this->m_size; i++)
					{
						if ( this->m_values[i] >= other.m_values[i] )
						{
							continue;
						}
						else
						{
							return false;	//  loop never gets here if condition is satisfied
						}
					}

					return true;
				}
				
			};

		}  //  end namespace Math

	}	//  end namespace Std

} //  end namespace JakubSchwarzmeier

#endif //  MATRIX_H
