/************************************************************
*
*  Copyright (c) 2002-2004 Jakub Schwarzmeier
*
************************************************************/

#ifndef QUATERNION_H
#define QUATERNION_H 

#include <schwarzmeier/std.h>

#include <stdexcept>
#include <iostream>
#include <iomanip>  //  setprecision
#include <float.h>

namespace JakubSchwarzmeier
{
	namespace Std
	{
		namespace Math
		{
			template <typename _T = double>
			class Quaternion
			{
			protected:

				_T			m_scalar;	//  scalar part
				Vector<_T>	m_vector;	//  vector part

			public:

				Quaternion()
				{
				}

				Quaternion<_T> & operator=( const Quaternion<_T> & right_side )
				{
					if ( this == &right_side ) return *this;
                    m_vector = right_side.m_vector;
					m_scalar = right_side.m_scalar;

					return *this;
				}
				
				Quaternion( const Vector<_T> & vec )
				{
                    operator=( vec );
				}
				
				Quaternion( const _T scalar, double v0, double v1, double v2 )
				{
					m_scalar = scalar;
                    m_vector[0] = v0;
                    m_vector[1] = v1;
                    m_vector[2] = v2;
				}

				Quaternion( Vector<_T> & axis, _T angle )
				{
					CreateFromEulerian( axis, angle );
				}

				Quaternion( double longitude, double latitude, double angle )
				{
					CreateFromSpherical( longitude, latitude, angle );
				}
				
				Quaternion<_T> & operator=( const Vector<_T> & vec )
				{
					//  imaginary quaternion
                    m_vector = vec;
					m_scalar = 0;

					return *this;
				}

				Vector<_T> GetVector()
				{
					return m_vector;
				}

				_T GetScalar()
				{
					return m_scalar;
				}

				void SetVector( Vector<_T> & vec )
				{
					m_vector = vec;
				}

				void SetScalar(_T scalar)
				{
					m_scalar = scalar;
				}

				void LoadIdentity()
				{
					m_scalar = 1;
					m_vector = 0;
				}

				//  reduces any normal quaternion to a unit quaternion, length = 1
				void Normalize()
				{
					// Calculate the length of the quaternion
					_T length = 0;
					for(int i = 0; i < 3; i++)	length += m_vector[i]*m_vector[i];
					length += m_scalar*m_scalar;
					length = sqrt( length );

					//  Keep the program from blowing up by vectors whose length 
					//  may be calculated too close to zero.
					if (length <= FLT_EPSILON)		//  == 0
						return;

					//  Dividing each element by the length will result in a unit
					//  quaternion.
					m_scalar /= length;
					m_vector /= length;
				}

				Quaternion<_T> GetConjugate()
				{
					Quaternion<_T> q;
					q.m_vector = -m_vector;
					q.m_scalar = m_scalar;

					return q;
				}

				//  Create quaternion from given rotational axis and rotational angle
				//  around that axis.
				//  Rotational axis must be normalized *before* it is passed to this 
				//  function.
				void CreateFromEulerian( Vector<_T> & axis, _T theta)
				{
					//  Calculate a scalar value by cos(theta / 2)
					m_scalar = Math::Functions::cos_angle( theta / 2.0 );

					//  Calculate the sin(theta / 2) once for optimization
					_T s = Math::Functions::sin_angle( theta / 2.0 );

					//  Calculate the x, x and z of the quaternion
					m_vector = s * axis;
				}

				//  Create quaternion from OpenGL suitable matrix.
				void CreateFromMatrix(_T matrix[])
				{
					_T m = 1 + matrix[0] + matrix[5] + matrix[10];				

					_T x, y, z;

					if (m > FLT_EPSILON)	//  m > 0, to avoid large distortions
					{
						_T t = sqrt(m) * 2;
						 x = ( matrix[9] - matrix[6] ) / t;
						 y = ( matrix[2] - matrix[8] ) / t;
						 z = ( matrix[4] - matrix[1] ) / t;
						m_scalar = 0.25 * t;
					}
					else
					{
						//  matrix == 0
						if ( matrix[0] > matrix[5] && matrix[0] > matrix[10] )
						{        
							//  first column
							_T t  = sqrt( 1.0 + matrix[0] - matrix[5] - matrix[10] ) * 2;
							 x = 0.25 * t;
							 y = (matrix[4] + matrix[1] ) / t;
							 z = (matrix[2] + matrix[8] ) / t;
							m_scalar = (matrix[9] - matrix[6] ) / t;
						}
						else if ( matrix[5] > matrix[10] ) 
						{                        
							//  second column
							_T t  = sqrt( 1.0 + matrix[5] - matrix[0] - matrix[10] ) * 2;
							 x = (matrix[4] + matrix[1] ) / t;
							 y = 0.25 * t;
							 z = (matrix[9] + matrix[6] ) / t;
							m_scalar = (matrix[2] - matrix[8] ) / t;
						} 
						else 
						{                                                
							//  third column
							_T t  = sqrt( 1.0 + matrix[10] - matrix[0] - matrix[5] ) * 2;
							_T x = (matrix[2] + matrix[8] ) / t;        
							_T y = (matrix[9] + matrix[6] ) / t;        
							_T z = 0.25 * t;        
							m_scalar = (matrix[4] - matrix[1] ) / t;    
						}
					}

					m_vector[0] = x;
					m_vector[1] = y;
					m_vector[2] = z;
				}

				//  Creates matrix from this quaternion.
				void CreateMatrix(_T matrix[])
				{
					//  first row
					matrix[0]  = 1.0 - 2.0 * ( m_vector[1] * m_vector[1] + m_vector[2] * m_vector[2] ); 
					matrix[1]  = 2.0 * (m_vector[0] * m_vector[1] + m_vector[2] * m_scalar);
					matrix[2]  = 2.0 * (m_vector[0] * m_vector[2] - m_vector[1] * m_scalar);
					matrix[3]  = 0.0;  

					//  second row
					matrix[4]  = 2.0 * ( m_vector[0] * m_vector[1] - m_vector[2] * m_scalar );  
					matrix[5]  = 1.0 - 2.0 * ( m_vector[0] * m_vector[0] + m_vector[2] * m_vector[2] ); 
					matrix[6]  = 2.0 * (m_vector[2] * m_vector[1] + m_vector[0] * m_scalar );  
					matrix[7]  = 0.0;  

					//  third row
					matrix[8]  = 2.0 * ( m_vector[0] * m_vector[2] + m_vector[1] * m_scalar );
					matrix[9]  = 2.0 * ( m_vector[1] * m_vector[2] - m_vector[0] * m_scalar );
					matrix[10] = 1.0 - 2.0 * ( m_vector[0] * m_vector[0] + m_vector[1] * m_vector[1] );  
					matrix[11] = 0.0;  

					//  fourth row
					matrix[12] = 0;  
					matrix[13] = 0;  
					matrix[14] = 0;  
					matrix[15] = 1.0;
				}

				void CreateMatrix( Matrix<_T> & matrix )
				{
                    CreateMatrix( const_cast<_T*>( matrix.GetMatrix() ) );
				}

				friend Quaternion<_T> operator+(const Quaternion<_T> & u, const Quaternion<_T> & v)
				{
					return Quaternion<_T>( u.m_scalar + v.m_scalar, 
						u.m_vector[0] + v.m_vector[0], 
						u.m_vector[1] + v.m_vector[1],
						u.m_vector[2] + v.m_vector[2] );
				}

				friend Quaternion<_T> operator/(const Quaternion &v, const double & denominator)
				{
					return Quaternion<_T>( v.m_scalar / denominator,
						v.m_vector[0] / denominator, 
						v.m_vector[1] / denominator, 
						v.m_vector[2] / denominator );
				}

				//  Multiply two quaternions according to formula:
				//  Qr = Q1*Q2 = ( s1 * s2 - v1 * v2, s1 * v2 + s2 * v1 + v1 m_vector[0] v2 )
				Quaternion<_T> operator*(Quaternion<_T> & other)
				{
					Quaternion<_T> result;

					result.m_scalar = m_scalar*other.m_scalar - scalar_product( m_vector, other.m_vector );
					result.m_vector = m_scalar * other.m_vector + other.m_scalar * m_vector + vector_product( m_vector, other.m_vector );
			  		
					return result;
				}

				friend Quaternion<_T> operator*(const Quaternion<_T> & q, const _T & scalar)
				{
					return Quaternion<_T>(  q.m_scalar * scalar, q.m_vector[0] * scalar, 
						q.m_vector[1] * scalar, q.m_vector[2] * scalar );
				}

				friend Quaternion<_T> operator*(const _T & scalar, const Quaternion<_T> & q)
				{
					return operator*(q, scalar);
				}

				friend Quaternion<_T> Slerp(const Quaternion<_T> & q0, 
					const Quaternion<_T> & q1, 
					const _T t)
				{
					_T cosTheta = q0.m_scalar * q1.m_scalar + 
						q0.m_vector[0] * q1.m_vector[0] + 
						q0.m_vector[1] * q1.m_vector[1] + 
						q0.m_vector[2] * q1.m_vector[2];

					if (fabs(1 - cosTheta) < 0.00001)
					{
						return q0 * (1 - t) + q1 * t;
					} 
					else 
					{
						_T theta = acos(cosTheta);
						return (q0 * sin((1 - t) * theta) + q1 * sin(t * theta)) / sin(theta);
					}
				}

				//  These three functions returns values that are stored in individual
				//  rows when calling CreateMatrix.
				//
				Vector<_T> GetXAxis()
				{
					_T /*x2 = m_vector[0] + m_vector[0],*/ y2 = m_vector[1] + m_vector[1], z2 = m_vector[2] + m_vector[2];
					_T xy = m_vector[0] * y2, xz = m_vector[0] * z2;
					_T yy = m_vector[1] * y2, zz = m_vector[2] * z2;
					_T wy = m_scalar * y2, wz = m_scalar * z2;
					return Vector<_T>(1-(yy+zz), xy+wz, xz-wy);
				}

				Vector<_T> GetYAxis()
				{
					_T x2 = m_vector[0] + m_vector[0], y2 = m_vector[1] + m_vector[1], z2 = m_vector[2] + m_vector[2];
					_T xx = m_vector[0] * x2, xy = m_vector[0] * y2;
					_T yz = m_vector[1] * z2, zz = m_vector[2] * z2;
					_T wx = m_scalar * x2, wz = m_scalar * z2;
					return Vector<_T>(xy-wz, 1-(xx+zz), yz+wx);
				}

				Vector<_T> GetZAxis()
				{
					_T x2 = m_vector[0] + m_vector[0], y2 = m_vector[1] + m_vector[1], z2 = m_vector[2] + m_vector[2];
					_T xx = m_vector[0] * x2, xz = m_vector[0] * z2;
					_T yy = m_vector[1] * y2, yz = m_vector[1] * z2;
					_T wx = m_scalar * x2, wy = m_scalar * y2;
					return Vector<_T>( xz+wy, yz-wx, 1-(xx+yy) );
				}

				double GetAngle()
				{
					return Math::Functions::acos_angle( m_scalar ) * 2;
				}

				double GetLongitude()
				{
					double cos_a  = m_scalar;
					double sin_a  = sqrt( 1.0 - cos_a * cos_a );

					if ( fabs( sin_a ) < 0.0005 ) sin_a = 1;

					double tx = m_vector[0] / sin_a;
					double tz = m_vector[2] / sin_a;

					double longitude;

					if ( tx * tx + tz * tz < DBL_EPSILON )
						longitude   = 0;
					else
						longitude  = Math::Functions::atan2_angle( tx, tz );

					if ( longitude < 0 )
						longitude += 360.0;

					return longitude;
				}

				double GetLatitude()
				{
					double cos_a  = m_scalar;
					double sin_a  = sqrt( 1.0 - cos_a * cos_a );

					if ( fabs( sin_a ) < DBL_EPSILON ) sin_a = 1;

					double ty = m_vector[1] / sin_a;

					return -Math::Functions::asin_angle( ty );
				}


				void CreateFromSpherical(double longitude, double latitude, double angle)
				{
					double sin_a    = Math::Functions::sin_angle( angle / 2 );
					double cos_a    = Math::Functions::cos_angle( angle / 2 );

					double sin_lat  = Math::Functions::sin_angle( latitude );
					double cos_lat  = Math::Functions::cos_angle( latitude );

					double sin_long = Math::Functions::sin_angle( longitude );
					double cos_long = Math::Functions::cos_angle( longitude );

					m_vector[0]       = sin_a * cos_lat * sin_long;
					m_vector[1]       = sin_a * sin_lat;
					m_vector[2]       = sin_a * sin_lat * cos_long;
					m_scalar          = cos_a;
				}

			};	//  end class Quaternion

		}  //  end namespace Math

	}	//  end namespace Std

} //  end namespace JakubSchwarzmeier

#endif //  QUATERNION_H
