/************************************************************
*
*  Copyright (c) 2002-2004 Jakub Schwarzmeier
*
************************************************************/

#ifndef BH_H
#define BH_H

#include <schwarzmeier/std.h>
#include <fstream>

namespace JakubSchwarzmeier
{
	namespace Std
	{
		namespace Simulations
		{
			namespace BarnesHut
			{
				using namespace JakubSchwarzmeier::Std::Math;
				
				class ICommonNode: public JakubSchwarzmeier::Std::Data::ICommonNode
				{
				public:
					
					ICommonNode(Vector<> n_position = 0, double n_mass = 0)
					{
						m_position = n_position;
						m_mass = n_mass;
						m_load = 0;
					}

					virtual ~ICommonNode() {}

					Vector<> m_position;
					double m_mass;
					double m_load;	//  costzones
				};
				
				class DataReader
				{
				public:

					DataReader()
					{
						m_stream = NULL;
					}

					virtual ~DataReader()
					{
						if (m_stream != NULL)  delete m_stream;
					}

					void OpenStream( std::fstream * stream )
					{
						m_stream = new std::istream( stream->rdbuf() );
//						m_stream->seekg( 0, std::ios_base::beg );
					}

					void Read(double * d)
					{
						m_stream->read( reinterpret_cast<char*>(d), sizeof(*d) );
						#ifdef _DEBUG
						if (m_stream->fail())  throw;
						#endif
					}

				protected:

					std::istream * m_stream;					
				};

				class DataWriter
				{
				public:

					DataWriter()
					{
						m_stream = NULL;
					}

					virtual ~DataWriter()
					{
						if (m_stream != NULL)  delete m_stream;
					}

					void OpenStream( std::ostream * stream )
					{
						m_stream = new std::ostream( stream->rdbuf() );
						m_stream->seekp( m_stream->tellp(), std::ios_base::beg );
					}

					void Write(double * d)
					{
						m_stream->write( reinterpret_cast<char*>(d), sizeof(*d) );
						#ifdef _DEBUG
						if ( m_stream->fail() )	throw;
						#endif
					}

				protected:

					std::ostream * m_stream;					
				};

				class Leaf: public ICommonNode
				{
				public:
					
					Leaf(): ICommonNode() 
					{						
						m_acceleration = 0;
					}

					Leaf(Vector<> n_position, Vector<> n_velocity, double n_mass): ICommonNode(n_position, n_mass)
					{
						m_velocity = n_velocity;
						m_acceleration = 0;
					}

					void Read(DataReader * dr)
					{
						dr->Read( &m_mass );

						dr->Read( &m_position[0] );
						dr->Read( &m_position[1] );
						dr->Read( &m_position[2] );

						dr->Read( &m_velocity[0] );
						dr->Read( &m_velocity[1] );
						dr->Read( &m_velocity[2] );
					}
					
					void Write(DataWriter * dw)
					{
						dw->Write( &m_mass );

						dw->Write( &m_position[0] );
						dw->Write( &m_position[1] );
						dw->Write( &m_position[2] );

						dw->Write( &m_velocity[0] );
						dw->Write( &m_velocity[1] );
						dw->Write( &m_velocity[2] );
					}

					virtual ~Leaf()
					{
					}

					Vector<> m_velocity;
					Vector<> m_acceleration;		//  neni treba, pocita se vzdy od znovu; u distributed obsahuje zrychleni udelene pouze vzdalenymi nody
				};

				class Node: public Data::INode<BarnesHut::ICommonNode, BarnesHut::Leaf >
				{
				public:

					Node(Data::INode<BarnesHut::ICommonNode, BarnesHut::Leaf > * n_ptrParent, Vector<> * n_ptrCorner, double n_bisect, Quaternion<> * n_ptrHilbertOrientation):
					  Data::INode<BarnesHut::ICommonNode, BarnesHut::Leaf >(n_ptrParent, n_ptrCorner, n_bisect, n_ptrHilbertOrientation)
					{}

					virtual ~Node() {}

				protected:

				};

				class Group
				{
				public:

					ICommonNode * m_element;	//  particle or node
					Vector<> m_acceleration, m_velocity, m_position;	//  far forces
					std::vector<ICommonNode*> m_nearList;	//  near forces

					Group( ICommonNode * n_element = NULL)
					{
						m_element = n_element;
						//  must explicitly set up all the matrices to zeros
						m_acceleration = 0;
						m_velocity = 0;
						m_position = 0;
					}
				};


				class Tree: public Std::Data::OctTree<ICommonNode, Leaf, Node>
				{
				public:

					Tree( std::vector<Leaf*> * n_ptrLeavesList, double n_theta, 
                        double n_epsilon, size_t n_groupSize = 32, double n_dt = 1,
						double n_G = 6.67259e-11 );
					void SetLeaves( std::vector<Leaf*> * n_ptrLeavesList );
					void Step();
					void Write(DataWriter * dw);

				protected:

					std::vector<Leaf*> * m_ptrLeavesList;
					std::vector<Group*> m_groupsList;

					double m_theta;
					double m_epsilon;
					size_t m_groupSize;
					double m_dt;
					double m_G;

				private:

					void CreateGroups(Node * node);

				};	//  end class Tree

			}	//  end namespace BarnesHut

		}	//  end namespace Simulations

	}	//  end namespace Std
	
} //  end namespace JakubSchwarzmeier

#endif //  BH_H
