#include "stdwx.h"

#include "glCanvasWin.h"
#include "OpenGLCanvas.h"
#include "SerializeReaderWriter.h"
#include <GL/wglew.h>
#include <crtdbg.h>


using namespace JakubSchwarzmeier::wx;

#if defined(__GNUG__) && !defined(NO_GCC_PRAGMA)
#pragma implementation "glcanvas.h"
#endif

#include "wx/wxprec.h"

#ifndef WX_PRECOMP
    #include "wx/frame.h"
    #include "wx/settings.h"
    #include "wx/intl.h"
    #include "wx/log.h"
    #include "wx/app.h"
#endif

#include "wx/module.h"

#include "wx/msw/private.h"

// DLL options compatibility check:
#include "wx/build.h"
WX_CHECK_BUILD_OPTIONS("wxGL")

#include "wx/glcanvas.h"


/*
  The following two compiler directives are specific to the Microsoft Visual
  C++ family of compilers

  Fundementally what they do is instruct the linker to use these two libraries
  for the resolution of symbols. In essence, this is the equivalent of adding
  these two libraries to either the Makefile or project file.

  This is NOT a recommended technique, and certainly is unlikely to be used
  anywhere else in wxWidgets given it is so specific to not only wxMSW, but
  also the VC compiler. However, in the case of opengl support, it's an
  applicable technique as opengl is optional in setup.h This code (wrapped by
  wxUSE_GLCANVAS), now allows opengl support to be added purely by modifying
  setup.h rather than by having to modify either the project or DSP fle.

  See MSDN for further information on the exact usage of these commands.
*/


/*
 * GLContext implementation
 */

IMPLEMENT_DYNAMIC_CLASS (wxMyGLContext, wxGLContext)

wxMyGLContext::wxMyGLContext(
               bool isRGB, wxGLCanvas *win,
               const wxPalette& palette,
               const wxMyGLContext *other  /*  for shared list */
			   )
{
	Create( isRGB, win, false, palette, other );
}

void wxMyGLContext::Create(
               bool isRGB, wxGLCanvas *win, bool bDeserialize,
               const wxPalette& palette,
               const wxMyGLContext *other  /*  for shared list */
			   )
{
	/*  Parent's ctor begin */
	m_window = win;

	m_hDC = win->GetHDC();

	m_glContext = wglCreateContext((HDC) m_hDC);
	wxCHECK_RET( m_glContext, wxT("Couldn't create OpenGL context") );
	/*  Parent's ctor end */

	SetCurrent();
	GLenum err1 = glewInit();
	GLenum err2 = wglewInit();
	UnsetCurrent();

	if ((GLEW_OK != err1) || (GLEW_OK != err2))
	{
		/* Problem: glewInit failed, something is seriously wrong. */
		wxString err("Error: %s", glewGetErrorString(err1) );
		wxCHECK_RET( 0, err.GetData() );
		return;
	}

	m_hDCWindow = m_hDC;

	m_image = NULL;
	m_bitmapData = NULL;

	if ((bDeserialize) && (m_hbuffer != NULL))
	{
		m_hbuffer = NULL;
		CreatePBuffer();
	}
	else
	{
		m_hbuffer = NULL;
	}
}


wxMyGLContext::~wxMyGLContext()
{
}

GLEWContext * wxMyGLContext::glewGetContext()
{
	return &m_glewCtx;
}

WGLEWContext * wxMyGLContext::wglewGetContext()
{
	return &m_wglewCtx;
}

void wxMyGLContext::CreatePBuffer()
{
	SetCurrent();
	reinterpret_cast<COpenGLCanvas*>( m_window )->OnContextTerminate();
	UnsetCurrent();

	int width, height;
	m_window->GetClientSize( &width, &height );
	if (width == 0) width = 1;
	if (height == 0) height = 1;
	int depth = wxDisplayDepth();		//  vrati 32
	depth = 24;		

	//  Create memory storage for PBuffer
	reinterpret_cast<COpenGLCanvas*>( m_window )->m_csRenderingBitmap.Enter();
	if ( m_bitmapData == NULL || width != m_image->GetWidth() || height != m_image->GetHeight() )
	{
		delete [] m_bitmapData;
		delete m_image;
		m_bitmapData = new unsigned char[3*width*height];
		m_image = new wxImage( width, height, m_bitmapData, true );
	}

	//  Create PBuffer with OpenGL

	if (!wglewIsSupported("WGL_ARB_pbuffer")) 
	{
		wxCHECK_RET( m_glContext, wxT("Error: WGL_ARB_pbuffer extension not available.") );
		return;
	}

	//  Attribute arrays must be "0" terminated
	int ifmtList[] = {
		WGL_DRAW_TO_PBUFFER_ARB,	GL_TRUE,  //  enable render to pbuffer
		WGL_SUPPORT_OPENGL_ARB,		GL_TRUE,  //  automatic, for NVIDIA for compatibility issues
		WGL_ACCELERATION_ARB,		WGL_FULL_ACCELERATION_ARB,
		WGL_DEPTH_BITS_ARB,         24,	//  the depth of the depth (z-axis) buffer; not working for 32
		WGL_COLOR_BITS_ARB,         depth,	
		WGL_ALPHA_BITS_ARB,         8,
		WGL_STENCIL_BITS_ARB,       0,
		WGL_DOUBLE_BUFFER_ARB,		GL_FALSE,	//  we dont require double buffering
		0										//  zero terminates the list
	};

	int pformats[256];
	unsigned int nformats;

	//  Ask the hardware to find pixel format defined in ifmtList.
	//  Note: Before this step must be set valid current 
	//  rendering context.
	BOOL ret = wglChoosePixelFormatARB( (HDC)m_hDCWindow, 
		ifmtList, NULL, 256, pformats, &nformats );

	if ( (ret == 0) || (nformats < 1) )
	{
		wxCHECK_RET( 0, wxT("Error: wglChoosePixelFormatARB failed.") );
	}

	//  Create pixel buffer (pbuffer) based on format.
	int iattributes[1] = {0};

	if (m_hbuffer != NULL)
	{
		wglReleasePbufferDCARB( m_hbuffer, (HDC) m_hDC );
		wglDestroyPbufferARB( m_hbuffer );
	}

	m_hbuffer =	wglCreatePbufferARB( (HDC)m_hDCWindow,	pformats[0], 
		width, height, iattributes );

	if (m_hbuffer == NULL)
	{
		wxString err;
		err.sprintf("Error: %d", ::GetLastError() & 0xFFFF);
		wxCHECK_RET( m_hbuffer, err.GetData() );
	}

	//  Create a device context for the newly created pbuffer.
	if (m_hDC != m_hDCWindow)
	{
		DeleteDC( (HDC)m_hDC );
	}
	m_hDC = wglGetPbufferDCARB( m_hbuffer );
	if (m_hDC == NULL)
	{
		wxCHECK_RET( m_hDC, wxT("Error: wglGetPbufferDCARB failed.") );
	}

	//  Create an OpenGL rendering context and associate it with 
	//  the handle for the pbuffer.
	HGLRC pbufglctx = wglCreateContext( (HDC)m_hDC );
	if (pbufglctx == NULL)
	{
		wxCHECK_RET( pbufglctx, wxT("Error: wglCreateContext for PBuffer failed.") );
	}

	//  Delete old rendering context.
	wglDeleteContext( m_glContext );

	//  save new rendering context
	m_glContext = pbufglctx;

	reinterpret_cast<COpenGLCanvas*>( m_window )->m_csRenderingBitmap.Leave();

	SetCurrent();
	reinterpret_cast<COpenGLCanvas*>( m_window )->PreRender();
	UnsetCurrent();
}

void wxMyGLContext::DontUsePBuffer()
{
	SetCurrent();
	reinterpret_cast<COpenGLCanvas*>( m_window )->OnContextTerminate();
	UnsetCurrent();

	if (m_image != NULL)
	{
		delete m_image;
		m_image = NULL;
		delete [] m_bitmapData;
		m_bitmapData = NULL;

		wglReleasePbufferDCARB( m_hbuffer, (HDC)m_hDC );
		wglDestroyPbufferARB( m_hbuffer );

		DeleteDC( (HDC)m_hDC );
		m_hDC = m_hDCWindow;

		m_glContext = wglCreateContext((HDC) m_hDC);
	}

	SetCurrent();
	reinterpret_cast<COpenGLCanvas*>( m_window )->PreRender();
	UnsetCurrent();
}

void wxMyGLContext::UpdatePBuffer()
{
	if (m_image != NULL)
	{
		reinterpret_cast<COpenGLCanvas*>( m_window )->m_csRenderingBitmap.Enter();

		int width = m_image->GetWidth();
		int height = m_image->GetHeight();
		unsigned char * flipped = new unsigned char[3 * width * height];
		
		glPixelStorei(GL_PACK_ALIGNMENT, 1);
		glReadPixels(0, 0, m_image->GetWidth(), m_image->GetHeight(),
			GL_RGB, GL_UNSIGNED_BYTE, flipped);

		for (int y = 0; y < height; y++)
		{
			memcpy(m_bitmapData + y * width * 3, flipped + (height - y - 1) * width * 3, width * 3);
		}

		delete flipped;

		reinterpret_cast<COpenGLCanvas*>( m_window )->m_csRenderingBitmap.Leave();
	}
	else
	{
		//  flush rendering pipe
		glFinish();
	}
}

void wxMyGLContext::ResizePBuffer()
{
	if (m_image != NULL)
	{
		CreatePBuffer();
	}
}

void wxMyGLContext::GetCopyImage(wxImage & newImage)
{
	if (m_image != NULL)
	{
		reinterpret_cast<COpenGLCanvas*>( m_window )->m_csRenderingBitmap.Enter();
		newImage = m_image->Copy();
		reinterpret_cast<COpenGLCanvas*>( m_window )->m_csRenderingBitmap.Leave();
	}
	else
	{
		int width, height;
		m_window->GetClientSize( &width, &height );

		unsigned char * flipped = new unsigned char[3 * width * height];
		SetCurrent();
		glPixelStorei(GL_PACK_ALIGNMENT, 1);
		glReadPixels(0, 0, width, height, GL_RGB, GL_UNSIGNED_BYTE, flipped);
		UnsetCurrent();

		unsigned char * pixels = (unsigned char*)malloc(3 * width * height);
		for (int y = 0; y < height; y++)
		{
			memcpy(pixels + y * width * 3, flipped + (height - y - 1) * width * 3, width * 3);
		}
		delete flipped;

		newImage.Destroy();
		newImage.Create( width, height, pixels );
	}
}

wxImage wxMyGLContext::GetImage()
{
	return wxImage( m_image->GetWidth(), m_image->GetHeight(), m_image->GetData(), true );
}

void wxMyGLContext::SetCurrent()
{
	reinterpret_cast<COpenGLCanvas*>( m_window )->m_csRenderingContext.Enter();
    if (! wglMakeCurrent((HDC) m_hDC, m_glContext) )
	     wxLogLastError(_T("wglMakeCurrent"));
}

void wxMyGLContext::UnsetCurrent()
{
    wglMakeCurrent( NULL, NULL );
	reinterpret_cast<COpenGLCanvas*>( m_window )->m_csRenderingContext.Leave();
}

void wxMyGLContext::SwapBuffers()
{
	if (m_image == NULL)
	{
		SetCurrent();
		if (! ::SwapBuffers((HDC) m_hDC) )    //blits the backbuffer into DC
			wxLogLastError(_T("SwapBuffers"));
		UnsetCurrent();
	}
	else
	{
		wxClientDC clientDC(m_window);
		reinterpret_cast<COpenGLCanvas*>( m_window )->m_csRenderingBitmap.Enter();
		wxBitmap bmp( *m_image, 24 );
		clientDC.DrawBitmap( bmp, 0, 0, false );
		reinterpret_cast<COpenGLCanvas*>( m_window )->m_csRenderingBitmap.Leave();
	}
}

bool wxMyGLContext::OnSerialize( JakubSchwarzmeier::wx::CSerializeReaderWriter * dir )
{
	wxGLContext::OnSerialize(dir);
	dir->Value( &m_hbuffer, wxString("m_hbuffer") );

	return true;
}
