#include "stdwx.h"

#include "ms3dmodel.h"

#include <fstream>
#include <schwarzmeier/std.h>

using namespace JakubSchwarzmeier::wx;
using namespace JakubSchwarzmeier::Std::Math;

CMS3DModel::CMS3DModel(COpenGLCanvas * openGLCanvas): CThreedModel(openGLCanvas)
{
	m_texture = 0;
}

CMS3DModel::~CMS3DModel()
{
	//  release textures
	for(int i = 0; i < nNumMaterials; i++)
	{
		glDeleteTextures( 1, &m_texture[i] );	//  delete one allocated texture under name of m_iTexture value
	}

	delete [] m_texture;
}

void CMS3DModel::Load( std::string filename )
{
	std::ifstream m_is;
	m_is.open( filename.c_str(), std::ios::in | std::ios::binary  );
	if (!m_is)	throw;

	JakubSchwarzmeier::Std::Math::Vector<> ** triangles;	//  size is given by m_nNumTriangles
	JakubSchwarzmeier::Std::Math::Vector<> ** normals;		//  normal for EVERY vertex
	TEX_TYPE ** texCoords;
	int m_nNumTriangles;

	//  read header (unused)
	m_is.read( reinterpret_cast<char*> (&ms3d_header_t.id), sizeof(ms3d_header_t.id) );
	m_is.read( reinterpret_cast<char*> (&ms3d_header_t.version), sizeof(ms3d_header_t.version) );

	//  read no of vertices
	m_is.read( reinterpret_cast<char*> (&nNumVertices), sizeof(nNumVertices) );
	//  allocate storage for vertices
	Vector<> * vertices = new Vector<>[nNumVertices];
	//  read all vertices into temporary array
	//  cause: We want triangles. But triangles are stored in form of indices (indexes)
	//  into vertices array. So read vertices temporary and then insert them according to
	//  triangles info into class array.
	for(int i = 0; i < nNumVertices; i++)
	{
		//  read vertices
		m_is.read( reinterpret_cast<char*> (&ms3d_vertex_t.flags), sizeof(ms3d_vertex_t.flags) );
		m_is.read( reinterpret_cast<char*> (&ms3d_vertex_t.vertex), sizeof(ms3d_vertex_t.vertex) );
		m_is.read( reinterpret_cast<char*> (&ms3d_vertex_t.boneId), sizeof(ms3d_vertex_t.boneId) );
		m_is.read( reinterpret_cast<char*> (&ms3d_vertex_t.referenceCount), sizeof(ms3d_vertex_t.referenceCount) );
		//  insert into my i-th vertex vector class object
		vertices[i][0] = ms3d_vertex_t.vertex[0];
		vertices[i][1] = ms3d_vertex_t.vertex[1];
		vertices[i][2] = ms3d_vertex_t.vertex[2];
	}

	//  read no of triangles
	m_is.read( reinterpret_cast<char*> (&nNumTriangles), sizeof(nNumTriangles) );
	m_nNumTriangles = nNumTriangles;

	//  allocate space for triangles (each triangle consist of 3 vertexes (= 3 vectors))
	triangles = new Vector<>*[m_nNumTriangles];
	normals = new Vector<>*[m_nNumTriangles];
	texCoords = new CThreedModel::TEX_TYPE*[m_nNumTriangles];
	for(int i = 0; i < m_nNumTriangles; i++)
	{
		triangles[i] = new Vector<>[3];
		normals[i] = new Vector<>[3];
		texCoords[i] = new CThreedModel::TEX_TYPE[3];
	}

	//  read all vertices and insert them into just allocated vertices array
	for(int i = 0; i < m_nNumTriangles; i++)
	{
		//  read next triangle
		m_is.read( reinterpret_cast<char*> (&ms3d_triangle_t.flags), sizeof(ms3d_triangle_t.flags) );
		m_is.read( reinterpret_cast<char*> (&ms3d_triangle_t.vertexIndices), sizeof(ms3d_triangle_t.vertexIndices) );
		m_is.read( reinterpret_cast<char*> (&ms3d_triangle_t.vertexNormals), sizeof(ms3d_triangle_t.vertexNormals) );
		m_is.read( reinterpret_cast<char*> (&ms3d_triangle_t.s), sizeof(ms3d_triangle_t.s) );
		m_is.read( reinterpret_cast<char*> (&ms3d_triangle_t.t), sizeof(ms3d_triangle_t.t) );
		for(int j = 0; j < 3; j++)
		{
			//  triangle
			//  i = triangle, j = vector in triangle or vertex in triangle
			triangles[i][j] = vertices[ ms3d_triangle_t.vertexIndices[j] ];
			//  normals
			normals[i][j][0] = ms3d_triangle_t.vertexNormals[j][0];
			normals[i][j][1] = ms3d_triangle_t.vertexNormals[j][1];
			normals[i][j][2] = ms3d_triangle_t.vertexNormals[j][2];
			//  texture coordinates
			texCoords[i][j].s = ms3d_triangle_t.s[j];
			texCoords[i][j].t = ms3d_triangle_t.t[j];
		}
		m_is.read( reinterpret_cast<char*> (&ms3d_triangle_t.smoothingGroup), sizeof(ms3d_triangle_t.smoothingGroup) );
		m_is.read( reinterpret_cast<char*> (&ms3d_triangle_t.groupIndex), sizeof(ms3d_triangle_t.groupIndex) );
	}

	delete [] vertices;

	//  read groups
	m_is.read( reinterpret_cast<char*> (&nNumGroups), sizeof(nNumGroups) );

	for(int i = 0; i < nNumGroups; i++)
	{
		m_is.read( reinterpret_cast<char*> (&ms3d_group_t.flags), sizeof(ms3d_group_t.flags) );
		m_is.read( reinterpret_cast<char*> (&ms3d_group_t.name), sizeof(ms3d_group_t.name) );
		m_is.read( reinterpret_cast<char*> (&ms3d_group_t.numtriangles), sizeof(ms3d_group_t.numtriangles) );
		for(int j = 0; j < ms3d_group_t.numtriangles; j++)
			m_is.read( reinterpret_cast<char*> (&ms3d_group_t.triangleIndices), sizeof(ms3d_group_t.triangleIndices) );
		
		m_is.read( reinterpret_cast<char*> (&ms3d_group_t.materialIndex), sizeof(ms3d_group_t.materialIndex) );
	}

	//  read materials
	m_is.read( reinterpret_cast<char*> (&nNumMaterials), sizeof(nNumMaterials) );

	m_texture = new GLuint[nNumMaterials];

	for(int i = 0; i < nNumMaterials; i++)
	{
		m_is.read( reinterpret_cast<char*> (&ms3d_material_t.name), sizeof(ms3d_material_t.name) );
		m_is.read( reinterpret_cast<char*> (&ms3d_material_t.ambient), sizeof(ms3d_material_t.ambient) );
		m_is.read( reinterpret_cast<char*> (&ms3d_material_t.diffuse), sizeof(ms3d_material_t.diffuse) );
		m_is.read( reinterpret_cast<char*> (&ms3d_material_t.specular), sizeof(ms3d_material_t.specular) );
		m_is.read( reinterpret_cast<char*> (&ms3d_material_t.emissive), sizeof(ms3d_material_t.emissive) );
		m_is.read( reinterpret_cast<char*> (&ms3d_material_t.shininess), sizeof(ms3d_material_t.shininess) );
		m_is.read( reinterpret_cast<char*> (&ms3d_material_t.transparency), sizeof(ms3d_material_t.transparency) );
		m_is.read( reinterpret_cast<char*> (&ms3d_material_t.mode), sizeof(ms3d_material_t.mode) );
		m_is.read( reinterpret_cast<char*> (&ms3d_material_t.texture), sizeof(ms3d_material_t.texture) );
		m_is.read( reinterpret_cast<char*> (&ms3d_material_t.alphamap), sizeof(ms3d_material_t.alphamap) );

		wxString strFileName;
		if ( wxFileName::FileExists( ms3d_material_t.texture ) )
		{
			strFileName = ms3d_material_t.texture;
		}
		else
		{
			wxFileName fn( filename.c_str() );
			fn.SetFullName( ms3d_material_t.texture );
			strFileName = fn.GetFullPath();
		}

		m_texture[i] = m_pOpenGLCanvas->LoadTexture( strFileName );
	}
	
	m_is.close();

	//  create display list
	m_displayList = glGenLists(1);
	if (m_displayList == 0)  throw;
	
	glNewList(m_displayList, GL_COMPILE);

	if (m_texture[0] != 0)  	glBindTexture(GL_TEXTURE_2D, m_texture[0]);

	glMatrixMode(GL_TEXTURE);
		glPushMatrix();
		glLoadIdentity();
		glRotated(180, 0, 1, 0);
		glRotated(180, 0, 0, 1);
	glMatrixMode(GL_MODELVIEW);

	glBegin(GL_TRIANGLES);

	for(int i = 0; i < m_nNumTriangles; i++)
	{			
		for(int j = 0; j < 3; j++)
		{
			if (m_texture[0]!=0) 
				glTexCoord2d(texCoords[i][j].s, texCoords[i][j].t);
			glNormal3d(normals[i][j][0], normals[i][j][1], normals[i][j][2]);
			glVertex3d(triangles[i][j][0], triangles[i][j][1], triangles[i][j][2]);
		}
	}

	glEnd();

	glMatrixMode(GL_TEXTURE);
		glPopMatrix();
	glMatrixMode(GL_MODELVIEW);

	glEndList();
	//  end create display list

	//  free allocated structures
	if (triangles != NULL)
	{
		//  free up space from triangles
		for(int i = 0; i < m_nNumTriangles; i++)
			delete [] triangles[i];
		delete [] triangles;
		
		//  free up space from normals
		for(int i = 0; i < m_nNumTriangles; i++)
			delete [] normals[i];
		delete [] normals;
	}

	if (texCoords !=NULL)
	{
		//  free up space from texture coords
		for(int i = 0; i < m_nNumTriangles; i++)
			delete [] texCoords[i];
		delete [] texCoords;
	}

}