/*
*
*  Copyright (c) 2002-2005 Jakub Schwarzmeier
*
************************************************************/

#include "stdwx.h"

#include "openglcanvas.h"

#include "SerializeReaderWriter.h"
#include <crtdbg.h>

using namespace JakubSchwarzmeier::wx;

#pragma comment(lib, "opengl32.lib")
#pragma comment(lib, "glu32.lib")

IMPLEMENT_DYNAMIC_CLASS (COpenGLCanvas, wxGLCanvas)

BEGIN_EVENT_TABLE(COpenGLCanvas, wxGLCanvas)
    EVT_PAINT(COpenGLCanvas::OnPaint)
    EVT_SIZE(COpenGLCanvas::OnSize)
    EVT_ERASE_BACKGROUND(COpenGLCanvas::OnEraseBkgnd)
    EVT_MOUSE_EVENTS(COpenGLCanvas::OnMouse)
    EVT_KEY_UP(COpenGLCanvas::OnKeyUp)
    EVT_KEY_DOWN(COpenGLCanvas::OnKeyDown)
	EVT_ICONIZE(COpenGLCanvas::OnMinimize)
END_EVENT_TABLE()


COpenGLCanvas::COpenGLCanvas(wxWindow *parent, wxWindowID id,
    const wxPoint& pos, const wxSize& size, long style, const wxString& name,
	int *attribList, const wxPalette& palette)
{
	Create( parent, false, id, pos, size, style, name, attribList, palette );
}

void COpenGLCanvas::Create(wxWindow *parent, bool bDeserialize, wxWindowID id,
    const wxPoint& pos, const wxSize& size, long style, const wxString& name,
	int *attribList, const wxPalette& palette)
{
	m_bSize = false;

	/*  Parent's ctor begin */
	bool ret = wxGLCanvas::Create(parent, id, pos, size, style, name);

	if ( ret )
	{
		SetBackgroundColour(wxSystemSettings::GetColour(wxSYS_COLOUR_WINDOW));
	}

	m_hDC = (WXHDC) ::GetDC((HWND) GetHWND());

	SetupPixelFormat(attribList);
//	SetupPalette(palette);
	/*  Parent's ctor end */

	m_quadric = NULL;
	m_lastList = 0;
	for(int i = 0; i < GL_MAX_LIGHTS; i++)		//  available number of lights
		m_freeLights[i] = true;

	m_pRenderingThread = NULL;
	m_bAutoRender = true;
	m_bWasAutoRender = true;

	m_internalObserver = new CObserver;
	SetObserver( m_internalObserver );

	m_nearClippingPlane = 1.0;
	m_farClippingPlane = 5000.0;
	m_fov = 50.0;
	m_aspect = 1.0;

	m_bTexturing = true;
	m_polygonMode = GL_FILL;
	m_bLighting = false;
	m_bColorMaterial = false;
	m_bCullFace = true;

	if (bDeserialize)
	{
		GetMyContext()->Create( false, this, true );
	}
	else
	{
        m_glContext = new wxMyGLContext(true, this, wxNullPalette);
//		GetMyContext()->CreatePBuffer();
	}

	//  threading
	m_pRenderingThread = new CRenderingThread;
	m_pRenderingThread->Create(this);
	m_pRenderingThread->Run();

	m_bSize = true;
}

//  NOTES:
//
// An OpenGL window must be created with the WS_CLIPSIBLINGS | WS_CLIPCHILDREN 
// flags and must not include CS_PARENTDC for the class style. 
// Refer to SetPixelFormat documentation in the "Remarks" section 
// for further information.
//
// Win95 require window class with CS_OWNDC style, because of limited GDI
// memory. If you don't have window with CS_OWNDC, Win95 can "steal" 
// your DC, which lead to MakeCurrent failure. Not implemented here.
//

COpenGLCanvas::~COpenGLCanvas()
{
    wxWindow* oldWindow = wxWindow::FindWindowByName( SYMBOL_COPENGLCANVASDLG_TITLE, GetParent()->GetParent() );
	if (oldWindow)  oldWindow->Destroy();
	
	//  prevent destruction of observers (dlgs)
    for ( wxWindowList::compatibility_iterator node = GetChildren().GetFirst();
          node;
          node = node->GetNext() )
    {
        wxWindow *win = node->GetData();
		win->SetParent( NULL );
	}

	GetChildren().Clear();

	delete m_internalObserver;
}

bool COpenGLCanvas::Destroy()
{
	m_pRenderingThread->m_bTerminateThread = true;
	ForceReRender();
	m_pRenderingThread->Wait();
	delete m_pRenderingThread;

	GetMyContext()->DontUsePBuffer();
	
	GetMyContext()->SetCurrent();
	OnContextTerminate();
	GetMyContext()->UnsetCurrent();

	delete this;

	return true;
}

/*virtual*/
void COpenGLCanvas::PreRender()
{
	//  global properties

	glClearColor ( 0.0, 0.0, 0.0, 0.0 );

	//  set depth (z-buffer) attributes
	glEnable(GL_DEPTH_TEST);			//  glClear must then have set GL_DEPTH_BUFFER_BIT flag, otherwise it cause displaying errors!!!
	glDepthFunc(GL_LEQUAL);
	glClearDepth(1.0f);	//  filled with a value that's as far from the viewpoint as possible, so any object is nearer than that

	// set up one-byte alignment for pixel storage (saves memory)
	//  solved problem with font
	glPixelStorei(GL_PACK_ALIGNMENT, 1);
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	//  performance and quality attributes
	glShadeModel(GL_SMOOTH);		// make the shading smooth
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	glEnable(GL_POINT_SMOOTH);
	glEnable(GL_CULL_FACE);

	//  projection properties

	wxRect rect = GetRect();
	GLint width = rect.GetWidth();
	GLint height = rect.GetHeight();

	glViewport(0, 0, width, height);		//  set transformation from device to window coordinates

	glMatrixMode(GL_PROJECTION); 
	glLoadIdentity();

	m_aspect = static_cast<GLfloat>(width)/static_cast<GLfloat>(height);

	//  point of this is that eye is every time in 0,0,0 
	//  glFrustum (-glAspect, glAspect, -1, 1, 1.0, 5000.0);
	//  multiply projection matrix 
	gluPerspective(/*FOV*/m_fov, m_aspect, m_nearClippingPlane, m_farClippingPlane);

	glMatrixMode   ( GL_MODELVIEW );
}

void COpenGLCanvas::OnContextTerminate()
{
}

void COpenGLCanvas::SetRenderingParameters()
{
	//  texturing
	if (m_bTexturing)	
		glEnable( GL_TEXTURE_2D );
	else
		glDisable( GL_TEXTURE_2D );

	//  wireframe/fill mode
	glPolygonMode(GL_FRONT_AND_BACK, m_polygonMode);

	//  light
	if (m_bLighting)	
		glEnable( GL_LIGHTING  );
	else
		glDisable( GL_LIGHTING );

	//  material
	if (m_bColorMaterial)	
		glEnable( GL_COLOR_MATERIAL );
	else
		glDisable( GL_COLOR_MATERIAL );

	//  face culling
	if (m_bCullFace)
		glEnable( GL_CULL_FACE );
	else
		glDisable(GL_CULL_FACE );
}

void COpenGLCanvas::OnRender()
{
	//  clear the color and depth buffers 
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); 

	//  modeling transformation in ModelView matrix

	glLoadIdentity();

	//  simulate viewing transformation in ModelView matrix
	glMultMatrixd( m_observer->GetCamera()->GetMatrix() );

	//  objects drawing

	glBegin( GL_QUADS );
		glColor3f( 1.0f, 0.0f, 0.0f ); glVertex2f( 1, 1 );
		glColor3f( 0.0f, 1.0f, 0.0f ); glVertex2f( -1, 1 );
		glColor3f( 0.0f, 0.0f, 1.0f ); glVertex2f( -1, -1 );
		glColor3f( 0.0f, 1.0f, 0.0f ); glVertex2f( 1, -1 );

		glColor3f( 1.0f, 0.0f, 0.0f ); glVertex2f( 1, 1 );
		glColor3f( 0.0f, 1.0f, 0.0f ); glVertex2f( 1, -1 );
		glColor3f( 0.0f, 0.0f, 1.0f ); glVertex2f( -1, -1 );
		glColor3f( 0.0f, 1.0f, 0.0f ); glVertex2f( -1, 1 );
	glEnd();

	glBegin( GL_LINES );
		glColor3f( 1.0f, 1.0f, 1.0f );
	for(int i = -40; i <= 40; i+=2)
	{
		//  perpendicular
		glVertex3f( 100, 0, i );	glVertex3f( -100, 0, i );
		glVertex3f( i, 0, 100 );	glVertex3f( i, 0, -100 );

		//  z
		glVertex3f( -15, -100, i );	glVertex3f( -15, 100, i );
		glVertex3f( -15, i, -100 );	glVertex3f( -15, i, 100 );
	}
		glEnd();

	//  y-axis
	glBegin( GL_LINES );
		glVertex3f( 0, 100, 0 );	glVertex3f( 0, -100, 0 );
	glEnd();


	//  flush rendering pipe
	glFinish();
}

//  Put backbuffer fastly on screen.
void COpenGLCanvas::OnDraw( wxDC * pDC )
{
	SwapBuffers();		//  blits the backbuffer into DC
}


void COpenGLCanvas::SetObserver( CObserver * observer )
{
	observer->OnCreateCamera();
	observer->SetOpenGLCanvas( this );
	m_observer = observer;
	//  Function SetObserver is called from COpenGLCanvas constructor when 
	//  no rendering thread exist yet.
	if (m_pRenderingThread!=NULL)
	{
		ReRender();	
	}
}

CObserver * COpenGLCanvas::GetObserver( )
{
	return m_observer;
}


GLUquadricObj * COpenGLCanvas::GetQuadric()
{
	if ( m_quadric == NULL )  m_quadric = gluNewQuadric();

	return m_quadric;
}

wxString COpenGLCanvas::GetLastError()
{
	return wxString( reinterpret_cast<const char *>( gluErrorString(glGetError()) ) );
}

int COpenGLCanvas::ReserveLight()
{
	for(int i = GL_LIGHT0; i < GL_LIGHT0 + GL_MAX_LIGHTS; i++)
	{
		if (m_freeLights[i-GL_LIGHT0])
		{
			m_freeLights[i-GL_LIGHT0] = false;
			
			return i;
		}
	}

	return -1;
}

void COpenGLCanvas::DisposeLight(int light)
{
	m_freeLights[light-GL_LIGHT0] = true;
}

GLuint COpenGLCanvas::GetList()
{
	return ++m_lastList;
}

void COpenGLCanvas::ShowOpenGLInfo()
{
    wxWindow* oldWindow = wxWindow::FindWindowByName( SYMBOL_COPENGLCANVASDLG_TITLE, GetParent()->GetParent() );
    if (oldWindow)
    {
		oldWindow->Show();
        return;
    }

    COpenGLCanvasDlg* window = new COpenGLCanvasDlg( GetParent()->GetParent(), this );
    window->Show(true);
}

void COpenGLCanvas::ShowObserver()
{
	m_observer->Show( /*parent*/this, /*canvas*/this );
}


void COpenGLCanvas::OnPaint(wxPaintEvent& event)
{
    wxPaintDC dc(this);		//  must always be here

	if (m_bSize)  OnDraw( &dc );
}

//  For the first time this is called after window creation
//  and before the window is made visible for the first time.
void COpenGLCanvas::OnSize(wxSizeEvent& event)
{
    // this is also necessary to update the context on some platforms
    wxGLCanvas::OnSize(event);

	wxRect rect = GetRect();
	GLint width = rect.GetWidth();
	GLint height = rect.GetHeight();

	if (m_bSize)
	{
		GetMyContext()->ResizePBuffer();
		ReRender();
	}
}

void COpenGLCanvas::OnEraseBkgnd(wxEraseEvent& event)
{
	//  pretend we handle it, do nothing, to avoid flashing on MSW
}

void COpenGLCanvas::OnMouse(wxMouseEvent & event)
{
	m_csRenderingContext.Enter();
	m_observer->OnMouse(event);
	m_csRenderingContext.Leave();
}

void COpenGLCanvas::OnKeyDown(wxKeyEvent & event)
{
	m_csRenderingContext.Enter();

	switch( event.GetKeyCode() )
	{
	case WXK_F1:

		ShowOpenGLInfo();
		break;

	case WXK_F5:

		ShowObserver();
		break;
	}

	m_observer->OnKeyDown(event);

	m_csRenderingContext.Leave();
}

void COpenGLCanvas::OnKeyUp(wxKeyEvent & event)
{
	m_csRenderingContext.Enter();
	m_observer->OnKeyUp(event);
	m_csRenderingContext.Leave();

}

void COpenGLCanvas::SetSignaled( bool signaled )
{
	m_pRenderingThread->m_bSignaled = signaled;
}

double COpenGLCanvas::GetFov()
{
	return m_fov;
}

double COpenGLCanvas::GetAspect()
{
	return m_aspect;
}

void COpenGLCanvas::GetCopyImage(wxImage & newImage)
{
	GetMyContext()->GetCopyImage(newImage);
}

void COpenGLCanvas::SetAutoRender(bool autoRender)
{
	m_bAutoRender = autoRender;
}

bool COpenGLCanvas::GetAutoRender()
{
	return m_bAutoRender;
}

GLuint COpenGLCanvas::LoadTexture(wxString fileName)
{
	wxImage m_image;
	bool b = wxFileName::FileExists( fileName );
	if (b) b = m_image.LoadFile( fileName );
	if(b)
	{
		GLboolean bTexture;
		if ( !( bTexture = glIsEnabled ( GL_TEXTURE_2D ) ) )
			glEnable( GL_TEXTURE_2D );

		GLuint textureID;
		glGenTextures(1, &textureID);	//  return one unused texture name
		glBindTexture(GL_TEXTURE_2D, textureID);	//  bind this texture name as current

		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE );	//  if GL_DECAL, it don't respect lighting
																		//  GL_MODULATE: blend the texture with the base colour of the object

		glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

		//  fastest (in meaning of speed) is to use glTextImage2D
		//  then is usable to use textures with power of two and gluBuild2DMipmaps
		//  and slowest is to use textures with no power of two and gluBuild2DMipmaps

		//  glTexImage2D requires textures in power of two; gluBuild2DMipmaps rescale size if needed
		//  glTexImage2D(GL_TEXTURE_2D, 0, 3, width, height, 0, GL_BGR_EXT, GL_UNSIGNED_BYTE,  bitmap.Scan0);

		int err = gluBuild2DMipmaps(GL_TEXTURE_2D, 3, m_image.GetWidth(), m_image.GetHeight(), GL_RGB, GL_UNSIGNED_BYTE, m_image.GetData() );
		if (err != 0)	throw;

		if (bTexture == GL_FALSE)	glDisable( GL_TEXTURE_2D );

		return textureID;
	}

	return 0;
}

void COpenGLCanvas::OnMinimize(wxIconizeEvent & event)
{
	if ( event.Iconized() )
	{
		m_bWasAutoRender = GetAutoRender();
		SetAutoRender( false );
	}
	else
	{
		SetAutoRender( m_bWasAutoRender );
	}

	event.Skip();
}

void COpenGLCanvas::ForceReRender()
{
	//  In simple terms following calls ThreadedReRender(), threaded 
	//  variant of ReRender.
	m_pRenderingThread->m_condition.Signal();
}

void COpenGLCanvas::ReRender()
{
	if (m_bAutoRender)  m_pRenderingThread->m_condition.Signal();
}

void COpenGLCanvas::ThreadedReRender()
{
	m_observer->UpdateInputs();

	//  Perform rendering
	GetMyContext()->SetCurrent();
	SetRenderingParameters();
	OnRender();
	GetMyContext()->UpdatePBuffer();
	GetMyContext()->UnsetCurrent();
	
	RenderFinished();
}

void COpenGLCanvas::RenderFinished()
{
	//  Call MDI Child window's OnPaint
	GetParent()->Refresh();
}

void COpenGLCanvas::SwapBuffers()
{
	GetMyContext()->SwapBuffers();
}

wxMyGLContext * COpenGLCanvas::GetMyContext()
{
	return reinterpret_cast<wxMyGLContext*>( GetContext() );
}

bool COpenGLCanvas::OnSerialize( JakubSchwarzmeier::wx::CSerializeReaderWriter * dir )
{
	wxGLCanvas::OnSerialize(dir);
	dir->Append( reinterpret_cast<wxObject **>( &m_glContext ), wxString("m_glContext") );

	return true;
}

CRenderingThread::CRenderingThread(): wxThread(wxTHREAD_JOINABLE), m_condition(m_mutex)
{ 
	m_glCanvas = NULL;
	m_bTerminateThread = false;
	m_bSignaled = true;
	if (!m_condition.IsOk())
	{
		wxCHECK_RET( 0, wxT("Error: condition object not ok.") );
	}
}

void CRenderingThread::Create(COpenGLCanvas * glCanvas, bool signaled)
{
	wxThread::Create();
	m_glCanvas = glCanvas;
	m_bSignaled = signaled;
}

void * CRenderingThread::Entry()
{ 
	while (1)
	{
		m_mutex.Lock();
		// Wait atomically releases the lock on the associated mutex
		// and starts waiting to be woken up by a Signal()/Broadcast()
		// once its signaled, then it will wait until it can reacquire
		// the lock on the associated mutex object, before returning.
		if (!m_bSignaled)
			m_condition.Wait();

		// check if we were asked to exit
        if ( TestDestroy() | m_bTerminateThread)
		{
			m_glCanvas->RenderFinished();
            break;
		}

		m_glCanvas->ThreadedReRender(); 
	}

	return NULL;
}

