#include "StdAfx.h"
#include ".\animation.h"
#include "sphericalcamera.h"

using namespace JakubSchwarzmeier::Std::Math;
using namespace JakubSchwarzmeier::Std::Graphics;

CAnimation::CAnimation(void)
{
}

CAnimation::~CAnimation(void)
{
	//  release video writer
	cvReleaseVideoWriter( &m_pVideoWriter );

	m_pRectangular->Destroy();
	m_pSpherical->Destroy();

	m_pVisualization->GetThreedCanvas()->SetAnimation( false );
}

size_t CAnimation::Create (CVisualization * vis, double fps, int w, int h)
{
	m_pVisualization = vis;
	m_pChunks = m_pVisualization->GetChunks();

	m_pRectangular = new CRectangularObserver;
	m_pRectangular->OnCreateCamera();
	m_pSpherical = new CSphericalObserver;
	m_pSpherical->OnCreateCamera();

	m_fps = fps;
	m_width = w;
	m_height = h;

	vis->GetThreedCanvas()->SetSize(m_width, m_height);
	vis->GetThreedCanvas()->SetAnimation( true );

	m_pVisualization->SetEt( m_pChunks->at( 0 )->GetStartTime() );
	vis->ClearTrajectories();

	//  Create video writer
	m_pVideoWriter = cvCreateVideoWriter( vis->GetOutputVideo().c_str(), 
	#if defined(__WXMSW__)
		-1,	/* let the user choose video codec */
	#else
        CV_FOURCC('P','I','M','1'), /* MPEG-1 codec */
	#endif
		m_fps, cvSize(m_width, m_height) );

	//  compute overall number of frames needed for production of whole animation
	size_t nTotalFrames = 0;
	std::vector<CChunkDesc*>::const_iterator it;
	for(it = m_pChunks->begin(); it != m_pChunks->end(); it++)
	{
		nTotalFrames += (*it)->GetDuration();
	}

	nTotalFrames *= m_fps;

	//  set initial values for animation generation
	if ( m_pChunks->empty() )
		return -1;
	m_iChunk = 0;
	m_iChunkPos = -1;

    return nTotalFrames;
}

void CAnimation::Update()
{
	//  render frame
	//

	//  get current chunk
	CChunkDesc * chunk = m_pChunks->at( m_iChunk );
	if ( m_iChunkPos > ( chunk->GetDuration() * m_fps ) )
	{
		m_iChunkPos = 0;
		m_iChunk++;

		if (m_iChunk >= m_pChunks->size())
			return;		//  we should already finish
		else
		{
			chunk = m_pChunks->at( m_iChunk );
		}
	}
	else
	{
		m_iChunkPos++;
	}

	if (chunk->GetDuration() == 0)	throw;
	//  total number of steps (frames) in this chunk
	double steps = chunk->GetDuration() * m_fps;

	//  set time
	double simulationTimeSpan = chunk->GetEndTime() - chunk->GetStartTime();
	double simulationTimeStep = simulationTimeSpan / (steps);
	
	double currentSimulationTime = simulationTimeStep * (double)m_iChunkPos + chunk->GetStartTime();
	m_pVisualization->SetEt( currentSimulationTime );

	//  set camera
	CCamera * startCamera = chunk->GetStartCamera();
	CCamera * endCamera = chunk->GetEndCamera();

	double t = (double)m_iChunkPos / steps;

	if ( startCamera->IsKindOf( CLASSINFO( CSphericalCamera ) ) )
	{
		CSphericalCamera* startCam = dynamic_cast<CSphericalCamera*>(startCamera);
		CSphericalCamera* endCam = dynamic_cast<CSphericalCamera*>(endCamera);

		double startDistance = startCam->GetDistance();
		double startLongitude = startCam->GetLongitude();
		double startLatitude = startCam->GetLatitude(); 

		double lenDistance = endCam->GetDistance() - startDistance;
		double lenLongitude = endCam->GetLongitude() - startLongitude;
		double lenLatitude = endCam->GetLatitude() - startLatitude; 

		//  linear interpolation
		double currentDistance = startDistance + t * lenDistance;
		double currentLongitude = startLongitude + t * lenLongitude;
		double currentLatitude = startLatitude + t * lenLatitude;

		CSphericalCamera camera = *(dynamic_cast<CSphericalCamera*>(startCamera));
		camera.SetLongitude( currentLongitude );
		camera.SetLatitude( currentLatitude );
		camera.SetDistance( currentDistance );
		m_pSpherical->SetCamera( &camera );

		m_pVisualization->GetThreedCanvas()->SetObserver( m_pSpherical );
	}
	else
	if ( startCamera->IsKindOf( CLASSINFO( CRectangularCamera ) ) )
	{
		RotationMatrix startOrbiter = dynamic_cast<CRectangularCamera*>(startCamera)->GetOrbiter();
		RotationMatrix endOrbiter = dynamic_cast<CRectangularCamera*>(endCamera)->GetOrbiter();

		Vector<> direction;
		direction[0] = endOrbiter[12] - startOrbiter[12];
		direction[1] = endOrbiter[13] - startOrbiter[13];
		direction[2] = endOrbiter[14] - startOrbiter[14];

		Vector<> start;
		start[0] = startOrbiter[12];
		start[1] = startOrbiter[13];
		start[2] = startOrbiter[14];

		//  linear interpolation
		Vector<> currentPosition = start + t * direction;
		Quaternion<> currentOrientation = Slerp( startOrbiter.GetQuaternion(),
			endOrbiter.GetQuaternion(), t );

		RotationMatrix currentOrbiter(currentPosition, currentOrientation);

		CRectangularCamera camera = *(dynamic_cast<CRectangularCamera*>(startCamera));
		camera.SetOrbiter( currentOrbiter );
		m_pRectangular->SetCamera( &camera );

		m_pVisualization->GetThreedCanvas()->SetObserver( m_pRectangular );
	}
	else
	{
	}

	m_pVisualization->GetThreedCanvas()->ForceReRender();

	//  get the image
	m_pVisualization->GetThreedCanvas()->GetCopyImage( m_image );

	//  generate corresponding OpenCV image
	CvSize size;
    size.width = m_image.GetWidth();
    size.height = m_image.GetHeight();

	if ( (m_width != size.width) || (m_height != size.height) )
	{
		throw;
	}

	int depth = IPL_DEPTH_8U;

    IplImage * i_image = cvCreateImage( size, depth, 3 );
 
	//  Create continuous block of memory
	char * continuous = i_image->imageData; //  Intel allocated raw bitmap storage for us
	unsigned int vsize = 3;		//  3 bytes for one pixel

	for(int i = 0; i < size.height; i++)
	{
		for(int j = 0; j < size.width; j++)
		{
			for(int k = 0, l = 2; k < vsize; k++, l--)
			{
				char * src = reinterpret_cast<char *>(m_image.GetData() + i * vsize * size.width + j * vsize + l );

				//  there is padding at the end of each row (widthStep - width)
				memcpy( /*dest*/continuous + i * i_image->width * vsize + i * (i_image->widthStep/vsize - i_image->width) + j * vsize + k, 
					/*src*/src, vsize);
			}
		}
	}

	//  write animation frame
	int e = cvWriteFrame( m_pVideoWriter, i_image );
	if (e != 1) throw;

	//  release OpenCV image
	cvReleaseImage(&i_image);
}