#include "StdAfx.h"
#include ".\bodytarget.h"

#include <spicelib/spicelib.h>

using namespace JakubSchwarzmeier::Std::Math;
using namespace JakubSchwarzmeier::Std::Graphics;

CBodyTarget::CBodyTarget(Body * body, CVisualization * visualization)
{
	m_pBody = body;
	m_pVisualization = visualization;

	//  trajectory
	Body * parent = m_pBody->GetParent();
	if (parent != NULL)
	{
		Vector<> posParent = parent->GetPosition();
		Vector<> velParent = parent->GetVelocity();

		Vector<> pos = m_pBody->GetPosition();
		Vector<> vel = m_pBody->GetVelocity();

		ConstSpiceDouble state[6] = {
		(pos[0]-posParent[0]) / 1000.0,
		(pos[1]-posParent[1]) / 1000.0,
		(pos[2]-posParent[2]) / 1000.0,
		(vel[0]-velParent[0]) / 1000.0,
		(vel[1]-velParent[1]) / 1000.0,
		(vel[2]-velParent[2]) / 1000.0 };

		double centralBodyMass = parent->GetMass();

		SpiceDouble mu = Simulation::GetG()  * centralBodyMass * 1e-9;
		wxDateTime et( m_pVisualization->GetEt() );

		SpiceDouble elts[8];

		try
		{
			oscelt_c ( state, ( et.GetJDN() - 2451545.0 ) * 86400.0, mu, elts );

			m_T = sqrt( ( 4 * M_PI * M_PI * pow( 1000.0*(elts[0]), 3 ) / 
				(Simulation::GetG() * ( centralBodyMass ) ) ) );

		}
		catch (...)
		{
			m_T = 365*24*60*60;
		}


		m_tLastSave = m_pVisualization->GetEt();
		m_positions.push_back( pos - posParent );
	}
}

CBodyTarget::~CBodyTarget(void)
{
}

void CBodyTarget::Render()
{
	CThreedCanvas * openGLCanvas = m_pVisualization->GetThreedCanvas();

	RotationMatrix objectInEye = GetObjectInEye();

	const CFrustum * frustum = openGLCanvas->GetFrustum();

	Trajectory();

	if ( frustum->PointInFrustum( objectInEye[12], objectInEye[13], objectInEye[14] ) )
	{
		//  draw body name
		DrawName();

		//  draw body
		glLoadMatrixd( objectInEye.GetMatrix() );

	    glEnable( GL_POINT_SPRITE_ARB );

		glPointSize( 5 );

		glPushAttrib( GL_ENABLE_BIT );
			glDisable( GL_TEXTURE_2D );
			glDisable( GL_LIGHTING );

			glBegin (GL_POINTS);
				glVertex2f( 0, 0 );
			glEnd();
		glPopAttrib();

	    glDisable( GL_POINT_SPRITE_ARB );

	}
}

void CBodyTarget::PreRender()
{
}

void CBodyTarget::OnContextTerminate()
{
}

RotationMatrix CBodyTarget::GetObjectInEye()
{
	CThreedCanvas * openGLCanvas = m_pVisualization->GetThreedCanvas();

	Vector<> pos = m_pBody->GetPosition();
	RotationMatrix objectInWCSCam;
	objectInWCSCam.SetPosition( pos );
	objectInWCSCam.Inverse();	//  modeling transformation

	return RotationMatrix( objectInWCSCam * *(openGLCanvas->GetObserver()->GetCamera()) );
}

double CBodyTarget::GetFurthermostDistance()
{
	RotationMatrix objectInEye = GetObjectInEye();
	return objectInEye[14] - fabs(objectInEye[14])/1e4;
}

double CBodyTarget::GetNearestDistance()
{
	RotationMatrix objectInEye = GetObjectInEye();
	return objectInEye[14] + fabs(objectInEye[14])/1e4;
}

bool CBodyTarget::IsLarge()
{
	return false;
}

void CBodyTarget::DrawName()
{
	GLdouble model[16];
	GLdouble proj[16];
	GLint view[4];

	glGetDoublev(GL_MODELVIEW_MATRIX, model);
	glGetDoublev(GL_PROJECTION_MATRIX, proj);
	glGetIntegerv(GL_VIEWPORT, view);

	RotationMatrix objectInEye = GetObjectInEye();

	//  position in pixels
	GLdouble wx, wy, wz;
	int e1 = gluProject(objectInEye[12], 
		objectInEye[13], 
		objectInEye[14],
		model, proj, view,
		&wx, &wy, &wz);

	int r, g, b;
	m_pBody->GetColor(&r, &g, &b);
	glColor3ub(r, g, b);	

	if (e1 == GL_TRUE)
		m_pVisualization->GetBodyFont()->PrintWindow(wx+5, view[3]-wy+3, m_pBody->GetName(), JakubSchwarzmeier::wx::CFreetype::TOP_LEFT);
}

void CBodyTarget::Trajectory()
{
	Body * parent = m_pBody->GetParent();
	if (parent != NULL)
	{
		//  remember position on body's path

		Vector<> vecCenter = 0, curPos = 0;

		//  if it is time, save the trajectory position
		if ( ((m_pVisualization->GetEt() - m_tLastSave)) >= (m_T/s_nPositionsPerPeriod)/(24*60*60) )
		{
			//  insert into other positions
			for(double dt = m_tLastSave; dt <= m_pVisualization->GetEt(); dt += (m_T/s_nPositionsPerPeriod)/(24*60*60) )
			{
				JakubSchwarzmeier::Std::Math::Vector<> dummyVel;
				try
				{
				m_pVisualization->GetSimulation()->GetSpice()->GetEphemerisEclipticRectangular( 
					0/*SSB*/, m_pBody->GetSpiceBodyID(), dt, curPos, dummyVel );
				m_pVisualization->GetSimulation()->GetSpice()->GetEphemerisEclipticRectangular( 
					0/*SSB*/, parent->GetSpiceBodyID(), dt, vecCenter, dummyVel );

				m_positions.push_back( curPos - vecCenter );
				}
				catch(...){
				
				}
				
			}

			//  assign new
			m_tLastSave = m_pVisualization->GetEt();

			//  delete trailing positions
			if ( m_positions.size() > s_nRememberPeriods * s_nPositionsPerPeriod )
			{
				std::vector< JakubSchwarzmeier::Std::Math::Vector<> >::iterator it;
				it = m_positions.begin();
				int i = m_positions.size();
				while(i > s_nRememberPeriods * s_nPositionsPerPeriod)
				{
					it++;
					i--;
				}

				m_positions.erase( m_positions.begin(), it );
			}
		}
	}
}

void CBodyTarget::DrawTrajectory()
{
	Body * parent = m_pBody->GetParent();
	if (parent != NULL)
	{
		//  draw path

		Vector<> vecCenter = parent->GetPosition();
		Vector<> curPos = m_pBody->GetPosition();

		int r, g, b;
		m_pBody->GetColor(&r, &g, &b);
		glColor3ub(r, g, b);

		CThreedCanvas * openGLCanvas = m_pVisualization->GetThreedCanvas();

		glPushAttrib( GL_ENABLE_BIT );
			glDisable( GL_TEXTURE_2D );
			glDisable( GL_LIGHTING );
//			glDisable( GL_DEPTH_TEST );

			glPushMatrix();
				glLoadMatrixd( openGLCanvas->GetObserver()->GetCamera()->GetMatrix() );

				glBegin(GL_LINE_STRIP);

					//  previous positions
					std::vector< JakubSchwarzmeier::Std::Math::Vector<> >::const_iterator it;
					for(it = m_positions.begin(); it != m_positions.end(); it++)
					{
						glVertex3d(vecCenter[0]+(*it)[0], vecCenter[1]+(*it)[1], 
							vecCenter[2] + (*it)[2] );
					}

					//  current position
					glVertex3d(curPos[0], curPos[1], curPos[2] );

				glEnd();
			glPopMatrix();

		glPopAttrib();
	}
}

void CBodyTarget::ClearTrajectory()
{
	m_positions.clear();
	m_tLastSave = m_pVisualization->GetEt();

	Body * parent = m_pBody->GetParent();
	if (parent != NULL)
	{
		Vector<> posParent = parent->GetPosition();
		Vector<> velParent = parent->GetVelocity();

		Vector<> pos, vel;

		JakubSchwarzmeier::Std::Math::Vector<> dummyVel;
		try
		{
		m_pVisualization->GetSimulation()->GetSpice()->GetEphemerisEclipticRectangular( 
			0/*SSB*/, m_pBody->GetSpiceBodyID(), m_tLastSave, pos, vel );
		m_pVisualization->GetSimulation()->GetSpice()->GetEphemerisEclipticRectangular( 
			0/*SSB*/, parent->GetSpiceBodyID(), m_tLastSave, posParent, velParent );

		m_positions.push_back( pos - posParent );

		//

		ConstSpiceDouble state[6] = {
		(pos[0]-posParent[0]) / 1000.0,
		(pos[1]-posParent[1]) / 1000.0,
		(pos[2]-posParent[2]) / 1000.0,
		(vel[0]-velParent[0]) / 1000.0,
		(vel[1]-velParent[1]) / 1000.0,
		(vel[2]-velParent[2]) / 1000.0 };

		double centralBodyMass = parent->GetMass();

		SpiceDouble mu = Simulation::GetG()  * centralBodyMass * 1e-9;
		wxDateTime et( m_pVisualization->GetEt() );

		SpiceDouble elts[8];

		try
		{
			oscelt_c ( state, ( et.GetJDN() - 2451545.0 ) * 86400.0, mu, elts );

			m_T = sqrt( ( 4 * M_PI * M_PI * pow( 1000.0*(elts[0]), 3 ) / 
				(Simulation::GetG() * ( centralBodyMass ) ) ) );

		}
		catch (...)
		{
			m_T = 365*24*60*60;
		}

		}
		catch(...){
		
		}
	}
}

size_t CBodyTarget::s_nRememberPeriods = 3;
size_t CBodyTarget::s_nPositionsPerPeriod = 100;