#include "stdafx.h"

#include "Comet.h"
#include "simulation.h"

#include <spicelib/spicelib.h>
#include <schwarzmeier/std.h>

using namespace JakubSchwarzmeier::Std::Math;

IMPLEMENT_DYNAMIC_CLASS (Comet, IrregularBody)

Comet::Comet()
{
	m_model = Marsden;
	m_eta = 0;
	m_tau = 0;
	for(int i = 0; i < 3; i++)  m_A[i] = 0;

    //  dispatch
	mapDispatch.insert( MAPDISPATCH::value_type( this->GetClassInfo() , Assign2 ) );
}

Comet::Comet( const Comet & original )
{
	operator=( original );
}

/* static */
void Comet::Assign2(Body * body, Body * right_side) 
{
	//  retype
	Comet * comet_left = dynamic_cast<Comet*>( body );
	Comet * comet_right = dynamic_cast<Comet*>( right_side );
	comet_left->operator=( *comet_right );
}

Comet & Comet::operator=( const Comet & right_side )
{
	if ( &right_side == this )  return (*this);

	IrregularBody::operator=( right_side );

	m_model = right_side.m_model;
	m_eta = right_side.m_eta;
	m_tau = right_side.m_tau;
	for(int i = 0; i < 3; i++)  m_A[i] = right_side.m_A[i];
	
	return *this;
}

void Comet::Update()
{
	Body * star = NULL;
	//  find star
	std::vector<Body*>::const_iterator it;
	for ( it = m_simulation->GetBodies()->begin( ) ; it != m_simulation->GetBodies()->end( ) ; it++ )
	{
		if ( (*it)->GetSpiceBodyID() == 10 )
		{
			star = (*it);
		}
	}

	//  Lsun
	double lambda = 
		m_simulation->GetSpice()->GetLSun( 
			m_spiceBodyID, m_simulation->GetEt().GetJDN()
			);
	
	//  Angle between plane of comet's orbit and her equator
	//

	double I;

	double latitude = GetPoleLatitude();
	double longitude = GetPoleLongitude();

    //  compute Keplerian parameter ``inclination''
	ConstSpiceDouble state[6] = {
		m_position[0] / 1000.0,
		m_position[1] / 1000.0,
		m_position[2] / 1000.0,
		m_velocity[0] / 1000.0,
		m_velocity[1] / 1000.0,
		m_velocity[2] / 1000.0 };

	SpiceDouble mu = Simulation::GetG() * star->GetMass() * 1e-9;

	SpiceDouble elts[8];

	oscelt_c ( state, ( m_simulation->GetEt().GetJDN() - 2451545.0 ) * 86400.0, mu, elts );

	//  inclination of body's orbit to ecliptic in degrees
	double inclKeplerian = elts[2] * dpr_c ();

	I = inclKeplerian - latitude;

	//
	//  end Angle between plane of comet's orbit and her equator

	//  compute directional cosines, Z. Sekanina (1981)
	double C1 = Functions::cos_angle(m_eta)+(1-Functions::cos_angle(m_eta))*pow(Functions::sin_angle(I),2)*pow(Functions::sin_angle(lambda),2);
	double C2 = Functions::sin_angle(m_eta)*Functions::cos_angle(I)+(1-Functions::cos_angle(m_eta))*pow(Functions::sin_angle(I),2)*Functions::sin_angle(lambda)*Functions::cos_angle(lambda);
	double C3 = -(Functions::sin_angle(m_eta)*Functions::cos_angle(lambda)-(1-Functions::cos_angle(m_eta))*Functions::cos_angle(I)*Functions::sin_angle(lambda))*Functions::sin_angle(I);

	//  radial vector from star to comet, using for determination of true anomaly too
	Vector<> radial = m_position - star->GetPosition();

	//  water ice sublimation rate

	double distance;
	if (m_tau != 0)
	{
		//  asymmetric with regard to perihelion

		wxDateTime atTime = m_simulation->GetEt() + wxTimeSpan(0, 0, m_tau );

		SpiceDouble state[6];
		conics_c( elts, ( atTime.GetJDN() - 2451545.0 ) * 86400.0, state );
		distance = 1000. * sqrt( pow(state[0], 2) + pow(state[1], 2)+pow(state[2], 2) );
	}
	else
	{
		distance = radial.GetLength();
	}

	//  water vaporization curve
	double rr0 = (distance/149.6e9)/2.808;			//  r/r0 [AU/AU]=[1]
	double g = 0.111262*pow(rr0, -2.15)*pow( 1+pow(rr0, 5.093) , -4.6142);

	//  compute radial unit vector (e1)
	radial.Normalize();

	//  compute transversal unit vector (e2)
	Vector<> transversal = m_velocity;
	transversal.Normalize();

	//  compute normal unit vector (e3)
	Vector<> normal = vector_product(radial, transversal);
	normal.Normalize();

	//  nongravitational acceleration
	Vector<> nongravitational = 0;
	switch (m_model)
	{
	case Marsden:
		{
		nongravitational = m_A[0]*radial*g;
		nongravitational += m_A[1]*transversal*g;
		nongravitational += m_A[2]*normal*g;
		
		break;
		}

	case noprecession:
	case linearprecession:
		{
		nongravitational = m_A[0]*C1*radial*g;
		nongravitational += m_A[1]*C2*transversal*g;
		nongravitational += m_A[2]*C3*normal*g;
			
		break;
		}

	}

	m_acceleration += 149.e9/pow(86400.,2.0) * nongravitational;	//  convert AU*day-2 to m*sec-2
}

Comet::MODEL Comet::GetModel()
{
	return m_model;
}

void Comet::SetModel(MODEL model)
{
	m_model = model;
}

double Comet::GetA1()
{
	return m_A[0];
}

void Comet::SetA1(double A1)
{
	m_A[0] = A1;
}

double Comet::GetA2()
{
	return m_A[1];
}

void Comet::SetA2(double A2)
{
	m_A[1] = A2;
}

double Comet::GetA3()
{
	return m_A[2];
}

void Comet::SetA3(double A3)
{
	m_A[2] = A3;
}

double Comet::GetEta()
{
	return m_eta;
}

void Comet::SetEta(double eta)
{
	m_eta = eta;
}

double Comet::GetTau()
{
	return m_tau;
}

void Comet::SetTau(double tau)
{
	m_tau = tau;
}

xmlNode * Comet::Serialize(xmlDoc * pXmlDoc, xmlNode * pXmlNode)
{
	//  call ancestor's implementation
	xmlNode * pCometNode = IrregularBody::Serialize(pXmlDoc, pXmlNode);		

	//  create non-gravitational node
	xmlNode * pNongravNode = xmlNewNode( NULL, (const xmlChar *)s_strNongrav.c_str() );
	xmlAddChild( pCometNode, pNongravNode );

	//  model
	xmlNode * pModelNode = xmlNewNode( NULL, (const xmlChar *)s_strModel.c_str() );
	xmlAddChild( pNongravNode, pModelNode );
	SetTextNode( pXmlDoc, pModelNode, m_model );

	//  non-gravitational parameters
	xmlNode * pParamsNode =  xmlNewNode( NULL, (const xmlChar *)s_strParameters.c_str() );
	xmlAddChild( pNongravNode, pParamsNode );
	SetAttribute( pParamsNode, std::string("A1"), m_A[0] );
	SetAttribute( pParamsNode, std::string("A2"), m_A[1] );
	SetAttribute( pParamsNode, std::string("A3"), m_A[2] );

	//  lag angle eta
	xmlNode * pLagAngleNode = xmlNewNode( NULL, (const xmlChar *)s_strEta.c_str() );
	xmlAddChild( pNongravNode, pLagAngleNode );
	SetTextNode( pXmlDoc, pLagAngleNode, m_eta );

	//  asymmetric to perihelion, tau
	xmlNode * pTauAsymmetryNode = xmlNewNode( NULL, (const xmlChar *)s_strTau.c_str() );
	xmlAddChild( pNongravNode, pTauAsymmetryNode );
	SetTextNode( pXmlDoc, pTauAsymmetryNode, m_tau );

	return pNongravNode;
}


void Comet::Deserialize(xmlDoc * pXmlDoc, xmlNode * pXmlNode )
{
	IrregularBody::Deserialize(pXmlDoc, pXmlNode);		//  call ancestor
	
	//  retrieve non-gravitational node
	xmlNode * pNongravNode = SelectSingleNode( pXmlDoc, pXmlNode, s_strNongrav );

	//  model
	std::string xpathModel( s_strModel );
	xpathModel += _T("/text()");
	std::string model = GetXPathResponse( pXmlDoc, pNongravNode, xpathModel );
	m_model = static_cast<Comet::MODEL>( atoi( model.c_str() ) );

	//  non-gravitational parameters
	xmlNode * paramsNode = SelectSingleNode( pXmlDoc, pNongravNode, s_strParameters );
	GetAttribute( paramsNode, std::string("A1"), &m_A[0] );
	GetAttribute( paramsNode, std::string("A2"), &m_A[1] );
	GetAttribute( paramsNode, std::string("A3"), &m_A[2] );

	//  lag angle eta
	std::string xpathEta( s_strEta );
	xpathEta += _T("/text()");
	std::string strEta = GetXPathResponse( pXmlDoc, pNongravNode, xpathEta );
	m_eta = _tstof( strEta.c_str() );	

	//  asymmetric to perihelion, tau
	std::string xpathTau( s_strTau );
	xpathTau += _T("/text()");
	std::string strTau = GetXPathResponse( pXmlDoc, pNongravNode, xpathTau );
	m_tau = _tstof( strTau.c_str() );	
}

std::string Comet::s_strNongrav( _T("nongravitational") );
std::string Comet::s_strModel( _T("model") );
std::string Comet::s_strParameters( _T("parameters") );
std::string Comet::s_strEta( _T("lagangle") );
std::string Comet::s_strTau( _T("tau") );
