#include "StdAfx.h"
#include ".\extendedbodytarget.h"
#include "extendedbody.h"

using namespace JakubSchwarzmeier::Std::Math;
using namespace JakubSchwarzmeier::Std::Graphics;

CExtendedBodyTarget::CExtendedBodyTarget(Body * body, CVisualization * visualization): 
	CBodyTarget( body, visualization )
{
	m_radius = dynamic_cast<ExtendedBody*>( m_pBody )->GetRadius();

	m_iTexture = 0;
	CThreedCanvas * openGLCanvas = m_pVisualization->GetThreedCanvas();
}

CExtendedBodyTarget::~CExtendedBodyTarget(void)
{
	if (m_iTexture != 0)
		throw;
}

void CExtendedBodyTarget::Render( )
{
	CThreedCanvas * openGLCanvas = m_pVisualization->GetThreedCanvas();

	RotationMatrix objectInEye = GetObjectInEye();

	const CFrustum * frustum = openGLCanvas->GetFrustum();

	Trajectory();

	if ( frustum->SphereInFrustum( objectInEye[12], objectInEye[13], objectInEye[14], m_radius ) )
	{
		if ( IsLarge() )
		{
			if ( GetSize() < 25 )	DrawName();

			//  prepare drawing
			int r, g, b;
			m_pBody->GetColor(&r, &g, &b);
			glColor3ub(r, g, b);	
			glLoadMatrixd( objectInEye.GetMatrix() );
			//  end prepare drawing			

			//  load texture if required
			if (m_iTexture == 0)
			{
				wxString strFn = dynamic_cast<ExtendedBody*>( m_pBody )->GetTexture().c_str();
				bool b = wxFileName::FileExists( strFn );

				if (b)  m_iTexture = openGLCanvas->LoadTexture( strFn );
			}

			//  bind my texture name as current
			glBindTexture(GL_TEXTURE_2D, m_iTexture);

			//  normals, needed for lighting computations
			gluQuadricNormals( openGLCanvas->GetQuadric(), GLU_SMOOTH);

			// enable texture on this quadric
			gluQuadricTexture( openGLCanvas->GetQuadric(), GL_TRUE);

			//  for correct texture mapping on GLU sphere
			glMatrixMode(GL_TEXTURE);
				glPushMatrix();
				glLoadIdentity();
				glRotated(180, 0, 1, 0);
				glRotated(180, 0, 0, 1);
			glMatrixMode(GL_MODELVIEW);
			//  end for correct texture mapping on GLU sphere

			//  render sphere with vertices, normals and texture
			glColor4f(1.0, 1.0, 1.0, 1.0);	//  modulate with white color
			gluSphere( openGLCanvas->GetQuadric(), m_radius, 30, 30 );

			//  return back to original texture orientation
			glMatrixMode(GL_TEXTURE);
				glPopMatrix();
			glMatrixMode(GL_MODELVIEW);

			RenderRotationAxis();
			RenderObjectCentricSystem();

			PrintOSD();
		}
		else
		{
			//  draw body name
			DrawName();

			//  prepare drawing
			int r, g, b;
			m_pBody->GetColor(&r, &g, &b);
			glColor3ub(r, g, b);	
			glLoadMatrixd( objectInEye.GetMatrix() );
			//  end prepare drawing		

			//  draw just point
			glPointSize( 3 );

			glPushAttrib( GL_ENABLE_BIT );
				glDisable( GL_TEXTURE_2D );
				glDisable( GL_LIGHTING );

				glBegin (GL_POINTS);
					glVertex2f( 0, 0 );
				glEnd();
			glPopAttrib();
		
		}	//  end else if ( IsLarge() )

	}	//  end SphereInFrustum
}

void CExtendedBodyTarget::OnContextTerminate()
{
	if (m_iTexture != 0)
	{
		glDeleteTextures( 1, &m_iTexture );	//  delete one allocated texture under name of m_iTexture value
		m_iTexture = 0;
	}
}

RotationMatrix CExtendedBodyTarget::GetObjectInEye()
{
	CThreedCanvas * openGLCanvas = m_pVisualization->GetThreedCanvas();

	RotationMatrix res;

	//  modeling transformation
	Vector<> pos = m_pBody->GetPosition();
	RotationMatrix objectInWCSCam;
	objectInWCSCam.SetPosition( pos );
	objectInWCSCam.ChangeRoll( dynamic_cast<ExtendedBody*>( m_pBody )->GetPoleLongitude() );
	objectInWCSCam.ChangePitch( dynamic_cast<ExtendedBody*>( m_pBody )->GetPoleLatitude() - 90 );
	objectInWCSCam.ChangeRoll( dynamic_cast<ExtendedBody*>( m_pBody )->GetPM() );

	res = objectInWCSCam;

	objectInWCSCam.Inverse();

	res = objectInWCSCam;

	res = *(openGLCanvas->GetObserver()->GetCamera());
	res.Inverse();

	RotationMatrix spherical;
	spherical = objectInWCSCam * *(openGLCanvas->GetObserver()->GetCamera());
	spherical.Inverse();

	return RotationMatrix( objectInWCSCam * *(openGLCanvas->GetObserver()->GetCamera()) );
}

double CExtendedBodyTarget::GetFurthermostDistance()
{
	RotationMatrix rot = GetObjectInEye();
	return rot[14] - 3*(m_radius + 0.001*m_radius);
}

double CExtendedBodyTarget::GetNearestDistance()
{
	RotationMatrix rot = GetObjectInEye();
	return rot[14] + 3*(m_radius + 0.001*m_radius);
}

bool CExtendedBodyTarget::IsLarge()
{
	return ( GetSize() > 1 );
}

double CExtendedBodyTarget::GetSize()
{
	GLdouble model[16];
	GLdouble proj[16];
	GLint view[4];

	glGetDoublev(GL_MODELVIEW_MATRIX, model);
	glGetDoublev(GL_PROJECTION_MATRIX, proj);
	glGetIntegerv(GL_VIEWPORT, view);

	//  how will be the sphere big in window?
	//

	//  center of object
	RotationMatrix objectInEye = GetObjectInEye();
	//  end center of object

	//  radius
	CThreedCanvas * openGLCanvas = m_pVisualization->GetThreedCanvas();

	Vector<> pos = m_pBody->GetPosition();
	pos += m_radius;
	RotationMatrix objectInWCSCam;
	objectInWCSCam.SetPosition( pos );
	objectInWCSCam.Inverse();

	RotationMatrix res = objectInWCSCam * *(openGLCanvas->GetObserver()->GetCamera());
	//  end radius

	GLdouble wx1, wy1, wz1;
	int e1 = gluProject(objectInEye[12], 
		objectInEye[13], 
		objectInEye[14],
		model, proj, view,
		&wx1, &wy1, &wz1);

	GLdouble wx2, wy2, wz2;
	int e2 = gluProject(res[12],
		res[13], 
		res[14],
		model, proj, view,
		&wx2, &wy2, &wz2);

	//  if some part is greater than 2 pixels

	double retSize = 0;

	if ( ( (e1 != GL_FALSE) && (e2 != GL_FALSE) ) && 
		 ( (fabs(wx1)< view[2]) || (fabs(wx2)< view[2]) ||
		   (fabs(wy1)< view[3]) || (fabs(wy2)< view[3]) ) )
	{
		retSize = MAX( fabs(wx1-wx2), fabs(wy1-wy2) );
	}

	return retSize;
}

void CExtendedBodyTarget::PrintOSD()
{
	if ( ( m_pVisualization->GetShowOsd() ) && ( GetSize() > 25 ) )
	{
		//  OSD
		m_pVisualization->AddHeading( m_pBody->GetName() );

		Vector<> pos, vel;
		pos = m_pBody->GetPosition();
		vel = m_pBody->GetVelocity();
		wxString strInfo;

		wxString u[7];
		for(int i = 0; i < 3; i++)
		{
			m_pVisualization->DistanceAutoUnit( pos[i], u[i] );
		}

		for(int i = 3; i < 6; i++)
		{
			u[i] = "m/s";
		}

		double radius = dynamic_cast<ExtendedBody*>( m_pBody )->GetRadius();
		m_pVisualization->DistanceAutoUnit( radius, u[6] );

		strInfo.sprintf( "px: %.18g %s\npy: %.18g %s\npz: %.18g %s\nvx: %.18g %s\nvy: %.18g %s\nvz: %.18g %s\nLon: %.2f, Lat: %.2f, PM: %.2f\nPolomr: %.18g %s",
			pos[0], u[0].c_str(), pos[1], u[1].c_str(), 
			pos[2], u[2].c_str(), vel[0], u[3].c_str(), 
			vel[1], u[4].c_str(), vel[2], u[5].c_str(), 
			dynamic_cast<ExtendedBody*>( m_pBody )->GetPoleLongitude(),
			dynamic_cast<ExtendedBody*>( m_pBody )->GetPoleLatitude(),
			dynamic_cast<ExtendedBody*>( m_pBody )->GetPM(),
			radius,
			u[6].c_str()
			);

		m_pVisualization->AddInformation( strInfo );
		
		//  end OSD
	}
}

void CExtendedBodyTarget::RenderRotationAxis()
{
	if ( m_pVisualization->GetRotationAxis() )
	{

		//  render rotation axis
		glPushAttrib( GL_ENABLE_BIT );
			glDisable( GL_TEXTURE_2D );
			glDisable( GL_LIGHTING );

			glBegin(GL_LINES);
				glColor3ub(0, 255, 0);
				glVertex3d(0, 0, 0);
				glVertex3d(0, 0, m_radius*3);

				glColor3ub(255, 255, 128);
				glVertex3d(0, 0, 0);
				glVertex3d(0, 0, -m_radius*3);
			glEnd();

		glPopAttrib();

	}
}

void CExtendedBodyTarget::RenderObjectCentricSystem()
{
	if ( m_pVisualization->GetBodycentricFabric() )
	{

		int lon_step = 5.0;
		double r_step = (3*m_radius - m_radius) / 15;

		glPushAttrib( GL_ENABLE_BIT );
			glDisable( GL_TEXTURE_2D );
			glDisable( GL_LIGHTING );

			glColor3ub(124, 124, 124);

		double startR = m_radius + 0.001*m_radius;
		for(double r = startR; r <= 3 * startR; r += r_step)
		{
			glBegin(GL_LINE_STRIP);
			for(int lon = 0; lon <= 360.0; lon += lon_step)
			{
				double x = r * Functions::sin_angle(lon);
				double y = r * Functions::cos_angle(lon);

				glVertex3d(x, y, 0);
			}
			glEnd();
		}

		glBegin(GL_LINES);
		for(int lon = 0; lon <= 360.0; lon += lon_step)
		{
			double sina = Functions::sin_angle(lon);
			double cosa = Functions::cos_angle(lon);

			double x = startR * sina;
			double y = startR * cosa;

			glVertex3d(x, y, 0);

			x = 3*m_radius * sina;
			y = 3*m_radius * cosa;

			glVertex3d(x, y, 0);
		}
		glEnd();


		glPopAttrib();

	}

}




