#include "StdAfx.h"

#include "rectangularcamera.h"
#include "PhaethonApp.h"
#include "simulation.h"
#include <vector>

using namespace JakubSchwarzmeier::Std::Math;
using namespace JakubSchwarzmeier::Std::Graphics;

IMPLEMENT_DYNAMIC_CLASS (CRectangularCamera, CCamera)

CRectangularCamera::CRectangularCamera()
{
	m_pBody = NULL;
	m_iNaifId = NULL;

    //  dispatch
	mapDispatch.insert( MAPDISPATCH::value_type( this->GetClassInfo() , Assign2 ) );
}

CRectangularCamera::~CRectangularCamera(void)
{
}

CRectangularCamera::CRectangularCamera( const CRectangularCamera & original )
{
	operator=( original );
}

/* static */
void CRectangularCamera::Assign2(CCamera * camera, CCamera * right_side) 
{
	//  retype
	CRectangularCamera * camera_left = dynamic_cast<CRectangularCamera*>( camera );
	CRectangularCamera * camera_right = dynamic_cast<CRectangularCamera*>( right_side );
	camera_left->operator=( *camera_right );
}

CRectangularCamera & CRectangularCamera::operator=( const CRectangularCamera & right_side )
{
	if ( &right_side == this )  return (*this);

	CCamera::operator=( right_side );

	SetCentralBody(right_side.m_iNaifId);
	m_orbiter = right_side.m_orbiter;

	Update();

	return *this;
}

int CRectangularCamera::GetCentralBody()
{
	return m_iNaifId;
}

void CRectangularCamera::SetCentralBody(int iNaifId)
{
	m_iNaifId = iNaifId;

	std::vector<Body*> * bodies = ::wxGetApp().GetDocument()->GetSimulation()->GetBodies();
	std::vector<Body*>::const_iterator it;
	bool found = false;
	for( it = bodies->begin(); it != bodies->end(); it++)
	{
		if ( (*it)->GetSpiceBodyID() == iNaifId )
		{
			m_pBody = (*it);
			found = true;
			break;
		}
	}

	if (!found)  m_pBody = NULL;

    Update();
}

void CRectangularCamera::Update()
{
	RotationMatrix centralBody;
	if (m_pBody!=NULL)  centralBody.SetPosition( m_pBody->GetPosition() );

	RotationMatrix res;
	res = centralBody * m_orbiter;
	for(int i = 0; i < m_size; i++)  
			m_values[i] = res.GetMatrix()[i];
}

void CRectangularCamera::ChangeHeading( double angle )
{
	m_orbiter.ChangeHeading( angle );
	Update();
}

void CRectangularCamera::ChangePitch( double angle )
{
	m_orbiter.ChangePitch( angle );
	Update();
}

void CRectangularCamera::ChangeRoll( double angle )
{
	m_orbiter.ChangeRoll( angle );
	Update();
}

void CRectangularCamera::ChangeRotation( Matrix<> & rotation )
{
	m_orbiter.ChangeRotation( rotation );
	Update();
}

void CRectangularCamera::ChangePosition( Vector<> & position )
{
	m_orbiter.ChangePosition( position );
	Update();
}

void CRectangularCamera::ChangeLambda( double angle )
{
	m_orbiter.ChangeLambda( angle );
	Update();
}

void CRectangularCamera::ChangePhi( double angle )
{
	m_orbiter.ChangePhi( angle );
	Update();
}

void CRectangularCamera::ChangeTheta( double angle )
{
	m_orbiter.ChangeTheta( angle );
	Update();
}

void CRectangularCamera::SetPosition( Vector<> & position )
{
	m_orbiter.SetPosition( position );
	Update();
}

const RotationMatrix & CRectangularCamera::GetOrbiter()
{
	return m_orbiter;
}

void CRectangularCamera::SetOrbiter(RotationMatrix & orbiter)
{
	m_orbiter = orbiter;
}

Vector<> CRectangularCamera::GetPosition() const
{
	return m_orbiter.GetPosition();
}

xmlNode * CRectangularCamera::Serialize(xmlDoc * pXmlDoc, xmlNode * pXmlNode)
{
	//  call ancestor's implementation
	xmlNode * pCameraNode = CCamera::Serialize(pXmlDoc, pXmlNode);		

	//  save central's body NAIF ID
	xmlNode * pNaifIdNode = xmlNewNode( NULL, (const xmlChar *)s_strCentralBodyNaifId.c_str() );
	xmlAddChild( pCameraNode, pNaifIdNode );
	SetTextNode( pXmlDoc, pNaifIdNode, m_iNaifId );

	//  save orbiter matrix
	xmlNode * pOrbiterMatrixNode = xmlNewNode( NULL, (const xmlChar *)s_strOrbiterMatrix.c_str() );
	xmlAddChild( pCameraNode, pOrbiterMatrixNode );
	SetTextNode( pXmlDoc, pOrbiterMatrixNode, m_orbiter );

	return pCameraNode;
}

void CRectangularCamera::Deserialize(xmlDoc * pXmlDoc, xmlNode * pXmlNode )
{
	CCamera::Deserialize(pXmlDoc, pXmlNode);		//  call ancestor
	
	//  retrieve central's body NAIF ID
	std::string xpathNaifId( s_strCentralBodyNaifId );
	xpathNaifId += _T("/text()");
	std::string strNaifId = GetXPathResponse( pXmlDoc, pXmlNode, xpathNaifId );
	int id = atoi( strNaifId.c_str() );
	SetCentralBody( id );

	//  retrieve orbiter matrix
	xmlNode * pOrbiterMatrixNode = SelectSingleNode( pXmlDoc, pXmlNode, s_strOrbiterMatrix );
	GetTextNode( pXmlDoc, pOrbiterMatrixNode, &m_orbiter );
}

Body * CRectangularCamera::GetCentralBodyPtr()
{
	return m_pBody;
}

std::string CRectangularCamera::s_strCentralBodyNaifId( _T("centralbodynaifid") );
std::string CRectangularCamera::s_strOrbiterMatrix( _T("orbitermatrix") );
