#include "stdafx.h"

#include "simulation.h"
#include <spicelib/spicelib.h>
#include "comet.h"
#include <schwarzmeier/excel.h>

IMPLEMENT_CLASS(Simulation, wxObject)

using namespace JakubSchwarzmeier::Std::Math;
using namespace JakubSchwarzmeier::Astronomy;

Simulation::Simulation(void)
{
	m_dt = 86400;
	m_epsilon = 0;
	m_bSPKNeedsUpdate = true;

	#if defined(__WXMSW__)
		m_bGenPlots = false;
	#endif
}

Simulation::~Simulation()
{
	std::vector<Body*>::const_iterator it;
	for ( it = m_vecBodies.begin( ) ; it != m_vecBodies.end( ) ; it++ )
	{
		delete (*it);
	}
	m_vecBodies.clear();

	m_spice.UnloadSpkFiles();
}

Simulation::Simulation( const Simulation & original )
{
	operator=( original );
}

Simulation & Simulation::operator=( const Simulation & right_side )
{
	if ( &right_side == this )  return (*this);

	//  release current bodies
	std::vector<Body*>::const_iterator it;
	for ( it = m_vecBodies.begin( ) ; it != m_vecBodies.end( ) ; it++ )
	{
		delete (*it);
	}
	m_vecBodies.clear();

	//  create new bodies and copy properties from old
	for ( it = right_side.m_vecBodies.begin( ) ; it != right_side.m_vecBodies.end( ) ; it++ )
	{
		Body * body = dynamic_cast<Body *>( (*it)->GetClassInfo()->CreateObject() );
		body->Assign( (*it) );
		
		Add( body );
	}

	//  Create new correct m_parent & m_childList
	//
	
	std::vector<Body*>::const_iterator it2;
	for ( it = right_side.m_vecBodies.begin( ), it2 = m_vecBodies.begin( ); it != right_side.m_vecBodies.end( ) ; it++, it2++ )
	{
		Body * bodyOriginal = (*it);
		Body * bodyNew = (*it2);

		//  Set m_parent

		if (bodyOriginal->m_parent == NULL)
			bodyNew->m_parent = NULL;
		else
		{
			std::vector<Body*>::const_iterator it3, it4;
			for ( it3 = right_side.m_vecBodies.begin( ), it4 = m_vecBodies.begin( ); it3 != right_side.m_vecBodies.end( ) ; it3++, it4++ )
			{
				if (bodyOriginal->m_parent == (*it3))
				{
					bodyNew->m_parent = (*it4);
					break;
				}
			}
		}

		//  Set m_childList
		if (bodyOriginal->m_childList.size() != 0)
		{
			std::vector<Body*>::const_iterator it5;
			for ( it5 = bodyOriginal->m_childList.begin(); it5 != bodyOriginal->m_childList.end( ); it5++ )
			{
				std::vector<Body*>::const_iterator it3, it4;
				for ( it3 = right_side.m_vecBodies.begin( ), it4 = m_vecBodies.begin( ); it3 != right_side.m_vecBodies.end( ) ; it3++, it4++ )
				{
					if ((*it5) == (*it3))
					{
						bodyNew->m_childList.push_back( (*it4) );
						break;
					}
				}
			}
		}

	}

	m_et = right_side.m_et;
	m_dt = right_side.m_dt;
	m_IntegrationMethod = right_side.m_IntegrationMethod;
	m_AccelerationMethod = right_side.m_AccelerationMethod;	
	m_epsilon = right_side.m_epsilon;

	m_setSPKFiles.clear();
	m_setSPKFiles.insert( right_side.m_setSPKFiles.begin(), right_side.m_setSPKFiles.end() );
	m_setPCKFiles.clear();
	m_setPCKFiles.insert( right_side.m_setPCKFiles.begin(), right_side.m_setPCKFiles.end() );

	m_bSPKNeedsUpdate = right_side.m_bSPKNeedsUpdate;
	m_outputFile = right_side.m_outputFile;
	
	return *this;
}


void Simulation::Add(Body * body)
{
	body->SetSimulation( this );

	return m_vecBodies.push_back( body );
}

void Simulation::Remove(Body * body)
{
	std::vector<Body*>::iterator it;
	for ( it = m_vecBodies.begin( ) ; it != m_vecBodies.end( ) ; it++ )
	{
		//  first find body in the list
		if ( (*it) == body )
		{
			//  erase body from list
			m_vecBodies.erase( it );

            //  set parents to children
			std::vector<Body*>::iterator it2;
			for ( it2 = body->m_childList.begin( ) ; it2 != body->m_childList.end( ) ; it2++ )
			{
				//  children's parent will be my parent
				(*it2)->m_parent = body->m_parent;
				//  and add to parent's childlist
				if (body->m_parent != NULL)	body->m_parent->m_childList.push_back( (*it2) );
			}

			//  release me from my parent
			if (body->m_parent != NULL)
			{
				for ( it2 = body->m_parent->m_childList.begin( ) ; it2 != body->m_parent->m_childList.end( ) ; it2++ )
				{
					if ((*it2) == body)	
					{
						body->m_parent->m_childList.erase( it2 );
						break;
					}
				}
			}

			//  free object
			delete body;

			break;
		}
	}
}

void Simulation::ComputeAcceleration()
{
	Barycenter();
	
	//  Compute Newtonian (non-relativistic) acceleration 
	//  needed in any case (for PPN too)

	//  all-pairs computation where 3rd Newton's Law is taken into account
	if ( !m_vecBodies.empty() )
	{
		for(size_t a = 0; a < (m_vecBodies.size()-1); a++)		//  no self interaction (-1)
		{
			Body * bodyA = m_vecBodies.at(a);

			for (size_t b = a + 1; b < m_vecBodies.size(); b++)
			{
				Body * bodyB = m_vecBodies.at(b);

				Vector<> rvect = bodyB->GetPosition() - bodyA->GetPosition();
				double d = rvect.GetLength() + m_epsilon;
				if (fabs(d) < DBL_EPSILON)	throw;

				// force that acts (b)-->(a); accel to (a) from (b)
				bodyA->m_acceleration += s_G * rvect * bodyB->m_mass / pow(d,3);
				// force that acts (b)-->(a); accel to (b) from (a)
				bodyB->m_acceleration += -s_G * rvect * bodyA->m_mass / pow(d,3);
			}
		}
	}

	if (m_AccelerationMethod == PPN)
	{
		//  Compute Post-Parametrized Newtonian (relativistic) acceleration
		//
		//  Based on following papers:
		//
		//	Newhall, XX, E. M. Standish, and J. G. Williams, 1983, "DE102:
		//  a numerically integrated ephemeris of the Moon and planets
		//  spanning forty-four centuries," Astronomy and Astrophysics,
		//  vol. 125, p. 150.
		//
		//  and
		//
		//  C. M. Will, K. Nordtvedt, 1972, "Conservation laws and preffered 
		//  frames in relativistic gravity", Astrophys. J., vol. 177, p. 757.
		//

		//  array for temporary storage of acceleration values
		Vector<> * accel = new Vector<>[ m_vecBodies.size() ];
		for(size_t a = 0; a < m_vecBodies.size(); a++)
			accel[a] = 0;

		//  PPN parameter measuring the nonlinearity in superposition in gravity
		double beta = 1.0;	
		//  PPN parameter measuring space curvature produced by unit rest mass
		double gama = 1.0;
		double c = 299792458.0;  //  m*s^-1, velocity of light in vacuum
		double cc = c * c;

		for(size_t i = 0; i  < m_vecBodies.size(); i++)		
		{
			Body * ithBody = m_vecBodies.at(i);
			double vi = ithBody->m_velocity.GetLength();

			//  ****  first great summation
			for(size_t j = 0; j  < m_vecBodies.size(); j++)		
			{
				if (i == j)  continue;	//  no self-interaction

				Body * jthBody = m_vecBodies.at(j);

				double mij = s_G * jthBody->m_mass;
				double rij = (jthBody->m_position - ithBody->m_position).GetLength();
				double vj = jthBody->m_velocity.GetLength();

				double sum1 = 0;
				for(size_t k = 0; k < m_vecBodies.size(); k++)
				{
					if (k == i)  continue;

					Body * kthBody = m_vecBodies.at(k);

					double mik = s_G * kthBody->m_mass;
					sum1 += mik / (kthBody->m_position - ithBody->m_position).GetLength();
				}

				double sum2 = 0;
				for(size_t k = 0; k < m_vecBodies.size(); k++)
				{
					if (k == j)  continue;

					Body * kthBody = m_vecBodies.at(k);

					double mik = s_G * kthBody->m_mass;
					sum2 += mik / (kthBody->m_position - jthBody->m_position).GetLength();
				}

				Vector<> t1 = mij * (jthBody->m_position - ithBody->m_position)/pow(rij, 3);

				double t2 = 2 * ( (beta + gama) / cc );

 				double t3 = ( 2 * beta - 1 ) / cc;

				double t4 = gama * pow( (vi/c) , 2 );

				double t5 = (1 + gama) * pow( (vj/c) , 2 );

				double t6 = ( 2 * ( 1 + gama ) / cc ) * scalar_product(ithBody->m_velocity, jthBody->m_velocity);

				double bracket = scalar_product( (ithBody->m_position - jthBody->m_position),  jthBody->m_velocity ) / rij;

				double t7 = ( 3 / ( 2 * cc) ) * pow( bracket, 2 );

				double t8 = ( 1/(2*cc) ) * scalar_product( ( jthBody->m_position - ithBody->m_position ), jthBody->m_acceleration );

				accel[i] += t1 * ( 1  - t2*sum1 - t3*sum2 + t4 + t5 - t6 - t7 + t8 );
			}
			//  ****  end first great summation

			//  ****  second great summation
			Vector<> accel2 = 0;
			for(size_t j = 0; j < m_vecBodies.size(); j++)		
			{
				if (i == j)  continue;	//  no self-interaction

				Body * jthBody = m_vecBodies.at(j);
				
				double mij = s_G * jthBody->m_mass;
				double rij = (jthBody->m_position - ithBody->m_position).GetLength();
				double vj = jthBody->m_velocity.GetLength();
			
				double bracket1 = scalar_product( (ithBody->m_position-jthBody->m_position), (2+2*gama)*ithBody->m_velocity - (1+2*gama)*jthBody->m_velocity );

				accel2 += ( mij/pow(rij,3) ) * bracket1 * (ithBody->m_velocity - jthBody->m_velocity);

			}
			//  ****  end second great summation

			accel[i] += ((1/cc)*accel2);

			//  ****  third great summation
			Vector<> accel3 = 0;
			for(size_t j = 0; j < m_vecBodies.size(); j++)
			{
				if (j == i)  continue;

				Body * jthBody = m_vecBodies.at(j);

				double mij = s_G * jthBody->m_mass;
				double rij = (jthBody->m_position - ithBody->m_position).GetLength();

				accel3 += mij * jthBody->m_acceleration / rij;
			}

			//  ****  end third great summation

			double term = (3 + 4 * gama)/(2*cc);

			accel[i] += term * accel3;
		}

		
		//  copy acceleration values from temporary array to body objects
		//
		std::vector<Body*>::const_iterator it;
		int n = 0;
		for ( it = m_vecBodies.begin( ) ; it != m_vecBodies.end( ) ; it++ )
		{
			(*it)->m_acceleration = accel[n++];
		}

		delete [] accel;
		//
		//  end copy acceleration values from temporary array to body objects

		//
		//  end Compute Post-Parametrized Newtonian (relativistic) acceleration
	
	}
}

void Simulation::Update()
{
	//  update position history (Body), rotation (ExtendedBody), 
	//  update with NG forces (Comet)
	std::vector<Body*>::const_iterator it;
	for ( it = m_vecBodies.begin( ) ; it != m_vecBodies.end( ) ; it++ )
	{
		(*it)->Update();
	}

	#if defined(__WXMSW__)
		//  Remember non-gravitational acceleration for graph
		int iPlotsPos = 0;
		if ((m_bGenPlots) && (m_plotWrites >= 60 * 60 * 24))
		{
			for ( it = m_comets.begin( ) ; it != m_comets.end( ) ; it++ )
			{
				m_plots[iPlotsPos]->AppendData( reinterpret_cast<void*>( const_cast<double *>( (*it)->GetAcceleration().GetMatrix() ) ), 3 * sizeof(double) );
				iPlotsPos++;
			}		
		}
	#endif

	//  trigger the right integration method
	if (m_IntegrationMethod == Euler)
	{
		ComputeAcceleration();

		//  Eulerian (leap frog) integration method
		for ( it = m_vecBodies.begin( ) ; it != m_vecBodies.end( ) ; it++ )
		{
			Body * body = *it;

			body->m_velocity += body->m_acceleration * m_dt;
			body->m_position += body->m_velocity * m_dt;
		}
	}
	else
	{
		//  Runge-Kutta method of 4th order

		//  allocating space
		Vector<> ** prevs = new Vector<>*[m_vecBodies.size() ];
		Vector<> ** k1 = new Vector<>*[ m_vecBodies.size() ];
		Vector<> ** k2 = new Vector<>*[ m_vecBodies.size() ];
		Vector<> ** k3 = new Vector<>*[ m_vecBodies.size() ];
		Vector<> ** k4 = new Vector<>*[ m_vecBodies.size() ];
		for (size_t i = 0; i < m_vecBodies.size(); i++)
		{ 
			prevs[i] = new Vector<>[3];
			k1[i] = new Vector<>[2];
			k2[i] = new Vector<>[2];
			k3[i] = new Vector<>[2];
			k4[i] = new Vector<>[2];
		}
		//  end allocating space

		//  save previous values
		for(size_t i = 0; i < m_vecBodies.size(); i++)
		{
			Body * body = m_vecBodies.at(i);
			
			prevs[i][0] = body->GetPosition();
			prevs[i][1] = body->GetVelocity();
			prevs[i][2] = body->m_acceleration;		//  for comets's acceleration
			body->m_acceleration = 0;				//  now clear, for comets
		}

		//  STEP 1
		ComputeAcceleration();

		for(size_t i = 0; i < m_vecBodies.size(); i++)
		{
			Body * body = m_vecBodies.at(i);
			k1[i][0] = m_dt * body->m_velocity;
			k1[i][1] = m_dt * (body->m_acceleration+prevs[i][2]);
			body->m_acceleration = 0;							//  clear acceleration
            
			body->SetPosition( prevs[i][0] + 0.5 * k1[i][0] );
			body->SetVelocity( prevs[i][1] + 0.5 * k1[i][1] );
		}

		//  STEP 2
		ComputeAcceleration();

		for(size_t i = 0; i < m_vecBodies.size(); i++)
		{
			Body * body = m_vecBodies.at(i);
			k2[i][0] = m_dt * body->m_velocity;
			k2[i][1] = m_dt * (body->m_acceleration+prevs[i][2]);
			body->m_acceleration = 0;							//  clear acceleration
            
			body->SetPosition( prevs[i][0] + 0.5 * k2[i][0] );
			body->SetVelocity( prevs[i][1] + 0.5 * k2[i][1] );
		}

		//  STEP 3
		ComputeAcceleration();

		for(size_t i = 0; i < m_vecBodies.size(); i++)
		{
			Body * body = m_vecBodies.at(i);
			k3[i][0] = m_dt * body->m_velocity;
			k3[i][1] = m_dt * (body->m_acceleration+prevs[i][2]);
			body->m_acceleration = 0;							//  clear acceleration
            
			body->SetPosition( prevs[i][0] + k3[i][0] );
			body->SetVelocity( prevs[i][1] + k3[i][1] );
		}

		//  STEP 4
		ComputeAcceleration();

		for(size_t i = 0; i < m_vecBodies.size(); i++)
		{
			Body * body = m_vecBodies.at(i);
			k4[i][0] = m_dt * body->m_velocity;
			k4[i][1] = m_dt * (body->m_acceleration+prevs[i][2]);
		}

		//  FINAL STEP
		for(size_t i = 0; i < m_vecBodies.size(); i++)
		{
			Body * body = m_vecBodies.at(i);
			Vector<> ser1 = (k1[i][0]+2.0*(k2[i][0]+k3[i][0])+k4[i][0])/6.0;
			body->SetPosition( prevs[i][0] + ser1 );
			Vector<> ser2 = (k1[i][1]+2.0*(k2[i][1]+k3[i][1])+k4[i][1])/6.0;
			body->SetVelocity( prevs[i][1]+ ser2 );
		}

        //  free allocated space
		for (size_t i = 0; i < m_vecBodies.size(); i++)
		{ 
			delete [] prevs[i];
			delete [] k1[i];
			delete [] k2[i];
			delete [] k3[i];
			delete [] k4[i];
		}

		delete [] prevs;
		delete [] k1;
		delete [] k2;
		delete [] k3;
		delete [] k4;
        //  end free allocated space
	}

	//  Remeber data for SPK
	//
	size_t nBufferSize = 0;

	//  copy each body to its stream buffer
	int i = 0;
	for ( it = m_vecBodies.begin( ); it != m_vecBodies.end( ) ; it++)
	{
		//  add to buffer
		m_bodyStream[i]->AppendData( reinterpret_cast<void*>( const_cast<double *>( (*it)->GetPosition().GetMatrix() ) ), 3 * sizeof(double) );
		m_bodyStream[i]->AppendData( reinterpret_cast<void*>( const_cast<double *>( (*it)->GetVelocity().GetMatrix() ) ), 3 * sizeof(double) );

		//  size of buffer
		nBufferSize += m_bodyStream[i]->GetDataLen();

		//  advance to buffer of next body
		i++;
	}

	m_nSavedSteps++;

	//  lets say that 25 MBs = 25 * 1024 * 1024 bytes is upper memory limit for buffer
	if ( nBufferSize > 25 * 1024 * 1024 ) 
	{
        //  save SPK file
		SaveSPK();

		//  free buffers
		for(size_t i = 0; i < m_vecBodies.size(); i++)
		{
			delete m_bodyStream[i];
		}
		delete [] m_bodyStream;

		//  allocate new buffers
		m_bodyStream = new wxMemoryBuffer * [ m_vecBodies.size() ];
		for(size_t i = 0; i < m_vecBodies.size(); i++)
		{
			m_bodyStream[i] = new wxMemoryBuffer( );
		}

	}

	//
	//  end Remeber data for SPK


	#if defined(__WXMSW__)
		//  Remember gravitational acceleration for graph
		if (m_bGenPlots)
		{
			if (m_plotWrites >= 60 * 60 * 24)
			{
				for ( it = m_vecBodies.begin( ) ; it != m_vecBodies.end( ) ; it++ )
				{
					m_plots[iPlotsPos]->AppendData( reinterpret_cast<void*>( const_cast<double *>( (*it)->GetAcceleration().GetMatrix() ) ), 3 * sizeof(double) );
					iPlotsPos++;
				}
				
				m_plotWrites = fabs(m_dt);
			}
			else
			{
				m_plotWrites += fabs(m_dt);
			}
		}
	#endif


	//  clear acceleration
	for(size_t i = 0; i < m_vecBodies.size(); i++)
	{
		m_vecBodies.at(i)->m_acceleration = 0;
	}

	//  increase ephemeris time
	m_et += wxTimeSpan(0, 0, m_dt, 0);
}

std::vector<Body*> * Simulation::GetBodies()
{
	return &m_vecBodies;
}

CDateTime * Simulation::GetDateTime()
{
	return &m_et;
}

void Simulation::SetDateTime( CDateTime & et )
{
	m_et = et;
}


Simulation::INTEGRATION_METHOD Simulation::GetIntegrationMethod()
{
	return m_IntegrationMethod;
}

Simulation::ACCELERATION_METHOD Simulation::GetAccelerationMethod()
{
	return m_AccelerationMethod;
}

void Simulation::SetSoftening(double epsilon)
{
	m_epsilon = epsilon;
}

double Simulation::GetSoftening()
{
	return m_epsilon;
}

void Simulation::SetIntegrationMethod(INTEGRATION_METHOD im)
{
	m_IntegrationMethod = im;
}

void Simulation::SetAccelerationMethod(ACCELERATION_METHOD am)
{
	m_AccelerationMethod = am;
}

void Simulation::Serialize(xmlDoc * pXmlDoc, xmlNode * branch)
{
	//  simulation
	xmlNode * pSimulationNode =  xmlNewNode( NULL, (const xmlChar *)s_strSimulation.c_str() );
	xmlAddChild( branch, pSimulationNode );

	//  datetime
	m_et.Serialize( pXmlDoc, pSimulationNode );

	//  integration method
	xmlNode * pIntegrationNode =  xmlNewNode( NULL, (const xmlChar *)s_strIntegration.c_str() );
	xmlAddChild( pSimulationNode, pIntegrationNode );
	SetAttribute( pIntegrationNode, s_strMethod, m_IntegrationMethod );

	//  acceleration method
	xmlNode * pAccelerationNode =  xmlNewNode( NULL, (const xmlChar *)s_strAcceleration.c_str() );
	xmlAddChild( pSimulationNode, pAccelerationNode );
	SetAttribute( pAccelerationNode, s_strMethod, m_AccelerationMethod );

	//  softening parameter epsilon
	xmlNode * epsilonNode = xmlNewNode( NULL, (const xmlChar *)s_strEpsilon.c_str() );
	xmlAddChild( pSimulationNode, epsilonNode );
	SetTextNode( pXmlDoc, epsilonNode, m_epsilon );

	//  timestep dt
	xmlNode * timestepNode = xmlNewNode( NULL, (const xmlChar *)s_strTimeStep.c_str() );
	xmlAddChild( pSimulationNode, timestepNode );
	SetTextNode( pXmlDoc, timestepNode, m_dt );

	//  output file
	//
	xmlNode * outputFileNode = xmlNewNode( NULL, (const xmlChar *)s_strOutputFile.c_str() );
	xmlAddChild( pSimulationNode, outputFileNode );

	//  create relative path from file name
	wxFileName outputFile( m_outputFile.c_str() );
	outputFile.MakeRelativeTo( wxFileName::GetCwd() );

	SetTextNode( pXmlDoc, outputFileNode, outputFile.GetFullPath().c_str() );
	//
	//  end output file

	//  SPICE
	//
	//
	xmlNode * spiceNode = xmlNewNode( NULL, (const xmlChar *)s_strSPICE.c_str() );
	xmlAddChild( pSimulationNode, spiceNode );

	//  SPK
	//
	xmlNode * spkNode = xmlNewNode( NULL, (const xmlChar *)s_strSPK.c_str() );
	xmlAddChild( spiceNode, spkNode );

	//  files
	std::set<std::string>::const_iterator it;
	for ( it = m_setSPKFiles.begin( ) ; it != m_setSPKFiles.end( ) ; it++ )
	{
		xmlNode * fileNode = xmlNewNode( NULL, (const xmlChar *)s_strFile.c_str() );
		xmlAddChild( spkNode, fileNode );

		//  create relative path from file name
		wxFileName spkFile( it->c_str() );
		spkFile.MakeRelativeTo( wxFileName::GetCwd() );

		xmlNode * textNode = xmlNewDocText( pXmlDoc, 
			(const xmlChar *)spkFile.GetFullPath().c_str() );
		xmlAddChild( fileNode, textNode );
	}

	//  PCK
	//
	xmlNode * pckNode = xmlNewNode( NULL, (const xmlChar *)s_strPCK.c_str() );
	xmlAddChild( spiceNode, pckNode );

	//  files
	for ( it = m_setPCKFiles.begin( ) ; it != m_setPCKFiles.end( ) ; it++ )
	{
		xmlNode * fileNode = xmlNewNode( NULL, (const xmlChar *)s_strFile.c_str() );
		xmlAddChild( pckNode, fileNode );

		//  create relative path from file name
		wxFileName pckFile( it->c_str() );
		pckFile.MakeRelativeTo( wxFileName::GetCwd() );

		xmlNode * textNode = xmlNewDocText( pXmlDoc, 
			(const xmlChar *)pckFile.GetFullPath().c_str() );
		xmlAddChild( fileNode, textNode );
	}

	//  bodies
//	xmlNode * bodiesNode = xmlNewNode( NULL, (const xmlChar *)s_strBodies.c_str() );
//	xmlAddChild( pSimulationNode, bodiesNode );

	//  Bodies cannot serialize their descendats by themselves, because serialization
	//  runs along hierarchy of derived classes up to Body class.
	//
	//  call serialization function on bodies
	std::vector<Body*>::const_iterator it2;
	for ( it2 = m_vecBodies.begin( ) ; it2 != m_vecBodies.end( ) ; it2++ )
	{
		Body * body = *it2;

		//  only for bodies within top-level XML
		if (body->m_parent == NULL)  LocalSerializeBodies(pXmlDoc, pSimulationNode, body);
	}
}

void Simulation::LocalSerializeBodies(xmlDoc * pXmlDoc, xmlNode * pSimulationNode, Body * body)
{
	xmlNode * parentNode = body->Serialize( pXmlDoc, pSimulationNode );
	
	//  serialize descendats
	std::vector<Body*>::const_iterator it;
	for ( it = body->m_childList.begin( ) ; it != body->m_childList.end( ) ; it++ )
	{
		Body * childBody = *it;
		LocalSerializeBodies( pXmlDoc, parentNode, childBody );
	}
}

void Simulation::Deserialize(xmlDoc * pXmlDoc, xmlNode * pXmlNode)
{
	xmlNode * simulationNode = SelectSingleNode( pXmlDoc, pXmlNode, s_strSimulation );
	
	//  get datetime
	m_et.Deserialize( pXmlDoc, simulationNode );

	//  integration method
	xmlNode * integrationNode = SelectSingleNode( pXmlDoc, simulationNode, s_strIntegration );
	GetAttribute( integrationNode, s_strMethod, reinterpret_cast<int*>( &m_IntegrationMethod ) );

	//  acceleration method
	xmlNode * accelerationNode = SelectSingleNode( pXmlDoc, simulationNode, s_strAcceleration );
	GetAttribute( accelerationNode, s_strMethod, reinterpret_cast<int*>( &m_AccelerationMethod ) );

	//  softening parameter epsilon
	std::string strEpsilon = GetXPathResponse( pXmlDoc, simulationNode, s_strEpsilon + "/text()" );
	m_epsilon = _tstof( strEpsilon.c_str() );

	//  softening parameter epsilon
	std::string strDT = GetXPathResponse( pXmlDoc, simulationNode, s_strTimeStep + "/text()" );
	m_dt = _tstof( strDT.c_str() );

	//  output file
	std::string strOutputFile = GetXPathResponse( pXmlDoc, simulationNode, s_strOutputFile + "/text()" );

	wxFileName outputFile( strOutputFile.c_str() );
	outputFile.MakeAbsolute( );

	SetOutputFile( outputFile.GetFullPath().c_str() );

	//  SPICE
	//
	//
	xmlNode * spiceNode = SelectSingleNode( pXmlDoc, simulationNode, s_strSPICE );

	//  SPK
	//
	xmlNode * spkNode = SelectSingleNode( pXmlDoc, spiceNode, s_strSPK );
	if (spkNode != NULL)
	{
		//  files
		xmlNodeSet * SPKfiles = SelectNodes( pXmlDoc, spkNode, s_strFile );
		if ( SPKfiles != NULL)
		{
			for(int i = 0; i < SPKfiles->nodeNr; i++)
			{
				xmlChar * text = xmlNodeListGetString(pXmlDoc, SPKfiles->nodeTab[i]->xmlChildrenNode, 1);

				wxFileName spkFile( text );
				spkFile.MakeAbsolute( );

				m_setSPKFiles.insert( spkFile.GetFullPath().c_str() );
				xmlFree( text );
			}
		}
	}

	//  PCK
	//
	xmlNode * pckNode = SelectSingleNode( pXmlDoc, spiceNode, s_strPCK );
	if (pckNode != NULL)
	{
		//  files
		xmlNodeSet * PCKfiles = SelectNodes( pXmlDoc, pckNode, s_strFile );
		if ( PCKfiles != NULL)
		{
			for(int i = 0; i < PCKfiles->nodeNr; i++)
			{
				xmlChar * text = xmlNodeListGetString(pXmlDoc, PCKfiles->nodeTab[i]->xmlChildrenNode, 1);

				wxFileName pckFile( text );
				pckFile.MakeAbsolute( );

				m_setPCKFiles.insert( pckFile.GetFullPath().c_str() );
				xmlFree( text );
			}
		}
	}

	//  deserialize bodies
	LocalDeserializeBodies(pXmlDoc, simulationNode, NULL);
}

void Simulation::LocalDeserializeBodies(xmlDoc * pXmlDoc, xmlNode * pSimulationNode, Body * parent)
{
	xmlNodeSet * bodiesNode = SelectNodes( pXmlDoc, pSimulationNode, s_strBody );
	if (bodiesNode != NULL)
	{
		for(int i = 0; i < bodiesNode->nodeNr; i++)
		{
			//  create object of the right class according to XML record of class name
			Body * body = Body::Deserialize( bodiesNode->nodeTab[i] ) ;

			Add(body);

			//  set relationships
			if ( parent != NULL )
			{
				parent->m_childList.push_back( body );
				body->m_parent = parent;
			}

			//  deserialize that object
			body->Deserialize( pXmlDoc, bodiesNode->nodeTab[i] );

			//  go deeper
			LocalDeserializeBodies( pXmlDoc, bodiesNode->nodeTab[i], body );
		}
	}

}


void Simulation::Barycenter()
{
	//  find barycenter
	//
	Vector<> numerator = 0;
	double denominator = 0;

	std::vector<Body*>::const_iterator it;
	for ( it = m_vecBodies.begin( ) ; it != m_vecBodies.end( ) ; it++ )
	{
		Body * body = *it;

		numerator += body->m_mass * body->m_position;
		denominator += body->m_mass;
	}

	Vector<> barycenterPosition = numerator / denominator;
	//
	//  end find barycenter

	//  update position of all bodies according to barycenter
	for ( it = m_vecBodies.begin( ) ; it != m_vecBodies.end( ) ; it++ )
	{
		Body * body = *it;

		body->m_position -= barycenterPosition;
	}
}

CDateTime Simulation::GetEt()
{
	return m_et;
}

std::set<std::string> * Simulation::GetSpkFiles()
{
	m_bSPKNeedsUpdate = true;
	return &m_setSPKFiles;
}

std::set<std::string> * Simulation::GetPckFiles()
{
	return &m_setPCKFiles;
}

JakubSchwarzmeier::Astronomy::CSpice * Simulation::GetSpice()
{
	if (m_bSPKNeedsUpdate)
	{
		// SPK

		//  first, unload all
		m_spice.UnloadSpkFiles();

		//  reload according to list
		std::set<std::string>::const_iterator it;
		for( it = m_setSPKFiles.begin(); it != m_setSPKFiles.end(); it++)
		{
			m_spice.LoadSpkFile(*it);
		}

		m_spice.LoadSpkFile( m_outputFile );

		//  PCK, text PCK cannot be unloaded
		for( it = m_setPCKFiles.begin(); it != m_setPCKFiles.end(); it++)
		{
			m_spice.LoadKernelFile(*it);
		}		

		m_bSPKNeedsUpdate = false;
	}

	return &m_spice;
}

double Simulation::GetDt()
{
	return m_dt;
}

void Simulation::SetEt( CDateTime dateTime )
{
	m_et = dateTime;
}

double Simulation::GetEpsilon()
{
	return m_epsilon;
}

void Simulation::SetEpsilon(double epsilon)
{
	m_epsilon = epsilon;
}

void Simulation::SetDt(double dt)
{
	m_dt = dt;
}

std::string Simulation::GetOutputFile()
{
	return m_outputFile;
}

void Simulation::SetOutputFile(std::string outputFile)
{
	m_outputFile = outputFile;
	m_spice.LoadSpkFile( m_outputFile );
}

void Simulation::StartSimulation()
{
	//  stream for each body in simulation
	m_bodyStream = new wxMemoryBuffer * [ m_vecBodies.size() ];

    for(size_t i = 0; i < m_vecBodies.size(); i++)
	{
		m_bodyStream[i] = new wxMemoryBuffer( );
	}

	m_dtSimStart = m_et;
	m_nSavedSteps = 0;

	//  unload output SPK file
	m_spice.UnloadSpkFile( m_outputFile );

	//  reload SPK files
	m_bSPKNeedsUpdate = true;
	GetSpice();
	//  reload ephemeris from SPK
	std::vector<Body*>::const_iterator it;
	for ( it = m_vecBodies.begin( ) ; it != m_vecBodies.end( ) ; it++ )
	{
		(*it)->InitializeFromSPICE();
	}

	#if defined(__WXMSW__)
		if (m_bGenPlots)
		{
			if (m_comets.size() > 0)  throw;

			std::vector<Body*>::const_iterator it;
			for ( it = m_vecBodies.begin( ) ; it != m_vecBodies.end( ) ; it++ )
			{
				if ( typeid( (*(*it)) ) == typeid( Comet ) )
				{
					m_comets.push_back( (*it) );
				}
			}

			size_t nGraphs = m_vecBodies.size() + m_comets.size();
			m_plots = new wxMemoryBuffer * [ nGraphs ];

			for(size_t i = 0; i < nGraphs; i++)
			{
				m_plots[i] = new wxMemoryBuffer( );
			}

			m_plotWrites = 60 * 60 * 24;
			m_startEt = m_et;
		}
	#endif
}

void Simulation::StopSimulation()
{
	SaveSPK();

    for(size_t i = 0; i < m_vecBodies.size(); i++)
	{
		delete m_bodyStream[i];
	}

	delete [] m_bodyStream;

	#if defined(__WXMSW__)
		if (m_bGenPlots)
		{
			GeneratePlots();

			for(size_t i = 0; i < m_comets.size() + m_vecBodies.size(); i++)
			{
				delete m_plots[i];
			}

			delete [] m_plots;
			
			m_comets.clear();
		}
	#endif

	m_spice.LoadSpkFile( m_outputFile );
}

void Simulation::SaveSPK()
{
	SpiceInt handle;

	if ( !wxFile::Exists(m_outputFile.c_str()) )
	{
		spkopn_c( m_outputFile.c_str(), "Phaethon SPK file", 5000, &handle );
		
		// 

		wxString strTxtName = wxFileName::CreateTempFileName( wxEmptyString );
		wxTextFile txtFile( strTxtName );
		
		wxString strComment;
		strComment << "BODIES = ";
		for ( std::vector<Body*>::const_iterator it = m_vecBodies.begin( ) ; it != m_vecBodies.end( ) ; it++)
		{
			strComment << (*it)->GetSpiceBodyID() << " ";
		}
		
		strComment << "\nSTART_TIME = " << m_dtSimStart.Format();
		strComment << "\nSTART_TIME_JD = " << m_dtSimStart.GetJDN();
		strComment << "\nTIME_STEP_SIZE = " << m_dt;
		strComment << "\nSOFTENING = " << m_epsilon;
		strComment << "\nINTEGRATION_METHOD = " << m_IntegrationMethod;
		strComment << "\nACCELERATION_METHOD = " << m_AccelerationMethod;
		strComment << "\nPURPOSE = For teaching purposes only.";
		strComment << "\nPRODUCER_ID = Jakub Schwarzmeier, e-mail: Jakub.Schwarzmeier@seznam.cz";

		txtFile.AddLine(strComment);
		txtFile.Write();

		//

        SpiceInt     unit;
		//  Open a new text file for read access, obtaining a Fortran logical unit.
		txtopr_  ( strTxtName.c_str(), &unit );

		spcac_( &handle, &unit, "", "" );

		//  Close the text file.
        ftncls_c ( unit );

		//

		remove( strTxtName.c_str() );
	}
	else
	{
		spkopa_c( m_outputFile.c_str(), &handle );
	}

	std::vector<Body*>::const_iterator it;
	int i = 0;
	for ( it = m_vecBodies.begin( ) ; it != m_vecBodies.end( ) ; it++, i++ )
	{
		double * ephStart = reinterpret_cast<double*>( m_bodyStream[i]->GetData() );

		//  convert to (m --> km, m/s --> km/s), divide all by 1000
		int j = 0;
		for( double * eph = ephStart, j = 0; j < m_nSavedSteps*3*2; eph++, j++ )
		{
            *eph /= 1000.0;
		}

		double start = ( m_dtSimStart.GetJDN() - 2451545.0 ) * 86400.0;
		//  -1.0  needed for aproximation, last states cannot be saved as expansion
		double stop = ( m_et.GetJDN() - 2451545.0 - 1.0 ) * 86400.0;	

		if (m_dt < 0.0)
		{
			start = ( m_et.GetJDN() - 2451545.0 ) * 86400.0;
			stop = ( m_dtSimStart.GetJDN() - 2451545.0 - 1.0 ) * 86400.0;	

			int k = 0;
			int end = m_nSavedSteps*3*2-1;
			double * eph = ephStart;
			double * eph2 = & ephStart[m_nSavedSteps*3*2-6];
			for( k = 0 ;k <= (m_nSavedSteps/2); k++ )
			{
				double tmp[6];
				for(int l = 0; l < 6; l++)
				{
                    tmp[l] = eph[l];
					eph[l] = eph2[l];
					eph2[l] = tmp[l];
				}

                eph += 6;
                eph2 -= 6;
			}
		}

		spkw08_( handle, (*it)->GetSpiceBodyID(), 0/*SSB*/, "ECLIPJ2000",
			start, stop, "Phaethon-SPK", 9, m_nSavedSteps,
            ephStart, start, fabs(m_dt) );

		m_bodyStream[i]->SetDataLen( 0 );		//  free data for this body
	}

	spkcls_c( handle );

	//  set starting ephemeris time of new segment
	m_dtSimStart = m_et + wxTimeSpan(0, 0, m_dt, 0);
	m_nSavedSteps = 0;
}

double Simulation::GetG()
{
	return s_G;
}

#if defined(__WXMSW__)
//	extern std::string C1R1ToExcel(int column, int row);

	void Simulation::SetGeneratePlots(bool bGenPlots)
	{
		m_bGenPlots = bGenPlots;
	}

	void Simulation::GeneratePlots()
	{
		Excel::_ApplicationPtr pExcel = CreateExcel();
		pExcel->PutVisible( 0, VARIANT_FALSE );
		Excel::_WorkbookPtr pWorkbook = AddWorkbook( pExcel );
		Excel::_WorksheetPtr pWorksheet = pExcel->ActiveSheet;		

		int colBase = 0;
		int rowBase = 0;

		int col = colBase;
		int row = rowBase;

		//   Create Tables
		//

		//  Top-Labels
		SetCellValue( pWorksheet, col++, row, _T("Date") );
		SetCellValue( pWorksheet, col++, row, _T("Time") );

		col++;			//  left one column space

		//  non-grav, comets
		std::vector<Body*>::const_iterator it;
		for ( it = m_comets.begin( ) ; it != m_comets.end( ) ; it++ )
		{
			std::string strNongrav( (*it)->GetName() );
			strNongrav += _T(" (non-gravitational)");
			SetCellValue( pWorksheet, col, row, strNongrav );

			SetCellValue( pWorksheet, col,   row+1, "a_jet_x" );
			SetCellValue( pWorksheet, col+1, row+1, "a_jet_y" );
			SetCellValue( pWorksheet, col+2, row+1, "a_jet_z" );
			SetCellValue( pWorksheet, col+3, row+1, "a_jet_total" );

			GetRange( pWorksheet, col, row, col+3, row )->Merge();

			Excel::RangePtr r = GetRange( pWorksheet, col, row, col, row );
			r->Interior->ColorIndex = 36;
			r->GetFont()->Bold = VARIANT_TRUE;
			r->VerticalAlignment = Excel::xlVAlignCenter;

			col += 5;
		}

		//  gravitational acceleration for all bodies
		for ( it = m_vecBodies.begin( ) ; it != m_vecBodies.end( ) ; it++ )
		{
			std::string strNongrav( (*it)->GetName() );
			strNongrav += _T(" (gravitational)");
			SetCellValue( pWorksheet, col, row, strNongrav );

			SetCellValue( pWorksheet, col,   row+1, "a_x" );
			SetCellValue( pWorksheet, col+1, row+1, "a_y" );
			SetCellValue( pWorksheet, col+2, row+1, "a_z" );
			SetCellValue( pWorksheet, col+3, row+1, "a_total" );

			GetRange( pWorksheet, col, row, col+3, row )->Merge();

			Excel::RangePtr r = GetRange( pWorksheet, col, row, col, row );
			r->Interior->ColorIndex = 36;
			r->GetFont()->Bold = VARIANT_TRUE; 
			r->VerticalAlignment = Excel::xlVAlignCenter;

			col += 5;
		}

		col = colBase;
		row +=2;

        //  list values
		wxDateTime et = m_startEt;
		m_plotWrites = 60 * 60 * 24;
		int line = 0;
		while (et < m_et)
		{
			if (m_plotWrites >= 60 * 60 * 24)
			{
				//  date/time
				
				GetRange( pWorksheet, col, row, col, row )->NumberFormat = _T("dd.mm.rrrr");
				SetCellValue( pWorksheet, col++, row, et.FormatDate().c_str() );

				SetCellValue( pWorksheet, col++, row, et.FormatTime().c_str() );

				col++;		//  left one column space

				for ( size_t i = 0; i < m_comets.size() + m_vecBodies.size(); i++ )
				{
					double * dStart = reinterpret_cast<double*>( m_plots[i]->GetData() );
					
					//  print components of vector
					double val = 0;
					for(int j = 0; j < 3; j++)
					{
						double comp = dStart[j + 3*line];
						SetCellValue( pWorksheet, col++, row, comp );
						val += pow( comp, 2 );
					}

					//  print absolute value of vector
					val = sqrt(val);
					SetCellValue( pWorksheet, col++, row, val );

					col++; 		//  left one column space
				}

				row++;
				col = colBase;
				line++;
				m_plotWrites = m_dt;
			}
			else
			{
				m_plotWrites += m_dt;
			}

			et += wxTimeSpan( 0, 0, m_dt );
		}

		//
		//   end Create Tables

		//  Create charts
		//

		for ( size_t i = 0; i < m_comets.size() + m_vecBodies.size(); i++ )
		{
			std::string chartTitle( GetCellValue(pWorksheet, 3 + i*5, 0) );
			
			std::string xlabel( GetCellValue(pWorksheet, 0, 0) );
			std::string ylabel( GetCellValue(pWorksheet, 6 + i*5, 1) );

			Excel::_ChartPtr pChart = CreateChart( pWorkbook, pWorksheet,
                chartTitle, xlabel, ylabel );


			std::string shortName = chartTitle.substr(0, 20);
			if (chartTitle.length() > 20)
				shortName += _T("...)");
			pChart->PutName( shortName.c_str() );

			Excel::SeriesCollectionPtr collection = pChart->SeriesCollection();
			Excel::SeriesPtr set = collection->Item(1);

			//  set x- and y-values
            SetSeriesValues( set, pWorksheet, 
				/* x-axis: date */	    0, 2, 0, line+1,
				/* y-axis: values */   6 + i*5, 2, 6 + i*5, line+1 );

			//  set marker style
			set->PutMarkerStyle( Excel::xlMarkerStyleNone );

			Excel::RangePtr r = GetRange( pWorksheet, 3 + i * 5, 2, 6 + i * 5, line+1 );
			r->Borders->Weight = Excel::xlThin;
			r->Interior->ColorIndex = 36;

			set->PutName( GetCellValue(pWorksheet, 6 + i*5, 1).c_str() );
		}

		Excel::RangePtr compRange = GetRange( pWorksheet, 0, 0, 3 + 5 * m_comets.size() + m_vecBodies.size(), line-1 );
		compRange->EntireColumn->AutoFit();
			
		//
		//  end Create charts

		//  Make sure Excel is visible and give the user control
		//  of Microsoft Excel's lifetime.
		pExcel->PutVisible(0, VARIANT_TRUE);
		pExcel->UserControl = VARIANT_TRUE;

		QuitExcel(pExcel);
	}
#endif


JakubSchwarzmeier::Astronomy::CSpice Simulation::m_spice;
const double Simulation::s_G = 6.67259e-11;

std::string Simulation::s_strSimulation( _T("simulation") );
std::string Simulation::s_strIntegration( _T("integration") );
std::string Simulation::s_strAcceleration( _T("acceleration") );
std::string Simulation::s_strEpsilon( _T("epsilon") );
std::string Simulation::s_strTimeStep( _T("timestep") );
std::string Simulation::s_strOutputFile( _T("outputfile") );
std::string Simulation::s_strBodies( _T("bodies") );
std::string Simulation::s_strMethod( _T("method") );
std::string Simulation::s_strBody( _T("body") );
std::string Simulation::s_strSPICE( _T("spice") );
std::string Simulation::s_strSPK( _T("spk") );
std::string Simulation::s_strPCK( _T("pck") );
std::string Simulation::s_strFile( _T("file") );
