#ifndef __SIMULATION__
#define __SIMULATION__

#include "body.h"
#include <vector>
#include <set>
#include <schwarzmeier/astronomy.h>

class Simulation: public wxObject
{
public:
	Simulation(void);
	Simulation( const Simulation & original );
	~Simulation(void);

	Simulation & operator=( const Simulation & right_side );

	void Add(Body * body);				//  add body to bodies list
	std::vector<Body*> * GetBodies();	//  get pointer vector of bodies
	void Remove(Body * body);			//  remove body from bodies list

	JakubSchwarzmeier::Astronomy::CDateTime * GetDateTime();
	void SetDateTime( JakubSchwarzmeier::Astronomy::CDateTime & et );

	void SetSoftening(double epsilon);
	double GetSoftening();

	std::set<std::string> * GetSpkFiles();
	std::set<std::string> * GetPckFiles();

	//  acceleration computation method
	enum ACCELERATION_METHOD {
		Newtonian = 0, PPN
	};

	//  integration scheme
	enum INTEGRATION_METHOD {
		Euler = 0, RK4
	};	

	ACCELERATION_METHOD GetAccelerationMethod();
	INTEGRATION_METHOD GetIntegrationMethod();

	void SetAccelerationMethod(ACCELERATION_METHOD am);
	void SetIntegrationMethod(INTEGRATION_METHOD im);

	virtual void Update();			//  update positions and velocities

	//  serialize simulation parameters and call bodies serialization
	virtual void Serialize(xmlDoc * pXmlDoc, xmlNode * branch);
	//  deserialize simulation parameters and call bodies deserialization
	virtual void Deserialize(xmlDoc * pXmlDoc, xmlNode * pXmlNode);

	JakubSchwarzmeier::Astronomy::CSpice * GetSpice();
	JakubSchwarzmeier::Astronomy::CDateTime GetEt();
	void SetEt( JakubSchwarzmeier::Astronomy::CDateTime dateTime );
	
	double GetDt();
	void SetDt(double dt);

	double GetEpsilon();
	void SetEpsilon(double epsilon);

	std::string GetOutputFile();
	void SetOutputFile(std::string outputFile);

	void StartSimulation();
	void StopSimulation();

	static double GetG();

	#if defined(__WXMSW__)
		void SetGeneratePlots(bool bGenPlots);
	#endif

protected:
	
	void Barycenter();
	void ComputeAcceleration();
//	void Integrate();				//  not implemented yet
	void SaveSPK();

	void LocalSerializeBodies(xmlDoc * pXmlDoc, xmlNode * pSimulationNode, Body * body);
	void LocalDeserializeBodies(xmlDoc * pXmlDoc, xmlNode * pSimulationNode, Body * parent);

	std::vector<Body*> m_vecBodies;	//  list of bodies
	JakubSchwarzmeier::Astronomy::CDateTime m_et;	//  ephemeris time
	double m_dt;					//  timestep in seconds
	INTEGRATION_METHOD m_IntegrationMethod;		//  integration method
	ACCELERATION_METHOD m_AccelerationMethod;	//  acceleration method
	double m_epsilon;				//  softening parameter
	static const double s_G;		//  gravitational constant

	std::set<std::string> m_setSPKFiles;	
	std::set<std::string> m_setPCKFiles;

	static JakubSchwarzmeier::Astronomy::CSpice m_spice;

	bool m_bSPKNeedsUpdate;
	std::string m_outputFile;

	//  SPK saving
	wxMemoryBuffer ** m_bodyStream;	//  bodies position & velocity buffers during simulating
	wxDateTime m_dtSimStart;		//  starting date/time of current segment
	long m_nSavedSteps;

	#if defined(__WXMSW__)
	//  Excel plots
		bool m_bGenPlots;
		int m_plotWrites;
		wxDateTime m_startEt;			//  starting date/time of simulation
		std::vector<Body*> m_comets;
		wxMemoryBuffer ** m_plots;
		void GeneratePlots();
	#endif

	static std::string s_strSimulation;
	static std::string s_strIntegration;
	static std::string s_strAcceleration;
	static std::string s_strEpsilon;
	static std::string s_strTimeStep;
	static std::string s_strOutputFile;
	static std::string s_strBodies;
	static std::string s_strMethod;
	static std::string s_strBody;
	static std::string s_strSPICE;
	static std::string s_strSPK;
	static std::string s_strPCK;
	static std::string s_strFile;
	
	DECLARE_CLASS(Simulation)

};	//  end class Simulation

#endif // __SIMULATION__