#include "StdAfx.h"
#include ".\sphericalcamera.h"

#include <schwarzmeier/std.h>

using namespace JakubSchwarzmeier::Std::Math;

IMPLEMENT_DYNAMIC_CLASS (CSphericalCamera, CRectangularCamera)

CSphericalCamera::CSphericalCamera()
{
	m_distance = 0;
	m_longitude.LoadIdentity();
	m_latitude.LoadIdentity();

    //  dispatch
	mapDispatch.insert( MAPDISPATCH::value_type( this->GetClassInfo() , Assign2 ) );
}

CSphericalCamera::~CSphericalCamera(void)
{
}

CSphericalCamera::CSphericalCamera( const CSphericalCamera & original )
{
	operator=( original );
}

/* static */
void CSphericalCamera::Assign2(CCamera * camera, CCamera * right_side) 
{
	//  retype
	CSphericalCamera * camera_left = dynamic_cast<CSphericalCamera*>( camera );
	CSphericalCamera * camera_right = dynamic_cast<CSphericalCamera*>( right_side );
	camera_left->operator=( *camera_right );
}

CSphericalCamera & CSphericalCamera::operator=( const CSphericalCamera & right_side )
{
	if ( &right_side == this )  return (*this);

	CRectangularCamera::operator=( right_side );

	m_longitude = right_side.m_longitude;
	m_latitude = right_side.m_latitude;
	m_distance = right_side.m_distance;

	Update();

	return *this;
}

void CSphericalCamera::ChangeLatitude( double angle )
{
	Quaternion<> q( Vector<>(1.0, 0.0, 0.0), angle );
	m_latitude = q * m_latitude;

	Update();
}

void CSphericalCamera::ChangeLongitude( double angle )
{
	Quaternion<> q( Vector<>(0.0, 1.0, 0.0), angle );
	m_longitude = q * m_longitude;

	Update();
}

void CSphericalCamera::ChangeDistance( double distance )
{
	m_distance += distance;
	Update();
}

void CSphericalCamera::SetLongitude( double angle )
{
	m_longitude.CreateFromEulerian( Vector<>(0.0, -1.0, 0.0), angle );
	Update();
}

void CSphericalCamera::SetLatitude( double angle )
{
	m_latitude.CreateFromEulerian( Vector<>(1.0, 0.0, 0.0), angle );
	Update();
}

void CSphericalCamera::SetDistance( double distance )
{
	m_distance = distance;
	Update();
}

double CSphericalCamera::GetLatitude()
{
	double lon = m_latitude.GetLongitude();
	if (lon == 270)	
		return -m_latitude.GetAngle();
	else
		return m_latitude.GetAngle();
}

double CSphericalCamera::GetLongitude()
{
	double lat = m_longitude.GetLatitude();
	if (lat == -90)	
		return -m_longitude.GetAngle();
	else
		return m_longitude.GetAngle();
}

double CSphericalCamera::GetDistance()
{
	return m_distance;
}

void CSphericalCamera::Update()
{
	RotationMatrix distance;
 	distance.SetPosition( Vector<>( 0, 0, m_distance ) );

	Quaternion<> c1;
	c1.CreateFromEulerian( Vector<>(-1.0, 0.0, 0.0), 90 );

	Quaternion<> c2;
	c2.CreateFromEulerian( Vector<>(0.0, 0.0, 1.0), 90 );

	Quaternion<> q;
	q = m_latitude * m_longitude * c1 * c2;

	double mat[16];
	q.CreateMatrix( mat );
	Matrix<> spherical(4, 4, mat);

	m_orbiter = spherical * distance;

	CRectangularCamera::Update();
}

xmlNode * CSphericalCamera::Serialize(xmlDoc * pXmlDoc, xmlNode * pXmlNode)
{
	//  call ancestor's implementation
	xmlNode * pCameraNode = CRectangularCamera::Serialize(pXmlDoc, pXmlNode);		

	//  save distance
	xmlNode * pDistanceNode = xmlNewNode( NULL, (const xmlChar *)s_strDistance.c_str() );
	xmlAddChild( pCameraNode, pDistanceNode );
	SetTextNode( pXmlDoc, pDistanceNode, m_distance );

	//  save longitude
	xmlNode * pLongitudeNode = xmlNewNode( NULL, (const xmlChar *)s_strLongitude.c_str() );
	xmlAddChild( pCameraNode, pLongitudeNode );
	SetTextNode( pXmlDoc, pLongitudeNode, m_longitude );

	//  save latitude
	xmlNode * pLatitudeNode = xmlNewNode( NULL, (const xmlChar *)s_strLatitude.c_str() );
	xmlAddChild( pCameraNode, pLatitudeNode );
	SetTextNode( pXmlDoc, pLatitudeNode, m_latitude );

	return pCameraNode;
}

void CSphericalCamera::Deserialize(xmlDoc * pXmlDoc, xmlNode * pXmlNode )
{
	CRectangularCamera::Deserialize(pXmlDoc, pXmlNode);		//  call ancestor
	
	//  retrieve distance
	std::string xpathDistance( s_strDistance );
	xpathDistance += _T("/text()");
	std::string strDistance = GetXPathResponse( pXmlDoc, pXmlNode, xpathDistance );
	m_distance = atof( strDistance.c_str() );

	//  retrieve longitude
	xmlNode * pLongitudeNode = SelectSingleNode( pXmlDoc, pXmlNode, s_strLongitude );
	GetTextNode( pXmlDoc, pLongitudeNode, &m_longitude );

	//  retrieve latitude
	xmlNode * pLatitudeNode = SelectSingleNode( pXmlDoc, pXmlNode, s_strLatitude );
	GetTextNode( pXmlDoc, pLatitudeNode, &m_latitude );
}

std::string CSphericalCamera::s_strDistance( _T("distance") );
std::string CSphericalCamera::s_strLongitude( _T("longitude") );
std::string CSphericalCamera::s_strLatitude( _T("latitude") );
