#include "StdAfx.h"

#include "visualization.h"
#include "comet.h"
#include "comettarget.h"
#include "star.h"
#include "startarget.h"

#include  "PhaethonApp.h"
#include  "SettingsDlg.h"

using namespace JakubSchwarzmeier::wx;

CVisualization::CVisualization(void)
{
	m_pBodyFont = NULL;
	m_pHeadingFont = NULL;
	m_pInformationFont = NULL;
	m_sfDisplayList = NULL;
	m_starTexture = 0;
}

void CVisualization::Create(Simulation * simulation, CThreedCanvas * pOpenGlCanvas)
{
	m_et = 0;
	m_dt = 1;
	m_fps = 25;
	m_bCoordSolar = true;

	m_pSimulation = simulation;
	m_pOpenGlCanvas = pOpenGlCanvas;
}

CVisualization::~CVisualization(void)
{
	std::vector<CChunkDesc*>::const_iterator it;
	for ( it = m_chunks.begin( ) ; it != m_chunks.end( ) ; it++ )
	{
		delete (*it);
	}
	m_chunks.clear();

	std::vector<CRenderingTarget*>::const_iterator it2;
	for ( it2 = m_renderingTargets.begin( ) ; it2 != m_renderingTargets.end( ) ; it2++ )
	{
		delete (*it2);
	}
	m_renderingTargets.clear();

	

	if ( (m_pBodyFont != NULL) || 
		 (m_pHeadingFont != NULL) || 
		 (m_pInformationFont != NULL)
	   )
	{
		throw;
	}

	if (m_starTexture != 0)
		throw;

}

void CVisualization::SetEt(double et)
{
	m_et = et;

	m_pOpenGlCanvas->AcquireRendering();

	m_pSimulation->SetEt( JakubSchwarzmeier::Astronomy::CDateTime( m_et ) );

	std::vector<Body*> * bodies = m_pSimulation->GetBodies();
	std::vector<Body*>::const_iterator it;
	for ( it = bodies->begin( ) ; it != bodies->end( ) ; it++ )
	{
		(*it)->InitializeFromSPICE();
	}

	CCamera* cam = dynamic_cast<CCamera*>( m_pOpenGlCanvas->GetObserver()->GetCamera() );
	if (cam != NULL)  cam->Update();

	m_pOpenGlCanvas->ReleaseRendering();

	m_pOpenGlCanvas->ReRender();
}

double CVisualization::GetEt()
{
	return m_et;
}

void CVisualization::SetDt(double dt)
{
	m_dt = dt;
}

double CVisualization::GetDt()
{
	return m_dt;
}

void CVisualization::SetOutputVideo(std::string outputVideo)
{
	m_outputVideo = outputVideo;
}

std::string CVisualization::GetOutputVideo()
{
	return m_outputVideo;
}

void CVisualization::SetFps(double fps)
{
	m_fps = fps;
}

double CVisualization::GetFps()
{
	return m_fps;
}

bool CVisualization::GetCoordSolar()
{
	return m_bCoordSolar;
}

void CVisualization::SetCoordSolar(bool coordSolar)
{
	m_bCoordSolar = coordSolar;
}

std::string CVisualization::GetCoordCustom()
{
	return m_coordCustom;
}

void CVisualization::SetCoordCustom(std::string coordCustom)
{
	m_coordCustom = coordCustom;
}

std::vector<CChunkDesc*> * CVisualization::GetChunks()
{
	return &m_chunks;
}

void CVisualization::Serialize(xmlDoc * pXmlDoc, xmlNode * branch)
{
	//  simulation
	xmlNode * pVisualizationNode =  xmlNewNode( NULL, (const xmlChar *)s_strVisualization.c_str() );
	xmlAddChild( branch, pVisualizationNode );

	//  time
	xmlNode * timeNode = xmlNewNode( NULL, (const xmlChar *)s_strTime.c_str() );
	xmlAddChild( pVisualizationNode, timeNode );
	SetTextNode( pXmlDoc, timeNode, m_et );

	//  timestep dt
	xmlNode * timestepNode = xmlNewNode( NULL, (const xmlChar *)s_strTimestep.c_str() );
	xmlAddChild( pVisualizationNode, timestepNode );
	SetTextNode( pXmlDoc, timestepNode, m_dt );

	//  fps
	xmlNode * fpsNode = xmlNewNode( NULL, (const xmlChar *)s_strFps.c_str() );
	xmlAddChild( pVisualizationNode, fpsNode );
	SetTextNode( pXmlDoc, fpsNode, m_fps );

	//  display coordinates as solar ecliptical
	xmlNode * solarNode = xmlNewNode( NULL, (const xmlChar *)s_strCoordSolar.c_str() );
	xmlAddChild( pVisualizationNode, solarNode );
	SetAttribute( solarNode, s_strUse, m_bCoordSolar );

	//  save cutom text for coordinate system
	xmlNode * customNode = xmlNewNode( NULL, (const xmlChar *)s_strCoordCustom.c_str() );
	xmlAddChild( pVisualizationNode, customNode );
	SetTextNode( pXmlDoc, customNode, m_coordCustom );

	//  video file
	//
	xmlNode * outputVideoNode = xmlNewNode( NULL, (const xmlChar *)s_strOutputVideo.c_str() );
	xmlAddChild( pVisualizationNode, outputVideoNode );

	//  create relative path from file name
	wxFileName outputVideo( m_outputVideo.c_str() );
	outputVideo.MakeRelativeTo( wxFileName::GetCwd() );

	SetTextNode( pXmlDoc, outputVideoNode, outputVideo.GetFullPath().c_str() );
	//
	//  end video file

	//  serialize cameras chunks
	xmlNode * cameraChunksNode = xmlNewNode( NULL, (const xmlChar *)s_strCameraChunks.c_str() );
	xmlAddChild( pVisualizationNode, cameraChunksNode );

	std::vector<CChunkDesc*>::const_iterator it;
	for ( it = m_chunks.begin( ) ; it != m_chunks.end( ) ; it++ )
	{
		(*it)->Serialize( pXmlDoc, cameraChunksNode );
	}

}

void CVisualization::Deserialize(xmlDoc * pXmlDoc, xmlNode * pXmlNode)
{
	xmlNode * visualizationNode = SelectSingleNode( pXmlDoc, pXmlNode, s_strVisualization );
	
	//  time
	std::string strET = GetXPathResponse( pXmlDoc, visualizationNode, s_strTime + "/text()" );
	double et = _tstof( strET.c_str() );
	SetEt( et );

	//  timestep dt
	std::string strDT = GetXPathResponse( pXmlDoc, visualizationNode, s_strTimestep + "/text()" );
	m_dt = _tstof( strDT.c_str() );

	//  fps
	std::string strFPS = GetXPathResponse( pXmlDoc, visualizationNode, s_strFps + "/text()" );
	m_fps = _tstof( strFPS.c_str() );

	//  display coordinates as solar ecliptical
	xmlNode * solarNode = SelectSingleNode(pXmlDoc, visualizationNode, s_strCoordSolar );
	if (solarNode != NULL)	GetAttribute( solarNode, s_strUse, &m_bCoordSolar );

	//  save cutom text for coordinate system
	m_coordCustom = GetXPathResponse( pXmlDoc, visualizationNode, s_strCoordCustom + "/text()" );

	//  output file
	std::string strOutputVideo = GetXPathResponse( pXmlDoc, visualizationNode, s_strOutputVideo + "/text()" );

	wxFileName outputVideo( strOutputVideo.c_str() );
	outputVideo.MakeAbsolute( );

	SetOutputVideo( outputVideo.GetFullPath().c_str() );

	//  deserialize cameras chunks
	xmlNode * camerasNode = SelectSingleNode( pXmlDoc, visualizationNode, s_strCameraChunks );
	if (camerasNode != NULL)
	{
		xmlNodeSet * cameraNode = SelectNodes( pXmlDoc, camerasNode, s_strCameraChunk );
		if ( cameraNode != NULL)
		{
			for(int i = 0; i < cameraNode->nodeNr; i++)
			{
				//  create object of the right class according to XML record of class name
				CChunkDesc * chunk = new CChunkDesc;
				chunk->Deserialize( pXmlDoc, cameraNode->nodeTab[i] );
				m_chunks.push_back( chunk );
			}
		}
	}

}

void CVisualization::CreateRenderingTargets()
{
	SetEt( m_et );
	m_pOpenGlCanvas->AcquireRendering();

	m_pOpenGlCanvas->GetMyContext()->SetCurrent();
	std::vector<CRenderingTarget*>::const_iterator it2;
	for ( it2 = m_renderingTargets.begin( ) ; it2 != m_renderingTargets.end( ) ; it2++ )
	{
		(*it2)->OnContextTerminate();
		delete (*it2);
	}
	m_pOpenGlCanvas->GetMyContext()->UnsetCurrent();
	m_renderingTargets.clear();

	std::vector<Body*> * bodies = m_pSimulation->GetBodies();
	std::vector<Body*>::const_iterator it;
	for ( it = bodies->begin( ) ; it != bodies->end( ) ; it++ )
	{
		CRenderingTarget * newTarget;

		if ( (*it)->GetClassInfo()->IsKindOf( CLASSINFO(Comet) ) )
		{
			newTarget = new CCometTarget( (*it), this );
		}
		else
		if ( (*it)->GetClassInfo()->IsKindOf( CLASSINFO(IrregularBody) ) )
		{
			newTarget = new CIrregularBodyTarget( (*it), this );
		}
		else
		if ( (*it)->GetClassInfo()->IsKindOf( CLASSINFO(Star) ) )
		{
			newTarget = new CStarTarget( (*it), this );
		}
		else
		if ( (*it)->GetClassInfo()->IsKindOf( CLASSINFO(ExtendedBody) ) )
		{
			newTarget = new CExtendedBodyTarget( (*it), this );
		}
		else
		if ( (*it)->GetClassInfo()->IsKindOf( CLASSINFO(Body) ) )
		{
			newTarget = new CBodyTarget( (*it), this );
		}
		else
		{
			throw;
		}

		m_renderingTargets.push_back( newTarget );
	}

	m_pOpenGlCanvas->ReleaseRendering();
}


class SmallerDistance : public std::binary_function< CRenderingTarget*, CRenderingTarget*, bool > 
{
public:
    bool operator()( CRenderingTarget * elem1, CRenderingTarget * elem2) const 
    {
		return elem1->GetFurthermostDistance() < elem2->GetFurthermostDistance();
    }
};

void CVisualization::Render()
{
	if (!m_renderingTargets.empty())
	{
		//  clear old headings
		m_headings.clear();
		m_informations.clear();

		//  draw starfield
		SetNearestAndFurthermost(-1, -10);
		glPushMatrix();
			JakubSchwarzmeier::Std::Graphics::RotationMatrix rm = *( m_pOpenGlCanvas->GetObserver()->GetCamera() );
			for(int i = 12; i < 15; i++)	rm[i] = 0;
			glLoadMatrixd( rm.GetMatrix() );
			glCallList(m_sfDisplayList);
		glPopMatrix();

		//  render bodies
		//  First, figure out, if at least one body is greater than one px or so
		//  the screen.
		bool large = false;
		std::vector<CRenderingTarget*>::const_iterator it;
		for ( it = m_renderingTargets.begin( ) ; it != m_renderingTargets.end( ) ; it++ )
		{
			if ( (*it)->IsLarge() )
			{
				large = true;
				break;
			}
		}

		if (large)
		{
			//  sort bodies according to z-distance in eye (opengl) coordinates
			//  from the furthermost to the closest
			sort( m_renderingTargets.begin( ), m_renderingTargets.end( ), SmallerDistance() );

			size_t i;
			for(i = 0; i < m_renderingTargets.size(); i++)
			{
				double furthermost = m_renderingTargets.at(i)->GetFurthermostDistance();
				if (furthermost >= 0)  break;
				double nearest =  MIN( -0.01, m_renderingTargets.at(i)->GetNearestDistance() );
				SetNearestAndFurthermost( nearest, furthermost );

				glPushMatrix();
				m_renderingTargets.at( i )->Render(  );
				glPopMatrix();
			}

			for(; i < m_renderingTargets.size(); i++)
			{
				glPushMatrix();
				m_renderingTargets.at( i )->Render(  );
				glPopMatrix();
			}

			if ( GetDrawTrajectories() )
			{
				SetNearestAndFurthermost(-1e2, -1e10);
				for ( it = m_renderingTargets.begin( ) ; it != m_renderingTargets.end( ) ; it++ )
				{
					glPushMatrix();
					(*it)->DrawTrajectory();
					glPopMatrix();
				}
			}
/*
			//  group bodies into same and overlapping distances and render them
			//
			//  starting values
			size_t starting = 0;
			double furthermost = m_renderingTargets.front()->GetFurthermostDistance();
			double nearest = MIN( -0.01, m_renderingTargets.front()->GetNearestDistance() );

			size_t i = 0;
			CRenderingTarget * renderingTarget;
			while ( (renderingTarget = m_renderingTargets.at( i ))->GetFurthermostDistance() < 0 )
			{
				double thisFurthermost = renderingTarget->GetFurthermostDistance();
				double thisNearest = MIN( -0.01, renderingTarget->GetNearestDistance() );

				if ( thisFurthermost <= nearest )	//  if (overlapping)
				{
					nearest = MAX( thisNearest, nearest );
				}
				else
				{
					SetNearestAndFurthermost( nearest, furthermost );

					for(size_t j = starting; j < i; j++)
					{
						glPushMatrix();
						m_renderingTargets.at( j )->Render(  );
						glPopMatrix();
					}

					starting = i;
					furthermost = thisFurthermost;
					nearest = thisNearest;
				}

				if ( i+1 < m_renderingTargets.size() )
					i++;
				else
					break;
			}

			//  render last group of visible objects
			if ((starting < i) || ( (starting == i) && (i != 0) ) )
			{
    			SetNearestAndFurthermost( nearest, furthermost );
				for(size_t j = starting; j < i+1; j++)
				{
					glPushMatrix();
					m_renderingTargets.at( j )->Render(  );
					glPopMatrix();
				}
				i++;
			}

			//  render z-positive (invisible) bodies
			SetNearestAndFurthermost( 1.0, 2.0 );	//  dummy, however valid values
			for(size_t j = i; j < m_renderingTargets.size(); j++)
			{
				glPushMatrix();
				m_renderingTargets.at( j )->Render(  );
				glPopMatrix();
			}
*/

			//  
			//  end group bodies into same and overlapping distances and render them

			if ( GetShowOsd() )
			{
				//  render labels
				//


				//  left
				//
				glColor3ub(255, 255, 255);

				int left = 10;
				int y = 10;

				for( size_t k = 0; k < m_headings.size(); k++)
				{
					m_pHeadingFont->PrintWindow( left, y, m_headings.at( k ).c_str(), CFreetype::TOP_LEFT );
					y += m_pHeadingFont->GetHeight() + 4;

					m_pInformationFont->PrintWindow( left, y,	m_informations.at( k ).c_str(), CFreetype::TOP_LEFT );
					for(size_t j = 0; j < m_informations.at( k ).length(); j++)
						if (m_informations.at( k )[j] == '\n')	
							y += m_pInformationFont->GetHeight();
					y += 4*m_pHeadingFont->GetHeight() + 4;
				}

				//
				//  end left


				//
				//  end render labels
			
			}	//  end if ( GetShowOsd() )

		//  end if (large)
		}	
		else	
		{
			SetNearestAndFurthermost(-1e5, -1e18);

			for ( it = m_renderingTargets.begin( ) ; it != m_renderingTargets.end( ) ; it++ )
			{
				glPushMatrix();
				(*it)->Render();
				glPopMatrix();

				if ( GetDrawTrajectories() )
				{
					glPushMatrix();
					(*it)->DrawTrajectory();
					glPopMatrix();
				}
			}
		}

		if ( GetShowOsd() )
		{

			glColor3ub(255, 255, 255);
			int left = 10;
			
			std::string * cs;
			if (m_bCoordSolar)
				cs = &s_strSolar;
			else
				cs = &m_coordCustom;

			m_pInformationFont->PrintWindow( left, 10, *cs, CFreetype::BOTTOM_LEFT );

			//  right OSD
			//

			int right = 10;
			int y = 10;
		
			JakubSchwarzmeier::wx::CObserver * obs = m_pOpenGlCanvas->GetObserver();

			if ( obs->IsKindOf( CLASSINFO( CSphericalObserver ) ) )
			{
				//  receive values for display
				CSphericalCamera * cam = dynamic_cast<CSphericalCamera*>( obs->GetCamera() );

				CSphericalCamera mat = *cam;
				mat.Inverse();

				double distance = cam->GetDistance();

				//  convert units
				wxString u[4];
				for(int i = 0; i < 3; i++)
				{
					DistanceAutoUnit( mat[i+12], u[i] );
				}

                DistanceAutoUnit(distance, u[3]);
			
				//  prepare printing string
				wxString camText;
	//			camText.sprintf("Spherical Camera\ndistance: %.18f\nlongitude: %.4f\u00b0\nlatitude: %.4f\u00b0\npx: %.18f\npy: %.18f\npz: %.18f\nX\nex: %.18f\ney: %.18f\nez: %.18f\nY\nex: %.18f\ney: %.18f\nez: %.18f\nZ\nex: %.18f\ney: %.18f\nez: %.18f\nFOV: %.3f\u00b0",
				camText.sprintf("Sfrick kamera\nvzdlenost: %.18g %s\ndlka: %.2f\u00b0\nka: %.2f\u00b0\npx: %.18g %s\npy: %.18g %s\npz: %.18g %s\nX\nex: %.18g\ney: %.18g\nez: %.18g\nY\nex: %.18g\ney: %.18g\nez: %.18g\nZ\nex: %.18g\ney: %.18g\nez: %.18g\nFOV: %.2f\u00b0",
					distance, u[3], cam->GetLongitude(), cam->GetLatitude(),
					mat[12], u[0], mat[13], u[1], mat[14], u[2],
					(*cam)[0], (*cam)[1], (*cam)[2], 
					(*cam)[4], (*cam)[5], (*cam)[6], 
					(*cam)[8], (*cam)[9], (*cam)[10],
					m_pOpenGlCanvas->GetFov()
					);

				glColor3ub( 255, 255, 255 );
				m_pInformationFont->PrintWindow( right, y, camText.c_str(), CFreetype::TOP_RIGHT );
			}
			else
			if ( obs->IsKindOf( CLASSINFO( CRectangularObserver ) ) )
			{
				//  receive values for display
				CRectangularCamera * cam = dynamic_cast<CRectangularCamera*>( obs->GetCamera() );

				JakubSchwarzmeier::Std::Graphics::RotationMatrix orbiter = cam->GetOrbiter();
				orbiter.Inverse();
				
				CRectangularCamera mat = *cam;
				mat.Inverse();

				//  convert units
				wxString u[6];
				for(int i = 0; i < 3; i++)
				{
					DistanceAutoUnit( mat[i+12], u[i] );
				}
				for(int i = 3; i < 6; i++)
				{
					DistanceAutoUnit( orbiter[i+9], u[i] );
				}

				//  prepare printing string
				wxString camText;
				camText.sprintf("Pravohl kamera\nbx: %.18g %s\nby: %.18g %s\nbz: %.18g %s\npx: %.18g %s\npy: %.18g %s\npz: %.18g %s\nX\nex: %.18g\ney: %.18g\nez: %.18g\nY\nex: %.18g\ney: %.18g\nez: %.18g\nZ\nex: %.18g\ney: %.18g\nez: %.18g\nFOV: %.2f\u00b0",
					orbiter[12], u[3], orbiter[13], u[4], orbiter[14], u[5],
					mat[12], u[0], mat[13], u[1], mat[14], u[2],
					(*cam)[0], (*cam)[1], (*cam)[2], 
					(*cam)[4], (*cam)[5], (*cam)[6], 
					(*cam)[8], (*cam)[9], (*cam)[10],
					m_pOpenGlCanvas->GetFov()
					);

				glColor3ub( 255, 255, 255 );
				m_pInformationFont->PrintWindow( right, y, camText.c_str(), CFreetype::TOP_RIGHT );
			}

			//  bottom
			glColor3ub( 255, 255, 255 );
			wxString dateTime;
			dateTime.sprintf("%s, %s UT\nJD %.10g", m_pSimulation->GetEt().FormatDateLong().c_str(),
				m_pSimulation->GetEt().FormatTimeLong().c_str(),
				m_pSimulation->GetEt().GetJDN() );

			m_pInformationFont->PrintWindow( right, y+2*m_pInformationFont->GetHeight(), 
				dateTime.c_str(), CFreetype::BOTTOM_RIGHT );
			
			//
			//  end right

		}	//  end if ( GetShowOsd() )

	}	//  end (!m_renderingTargets.empty())

}

void CVisualization::SetNearestAndFurthermost( double nearest, double furthermost )
{
	if (nearest == furthermost)
	{
		nearest += fabs(nearest/1e6);
		furthermost -= fabs(furthermost/1e6);
	}

	glMatrixMode(GL_PROJECTION); 
	glLoadIdentity();
	gluPerspective(m_pOpenGlCanvas->GetFov(), 
		m_pOpenGlCanvas->GetAspect(), 
		-nearest, -furthermost);		//  gluPerspective wants positive numbers
	glMatrixMode   ( GL_MODELVIEW );
	glClear( GL_DEPTH_BUFFER_BIT );
}

CThreedCanvas * CVisualization::GetThreedCanvas()
{
	return m_pOpenGlCanvas;
}

void CVisualization::AddHeading(std::string heading)
{
	m_headings.push_back(heading);
}

void CVisualization::AddInformation(wxString info)
{
	m_informations.push_back(info.c_str());
}

JakubSchwarzmeier::wx::CFreetype * CVisualization::GetBodyFont()
{
	return m_pBodyFont;
}

Simulation * CVisualization::GetSimulation()
{
	return m_pSimulation;
}

void CVisualization::PreRender()
{
	wxFileName fn( ::wxGetApp().argv[0] );
	fn.SetFullName( _T("pala.ttf") );
	
	//  create font used by RenderingTargets for displaying labels to bodies
	if (m_pBodyFont != NULL)  throw;
	m_pBodyFont = new JakubSchwarzmeier::wx::CFreetype;
	m_pBodyFont->Create( fn.GetFullPath().c_str(), 13);

	//  create font for OSD
	if (m_pHeadingFont != NULL)  throw;
	m_pHeadingFont = new JakubSchwarzmeier::wx::CFreetype;
	m_pHeadingFont->Create( fn.GetFullPath().c_str(), 13 );
	if (m_pInformationFont != NULL)  throw;
	m_pInformationFont = new JakubSchwarzmeier::wx::CFreetype;
	m_pInformationFont->Create( fn.GetFullPath().c_str(), 10 );

	//  call prerender on RenderingTargets
	std::vector<CRenderingTarget*>::const_iterator it;
	for ( it = m_renderingTargets.begin( ) ; it != m_renderingTargets.end( ) ; it++ )
	{
		(*it)->PreRender();
	}

	CreateStarField();
}

void CVisualization::OnContextTerminate()
{
	//  
	if (m_pBodyFont != NULL)
	{
		delete m_pBodyFont;
		m_pBodyFont = NULL;
	}

	if (m_pHeadingFont != NULL)
	{
		delete m_pHeadingFont;
		m_pHeadingFont = NULL;
	}

	if (m_pInformationFont != NULL)
	{
		delete m_pInformationFont;
		m_pInformationFont = NULL;
	}


	if (m_sfDisplayList != 0)
		glDeleteLists(m_sfDisplayList, 1);

	if (m_starTexture != 0)
	{
		glDeleteTextures( 1, &m_starTexture );
		m_starTexture = 0;
	}

	//  call OnContextTerminate on RenderingTargets
	std::vector<CRenderingTarget*>::const_iterator it;
	for ( it = m_renderingTargets.begin( ) ; it != m_renderingTargets.end( ) ; it++ )
	{
		(*it)->OnContextTerminate();
	}
}

void CVisualization::CreateStarField()
{
	m_sfDisplayList = glGenLists(1);
	if (m_sfDisplayList == 0)  throw;

	if (m_starTexture != 0)	throw;

	m_starTexture = m_pOpenGlCanvas->LoadTexture( "star.jpg" );

	glNewList ( m_sfDisplayList, GL_COMPILE );
		glPushAttrib( GL_ENABLE_BIT );

			glEnable( GL_TEXTURE_2D );
			glEnable( GL_BLEND );
			glBlendFunc( GL_SRC_ALPHA, GL_ONE );
			glDisable( GL_LIGHTING );
			glDisable(GL_DEPTH_TEST);

	// Set up the OpenGL state machine for using point sprites...
	//

    // This is how will our point sprite's size will be modified by 
    // distance from the viewer
    float quadratic[] =  { 1.0f, 0.0f, 0.01f };
    glPointParameterfvARB( GL_POINT_DISTANCE_ATTENUATION_ARB, quadratic );

    // Query for the max point size supported by the hardware
    float maxSize = 0.0f;
    glGetFloatv( GL_POINT_SIZE_MAX_ARB, &maxSize );

    // Clamp size to 100.0f or the sprites could get a little too big on some  
    // of the newer graphic cards. My ATI card at home supports a max point 
    // size of 1024.0f!
    if( maxSize > 100.0f )
        maxSize = 100.0f;

    glPointSize( maxSize );

    // The alpha of a point is calculated to allow the fading of points 
    // instead of shrinking them past a defined threshold size. The threshold 
    // is defined by GL_POINT_FADE_THRESHOLD_SIZE_ARB and is not clamped to 
    // the minimum and maximum point sizes.
    glPointParameterfARB( GL_POINT_FADE_THRESHOLD_SIZE_ARB, 60.0f );

    glPointParameterfARB( GL_POINT_SIZE_MIN_ARB, 1.0f );
    glPointParameterfARB( GL_POINT_SIZE_MAX_ARB, maxSize );

	glBindTexture (GL_TEXTURE_2D, m_starTexture);

    // Specify point sprite texture coordinate replacement mode for each 
    // texture unit
    glTexEnvf( GL_POINT_SPRITE_ARB, GL_COORD_REPLACE_ARB, GL_TRUE );

		    glEnable( GL_POINT_SPRITE_ARB );

				glPointSize( 1 );

				glBegin ( GL_POINTS );
					for ( int i = 0; i < 500; i++ )
					{
						CreateStar();
					}
					
				glEnd ( );

				glPointSize( 4 );

				glBegin ( GL_POINTS );
					for ( int i = 0; i < 80; i++ )
					{
						CreateStar();
					}
					
				glEnd ( );

			glDisable( GL_POINT_SPRITE_ARB );


		glPopAttrib();

	glEndList ( );
}

void CVisualization::CreateStar()
{
	static const double STAR_RADIUS = 5;

	GLubyte g, b;
	g = 180.0 + JakubSchwarzmeier::Std::Math::Functions::Random( 1 ) * ( 255.0 - 180.0 );
	b = 180.0 + JakubSchwarzmeier::Std::Math::Functions::Random( 1 ) * ( 255.0 - 180.0 );
	glColor3ub(252, g, b );

	double x, y, z, elevation, azimuth;

	elevation  = JakubSchwarzmeier::Std::Math::Functions::Random( 1 ) * 2 * M_PI;
	azimuth    = JakubSchwarzmeier::Std::Math::Functions::Random( 1 ) * 2 * M_PI;

	y = STAR_RADIUS * sin ( elevation );
	x = STAR_RADIUS * cos ( elevation ) * cos ( azimuth );
	z = STAR_RADIUS * cos ( elevation ) * sin ( azimuth );

	glVertex3d ( x, y, z );
}

bool CVisualization::GetShowOsd()
{
	CSettingsDlg settings;
	return settings.GetShowOsd();
}

bool CVisualization::GetAutoUnits()
{
	CSettingsDlg settings;
	return settings.GetAutoUnits();
}

bool CVisualization::GetRotationAxis()
{
	CSettingsDlg settings;
	return settings.GetRotationAxis();
}

bool CVisualization::GetBodycentricFabric()
{
	CSettingsDlg settings;
	return settings.GetBodycentricFabric();
}

bool CVisualization::GetDrawTrajectories()
{
	CSettingsDlg settings;
	return settings.GetDrawTrajectories();
}

void CVisualization::DistanceAutoUnit( double & val, wxString & strUnit )
{
	if ( GetAutoUnits() )
	{
		static const double AU = 149597870e3;

		if ( val > 1.5 * AU )
		{
			val /= AU;
			strUnit = "AU";
		}
		else
		if ( val > 1e4 )
		{
			val /= 1e3;
			strUnit = "km";
		}
		else
		{
			strUnit = "m";
		}
	}
	else
			strUnit = "m";
}

GLEWContext * CVisualization::glewGetContext()
{
	return m_pOpenGlCanvas->GetMyContext()->glewGetContext();
}

WGLEWContext * CVisualization::wglewGetContext()
{
	return m_pOpenGlCanvas->GetMyContext()->wglewGetContext();
}

void CVisualization::ClearTrajectories()
{
	std::vector<CRenderingTarget*>::const_iterator it2;
	for ( it2 = m_renderingTargets.begin( ) ; it2 != m_renderingTargets.end( ) ; it2++ )
	{
		(*it2)->ClearTrajectory();
	}
}

std::string CVisualization::s_strVisualization( _T("visualization") );
std::string CVisualization::s_strTimestep( _T("timestep") );
std::string CVisualization::s_strTime( _T("time") );
std::string CVisualization::s_strFps( _T("fps") );
std::string CVisualization::s_strCoordSolar( _T("coordSolar") );
std::string CVisualization::s_strCoordCustom( _T("coordCustom") );
std::string CVisualization::s_strOutputVideo( _T("outputVideo") );
std::string CVisualization::s_strCameraChunks( _T("camerachunks") );
std::string CVisualization::s_strCameraChunk( _T("camerachunk") );
std::string CVisualization::s_strUse( _T("use") );

std::string CVisualization::s_strSolar( _("Souadnicov systm: Slunen ekliptikln/ICRF/J2000.0") );
